/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic

import com.netapp.df.uba.server.common.elastic.BulkFlushBackoffMode.BulkFlushBackoffMode
import com.netapp.df.uba.server.common.misc.utils.SettingUtils
import com.netapp.df.uba.server.common.misc.utils.SettingUtils._
import com.typesafe.config.Config

import scala.concurrent.duration._

/**
  * Indicates the mode for flush backoff when there are failures in bulk requests
  * Used to configure the right backoff policy in the bulk processor
  */
object BulkFlushBackoffMode extends Enumeration {
  type BulkFlushBackoffMode = Value

  val NoBackoff: BulkFlushBackoffMode = Value("nobackoff")
  val Constant: BulkFlushBackoffMode = Value("constant")
  val Exponential: BulkFlushBackoffMode = Value("exponential")
}

/**
  * BulkFlushBackoffStrategy defines the overall strategy for configuring
  * the backoff policy on the bulk processor
  */
object BulkFlushBackoffStrategy {

  def apply(config: Config): BulkFlushBackoffStrategy = {
    val bulkFlushBackoffMode: BulkFlushBackoffMode = BulkFlushBackoffMode.withName(config.getString("mode"))

    bulkFlushBackoffMode match {
      case BulkFlushBackoffMode.NoBackoff =>
        BulkFlushBackoffStrategy(
          bulkFlushBackoffMode,
          None,
          None
        )

      case BulkFlushBackoffMode.Constant =>
        // here retry count is mandatory
        val backoffMaxRetryCount: Int = config.getInt("retry.count")
        if (backoffMaxRetryCount < 0) throw new IllegalArgumentException("Bulk flush backoff max retry count needs to be greater than 0.")

        BulkFlushBackoffStrategy(
          bulkFlushBackoffMode,
          Some(backoffMaxRetryCount),
          None
        )

      case BulkFlushBackoffMode.Exponential =>
        // Here delay is mandatory
        val delay: FiniteDuration = SettingUtils.validMinDuration(config, "delay", 100.millis)
        BulkFlushBackoffStrategy(
          bulkFlushBackoffMode,
          None,
          Some(delay)
        )
    }
  }
}

case class BulkFlushBackoffStrategy private(backoffMode: BulkFlushBackoffMode,
                                            backoffMaxRetryCount: Option[Int],
                                            delay: Option[FiniteDuration]) extends Serializable {
  //noinspection ScalaUnusedSymbol
  private final val serialVersionUID = 1L
}

/**
  * BulkFlushSettings represent the bulk processor flush settings to be used in a flink elastic sink.
  */
object BulkFlushSettings {

  def apply(config: Config): BulkFlushSettings = {

    val maxActions: Option[Int] = SettingUtils.configValueWithEmptyAsNoneInt(config, "max.actions")
    val maxSizeMb: Option[Int] = SettingUtils.configValueWithEmptyAsNoneInt(config, "max.size.mb")

    val intervalStringOption = SettingUtils.configValueWithEmptyAsNone(config, "interval")
    val intervalDuration: Option[FiniteDuration] = intervalStringOption match {
      case None => None
      case Some(_) => Some(config.getDuration("interval"))
    }
    BulkFlushSettings(
      maxActions,
      maxSizeMb,
      intervalDuration,
      BulkFlushBackoffStrategy(config.getConfig("backoff")))
  }
}

/**
  * A case class to represent the bulk flush options for the bulk processor of Elasticsearch.
  * BulkProcessor class offers a simple interface to flush bulk operations automatically
  * based on the number or size of requests, or after a given period which is represented here.
  * It also maintains the backoff strategy options
  */
case class BulkFlushSettings(maxActions: Option[Int],
                             maxSizeMb: Option[Int],
                             interval: Option[FiniteDuration],
                             bulkFlushBackoffStrategy: BulkFlushBackoffStrategy) extends java.io.Serializable {

  //noinspection ScalaUnusedSymbol
  private final val serialVersionUID = 1L
}