/*
 * $Id$
 *
 * Copyright (c) 2021 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.df.uba.server.common.elastic

import com.netapp.df.uba.server.common.elastic.ElasticUtils.DEFAULT_SOCKET_TIMEOUT_MILLIS
import com.netapp.df.uba.server.common.elastic.model.ElasticClientInfo
import com.netapp.df.uba.server.common.misc.log.{LogTrait, MdcUtils}
import com.netapp.df.uba.server.common.misc.utils.GenericUtils
import com.netapp.oci.tenants.TenantsRegistry
import com.sksamuel.elastic4s.http.{ElasticClient, ElasticNodeEndpoint, NoOpHttpClientConfigCallback}
import org.apache.http.HttpHost
import org.elasticsearch.client.RestClient

import scala.util.{Failure, Success, Try}

object ElasticClientFactory extends LogTrait {

  val GLOBAL_TENANT_ID = "tenant_id_global"
  val GLOBAL_CLUSTER_ID = "cluster_id_global"

  def getCsElasticClient(tenantId: String, elasticSettings: ElasticSettings): Try[ElasticClientInfo] = Try {
    if (GLOBAL_TENANT_ID == tenantId) {
      //Special case of Global tenant Id, where we skip tenant cache lookup. In this case, ES details are picked from
      //the caller provided elasticSettings
      logDebugT(tenantId, "Creating elastic client for the global tenant")

      val csElasticClient: CsElasticClient = createCsElasticClient(elasticSettings)
      MdcUtils.setElasticContextForLogging(GLOBAL_CLUSTER_ID)
      ElasticClientInfo(GLOBAL_CLUSTER_ID, List(tenantId), csElasticClient)
    } else {

      val tenantOption = Option(GenericUtils.retry(5, "Getting tenant details from tenants registry")(TenantsRegistry.getInstance().get(tenantId)))

      tenantOption match {
        case Some(tenant) =>
          val tenantEsComponent = tenant.getLocations.getElasticsearch
          logDebugT(tenantId, "Creating elastic client for the tenant")
          val csElasticClient: CsElasticClient = createCsElasticClient(tenantEsComponent.getHosts, tenantEsComponent.getHttpPort, elasticSettings)
          MdcUtils.setElasticContextForLogging(tenantEsComponent.getClusterName)
          ElasticClientInfo(tenantEsComponent.getClusterName, List(tenantId), csElasticClient)

        case None =>
          throw new Exception(s"Failed to get tenant details for tenant $tenantId from TenantRegistry.")
      }
    }
  }

  def closeCsElasticClient(csElasticClientInfoTry: Try[ElasticClientInfo]): Unit = {
    csElasticClientInfoTry match {
      case Failure(ex) => logger.warn(s"Failed to close the elastic client. Reason: ${ex.getMessage}")
      case Success(csElasticClientInfo) => csElasticClientInfo.client.close()
    }
  }

  private def createCsElasticClient(hosts: Seq[String], port: Int, elasticSettings: ElasticSettings): CsElasticClient = {
    val elasticHosts = getHttpHostEndpoints("http", hosts, port)
    val restClient = createElasticRestClient(elasticHosts, elasticSettings)
    CsElasticClient(ElasticClient.fromRestClient(restClient))
  }

  private def createElasticRestClient(elasticHosts: Seq[HttpHost], elasticSettings: ElasticSettings): RestClient = {

    logger.debug(s"Elastic search cluster endpoints : $elasticHosts")
    /*
     A copy of the ElasticClient apply method to ensure the setting of max retry timeout
     which is necessary for scenarios as explained in https://github.com/elastic/elasticsearch/issues/33342

     However, note that setting of maxRetryTimeout has been deprecated from Elastic version 6.7.0
     and also the default value for this timeout has been increased to 90 seconds
     Refer to https://github.com/elastic/elasticsearch/pull/38425 and
     https://github.com/elastic/elasticsearch/pull/38085
    */
    GenericUtils.retry(5, "Creating elastic rest client", 2000) {
      RestClient
        .builder(elasticHosts: _*)
        .setRequestConfigCallback(requestConfigBuilder =>
          //Reference: https://github.com/elastic/elasticsearch/issues/24069. Setting
          //connection request timeout as 0 to address random Socket timeout exceptions on client.
          //Also increased connection timeout from default 1s, to keep it same as socket timeout.
          requestConfigBuilder
            .setConnectTimeout(DEFAULT_SOCKET_TIMEOUT_MILLIS)
            .setSocketTimeout(DEFAULT_SOCKET_TIMEOUT_MILLIS)
            .setConnectionRequestTimeout(0))
        .setHttpClientConfigCallback(NoOpHttpClientConfigCallback)
        .setMaxRetryTimeoutMillis(elasticSettings.retryTimeout.toMillis.toInt)
        .build()
    }
  }

  private def createCsElasticClient(elasticSettings: ElasticSettings): CsElasticClient = {
    if (elasticSettings.httpHostsProps.isEmpty) throw new IllegalArgumentException("Missing http hosts to create elastic client")

    val elasticHosts = elasticSettings.httpHostsProps.get.endpoints.map {
      case ElasticNodeEndpoint(protocol, host, port, _) => new HttpHost(host, port, protocol)
    }
    val restClient = createElasticRestClient(elasticHosts, elasticSettings)
    CsElasticClient(ElasticClient.fromRestClient(restClient))
  }

  private def getHttpHostEndpoints(protocol: String, hosts: Seq[String], port: Int): Seq[HttpHost] = {
    val httpHosts: Seq[HttpHost] = hosts
      .map(hostName => new HttpHost(hostName, port, protocol))
    httpHosts
  }
}
