/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic

import com.netapp.df.uba.server.common.elastic.ElasticUtils.logger
import com.netapp.df.uba.server.common.elastic.model.TimeFilter
import com.netapp.df.uba.server.common.elastic.utils.FieldConstants
import com.netapp.df.uba.server.common.misc.api.JsonFormats
import com.netapp.df.uba.server.common.misc.utils.GenericUtils
import com.sksamuel.elastic4s.http.ElasticClient
import com.sksamuel.elastic4s.http.ElasticDsl.{getIndex, _}
import com.sksamuel.elastic4s.http.index.GetIndexResponse
import spray.json._

import java.text.SimpleDateFormat
import java.time.format.DateTimeFormatter
import java.time.temporal.ChronoUnit
import java.time.{Instant, ZoneId}
import java.util.concurrent.TimeUnit
import java.util.{Calendar, TimeZone}
import scala.concurrent.ExecutionContext.Implicits.global
import scala.concurrent.duration._
import scala.concurrent.{Await, Future}
import scala.util.{Failure, Success, Try}

/*
  Case classes and JSON format classes for _cat/indices and _cat calls
 */
case class IndexAndCreationDate(idx: String, cd: String)

//Buffer times for index selection:
//bufferForIndexSelection - Buffer for running cat index calls
//bufferForFinalIndexSet - Buffer for returning final indices list
case class IndexBufferTimes(bufferForIndexSelection: FiniteDuration, bufferForFinalIndexSet: FiniteDuration)

class CatIndicesJsonFormats extends JsonFormats {
  implicit val indexAndCreationDateFormatter: RootJsonFormat[IndexAndCreationDate] = jsonFormat2(IndexAndCreationDate.apply)
}

object ElasticIndexSelector {
  /*
   Run get index call with specific base name as base and return the result map where
   the actual index name is the key and settings of the index as the value in the map.
   */
  def invokeGetIndex(elasticClient: ElasticClient, indexBaseName: String): Try[Map[String, GetIndexResponse]] = {
    try {
      val apiResponse = elasticClient.execute {
        getIndex(s"$indexBaseName-*")
      }.await

      ElasticUtils.logAndGetResult(s"$indexBaseName", apiResponse)
    } catch {
      case ex: Exception =>
        logger.error(s"Exception in get index for $indexBaseName. Reason: ${ex.getMessage}.")
        Failure(ex)
    }
  }

  /*
   Run _cat/indices/<one or more index patterns>?h=index,cd with specific index patterns and return the actual indices
   with their creation date.
   */
  private def computeRelevantIndicesAndCreationDate(indexBaseName: String, elasticClient: ElasticClient, indicesList: List[String]): Try[List[IndexAndCreationDate]] = {
    implicit val catIndicesJsonFormats: CatIndicesJsonFormats = new CatIndicesJsonFormats()
    import catIndicesJsonFormats._

    //For indices which get migrated from another ES cluster, correct creation time should be referred from original source index.
    //In such case, creation time based on index.creation_date will be wrong and instead is part of migrated index name itself as
    //below and hence extracted from index name in such case:
    //(index name prefix length) + (1 for hyphen) + (date length) + (1 for hyphen) + (13 for creation epoch timestamp)
    //E.g. audit-events-2021.10.28-1635437229278
    val indexNamePrefixLength: Int = indexBaseName.length + 1
    val indexNameWithCreationDateLength: Int = indexNamePrefixLength + ElasticUtils.DatePattern.length + 1 + 13

    val indicesListStr = indicesList.mkString(",")
    try {
      /*
       Returns the response in JSON format
       [
        {
          "idx": "audit-events-2020.03.15-000001",
          "cd": "1592330102936"
        },
        {
          "index": "audit-events-2020.03.09-000001",
          "cd": "1592330104420"
        }
       ]
       Note: Using read alias incurs heavy overhead on CAT APIs if number of indices are too much. Hence optimizing to use
       specific indices list.
       */
      val catIndicesResponseAsString: String = GenericUtils.execAndLogTime({
        val catIndicesFuture: Future[String] = ElasticUtils.invokeRestApiWithElastic4sClient(
          elasticClient,
          "GET",
          s"/_cat/indices/$indicesListStr?h=idx,cd&&format=json"
        )
          .map(httpResponse => httpResponse.entity.get.content)

        // Await till the request is completed
        Await.result(catIndicesFuture, 10.seconds)
      }, s"Running cat index on indices $indicesListStr")

      val allIndicesAndCreationDate: List[IndexAndCreationDate] = catIndicesResponseAsString
        .trim
        .parseJson
        .convertTo[List[IndexAndCreationDate]]

      val indicesRelevantForSearch = allIndicesAndCreationDate
        .filterNot(indexAndCreationDate => indexAndCreationDate.idx.contains(ElasticUtils.LatestSuffix)) // Does it even come in get index response, but just being safe
        .map(indexAndCreationDate => {
          //If its a migrated or coalesced index, creation time based on index.creation_date will be wrong. Instead creation time of
          //original index is picked from last part of migrated or coalesced index name string.
          val indexCreationTimestamp: String = if (indexNameWithCreationDateLength == indexAndCreationDate.idx.length) indexAndCreationDate.idx.split("-").last
          else indexAndCreationDate.cd
          IndexAndCreationDate(indexAndCreationDate.idx, indexCreationTimestamp)
        })

      logger.info(s"Cat index response - total indices: ${indicesRelevantForSearch.size}")
      Success(indicesRelevantForSearch)

    } catch {
      case ex: Exception =>
        logger.error(s"Exception in indices API call for $indicesListStr. Reason: ${ex.getMessage}.")
        Failure(ex)
    }
  }

  def filterRelevantIndicesForSearch(indexBaseName: String,
                                     indexWithCreationDateTupleList: List[IndexAndCreationDate],
                                     bufferedLowerEndTimestampInMillis: Long,
                                     bufferedHigherEndTimestampInMillis: Long): List[String] = {
    /*
         For migrated indices from other ES cluster, the creation date is part of index name itself.
         Now there are 2 groups of indices created:
         I. ONE with true creation date embedded as part of index name (Migrated indices) and
         II. SECOND with creation date as usual part of index setting index.creation_date (Normal indices).

         PARTITION LOGIC FOR NORMAL INDICES (group II):
         Partition the list of normal indices as:
         First partition:
          - Indices with creation date <= bufferedLowerEndTimestampInMillis.
          - The last element in this list needs to be added to the final list
         Second partition:
          - Indices with creation date > bufferedLowerEndTimestampInMillis
          - A further filter of this list to include only ones <= bufferedHigherEndTimestamp needs to be done

         NOTE: Partition logic is NOT applied for migrated indices (i.e. the ones with creation date part of index name itself).
         For migrated indices, all indices are considered, and no creation date filtering is applied.

          Example A: Indices are as follows, all normal indices with creation date part of index setting:
            a. audit-event-2020.07.10 -> 3:00 AM
            b. audit-event-2020.07.10 -> 4:00 AM
            c. audit-event-2020.07.10 -> 5:30 AM
            d. audit-event-2020.07.10 -> 6:00 AM
            e. audit-event-2020.07.10 -> 7:30 AM

          If data has to be loaded for 4:15 AM to 6:15 AM, return value is: 'b','c','d' and not 'a','e'

          First partition will contain "a" and "b" ==> the last will be "b"
          Second partition will contain "c", "d" and "e" ==> we will filter out ones > endTime i.e 6:15 which is "e" ==> So this will be "c" and "d"
          overall = List(b) and List(c, d)

          Example B: Indices are as follows, normal indices (creation date part of index setting), or migrated indices
          (creation date part of index name):
            a. audit-event-2020.07.10         -> 3:00 AM
            b. audit-event-2020.07.10         -> 4:00 AM
            c. audit-event-2020.07.10-4:10 AM -> 4:10 AM (Migrated index- true creation date part of index name)
            d. audit-event-2020.07.10         -> 5:30 AM
            e. audit-event-2020.07.10         -> 6:00 AM
            f. audit-event-2020.07.10-6:10 AM -> 6:10 AM (Migrated index- true creation date part of index name)
            g. audit-event-2020.07.10-6:30 AM -> 6:30 AM (Migrated index- true creation date part of index name)
            h. audit-event-2020.07.10         -> 7:30 AM

          If data has to be loaded for 4:15 AM to 6:15 AM, return value is: 'b','c','d', 'e', 'f', 'g' and not 'a', 'h'.
          For normal indices, First partition will contain last index viz "b".
          Second partition will filter out indices whose > endTime i.e 6:15 viz. "h".
          So final list is:
          Normal indices: List(b) + List(d,e) +
          Migrated indices: List(c) + List(f) + List(g)
          = List(b,c,d,e,f,g)
         */
    val indexNamePrefixLength: Int = indexBaseName.length + 1
    //Migrated indices
    val indexNameWithCreationDateLength: Int = indexNamePrefixLength + ElasticUtils.DatePattern.length + 1 + 13

    val indicesGroupedByCreationDateFlag = indexWithCreationDateTupleList.groupBy(indexAndCreationDate => {
      indexNameWithCreationDateLength == indexAndCreationDate.idx.length
    })

    indicesGroupedByCreationDateFlag.flatMap(
        indicesByCreationDateFlag => {
          val creationDateFlag = indicesByCreationDateFlag._1
          val indicesWithCreationDateTupleList = indicesByCreationDateFlag._2
          if (creationDateFlag) { //For migrated indices, return all the indices
            indicesWithCreationDateTupleList
              .sortBy(_.cd.toLong)
              .map(indexAndCreationDate => indexAndCreationDate.idx)
          } else { //For normal indices, apply creation date filtering
            val indicesBasedOnTimestamp: (List[IndexAndCreationDate], List[IndexAndCreationDate]) = indicesWithCreationDateTupleList
              .sortBy(_.cd.toLong) // Sort it by creation date timestamp
              .partition(_.cd.toLong <= bufferedLowerEndTimestampInMillis)

            // Lower end of the list needs to take the last index in the first partition and create it as a list.
            val lowerList = indicesBasedOnTimestamp
              ._1
              .lastOption
              .map(indexAndCreationDate => List(indexAndCreationDate.idx))
              .getOrElse(Nil)

            // Upper end of the list needs to take only indices whose creationDateTimestamp is not greater than the bufferedHigherEndTimestamp
            val upperList = indicesBasedOnTimestamp
              ._2
              .filterNot(_.cd.toLong > bufferedHigherEndTimestampInMillis) // Eliminate all indices created post the higher end timestamp
              .map(indexAndCreationDate => indexAndCreationDate.idx)

            // Create a combined list
            val finalList = lowerList ++ upperList
            finalList
          }
        }
    ).toList
  }

  /**
    * Restrict the number of timestamp/dateMath mode indices to perform search on using createdDate of the index.
    * Example: Find indices relevant for searching based on the given timestamps scoped for search
    * NOTE: TO BE USED ONLY FOR TIMESTAMP/DATEMATH MODE INDICES, AND NOT FOR NUMBER MODE.
    */
  def getListOfIndicesToSearch(elasticClient: ElasticClient,
                               indexBaseName: String,
                               gteTimestampInMillis: Long,
                               lteTimestampInMillis: Long,
                               indexBufferTimesOption: Option[IndexBufferTimes] = None): Try[List[String]] = {

    // Determine how much buffer to use for cat index selection, and final indices set.
    // A. Audit-events:
    // For audit indices, we need +-1 day buffer to account for case where yesterday's index may not have rolled over yet.
    // However for cold audit indices which get coalesced, the coalesced index at time T can contain data for T, T+1, T+2 indices.
    // Hence adding 2 days buffer for such cases.
    // B. Opsrate-metrics:
    // For opsrate indices, we need a buffer of at least 1 week (chosen 10 days here) to include last and this week indices
    // due to maxAge of 7 days.
    //
    // Final index set buffer is still 1 hour for ALL types of indices to exclude extra indices in final set.
    val isSearchStartInColdTier: Boolean =
    FiniteDuration(Instant.now().toEpochMilli - gteTimestampInMillis, TimeUnit.MILLISECONDS).compare(FieldConstants.ColdSearchDuration) >= 0

    val indexBufferTimes = indexBufferTimesOption match {
      case Some(_) => indexBufferTimesOption.get
      case None => if (indexBaseName == MappingTemplates.INDEX_AUDIT_EVENTS_KEY) {
        //for cold indices which get coalesced, the coalesced index at time T can contain data for T, T+1, T+2 indices.
        if (isSearchStartInColdTier) IndexBufferTimes(2.days, 1.hour) else IndexBufferTimes(1.days, 1.hour)
      }
      else if (indexBaseName == MappingTemplates.INDEX_OPSRATE_METRICS_KEY) IndexBufferTimes(10.days, 1.hour)
      else IndexBufferTimes(1.hour, 1.hour)
    }

    val utcTimeZone = TimeZone.getTimeZone("UTC")
    val fromDate = Calendar.getInstance(utcTimeZone)
    fromDate.setTimeInMillis(gteTimestampInMillis - indexBufferTimes.bufferForIndexSelection.toMillis)

    //For end date, 1 day buffer should be good enough, as current day's data might at the max spill over to only next day index in case of late arrival.
    val toDate = Calendar.getInstance(utcTimeZone)
    toDate.setTimeInMillis(lteTimestampInMillis + 1.days.toMillis)

    /*
      Index date format is based on daily rollover "yyyy.MM.dd"
       ex- audit-events-2020.01.27-000001
     */
    val indexDateFormat = new SimpleDateFormat("yyyy.MM.dd")
    indexDateFormat.setTimeZone(utcTimeZone)
    var indices: List[String] = List.empty
    while (!isSameDay(fromDate, toDate)) {
      // There could be multiple indices for same day (audit-events-2020.01.27-000001, audit-events-2020.01.27-000002 etc.
      // Hence using * after date prefix
      indices = indices ::: List(s"$indexBaseName-${indexDateFormat.format(fromDate.getTime)}-*")
      fromDate.add(Calendar.DATE, +1)
    }

    indices = indices ::: List(s"$indexBaseName-${indexDateFormat.format(fromDate.getTime)}-*") // include the toDate as well

    val indexWithCreationDateTupleListTry: Try[List[IndexAndCreationDate]] = computeRelevantIndicesAndCreationDate(
      indexBaseName,
      elasticClient,
      indices
    )
    indexWithCreationDateTupleListTry match {
      case Success(indexWithCreationDateTupleList) =>
        //For cold searches, creation date is not checked for index filtering, since certain cold indices can get coalesced
        //into a new index, resulting in wrong creation date in index settings. Hence directly return indices sorted by their
        //index names for cold searches.
        //This also applies to searches spanning warm and cold boundaries (like Last 100 to Last 70 days old searches)
        if (isSearchStartInColdTier) {
          val finalList = indexWithCreationDateTupleList
            .sortBy(_.idx)
            .map(_.idx)
          return Success(finalList)
        }

        // In order to take into account processing of late records, a buffer of 1 hour is subtracted/added from/to fromTime/toTime
        val bufferedLowerEndTimestampInMillis = gteTimestampInMillis - indexBufferTimes.bufferForFinalIndexSet.toMillis
        val bufferedHigherEndTimestampInMillis = lteTimestampInMillis + indexBufferTimes.bufferForFinalIndexSet.toMillis

        Success(
          filterRelevantIndicesForSearch(
            indexBaseName,
            indexWithCreationDateTupleList,
            bufferedLowerEndTimestampInMillis,
            bufferedHigherEndTimestampInMillis
          )
        )
      case Failure(ex) =>
        logger.error(s"Error when getting list of indices for $indexBaseName. Reason: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Restrict the number of indices to perform search on. Since the index retention for the base indices is 190 days,
    * the number of indices are quite high, though the need to search (search space) is limited from requirements
    * perspective.
    * - This method gives a list of indices given the base index name to search on.
    * As number of indices within a day can also be > 1 (+/- 12 hours rollover check),
    * it uses a wildcard at the end of the day specification.
    * - This method also adds a buffer day, if searchSpaceInDays is 3,
    * it would give a buffer of 1 day + 1 day considering rollover
    * Return: 0, -1, -2, -3, -4 day index names.
    * To fetch indices relevant for current day, searchSpaceInDays is 0. In this case, it returns 0 and -1 day indices.
    */
  def getRecentListOfIndicesToSearch(indexBaseName: String, searchSpaceInDays: Int, addOneDayBuffer: Boolean = true): List[String] = {

    val indexReadAlias = ElasticUtils.ReadIndexAlias(indexBaseName)
    val indexDateFmt = DateTimeFormatter.ofPattern("yyyy.MM.dd").withZone(ZoneId.of("UTC"))
    val latestEpochInstant = Instant.now.atZone(ZoneId.of("UTC"))
    val searchSpaceInDaysWithBuffer = if (addOneDayBuffer) searchSpaceInDays + 1 else searchSpaceInDays

    if (searchSpaceInDays < 0)
      List(indexReadAlias)
    else {
      // Add an extra day buffer over the searchSpaceInDays
      (0 to searchSpaceInDaysWithBuffer)
        .map(
          /*
           There could be multiple indices for same day (users-2020.01.27-000001, users-2020.01.27-000002 etc.
           Hence there is a * after date prefix
           */
          age => s"$indexBaseName-${latestEpochInstant.minus(age, ChronoUnit.DAYS).format(indexDateFmt)}-*"
        ).toList
    }
  }

  /**
    * @param indexName  Index read alias to be appended with rollover date pattern
    * @param timeFilter Applied time filter
    * @return Optionally returns list of indices within the given time range
    *
    *         Since, an index can have +/-12hours of documents, we are adding 1 day buffer in the beginning and end of time range
    *         This method uses days based logic for indices selection and this is used when ElasticIndexSelector.getListOfIndicesToSearch
    *         returns any failure.
    */
  def getSearchIndicesBasedOnDays(indexName: String, timeFilter: TimeFilter): List[String] = {

    val utcTimeZone = TimeZone.getTimeZone("UTC")
    val oneDayBufferInMillis = 1.days.toMillis

    val fromDate = Calendar.getInstance(utcTimeZone)
    fromDate.setTimeInMillis(timeFilter.fromTime - oneDayBufferInMillis)

    val toDate = Calendar.getInstance(utcTimeZone)
    toDate.setTimeInMillis(timeFilter.toTime + oneDayBufferInMillis)

    /*
      Index date format is based on daily rollover "yyyy.MM.dd"
       ex- audit-events-2020.01.27-000001
     */
    val indexDateFormat = new SimpleDateFormat("yyyy.MM.dd")
    indexDateFormat.setTimeZone(TimeZone.getTimeZone("UTC"))
    var indices: List[String] = List.empty
    while (!isSameDay(fromDate, toDate)) {
      // There could be multiple indices for same day (audit-events-2020.01.27-000001, audit-events-2020.01.27-000002 etc.
      // Hence using * after date prefix
      indices = indices ::: List(s"$indexName-${indexDateFormat.format(fromDate.getTime)}-*")
      fromDate.add(Calendar.DATE, +1)
    }
    indices = indices ::: List(s"$indexName-${indexDateFormat.format(fromDate.getTime)}-*") // include the toDate as well
    indices
  }

  private def isSameDay(calendar1: Calendar, calendar2: Calendar): Boolean = {
    // Checking for year, month and date as rollover date format contains the same
    (calendar1.get(Calendar.YEAR).equals(calendar2.get(Calendar.YEAR))
      && calendar1.get(Calendar.MONTH).equals(calendar2.get(Calendar.MONTH))
      && calendar1.get(Calendar.DATE).equals(calendar2.get(Calendar.DATE)))
  }
}
