/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic

import akka.actor.{Actor, ActorLogging, ActorRef, Props}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.misc.model._

import scala.util.Try

/**
  * ElasticInfraChecker is the actor that is responsible for checking connectivity to elastic cluster
  */
object ElasticInfraChecker {
  def props(elasticSettings: ElasticSettings,
            requestorRef: ActorRef): Props = Props(new ElasticInfraChecker(elasticSettings, requestorRef))

  final val ContextElasticInfraCheck: String = "elasticInfraCheck"
}

class ElasticInfraChecker(elasticSettings: ElasticSettings, requestorRef: ActorRef) extends Actor with ActorLogging {

  import ElasticInfraChecker._

  override def preStart(): Unit = {
    // Send a self message
    self ! CheckInfra
  }

  // Receive case for elastic check
  override def receive: Receive = LoggingReceive {

    // Message to check for elastic cluster connectivity
    case CheckInfra =>
      checkElasticConnection() match {
        case scala.util.Success(_) =>
          log.info(s"Elastic cluster connection check to ${elasticSettings.httpHostsProps.get.endpoints.mkString(",")} succeeded.")

          // Send action done to parent
          requestorRef ! ActionSucceeded(ContextElasticInfraCheck)

        // Kafka broker check failed
        case scala.util.Failure(ex) =>
          val endPoints = if (elasticSettings.httpHostsProps.isDefined) elasticSettings.httpHostsProps.get.endpoints.mkString(",") else ""
          val errorMessage = s"Elastic cluster connection check to $endPoints failed with message: ${ex.getMessage}"
          log.error(errorMessage)

          // Send action done to parent
          requestorRef ! ActionFailed(ContextElasticInfraCheck, Error(ErrorCode.ConnectionNotEstablished, errorMessage, endPoints))
      }
    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message that is not handled: " + somethingElse)
  }

  // Code to check for elastic connection, kept separate so that unit tests can override it
  protected def checkElasticConnection(): Try[Boolean] =
    ElasticUtils.checkConnection(elasticSettings)
}