/*
 * $Id$
 *
 *  Copyright (c) 2019 NetApp, Inc.
 *  All rights reserved.
 */

package com.netapp.df.uba.server.common.elastic

import com.netapp.df.uba.server.common.elastic.aggregations.CardinalityAggregationInfo
import com.netapp.df.uba.server.common.elastic.utils.FieldConstants
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.sksamuel.elastic4s.http.ElasticDsl._
import com.sksamuel.elastic4s.http.search.{MultiSearchResponse, SearchResponse}
import com.sksamuel.elastic4s.http.update.UpdateResponse
import com.sksamuel.elastic4s.http.{ElasticClient, Response}
import com.sksamuel.elastic4s.searches.collapse.CollapseRequest
import com.sksamuel.elastic4s.searches.queries.Query
import com.sksamuel.elastic4s.searches.sort.FieldSort
import com.sksamuel.elastic4s.searches.{MultiSearchRequest, SearchRequest}
import com.sksamuel.elastic4s.update.UpdateRequest

import scala.concurrent.duration._
import scala.util.{Failure, Success, Try}

/**
  * ElasticSearch Handler handles the following:
  * - Actual invocation of search request on elastic cluster
  * - Building a bool
  */
case object ElasticSearchHandler extends LogTrait with ElasticSearchHandlerTrait {

  //Used  in elasticClient.execute call
  implicit val TimeOutDuration: FiniteDuration = 100.seconds

  /**
    * Use this method when collapse parameter is to be applied on a search query (which is usually done to avoid duplicate records)
    * This allows to collapse search results based on field values.This is applied to avoid loading duplicate records.
    * The collapsing is done by selecting only the top sorted document collapse key ("id" field)
    */
  def createBasicSearchRequestWithCollapse(searchQuery: Query,
                                           indexReadAlias: String,
                                           tenantId: String,
                                           sourceIncludeList: List[String] = List.empty,
                                           sourceExcludeList: List[String] = List.empty,
                                           indicesOption: Option[List[String]] = None): SearchRequest = {
    createBasicSearchRequest(searchQuery, indexReadAlias, tenantId, sourceIncludeList, sourceExcludeList, indicesOption)
      .collapse(CollapseRequest(FieldConstants.IdFieldName))
  }

  /**
    * Use this method to create a search request with tenant id filter
    * Note: Whenever using sourceIncludeList/sourceExcludeList the object json formatting will fail.
    * Deal with the Source hits directly.
    */
  def createBasicSearchRequest(searchQuery: Query,
                               indexReadAlias: String,
                               tenantId: String,
                               sourceIncludeList: List[String] = List.empty,
                               sourceExcludeList: List[String] = List.empty,
                               indicesOption: Option[List[String]] = None): SearchRequest = {
    // Note: Adding clientCoordinates field of audit-events index to exclusion list to avoid parsing errors observed after upgrading to ES v6.8.23.
    // This can be removed if we decide to drop this field in future.
    val fieldsToExcludeFromSource: List[String] = if (indexReadAlias.equals(MappingTemplates.INDEX_AUDIT_EVENTS_KEY))
      sourceExcludeList :+ "clientCoordinates"
    else sourceExcludeList

    val searchRequest = search(indicesOption.getOrElse(List(indexReadAlias))) // Note: indicesOption should contain valid indices or contain wildcard * (asterisk)
      .sourceInclude(sourceIncludeList)
      .sourceExclude(fieldsToExcludeFromSource)
      .query(
        boolQuery()
          .filter(
            searchQuery,
            //Tenant Id filter
            termQuery(FieldConstants.TenantIdFieldName, tenantId)
          ))
    searchRequest
  }

  /**
    * Builds search query with tenant id filter and optional collapse to avoid duplicate object loading
    * At any cost, data of one tenant cannot be mixed with another.
    * So, keeping this method close to base/infra layer so that it is not missed.
    */
  def buildPaginationSearchRequest(searchQuery: com.sksamuel.elastic4s.searches.queries.Query,
                                   tenantId: String,
                                   offset: Int,
                                   limit: Int,
                                   sortFields: List[FieldSort],
                                   indexName: String,
                                   collapseWithCardinalityNeeded: Boolean,
                                   includeVersion: Boolean,
                                   indicesOption: Option[List[String]] = None,
                                   trackTotalHits: Boolean = true): SearchRequest = {
    if (collapseWithCardinalityNeeded) {
      /*
        Adding cardinality aggregation to get distinct document count for indices that need it.
        https://www.elastic.co/guide/en/elasticsearch/reference/current/search-aggregations-metrics-cardinality-aggregation.html
        */
      createBasicSearchRequestWithCollapse(searchQuery, indexName, tenantId, indicesOption = indicesOption)
        .from(offset)
        .size(limit)
        .sortBy(sortFields)
        .version(includeVersion)
        .aggs(CardinalityAggregationInfo(FieldConstants.IdFieldName).buildAggregationSnippet())
        .trackTotalHits(trackTotalHits)
    } else {
      createBasicSearchRequest(searchQuery, indexName, tenantId, indicesOption = indicesOption)
        .from(offset)
        .size(limit)
        .sortBy(sortFields)
        .version(includeVersion)
        .trackTotalHits(trackTotalHits)
    }
  }

  // Runs an actual search request in elastic and returns the raw search response
  def runQuery(tenantId: String, context: String, elasticClient: ElasticClient, searchRequest: SearchRequest): Try[SearchResponse] = {

    val responseTry: Try[Response[SearchResponse]] = Try(elasticClient.execute(searchRequest).await)

    responseTry match {
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed to execute the search request", ex, detail = true)
        Failure(ex)
      case Success(searchResponse) =>
        // Log and try to get results
        ElasticUtils.logAndGetResult(s"Tenant: $tenantId; $context. Search request : ${searchRequest.show}", searchResponse)
    }
  }

  // Runs multi search request in elastic and returns the raw search response
  def runQuery(tenantId: String, context: String, elasticClient: ElasticClient, multiSearchRequest: MultiSearchRequest): Try[MultiSearchResponse] = {

    val multiSearchResponseTry: Try[Response[MultiSearchResponse]] = Try(elasticClient.execute(multiSearchRequest).await)

    multiSearchResponseTry match {
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed to execute the Multi search request", ex, detail = true)
        Failure(ex)

      case Success(multiSearchResponse) =>
        // Log and try to get results
        ElasticUtils.logAndGetResult(s"Tenant: $tenantId; $context. Multi search request : ${multiSearchRequest.show}", multiSearchResponse)
    }
  }

  // Runs update request in elastic and returns the raw search response
  def runQuery(tenantId: String, context: String, elasticClient: ElasticClient, updateRequest: UpdateRequest): Try[UpdateResponse] = {

    val updateResponseTry: Try[Response[UpdateResponse]] = Try(elasticClient.execute(updateRequest).await)
    updateResponseTry match {
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed to execute the Update request", ex, detail = true)
        Failure(ex)
      case Success(updateResponse) =>
        // Log and try to get results
        ElasticUtils.logAndGetResult(s"Tenant: $tenantId; $context. Update request : ${updateRequest.show}", updateResponse)
    }
  }
}

/**
  * Trait defined for Unit testing.
  * Otherwise it is difficult to test singleton object
  */
trait ElasticSearchHandlerTrait {
  def runQuery(tenantId: String, context: String, elasticClient: ElasticClient, multiSearchRequest: MultiSearchRequest): Try[MultiSearchResponse]
}
