/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic

import com.netapp.df.uba.server.common.elastic.PruneConfigMode.PruneConfigMode
import com.netapp.df.uba.server.common.elastic.RollOverMode.RollOverMode
import com.netapp.df.uba.server.common.elastic.utils.IndexRouteSettings
import com.netapp.df.uba.server.common.misc.utils.SettingUtils
import com.sksamuel.elastic4s.http.ElasticProperties
import com.typesafe.config.{Config, ConfigException}

import scala.concurrent.duration._

/**
  * A marker trait for case classes that represent the indices
  * to which data is written or read from in a micro-service
  */
trait IndexTrait {
  def baseName: String
}

/**
  * A case class for indices that are read in a micro-service
  */
object ReadIndexConfig {
  def apply(config: Config): ReadIndexConfig = ReadIndexConfig(SettingUtils.mandatoryConfigValue(config, "base-name"))
}

case class ReadIndexConfig(baseName: String) extends IndexTrait

/**
  * Contains all the settings for an index.
  * Includes the base name, shards, replicas and the prune and rollover configuration.
  * Note: Rollover and pruning is not mandatory
  */
object WriteIndexConfig {
  def apply(config: Config): WriteIndexConfig = {

    val baseName = SettingUtils.mandatoryConfigValue(config, "base-name")

    val rolloverConfig = if (config.hasPath("rollover")) Some(RolloverConfig(config.getConfig("rollover")))
    else None

    // Determine which prune config option is used for data in the index, index level deletion or document level deletion
    val pruneConfig = if (config.hasPath("prune")) {
      val pruneConfigMode = PruneConfigMode.withName(config.getString("prune.mode"))
      pruneConfigMode match {
        case PruneConfigMode.Index => Some(PruneIndexConfig(config.getConfig("prune")))
        case PruneConfigMode.Document => Some(PruneDocumentConfig(config.getConfig("prune")))
      }
    } else None

    val baseAsIndexName: String = SettingUtils.configValueWithEmptyAsNone(config, "base-as-indexname").getOrElse("false")
    val baseAsIndexBool = try {
      baseAsIndexName.toBoolean
    } catch {
      case ex: IllegalArgumentException => throw new ConfigException.BadValue("base-as-indexname", ex.getMessage)
    }

    val indexRoutingSettings = IndexRouteSettings(config.getConfig("routing"))

    new WriteIndexConfig(
      baseName,
      SettingUtils.validMinInt(config, "shards", 1),
      SettingUtils.validMinInt(config, "replicas", 1),
      SettingUtils.validMinInt(config, "max-result-window", 10000),
      SettingUtils.configValueWithEmptyAsNone(config, "auto-expand-replicas"),
      rolloverConfig,
      pruneConfig,
      baseAsIndexBool,
      indexRoutingSettings
    )
  }
}

/**
  * A case class for indices that are read in a micro-service, but need not be checked for existence
  */
object InfoIndexConfig {
  def apply(config: Config): InfoIndexConfig = InfoIndexConfig(SettingUtils.mandatoryConfigValue(config, "base-name"))
}

case class InfoIndexConfig(baseName: String) extends IndexTrait


case class WriteIndexConfig(baseName: String,
                            shards: Int,
                            replicas: Int,
                            maxResultsWindowSize: Int,
                            autoExpandReplicas: Option[String],
                            rolloverConfigOption: Option[RolloverConfig],
                            pruneConfigOption: Option[PruneConfigTrait],
                            useBaseNameAsIndexName: Boolean,
                            indexRoutingSettings: IndexRouteSettings
                           ) extends IndexTrait

/**
  * Rollover config contains the configuration for rolling over the index
  * based on configuration set for the index.
  * maxAge - Maintain an index for a maximum of this time (1d, 5d etc)
  * maxSize - Rollover to a new index if the size has crossed this number
  * maxDocs - Rollover to a new index if the number of documents in the index has crossed this number.
  * mode - Defines the roll over index name pattern. It can have two values 'number' or 'dateMath'
  * if mode is number => roll over index name will be logs-000001
  * if mode is dateMath => roll over index name will be logs-{now/d}-000001
  *
  * Not all options are required, but atleast one is required if rollover is being configured.
  */
object RolloverConfig {
  def apply(config: Config): RolloverConfig = {
    val maxAge = if (config.hasPath("max-age")) Some(config.getString("max-age")) else None
    val maxSize = if (config.hasPath("max-size")) Some(config.getString("max-size")) else None
    val maxDocs = if (config.hasPath("max-docs")) Some(config.getLong("max-docs")) else None
    val mode = if (config.hasPath("mode")) Some(config.getString("mode")) else None

    if (maxAge.isEmpty && maxSize.isEmpty && maxDocs.isEmpty)
      throw new ConfigException.BadValue("rollover-config", "At least one of max age, size or docs should be configured.")

    // Frequency of checking for rollover conditions.
    val checkInterval = SettingUtils.validMinDuration(config, "check-interval", 10.minutes)

    // Checking mode configuration
    val rollOverMode = validateRollOverMode(mode)

    RolloverConfig(maxAge, maxSize, maxDocs, checkInterval, rollOverMode)
  }

  def validateRollOverMode(modeConfig: Option[String]): RollOverMode = {
    modeConfig match {
      case Some(value) =>
        if (value.equals("number") | value.equals("dateMath")) {
          RollOverMode.withName(value)
        } else {
          throw new ConfigException.BadValue("rollover-config", "Wrong configuration for mode. mode can be either 'number' or 'dateMath' ")
        }
      case None =>
        throw new ConfigException.BadValue("rollover-config", "mode should be configured.")
    }
  }
}

case class RolloverConfig(maxAge: Option[String], maxSize: Option[String], maxDocs: Option[Long],
                          checkInterval: FiniteDuration,
                          mode: RollOverMode)

object RollOverMode extends Enumeration {
  type RollOverMode = Value

  val Number: RollOverMode = Value("number")
  val DateMath: RollOverMode = Value("dateMath")
}

object PruneConfigMode extends Enumeration {
  type PruneConfigMode = Value

  val Index: PruneConfigMode = Value("index")
  val Document: PruneConfigMode = Value("document")
}

/**
  * A marker trait
  */
trait PruneConfigTrait {
  def mode: PruneConfigMode

  def maxRetention: FiniteDuration

  def checkInterval: FiniteDuration
}

/**
  * PruneIndexConfig contains the settings for pruning the indices of a specific index base.
  * maxRetentionDays: is for how many days an index should be retained
  * checkInterval: is about how frequently it should be checked if there are indices that have to be deleted.
  */
object PruneIndexConfig {
  def apply(config: Config): PruneIndexConfig = PruneIndexConfig(
    PruneConfigMode.withName(config.getString("mode")),
    SettingUtils.validMinDuration(config, "max-retention", 2.hours), // Small number here to handle testing scenarios
    SettingUtils.validMinDuration(config, "check-interval", 1.hour) // Frequency of checking for pruning condition
  )
}

case class PruneIndexConfig(mode: PruneConfigMode, maxRetention: FiniteDuration, checkInterval: FiniteDuration) extends PruneConfigTrait

/**
  * PruneDocumentConfig contains the settings for deleting documents from indices.
  * Indices that need document deletion based on age of the document decided
  * by some field in the document can use this.
  * maxRetention: Retention interval for the document
  * timestampFieldName: Timestamp field in the document that determines the age
  * checkInterval: is about how frequently it should be checked if there are indices that have to be deleted.
  */
object PruneDocumentConfig {
  def apply(config: Config): PruneDocumentConfig = PruneDocumentConfig(
    PruneConfigMode.withName(config.getString("mode")),
    SettingUtils.validMinDuration(config, "max-retention", 2.hours),
    config.getString("timestamp-field-name"),
    SettingUtils.validMinDuration(config, "check-interval", 1.hour)
  )
}

case class PruneDocumentConfig(mode: PruneConfigMode,
                               maxRetention: FiniteDuration,
                               timestampFieldName: String,
                               checkInterval: FiniteDuration) extends PruneConfigTrait

/**
 * Settings specific on connecting to elastic server and reading/writing to indices
 * Contains a set of indices which will be either read or written to by the micro-service
 * httpHosts should be of the format:
 * http(s)://host:port,host:port(/prefix)?querystring
 */
object ElasticSettings {

  def apply(config: Config,
            writeIndices: List[String] = Nil,
            readIndices: List[String] = Nil,
            infoIndices: List[String] = Nil): ElasticSettings = {

    // Construct the ElasticProperties out of it. Hosts is optional since its usually derived automatically from tenant cache.
    val httpHostsProps = if (config.hasPath(s"http.hosts")) {
      // Get the http hosts string as a mandatory config value, missing exception thrown
      val httpHostsRawString = SettingUtils.mandatoryConfigValue(config, s"http.hosts").trim
      try {
        Some(ElasticProperties(httpHostsRawString))
      } catch {
        case ex: Exception =>
          throw new ConfigException.BadValue("http.hosts", ex.getMessage)
      }
    } else None

    val retryTimeout = SettingUtils.validMinDuration(config, s"retry.timeout", 5.seconds)

    // For each of the index to which the micro-service is writing, create the IndexConfig with the full configuration
    val writerIndexMap: Map[String, IndexTrait] = if (config.hasPath("index.write")) {
      // Get the main write index configuration
      val writeIndexConfigs = config.getConfig("index.write")

      writeIndices
        .map(writeIndexKey => writeIndexKey -> WriteIndexConfig(writeIndexConfigs.getConfig(writeIndexKey)))
        .toMap
    } else Map.empty

    // For each of the index to which the micro-service is reading from, create the ReadIndexConfig with the name
    val readIndexMap: Map[String, IndexTrait] = if (config.hasPath("index.read")) {
      // Get the main read index configuration
      val readIndexConfigs = config.getConfig("index.read")

      readIndices
        .map(readIndexKey => readIndexKey -> ReadIndexConfig(readIndexConfigs.getConfig(readIndexKey)))
        .toMap
    } else Map.empty

    // For each of the index to which the module needs info, create InfoIndexConfig
    val infoIndexMap: Map[String, IndexTrait] = if (config.hasPath("index.info")) {
      // Get the main info index configuration
      val infoIndexConfigs = config.getConfig("index.info")

      infoIndices
        .map(infoIndexKey => infoIndexKey -> InfoIndexConfig(infoIndexConfigs.getConfig(infoIndexKey)))
        .toMap
    } else Map.empty

    // Full map is a union of the two
    val indexMap: Map[String, IndexTrait] = writerIndexMap ++ readIndexMap ++ infoIndexMap

    new ElasticSettings(
      httpHostsProps,
      retryTimeout,
      indexMap)
  }
}

case class ElasticSettings(httpHostsProps: Option[ElasticProperties],
                           retryTimeout: FiniteDuration,
                           indexMap: Map[String, IndexTrait])

