/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic

import com.netapp.df.uba.server.common.elastic.model.ElasticClientInfo
import com.netapp.df.uba.server.common.misc.api.JsonFormats
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model.{ActionFailed, ActionStatus, ActionSucceeded, Error, ErrorCode}
import com.sksamuel.elastic4s.http.ElasticDsl._
import com.sksamuel.elastic4s.http._
import com.sksamuel.elastic4s.http.bulk.BulkResponse
import com.sksamuel.elastic4s.http.cluster.ClusterHealthResponse
import com.sksamuel.elastic4s.http.count.CountResponse
import com.sksamuel.elastic4s.http.delete.{DeleteByQueryResponse, DeleteResponse}
import com.sksamuel.elastic4s.http.index._
import com.sksamuel.elastic4s.http.index.admin.{AliasActionResponse, DeleteIndexResponse}
import com.sksamuel.elastic4s.http.search.{ClearScrollResponse, MultiSearchResponse, SearchHit, SearchResponse}
import com.sksamuel.elastic4s.http.update.{UpdateByQueryResponse, UpdateResponse}
import com.sksamuel.elastic4s.script.Script
import com.sksamuel.elastic4s.searches.SearchRequest
import com.sksamuel.elastic4s.searches.sort.FieldSort

import java.text.SimpleDateFormat
import scala.collection.mutable.ListBuffer
import scala.concurrent.duration._
import scala.concurrent.{Future, Promise}
import scala.util.{Failure, Success, Try}

// A message to send to the actor to start the index management
case object ManageIndices

case class CsElasticClient(client: ElasticClient) extends LogTrait {
  def close(): Unit = {
    try {
      client.close()
    } catch {
      case ex: Exception =>
        logger.warn(s"Failed to close the elastic client. Reason: ${ex.getMessage}")
    }
  }
}

/**
  * A custom exception to be used for elastic issues
  */
case class ElasticIssue(httpStatusCode: Int, message: String) extends Exception(message)

/**
  * ElasticUtils is a utility class to support
  * logging the responses for various elasticsearch API calls, connection checks
  */
object ElasticUtils extends LogTrait with JsonFormats {

  final val DEFAULT_SOCKET_TIMEOUT_MILLIS = 75000

  // Context used for action on indices
  final val ContextIndexManagement: String = "indices"

  final val OneHourInMillis: Long = 3600000L
  final val OneDayInMillis: Long = 86400000L

  /*
    Lock namespace for index, note: this is appended to /oci/locks used in DistributedSynchronousTask taken from OCI_ commons
    Defining it here as it is to be used in index creator, index rollover and pruner.
  */
  final val IndexLockNamespace = "index/"

  // Lock name used to synchrnoize the index management across micro-services includes a namespace of index/
  def lockNameForIndex(index: String) = s"$IndexLockNamespace$index"

  // Used across the service, some constants
  final val DatePattern: String = "YYYY-MM-dd"
  final val DateFormatter = new SimpleDateFormat(DatePattern)
  final val LatestSuffix: String = "latest"

  // Latest index name with base
  def LatestIndexAlias(indexBaseName: String): String = s"$indexBaseName-$LatestSuffix"

  // Read index name alias
  def ReadIndexAlias(indexBaseName: String): String = indexBaseName

  /*
    Set of fields in sequences index that is used for auto-incrementing
   */
  private val SequenceFields: Set[String] = Set("nextCommunityId", "nextAlertNumber")

  def createElasticConnectionString(protocol: String, hosts: Seq[String], port: Int): String = {
    val hostsString = hosts
      .map(hostName => s"$hostName:$port")
      .mkString(",")

    //The format of the url (expected by elastic4s client) is: http(s)://host:port,host:port(/prefix)?querystring
    s"$protocol://$hostsString"
  }

  /**
    * Connection checker for elastic. Uses ElasticDsl to make a call _cluster/health API
    */
  def checkConnection(elasticSettings: ElasticSettings): Try[Boolean] = {
    try {
      if (elasticSettings.httpHostsProps.isEmpty) throw new IllegalArgumentException("Missing http hosts to create elastic client")

      // Create the elastic client
      val elasticClient = ElasticClient(elasticSettings.httpHostsProps.get)

      // Invoke _cluster/health API
      val apiResponse = elasticClient.execute {
        clusterHealth()
      }.await

      // Log and get the unwrapped result as a Try object
      val parsedResult = logAndGetResult("CheckConnectivity", apiResponse)

      // Close the elastic client
      elasticClient.close()

      parsedResult match {
        // If the cluster health is red, mark failure
        case Success(clusterHealth) =>
          if (clusterHealth.status == "red") Failure(new Exception("Cluster health reports red."))
          else Success(true)

        case Failure(ex) => Failure(ex)
      }
    } catch {
      case ex: Exception =>
        Failure(ex)
    }
  }

  /**
    * Utility to see if the response from Elastic is a error or a success
    * and log appropriately. In case of success, the result object is returned as a Some(object)
    * In case of error, the issue is returned as Failure(ElasticIssue)
    */
  def logAndGetResult[U](context: String, result: com.sksamuel.elastic4s.http.Response[U]): Try[U] = {
    if (result.isSuccess) {
      logResponseSuccess(context, result.result)
      Success(result.result)

    } else {
      val errorMessage = s"Received failure for the elastic operation $context - Status: ${result.status} Type: ${result.error.`type`} Reason: ${result.error.reason}"
      logger.error(errorMessage)
      logger.error(result.error.rootCause.map(_.reason).mkString("[", ", ", "]"))
      Failure(ElasticIssue(result.status, errorMessage))
    }
  }

  /**
    * Utility method to check existence of indices
    * Expected to be called by micro-services as part of initialization if they depend on data in those indices
    * UT as part of IndexChecker
    */
  def checkReadIndices(csElasticClientInfoTry: Try[ElasticClientInfo], readIndices: Map[String, ReadIndexConfig]): ActionStatus = {
    // If there are no read indices to be checked for, mark Success
    if (readIndices.isEmpty) ActionSucceeded(ContextIndexManagement)
    else {
      csElasticClientInfoTry match {
        case Failure(ex) =>
          // Failed to get the elastic client itself
          val errorObject = Error(ErrorCode.InternalError, s"Index existence check failed. Reason: ${ex.getMessage}", ContextIndexManagement)
          ActionFailed(ContextIndexManagement, errorObject)

        case Success(csElasticClientInfo) =>

          val failedErrorObjects: List[Error] = readIndices
            .values
            .map(_.baseName)
            .map(indexName => indexName -> checkIndex(csElasticClientInfo.client.client, indexName))
            .filterNot(indexToResultEntry => indexToResultEntry._2.isSuccess) // Rest of the logic requires failed indices only.
            .map(entry => (entry._1, entry._2.asInstanceOf[Failure[Boolean]]))
            .map(entry => Error(ErrorCode.InternalError, s"Index check failed for ${entry._1}. Reason: ${entry._2.exception.getMessage}", entry._1))
            .toList

          if (failedErrorObjects.isEmpty) ActionSucceeded(ContextIndexManagement)
          else {
            // Send the action failed event back to the micro-service
            ActionFailed(ContextIndexManagement,
              Error(ErrorCode.InternalError, "Failed in checking indices", ContextIndexManagement, Some(failedErrorObjects)))
          }
      }
    }
  }

  /**
    * Private method used by checkReadIndices
    * UT as part of IndexChecker
    */
  private def checkIndex(elasticClient: ElasticClient, indexBaseName: String): Try[Boolean] = {

    try {
      val apiResponse = elasticClient.execute {
        /*
         Using a regular expression here.
         In that case, an empty map is returned when no indices match
          */
        getIndex(s"$indexBaseName-*")
      }.await

      val parsedResultTry = ElasticUtils.logAndGetResult(s"IndexChecker for $indexBaseName", apiResponse)
      parsedResultTry match {
        case Success(indexMap) =>
          if (indexMap.isEmpty) Failure(new Exception(s"No indices with the base name as $indexBaseName."))
          else Success(true)

        case Failure(ex) =>
          Failure(ex)
      }
    } catch {
      case ex: Exception =>
        logger.error(s"Exception when checking index $indexBaseName. Reason: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Utility to log the result of an elastic REST API call which was successful.
    * Logs some important points from the response
    */
  private def logResponseSuccess(context: String, result: Any): Unit =

    result match {
      case clusterHealthResponse: ClusterHealthResponse =>
        logger.info(s"$context - Cluster health response - name: ${clusterHealthResponse.clusterName}, status: ${clusterHealthResponse.status}")
        if (clusterHealthResponse.status == "red") logger.error(s"$clusterHealthResponse")

      case getIndexTemplates: GetIndexTemplates =>
        logger.info(s"$context - GetIndexTemplate response - size: ${getIndexTemplates.templates.size}")

        getIndexTemplates
          .templates
          .foreach(entry => logger.debug(s"$context - Name: ${entry._1} Template: ${entry._2.toString}"))

      case createIndexTemplateResponse: CreateIndexTemplateResponse =>
        logger.info(s"$context - CreateIndexTemplate response - acknowledged: ${createIndexTemplateResponse.acknowledged}")

      case getIndexResponseMap: Map[String, GetIndexResponse] =>
        logger.info(s"$context - GetIndex map response - size: ${getIndexResponseMap.size}, keys: ${getIndexResponseMap.keys.mkString(", ")}")

      case indexResponse: IndexResponse =>
        logger.info(s"$context - GetIndex response index: ${indexResponse.index}, shards: ${indexResponse.shards}, version: ${indexResponse.version}, result: ${indexResponse.result}")

      case createIndexResponse: CreateIndexResponse =>
        logger.info(s"$context - CreateIndex response - acknowledged: ${createIndexResponse.acknowledged}, shards_acknowledged: ${createIndexResponse.shards_acknowledged}")

      case rolloverResponse: RolloverResponse =>
        logger.info(s"$context - Rollover response - rolledOver: ${rolloverResponse.rolledOver}, newIndex: ${rolloverResponse.newIndex}")

      case deleteIndexResponse: DeleteIndexResponse =>
        logger.info(s"$context - DeleteIndex response - acknowledged: ${deleteIndexResponse.acknowledged}")

      case deleteResponse: DeleteResponse =>
        logger.info(s"$context - Delete response - index: ${deleteResponse.index}, id: ${deleteResponse.id}, result: ${deleteResponse.result}")

      case updateResponse: UpdateResponse =>
        logger.info(s"$context - Update response - result: ${updateResponse.result}, found: ${updateResponse.found}, source: ${updateResponse.source}")

      case updateByQueryResponse: UpdateByQueryResponse =>
        logger.info(s"$context - UpdateByQuery response - total: ${updateByQueryResponse.total}, updated: ${updateByQueryResponse.updated}, " +
          s"versionConflicts: ${updateByQueryResponse.versionConflicts}, noops: ${updateByQueryResponse.noops}, took(ms): ${updateByQueryResponse.took}")

      case deleteByQueryResponse: DeleteByQueryResponse =>
        logger.info(s"$context - DeleteByQuery response - total: ${deleteByQueryResponse.total}, deleted: ${deleteByQueryResponse.deleted}, " +
          s"versionConflicts: ${deleteByQueryResponse.versionConflicts}, noops - ${deleteByQueryResponse.noops}, took(ms): ${deleteByQueryResponse.took}")

      case searchResponse: SearchResponse =>
        logger.info(s"$context - SearchRequest response - totalHits: ${searchResponse.totalHits}, timedOut: ${searchResponse.isTimedOut}, terminatedEarly: ${searchResponse.isTerminatedEarly}, idsSize: ${searchResponse.ids.size}, " +
          s"shards: total(${searchResponse.shards.total}), failed(${searchResponse.shards.failed}), successful(${searchResponse.shards.successful}), took(ms): ${searchResponse.took}")
        logger.debug(s"$context - SearchRequest response - totalHits: ${searchResponse.totalHits}, timedOut: ${searchResponse.isTimedOut}, terminatedEarly: ${searchResponse.isTerminatedEarly}, idsSize: ${searchResponse.ids.size}, took(ms): ${searchResponse.took}, ids: ${searchResponse.ids.mkString(",")}")

      case clearScrollResponse: ClearScrollResponse =>
        logger.info(s"$context - ClearScroll response - succeeded: ${clearScrollResponse.succeeded}, num_freed: ${clearScrollResponse.num_freed}.")

      case bulkResponse: BulkResponse =>
        logger.info(s"$context - BulkResponse - items.size: ${bulkResponse.items.size}, errors: ${bulkResponse.errors}, failures.size: ${bulkResponse.failures.size} took(ms): ${bulkResponse.took}")

      case countResponse: CountResponse =>
        logger.info(s"$context - CountResponse - total count: ${countResponse.count}")

      case multiSearchResponse: MultiSearchResponse =>
        val isTimedOut = if (multiSearchResponse.failures.isEmpty) multiSearchResponse.successes.exists(searchResp => searchResp.isTimedOut) else false
        val isTerminatedEarly = if (multiSearchResponse.failures.isEmpty) multiSearchResponse.successes.exists(searchResp => searchResp.isTerminatedEarly) else false
        val timeTakenMs = if (multiSearchResponse.successes.nonEmpty) multiSearchResponse.successes.maxBy(resp => resp.took).took else 0L
        logger.info(s"$context - MultiSearchResponse - items.size: ${multiSearchResponse.size}, timedOut: $isTimedOut, terminatedEarly: $isTerminatedEarly, errors: ${multiSearchResponse.failures.map(searchError => s"Index- ${searchError.index}; Type- ${searchError.`type`}; Reason- ${searchError.reason}").mkString("\n")}, failures.size: ${multiSearchResponse.failures.size}, took(ms): $timeTakenMs")

      case aliasActionResponse: AliasActionResponse =>
        logger.info(s"$context - AliasActionResponse - acknowledged: ${aliasActionResponse.acknowledged}")

      case somethingElse =>
        logger.warn(s"Received something else $somethingElse in logResponseSuccess.")
    }

  /**
    * Generic code to get the next number in a sequence using elastic.
    * Reference documentation used for implementation:
    * https://www.lovelysystems.com/blog/es-iid-generator/
    */
  def getNextIdInSequence(elasticClient: ElasticClient, indexName: String, documentId: String, fieldName: String, bulkSize: Long = 1): Try[Long] = {

    // Try creating some values in case of sequences using scripted approach
    val scriptObj = Script(s"ctx._source.$fieldName += params.bulk_size")
      .lang("painless")
      .param("bulk_size", bulkSize)

    /*
      If this call is for a specific sequence field, then:
      - That specific field will have the upsert value as 1
      - The rest of the sequence fields will have an upsert value as 0
     */
    val upsertOtherFields = SequenceFields
      .filter(_ != fieldName)
      .map(_ -> 0L)
      .toMap

    val upsertAllFields = upsertOtherFields ++ Map(fieldName -> 1L)

    try {
      val apiResponse = elasticClient.execute {
        updateById(indexName, MappingTemplates.IndexTypeName, documentId)
          .script(scriptObj)
          .upsert(upsertAllFields)
          .retryOnConflict(5)
          .fetchSource(true)
      }.await

      val updateSeqResultTry = ElasticUtils.logAndGetResult(s"GetNextId in sequence for document: $documentId, field: $fieldName, index: $indexName", apiResponse)
      updateSeqResultTry match {
        case Success(updateSeqResult) =>
          val sequenceValue = updateSeqResult.source(fieldName)
          Success(sequenceValue.toString.toLong)

        case Failure(ex) =>
          // Already logged
          Failure(ex)
      }
    } catch {
      case ex: Exception =>
        logErrorT(documentId, s"Exception in getNextIdInSequence for index $indexName and documentId: $documentId and fieldName: $fieldName. Reason: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Utility method to get a list of all search hits given a search query.
    * This should be used for scenarios where it is known to have limited number of results
    * and not for iterative forensic APIs
    */
  def searchAllHits(elasticClient: ElasticClient,
                    searchRequest: SearchRequest,
                    searchContext: String,
                    maxRecordsInIteration: Int = 100,
                    version: Boolean = false,
                    awaitDuration: Duration = 60.seconds): Try[List[SearchHit]] = {
    // Initialize the search hits
    var searchHits: List[SearchHit] = Nil

    // Maintain a counter on the hits traversed till now
    var hitsTraversed: Long = 0

    /*
     A variable to control the loop,
     Success(true) means continue iteration,
     Success(false) means stop the loop but return search hits to the caller
     Failure(ex) means there is an issue in executing the search query
     */
    var continueLoop: Try[Boolean] = Success(true)

    /*
      Control of from and offset
     */
    var fromCounter: Int = 0

    // Run the loop
    while (continueLoop.isSuccess && continueLoop.get) {

      try {
        // Execute the given search query along with controls on limit, from and size
        val apiResponse = elasticClient.execute {
          searchRequest
            .limit(maxRecordsInIteration)
            .from(fromCounter)
            .size(maxRecordsInIteration)
            .version(version)
        }.await(awaitDuration)

        val apiResultTry = ElasticUtils.logAndGetResult(searchContext, apiResponse)
        apiResultTry match {

          case Success(apiResult) =>
            // No need to continue the loop, but its a success
            if (apiResult.totalHits == 0) continueLoop = Success(false)

            else {
              // Get the hits in this result
              val hitsInThisResult: List[SearchHit] = apiResult.hits.hits.toList
              val hitsInThisResultSize: Integer = hitsInThisResult.size

              // Append to the overall search hits result
              searchHits = searchHits ++ hitsInThisResult

              // Update the counter on traversed number of hits till now
              hitsTraversed = hitsTraversed + hitsInThisResultSize

              // If all the hits are traversed, OR if we get less than required number of hits in this iteration,
              // mark the loop as done as search is successful and complete
              if ((hitsTraversed >= apiResult.totalHits) || (hitsInThisResultSize < maxRecordsInIteration)) {
                continueLoop = Success(false)
              } else {
                // Continue the loop, increase the fromCounter
                fromCounter = fromCounter + hitsInThisResultSize
              }
            }
          case Failure(ex) =>
            // Failure, so return immediately.
            continueLoop = Failure(ex)
        }
      } catch {
        case ex: Exception =>
          logger.error(s"Exception in searchAllHits for context $searchContext. Reason: ${ex.getMessage}")
          continueLoop = Failure(ex)
      }
    }

    // Return the right results
    continueLoop match {
      case Success(_) => Success(searchHits)
      case Failure(ex) => Failure(ex)
    }
  }

  /**
    * Scrolled search version to get a list of all search hits given a search query.
    * Recommended for large result-sets, though can cause memory issue while loading all records in memory.
    * NOTE:
    * 1. NOT recommended for real time searches.
    * 2. This overwrites the caller sorting to always sort based on _doc for performance reason. Caller should sort all
    * the returned results at its end appropriately.
    * 3. Cannot be used with Collapse.
    * maxRecordsInIteration - Scroll size
    * maxRecordsOverall - Max records to load, useful when memory is constraint. By default all records are loaded.
    */
  def scrolledSearchAllHits(elasticClient: ElasticClient,
                            searchRequest: SearchRequest,
                            searchContext: String,
                            maxRecordsInIteration: Int = 100,
                            maxRecordsOverall: Long = -1,
                            version: Boolean = false): Try[List[SearchHit]] = {

    val searchHits: ListBuffer[SearchHit] = ListBuffer.empty

    //scroll Ids may change across scrolls, hence keeping track of all scroll Ids.
    var scrollIds: scala.collection.mutable.Set[String] = scala.collection.mutable.Set.empty

    val scrolledSearchRequest = searchRequest
      .size(maxRecordsInIteration)
      .version(version)
      // This time is only for the processing of the current batch
      .scroll("5m")
      // Scroll offers some optimizations for _doc sorting
      .sortBy(FieldSort("_doc"))

    var scrollSearchApiResponse = try {
      // Initial search request after which it is only scrolling
      elasticClient.execute {
        scrolledSearchRequest
      }.await(2.minutes)
    } catch {
      case ex: Exception =>
        // No point in continuing in this case.
        logger.error(s"$searchContext - Exception when running the scrolled search request: ${ex.getMessage}.")
        return Failure(ex)
    }

    // Iteration controls
    var continueLoop: Try[Boolean] = Success(true)
    var hitsTraversed = 0 // To know the total number of docs
    var scrollId: Option[String] = None // scrollId though same across iteration is recommended to take from the latest result always

    while (continueLoop.isSuccess && continueLoop.get) {

      // Parse the result
      val apiResultTry = ElasticUtils.logAndGetResult(searchContext, scrollSearchApiResponse)

      apiResultTry match {

        case Success(apiResult) =>
          // No need to continue the loop, but its a success
          // Update the scroll id
          scrollId = apiResult.scrollId
          if (scrollId.isDefined) scrollIds += scrollId.get

          if (apiResult.totalHits == 0 || apiResult.hits.hits.isEmpty) {
            // Mark the loop continuation as false
            continueLoop = Success(false)
          } else {
            // Get the hits in this result
            val hitsInThisResult = apiResult.hits.hits

            searchHits ++= hitsInThisResult

            // Update the counter on traversed number of hits till now
            hitsTraversed = hitsTraversed + hitsInThisResult.length

            logger.info(s"$searchContext - Iteration of documents scrolling: " +
              s"content-length: ${scrollSearchApiResponse.headers.getOrElse("content-length", "NotPresent")}, hitsTraversed: $hitsTraversed.")

            if (scrollId.isEmpty) {
              // Do not continue the loop if scroll is over naturally
              continueLoop = Success(false)
            } else if (hitsTraversed >= apiResult.totalHits) {
              //break out if all hits are traversed.
              continueLoop = Success(false)
            } else if (maxRecordsOverall != -1 && hitsTraversed >= maxRecordsOverall) {
              // Do not cross the limits of maxRecordsOverall
              logger.warn(s"$searchContext - Documents count has reached or crossed the limit $hitsTraversed >= $maxRecordsOverall.")

              continueLoop = Success(false)

            } else {
              try {
                // Continue to scroll as long as scroll id is returned
                scrollSearchApiResponse = elasticClient.execute {
                  searchScroll(scrollId.get)
                    .keepAlive("5m")
                }.await(2.minutes)
              } catch {
                case ex: Exception =>
                  logger.error(s"$searchContext - Failure to load documents in search scroll API." +
                    s" Reason: ${ex.getMessage}")
                  continueLoop = Failure(ex)
              }
            }
          }
        case Failure(ex) =>
          // Failure, so return immediately.
          logger.error(s"$searchContext - Failure to load documents in search scroll API. Reason: ${ex.getMessage}")
          continueLoop = Failure(ex)
      }
    }

    // Clear all the scrolls
    if (scrollIds.nonEmpty) {
      try {
        val clearScrollResponse = elasticClient.execute {
          clearScroll(scrollIds)
        }.await

        ElasticUtils.logAndGetResult(searchContext, clearScrollResponse)

      } catch {
        case ex: Exception =>
          // Exception in close of scroll, can ignore and continue
          logger.error(s"$searchContext - Exception when closing the scroll for the search request: ${ex.getMessage}")
      }
    }

    // If continueLoop is Failure, it means we hit a failure in Elastic call
    continueLoop match {
      case Failure(ex) => Failure(ex)
      case Success(_) =>
        logger.info(s"$searchContext - Total count of documents: $hitsTraversed")
        Success(searchHits.toList)
    }
  }

  /**
    * A utility method to call any REST API in Elastic for which there is no direct support in elastic4s DSL
    */
  def invokeRestApiWithElastic4sClient(elasticClient: ElasticClient,
                                       method: String,
                                       restEndpoint: String): Future[HttpResponse] = {
    val restHttpClient = elasticClient.client

    // Construct the ElasticRequest
    val elasticRequest = ElasticRequest(method, restEndpoint)

    val promise = Promise[HttpResponse]()

    val callback: Either[Throwable, HttpResponse] => Unit = {
      case Left(throwable) => promise.tryFailure(throwable)
      case Right(response) => promise.trySuccess(response)
    }

    restHttpClient.send(elasticRequest, callback)
    promise.future
  }

  // A utility to get the read index alias given the mapping template key
  def readIndexAlias(elasticSettings: ElasticSettings, mappingTemplateKey: String): String = {
    ReadIndexAlias(elasticSettings.indexMap(mappingTemplateKey).baseName)
  }

  def AlertActionResultsIndexReadAlias(elasticSettings: ElasticSettings): String = readIndexAlias(elasticSettings, MappingTemplates.INDEX_ALERT_ACTION_RESULTS_KEY)
  def DataSourcesIndexReadAlias(elasticSettings: ElasticSettings): String = readIndexAlias(elasticSettings, MappingTemplates.INDEX_DATASOURCES_KEY)
}