/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic

import akka.actor.{Actor, ActorLogging, ActorRef, Props}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.elastic.ElasticClientFactory.GLOBAL_TENANT_ID
import com.netapp.df.uba.server.common.elastic.ElasticUtils.ContextIndexManagement
import com.netapp.df.uba.server.common.elastic.model.ElasticClientInfo
import com.netapp.df.uba.server.common.misc.model._

import scala.util.Try

/**
  * IndexChecker handles the checks of presence of indices only
  *
  * Note: Index creation, rollover/prune management is handled by service-flink-job-manager directly
  * using a different set of actors.
  */
object IndexChecker {
  def props(elasticSettings: ElasticSettings, requestorRef: ActorRef): Props = Props(new IndexChecker(elasticSettings, requestorRef))
}

class IndexChecker(elasticSettings: ElasticSettings, requestorRef: ActorRef) extends Actor with ActorLogging {

  // Get only the list of read indices i.e the ones whose existence has to be checked.
  val readIndices: Map[String, ReadIndexConfig] = elasticSettings.indexMap
    .filter(_._2.isInstanceOf[ReadIndexConfig])
    .map(_._2.asInstanceOf[ReadIndexConfig])
    .map(readIndexConfig => readIndexConfig.baseName -> readIndexConfig)
    .toMap

  // Create the elastic client
  private val csElasticClientInfoTry: Try[ElasticClientInfo] = ElasticClientFactory.getCsElasticClient(GLOBAL_TENANT_ID, elasticSettings)

  // Close the elastic client
  override def postStop(): Unit = ElasticClientFactory.closeCsElasticClient(csElasticClientInfoTry)

  /**
    * receive handles the index existence check
    */
  override def receive: Receive = LoggingReceive {

    // Message to start the checks
    case ManageIndices =>

      try {
        // First check all indices that are being read. Only if this passes, will the creation be handled.
        requestorRef ! ElasticUtils.checkReadIndices(csElasticClientInfoTry, readIndices)
      } catch {
        case ex: Exception =>
          requestorRef ! ActionFailed(ContextIndexManagement, Error(ErrorCode.InternalError, s"Failed to check for indices. Reason: ${ex.getMessage}.", "indices"))
      }

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message that is not handled: " + somethingElse)
  }
}