/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic

import com.sksamuel.elastic4s.analyzers._
import com.sksamuel.elastic4s.http.ElasticDsl._
import com.sksamuel.elastic4s.mappings.MappingDefinition
import com.sksamuel.elastic4s.mappings.dynamictemplate.DynamicMapping

/**
  * Maintains the _doc template mapping for each of the index used in Contrail
  */
object MappingTemplates {
  final val IndexTypeName: String = "_doc"

  final val INDEX_AUDIT_EVENTS_KEY = "audit-events"
  final val INDEX_USERS_KEY = "users"
  final val INDEX_DATASOURCES_KEY = "datasources"
  final val INDEX_DATA_VOLUMES_KEY = "data-volumes"
  final val INDEX_USER_DIRECTORIES_KEY = "user-directories"
  final val INDEX_SHARES_KEY = "shares"
  final val INDEX_EXPORTS_KEY = "exports"
  final val INDEX_COMMUNITIES_KEY = "communities"
  final val INDEX_COMMUNITY_MAPPINGS_KEY = "community-mappings"
  final val INDEX_SEQUENCES_KEY = "sequences"
  final val INDEX_OPSRATE_METRICS_KEY = "opsrate-metrics"
  final val INDEX_USER_COMMUNITY_COUNT_METRICS_KEY = "user-community-count-metrics"
  final val INDEX_ALERTS_KEY = "alerts"
  final val INDEX_ALERT_ACTION_RESULTS_KEY = "alert-action-results"
  final val INDEX_ANOMALY_DETECTOR_STATE_KEY = "ad-state"

  private val IgnoreAboveLength = 512

  private val PathAnalyzer = "contrail-path-analyzer"

  private val PathAnalyzerReversed = "contrail-path-analyzer-reversed"

  private val LowerCaseNormalizer = "contrail-lowercase-normalizer"

  private val ColonAnalyzer = "contrail-colon-analyzer"

  private val ColonTokenizer = "contrail-colon-tokenizer"

  private val DotAnalyzer = "contrail-dot-analyzer"
  private val DotTokenizer = "contrail-dot-tokenizer"
  /*
    A custom analyser for all fields following a unix path or hierarchy kind of format. E.g., a path field like
    /User/alice/photos/2017/my_photo2.jpg will be analyzed into below terms using / path delimiter:
    [/User, /User/alice, /User/alice/photos, /User/alice/photos/2017, /User/alice/photos/2017/my_photo2.jpg]
    This helps in all path wildcard queries using elasticsearch query_string, e.g. *ice/pho* OR /User/alice/photo*. Can
    also do lookup of all files/folders within a particular directory using a search query /User/alice. Currently this
    works for all alpha-numeric characters and special characters in path names, EXCEPT literal backslash(\) and asterix
    (*).
     */
  private val CustomPathAnalyzer: AnalyzerDefinition = CustomAnalyzerDefinition(PathAnalyzer,
    PathHierarchyTokenizer("contrail-path-tokenizer", delimiter = '/', replacement = '/', reverse = false),
    LowercaseTokenFilter
  )

  /*
  *  With the reverse parameter for this tokenizer, it’s also possible to match from the other end of the file path,
  *  such as individual file names or a deep level subdirectory
  *
  *  eg:
  * displayPath "/Brochures/nested/closures999.txt.wncry", is analyzed into following tokens
  * /Brochures/nested/closures999.txt.wncry
  * Brochures/nested/closures999.txt.wncry
  * nested/closures999.txt.wncry
  * closures999.txt.wncry
  *
  * This helps in searching with partial paths in reverse order.
  * This is used in ontap-arw integration user resolution query.
  * Suspect path will not have mount path prefixed. (only /nested/closures999.txt.wncry is received from ontap)

  * */
  private val CustomPathAnalyzerReversed: AnalyzerDefinition = CustomAnalyzerDefinition(PathAnalyzerReversed,
    PathHierarchyTokenizer("contrail-path-tokenizer-reversed", delimiter = '/', replacement = '/', reverse = true),
    LowercaseTokenFilter
  )

  /*
    A custom normalizer to index keyword type of fields in lowercase to allow alphabetical sort on them. E.g. a
    keyword field canonical path /User/Alice/photos/2017 will be indexed as /user/alice/photos/2017.
    */
  private val ContrailLowercaseNormalizer: NormalizerDefinition = CustomNormalizerDefinition(LowerCaseNormalizer,
    LowercaseTokenFilter)

  /*
  A custom normalizer to analyze the field by splitting it by ":". This is used for fields like reportedId or unixReportedId
  eg: reportedId field having a value ldap:child1.parent.cs.local:S-1-5-21-2783769067-3909118257-1748686977-1111 will be analyzed and indexed as below tokens
    ldap
    child1.parent.cs.local
    s-1-5-21-2783769067-3909118257-1748686977-1111 => lower case token filter is applied
  */
  private val CustomColonAnalyzer = CustomAnalyzerDefinition(ColonAnalyzer,
    PatternTokenizer(ColonTokenizer, pattern = ":"),
    LowercaseTokenFilter
  )

  /*
  A custom analyzer to analyze the domain field by splitting it by "." (Dot). This is used for user's domain field.
  e.g: userDomain field having a value NETAPP-DOMAIN01.NTAP-CLOUD.COM will be analyzed and indexed as-
  netapp-domain01
  ntap-cloud
  com
   */
  private val CustomDotAnalyzer = CustomAnalyzerDefinition(DotAnalyzer,
    PatternTokenizer(DotTokenizer, pattern = "\\."),
    LowercaseTokenFilter
  )


  // Audit events stored from audit-events.enriched topic
  private val AuditEventDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .fields(
        keywordField("id"),
        keywordField("tenantId"),
        keywordField("datasourceId"),
        dateField("eventTimestamp")
          .format("epoch_millis"),
        keywordField("eventType"),
        keywordField("objectType"),
        keywordField("canonicalPath")
          .normalizer(LowerCaseNormalizer),
        textField("displayPath")
          .analyzer(PathAnalyzer)
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength),
            textField("reversed")
              .analyzer(PathAnalyzerReversed)
          )
          .norms(false),
        textField("entityName")
          .analyzer("pattern")
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          ),
        keywordField("originalCanonicalPath")
          .normalizer(LowerCaseNormalizer),
        textField("originalDisplayPath")
          .analyzer(PathAnalyzer)
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          ),
        textField("userDisplayName")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          ),
        textField("objectExtension")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          ),
        keywordField("mountPath")
          .normalizer(LowerCaseNormalizer),
        ipField("clientIpAddress"),
        geopointField("clientCoordinates"),
        longField("sizeBytes"),
        keywordField("dataVolumeId"),
        keywordField("userId"),
        keywordField("rawUserId"),
        objectField("protocol")
          .fields(
            keywordField("name")
              .normalizer(LowerCaseNormalizer),
            intField("majorVersion"),
            intField("minorVersion")
          )
          .dynamic("strict"),
        keywordField("entityId"),
        keywordField("originalEntityId"),
        booleanField("justCreatedEntityEvent"),
        booleanField("isTemporary"),
        booleanField("isHomeDirectoryEvent")
      )

  // Dynamic template is used for open ended attributes of a document that can be datasource specific.
  private val dynamicTemplateForAttributes =
    dynamicTemplate("strings")
      .mapping(
        dynamicTextField()
          .analyzer("pattern")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false)
      )
      .matchMappingType("string")

  // Users stored from users.raw topic
  private val UserDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("id"),
        textField("reportedId")
          .analyzer(ColonAnalyzer)
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          ).norms(false),
        keywordField("tenantId"),
        keywordField("datasourceId"),
        textField("name")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("displayName")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("userIdInDirectory")
          .analyzer(ColonAnalyzer)
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          ).norms(false),
        keywordField("userDirectoryId"),
        objectField("attributes")
          .dynamic(true),
        dateField("lastReportedTimestamp")
          .format("epoch_millis"),
        longField("communityCount"),
        dateField("communityUpdateTimestamp")
          .format("epoch_millis"),
        textField("unixReportedId")
          .analyzer(ColonAnalyzer)
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          ).norms(false)
      )

  // Datasources stored from inventory.raw topic
  private val DatasourceDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("id"),
        keywordField("reportedId"),
        keywordField("tenantId"),
        keywordField("datasourceId"),

        textField("name")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("nameFromDatasource")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("locationFromDatasource")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        booleanField("isSubDatasource"),
        keywordField("connector"),
        keywordField("connectorType"),
        objectField("attributes")
          .dynamic(true),
        dateField("lastReportedTimestamp")
          .format("epoch_millis")
      )

  // Data Volumes stored from inventory.raw topic
  private val DataVolumeDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("id"),
        keywordField("reportedId"),
        keywordField("tenantId"),
        keywordField("datasourceId"),

        textField("name")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        keywordField("dataVolumeType"),
        keywordField("state"),
        textField("path")
          .analyzer("pattern")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        keywordField("caseSensitiveNature"),
        booleanField("isRoot"),

        textField("nameFromDatasource")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("locationFromDatasource")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        objectField("services")
          .dynamic(true),
        objectField("attributes")
          .dynamic(true),
        dateField("lastReportedTimestamp")
          .format("epoch_millis")
      )

  // User directories stored from inventory.raw topic
  private val UserDirectoryDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("id"),
        keywordField("reportedId"),
        keywordField("tenantId"),
        keywordField("datasourceId"),

        keywordField("directoryType"),
        keywordField("directoryLocation"),
        objectField("attributes")
          .dynamic(true),
        dateField("lastReportedTimestamp")
          .format("epoch_millis")
      )

  // Shares stored from inventory.raw topic
  private val ShareDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("id"),
        keywordField("reportedId"),
        keywordField("tenantId"),
        keywordField("datasourceId"),

        textField("name")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("cifsServer")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("path")
          .analyzer("pattern")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        keywordField("dataVolumeId"),
        objectField("attributes")
          .dynamic(true),
        dateField("lastReportedTimestamp")
          .format("epoch_millis")
      )

  // Exports stored from inventory.raw topic
  private val ExportDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("id"),
        keywordField("reportedId"),
        keywordField("tenantId"),
        keywordField("datasourceId"),

        textField("name")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        keywordField("objectType"),
        objectField("attributes")
          .dynamic(true),
        dateField("lastReportedTimestamp")
          .format("epoch_millis")
      )

  // communities stored from communities.resolved
  private val CommunityDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("id"),
        keywordField("tenantId"),

        textField("name")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        keywordField("userIds"),
        longField("usersCount"),
        keywordField("entityIds"),
        longField("entitiesCount"),
        keywordField("userIdsHash"),
        objectField("attributes")
          .dynamic(true),
        dateField("firstReportedTimestamp")
          .format("epoch_millis"),
        dateField("lastReportedTimestamp")
          .format("epoch_millis"),
        longField("mappingsCount")
      )

  // community mappings stored from communities.resolved
  private val CommunityMappingDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .dynamicTemplates(dynamicTemplateForAttributes)
      .fields(
        keywordField("tenantId"),
        keywordField("communityId"),
        keywordField("userId"),
        dateField("userJoiningTimestamp")
          .format("epoch_millis"),
        keywordField("entityId"),
        dateField("entityJoiningTimestamp")
          .format("epoch_millis"),
        longField("accessCount"),
        doubleField("prior"),
        longField("runId"),
        objectField("attributes")
          .dynamic(true),
        dateField("lastReportedTimestamp")
          .format("epoch_millis")
      )

  // sequences for generating auto-incrementing ids
  private val SequenceDocMapping =
    mapping(IndexTypeName)
      .dynamic(DynamicMapping.Strict)
      .fields(
        longField("nextCommunityId"),
        longField("nextAlertNumber")
      )

  // Time-series data of metrics
  private val OpsRateMetricsMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .fields(
        keywordField("id"),
        keywordField("tenantId"),
        keywordField("userId"),
        keywordField("communityId"),
        dateField("metricTimestamp")
          .format("epoch_millis"),
        doubleField("metricValue"),
        longField("rawAccessCount"),
        doubleField("userMean"),
        doubleField("communityMedian"),
        objectField("rawAccessCountByEventType")
          .dynamic(true)
      )

  // Time-series data of user-community-count metrics
  private val UserCommunityCountMetricsMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .fields(
        keywordField("id"),
        keywordField("tenantId"),
        keywordField("userId"),
        doubleField("metricValue"),
        dateField("metricTimestamp")
          .format("epoch_millis")
      )

  // Alerts created based on data in anomaly.generated topic
  private val AlertDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .fields(
        keywordField("id"),
        keywordField("displayId"),
        keywordField("tenantId"),
        keywordField("userId"),
        textField("userName")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .normalizer(LowerCaseNormalizer)
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        keywordField("communityId"),
        textField("communityName")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        dateField("alertTimestamp")
          .format("epoch_millis"),
        objectField("attributes")
          .dynamic(true),
        textField("synopsis")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        keywordField("alertType"),
        keywordField("severity"),
        keywordField("status"),
        keywordField("action"),
        textField("actionDetails")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        textField("note")
          .analyzer("standard")
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        nestedField("transitions")
          .fields(
            keywordField("severity"),
            keywordField("status"),
            dateField("alertTimestamp")
              .format("epoch_millis"),
            keywordField("alertType")
          ),
        doubleField("changePercentage"),
        booleanField("isFalsePositive"),
        dateField("patternStartTimestamp")
          .format("epoch_millis"),
        dateField("patternEndTimestamp")
          .format("epoch_millis"),
        textField("userDomain")
          .analyzer(DotAnalyzer)
          .fields(
            keywordField("keyword")
              .ignoreAbove(IgnoreAboveLength)
          )
          .norms(false),
        nestedField("dataVolumes")
          .fields(
            keywordField("dataSourceId"),
            keywordField("dataVolumeId"),
            keywordField("clientIps"),
            dateField("volumeEncryptionTimestamp")
              .format("epoch_millis")
          ),
        keywordField("alertTrigger"),
        keywordField("trustScore")
      )

  // alert action result created from alert-action.results topic
  private val AlertActionResultsDocMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .fields(
        keywordField("tenantId"),
        keywordField("actionId"),
        keywordField("actionType"),
        keywordField("alertId"),
        keywordField("userId"),
        booleanField("isAutomatic"),
        dateField("startTimestamp")
          .format("epoch_millis"),
        dateField("endTimestamp")
          .format("epoch_millis"),
        keywordField("createdBy"),
        nestedField("subActions")
          .fields(
            keywordField("deviceId"),
            keywordField("volumeId"),
            keywordField("policyId"),
            textField("policyName")
              .analyzer("standard")
              .fields(
                keywordField("keyword")
                  .ignoreAbove(IgnoreAboveLength)
              )
              .norms(false),
            keywordField("actionStatus"),
            objectField("actionResult")
              .dynamic(true)
          )
      )

  private val AdStateMapping =
    mapping(IndexTypeName)
      .source(true)
      .dynamic(DynamicMapping.Strict)
      .fields(
        keywordField("tenantId"),
        binaryField("state"),
        dateField("lastUpdatedTimestamp")
          .format("epoch_millis")
      )

  val DocMap: Map[String, MappingDefinition] = Map(
    INDEX_AUDIT_EVENTS_KEY -> AuditEventDocMapping,
    INDEX_USERS_KEY -> UserDocMapping,
    INDEX_DATASOURCES_KEY -> DatasourceDocMapping,
    INDEX_DATA_VOLUMES_KEY -> DataVolumeDocMapping,
    INDEX_USER_DIRECTORIES_KEY -> UserDirectoryDocMapping,
    INDEX_SHARES_KEY -> ShareDocMapping,
    INDEX_EXPORTS_KEY -> ExportDocMapping,
    INDEX_COMMUNITIES_KEY -> CommunityDocMapping,
    INDEX_COMMUNITY_MAPPINGS_KEY -> CommunityMappingDocMapping,
    INDEX_SEQUENCES_KEY -> SequenceDocMapping,
    INDEX_OPSRATE_METRICS_KEY -> OpsRateMetricsMapping,
    INDEX_USER_COMMUNITY_COUNT_METRICS_KEY -> UserCommunityCountMetricsMapping,
    INDEX_ALERTS_KEY -> AlertDocMapping,
    INDEX_ALERT_ACTION_RESULTS_KEY -> AlertActionResultsDocMapping,
    INDEX_ANOMALY_DETECTOR_STATE_KEY -> AdStateMapping
  )

  val AnalyzerMap: Map[String, List[AnalyzerDefinition]] = Map(
    INDEX_AUDIT_EVENTS_KEY -> List(CustomPathAnalyzer, CustomPathAnalyzerReversed),
    INDEX_USERS_KEY -> List(CustomColonAnalyzer),
    INDEX_ALERTS_KEY -> List(CustomDotAnalyzer)
  )

  val NormalizerMap: Map[String, List[NormalizerDefinition]] = Map(
    INDEX_AUDIT_EVENTS_KEY -> List(ContrailLowercaseNormalizer),
    INDEX_USERS_KEY -> List(ContrailLowercaseNormalizer),
    INDEX_ALERTS_KEY -> List(ContrailLowercaseNormalizer)
  )
}
