/*
 * $Id$
 *
 *  Copyright (c) 2019 NetApp, Inc.
 *  All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic.aggregations

import com.netapp.df.uba.server.common.elastic.model.{SortInfo, TimeFilter}
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.sksamuel.elastic4s.http.ElasticDsl._
import com.sksamuel.elastic4s.http.search._
import com.sksamuel.elastic4s.script.Script
import com.sksamuel.elastic4s.searches.aggs.{Aggregation, _}
import com.sksamuel.elastic4s.searches.queries.Query
import com.sksamuel.elastic4s.searches.sort.{Sort, SortOrder}
import org.joda.time.DateTimeZone

// Trait to indicate how to generate a name, build the aggregation snippet and parse the results
trait AggregationInfo {

  // Generates a name for the aggregation
  def aggregationName: String

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation

  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[Any]

  // Context of the aggregation
  def context: String
}

/**
 * Concrete class for terms based aggregation that supports both field and script
 */
case class TermsAggregationInfo(fieldName: String,
                                size: Int = AggregationFieldConstants.MAX_AGG_BUCKET_SIZE,
                                order: TermsOrder = TermsOrder("_count", asc = false),
                                scriptOption: Option[Script] = None,
                                subAggregationInfoOption: Option[AggregationInfo] = None) extends AggregationInfo with LogTrait {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-terms-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    val termsAggregationSnippet = termsAggregation(aggregationName)
      .size(size)
      .order(order)

    // We can either use a direct field or a scripted field
    val termsAgg = if (scriptOption.isDefined) termsAggregationSnippet.script(scriptOption.get)
    else termsAggregationSnippet.field(fieldName)

    if (subAggregationInfoOption.isDefined) termsAgg.addSubagg(subAggregationInfoOption.get.buildAggregationSnippet())
    else termsAgg
  }

  // Context of the aggregation
  def context: String = s"Terms aggregation on $fieldName"

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[TermsAggResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      val termsAggregationResult = searchResponse.aggregations.terms(aggregationName)
      //log doc count errors observed in terms aggregation
      logDocCountError(
        tenantId,
        termsAggregationResult.docCountErrorUpperBound,
        termsAggregationResult.otherDocCount,
        size,
        context
      )
      Some(termsAggregationResult)
    } else {
      None
    }
  }

  /**
   * For details on docCountError
   * Please refer https://www.elastic.co/guide/en/elasticsearch/reference/current/search-aggregations-bucket-terms-aggregation.html
   *
   * NOTE: When the aggregation is either sorted by a sub aggregation or in order of ascending document count,
   * the error in the document counts cannot be determined and is given a value of -1 to indicate this.
   *
   * WARNING: Sorting by ascending _count or by sub aggregation is discouraged as it increases the error on document counts.
   * It is fine when a single shard is queried, or when the field that is being aggregated was used as a routing key at index time:
   * in these cases results will be accurate since shards have disjoint values. However otherwise, errors are unbounded.
   * One particular case that could still be useful is sorting by min or max aggregation: counts will not be accurate
   * but at least the top buckets will be correctly picked.
   */
  def logDocCountError(tenantId: String, docCountErrorUpperBound: Long, sumOfOtherDocCount: Long, size: Int, context: String): Unit = {
    if (docCountErrorUpperBound != 0) logErrorT(tenantId, s"$context; Aggregation size: $size; Error in document count with maximum error margin of $docCountErrorUpperBound documents.")
    if (sumOfOtherDocCount > 0) logErrorT(tenantId, s"$context; Aggregation size: $size; The sum of the document counts for all buckets that are not part of the response : $sumOfOtherDocCount.")
  }

}

/**
 * Concrete class for top hits based aggregation
 */
case class TopHitsAggregationInfo(name: String = "default",
                                  size: Int,
                                  sort: List[Sort] = List.empty,
                                  isSourceIncluded: Boolean = true,
                                  sourceIncludeArray: Array[String] = Array.empty,
                                  sourceExcludeArray: Array[String] = Array.empty,
                                  scriptOption: Option[Script] = None,
                                 ) extends AggregationInfo {
  // Generates a name for the aggregation
  def aggregationName: String = s"$name-top-hits-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    val topHitsAggregationSnippet = topHitsAggregation(aggregationName)
      .size(size)
      .sortBy(sort)

    // Note: Passing both sourceIncludeArray and sourceExcludeArray as empty will fetch the source
    val topHitsAggWithSourceInfo = if (isSourceIncluded) topHitsAggregationSnippet.fetchSource(sourceIncludeArray, sourceExcludeArray)
    else topHitsAggregationSnippet.fetchSource(isSourceIncluded)

    val topHitsAgg = if (scriptOption.isDefined) topHitsAggWithSourceInfo.script("top-hits-script", scriptOption.get)
    else topHitsAggWithSourceInfo

    topHitsAgg
  }

  // Context of the aggregation
  def context: String = s"Top hits aggregation - $aggregationName"

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[TopHitsResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      Some(searchResponse.aggregations.tophits(aggregationName))
    } else {
      None
    }
  }
}

/**
 * Concrete class for filters based aggregation, to be used for
 */
case class FiltersAggregationInfo(fieldName: String, validValues: List[String], subAggregationInfoOption: Option[AggregationInfo] = None) extends AggregationInfo {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-filters-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {

    val validValuesFilterQueries: List[(String, Query)] = validValues
      .map(validValue => (validValue, termQuery(fieldName, validValue)))

    val filtersAgg = filtersAggregation(aggregationName)
      .queries(validValuesFilterQueries)

    if (subAggregationInfoOption.isDefined) filtersAgg.addSubagg(subAggregationInfoOption.get.buildAggregationSnippet())
    else filtersAgg
  }

  // Context of the aggregation
  def context: String = s"Filters aggregation on $fieldName"


  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[FiltersAggregationResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      Some(searchResponse.aggregations.filters(aggregationName))
    } else {
      None
    }
  }
}

/**
 * Concrete class for date histogram based aggregation
 */
object DateHistogramAggregationInfo {

  def apply(fieldName: String,
            interval: Long,
            minimumDocCount: Long,
            timeZone: Option[DateTimeZone],
            timeFilter: TimeFilter,
            sortInfo: SortInfo): DateHistogramAggregationInfo = {

    // Use sort info to convert to histogram order.
    val histogramOrder = (sortInfo.primarySort.field, sortInfo.primarySort.order) match {
      case ("time", SortOrder.Asc) => HistogramOrder.KEY_ASC
      case ("time", SortOrder.Desc) => HistogramOrder.KEY_DESC
      case ("value", SortOrder.Asc) => HistogramOrder.COUNT_ASC
      case ("value", SortOrder.Desc) => HistogramOrder.COUNT_DESC
      case (_, _) => HistogramOrder.KEY_ASC
    }

    DateHistogramAggregationInfo(
      fieldName,
      interval,
      minimumDocCount,
      timeFilter,
      histogramOrder,
      timeZone
    )
  }
}

case class DateHistogramAggregationInfo(fieldName: String,
                                        interval: Long,
                                        minimumDocCount: Long,
                                        timeFilter: TimeFilter,
                                        histogramOrder: HistogramOrder,
                                        timeZone: Option[DateTimeZone] = None,
                                        subAggregationInfoOption: Option[AggregationInfo] = None) extends AggregationInfo {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-datehistogram-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    val extBounds = LongExtendedBounds(timeFilter.fromTime, timeFilter.toTime)


    // Includes the base date histogram aggregation snippet
    val baseAggregationObject = dateHistogramAggregation(aggregationName)
      .field(fieldName)
      .interval(interval)
      .minDocCount(minimumDocCount)
      .timeZone(timeZone.getOrElse(DateTimeZone.UTC)) // Refer: https://www.elastic.co/guide/en/elasticsearch/reference/6.3/search-aggregations-bucket-datehistogram-aggregation.html#_timezone

    // Add external bounds to the query so that all buckets are returned even if count is 0
    val withExtBounds = baseAggregationObject.extendedBounds(extBounds)

    // Add histogram order to the query
    val withHistogramOrder = withExtBounds
      .order(histogramOrder)

    if (subAggregationInfoOption.isDefined) withHistogramOrder.addSubagg(subAggregationInfoOption.get.buildAggregationSnippet())
    else withHistogramOrder
  }

  // Context of the aggregation
  def context: String = s"DateHistogram aggregation on $fieldName"

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[DateHistogramAggResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      Some(searchResponse.aggregations.dateHistogram(aggregationName))
    } else {
      None
    }
  }
}


/**
 * Concrete class for value count based aggregation, to be used for
 */
case class ValueAggregationInfo(fieldName: String, subAggregationInfoOption: Option[AggregationInfo] = None) extends AggregationInfo {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-value-count-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    val valueCountAgg = valueCountAggregation(aggregationName)
      .field(fieldName)

    if (subAggregationInfoOption.isDefined) valueCountAgg.addSubagg(subAggregationInfoOption.get.buildAggregationSnippet())
    else valueCountAgg
  }

  // Context of the aggregation
  def context: String = s"Value count aggregation on $fieldName"

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[ValueCountResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      Some(searchResponse.aggregations.valueCount(aggregationName))
    } else {
      None
    }
  }
}

/**
 * Concrete class for cardinality based aggregation
 */
case class CardinalityAggregationInfo(fieldName: String,
                                      precisionThreshold: Int = AggregationFieldConstants.DEFAULT_CARDINALITY_PRECISION_THRESHOLD)
  extends AggregationInfo {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-cardinality-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    cardinalityAggregation(aggregationName)
      .field(fieldName)
      .precisionThreshold(precisionThreshold)
  }

  // Context of the aggregation
  def context: String = s"Cardinality aggregation on $fieldName"

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[CardinalityAggResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      Some(searchResponse.aggregations.cardinality(aggregationName))
    } else {
      None
    }
  }
}

/**
 * Concrete class for nested aggregation
 */
case class NestedAggregationInfo(fieldName: String = "default",
                                 path: String,
                                 subAggregationInfoOption: Option[AggregationInfo])
  extends AggregationInfo {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-nested-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    val nestedAggregationSnippet = nestedAggregation(aggregationName, path)

    val nestedAgg = if (subAggregationInfoOption.isDefined) nestedAggregationSnippet.addSubagg(subAggregationInfoOption.get.buildAggregationSnippet())
    else nestedAggregationSnippet

    nestedAgg
  }

  // Context of the aggregation
  def context: String = s"Nested aggregation on path: $path"

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[NestedAggResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      Some(searchResponse.aggregations.nested(aggregationName))
    } else {
      None
    }
  }
}

/**
 * Concrete class for reverse nested aggregation
 */
case class ReverseNestedAggregationInfo(fieldName: String = "default",
                                        subAggregationInfoOption: Option[AggregationInfo])
  extends AggregationInfo {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-reverse-nested-aggregation"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    val reverseNestedAggregationSnippet = reverseNestedAggregation(aggregationName)

    val reverseNestedAgg = if (subAggregationInfoOption.isDefined) reverseNestedAggregationSnippet.addSubagg(subAggregationInfoOption.get.buildAggregationSnippet())
    else reverseNestedAggregationSnippet

    reverseNestedAgg
  }

  // Context of the aggregation
  def context: String = s"Reverse nested aggregation is being performed with aggregationName: $aggregationName"

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[ReverseNestedAggResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      Some(searchResponse.aggregations.reverseNested(aggregationName))
    } else {
      None
    }
  }
}

/**
 * Concrete class for composite aggregation that supports both field and script.
 * For performance reasons, by default the composite buckets are sorted by their natural ordering.
 */
case class CompositeAggregationInfo(fieldName: String,
                                    size: Int = AggregationFieldConstants.MAX_AGG_BUCKET_SIZE,
                                    scriptOption: Option[Script] = None,
                                    subAggregationInfoOption: Option[AggregationInfo] = None) extends AggregationInfo with LogTrait {

  // Generates a name for the aggregation
  def aggregationName: String = s"$fieldName-composite-aggregation"

  def valueSourceName: String = s"$fieldName-value-source"

  // Context of the aggregation
  def context: String = s"Composite aggregation on $fieldName"

  // Builds the aggregation snippet
  def buildAggregationSnippet(): Aggregation = {
    // We can either use a direct field or a scripted field
    val compositeAgg = if (scriptOption.isDefined) {
      CompositeAggregation(aggregationName,
        List(TermsValueSource(valueSourceName, script = scriptOption)),
        Some(size)
      )
    } else {
      CompositeAggregation(aggregationName,
        List(TermsValueSource(valueSourceName, field = Some(fieldName))),
        Some(size)
      )
    }

    if (subAggregationInfoOption.isDefined) compositeAgg.addSubagg(subAggregationInfoOption.get.buildAggregationSnippet())
    else compositeAgg
  }

  // Extract the aggregation result
  def extractAggregationResult(tenantId: String, searchResponse: SearchResponse): Option[CompositeAggResult] = {
    if (searchResponse.aggregations.contains(aggregationName)) {
      val compositeAggregationResult = CompositeAggResult(aggregationName, searchResponse.aggregations.dataAsMap(aggregationName).asInstanceOf[Map[String, Any]])
      Some(compositeAggregationResult)
    } else {
      None
    }
  }

  def getCompositeBucketKey(bucket: CompositeBucket): Any = bucket.key(valueSourceName)
}

//elastic4s doesnt support result objects for composite aggregation currently. Hence defining these newly.
case class CompositeBucket(key: Map[String, Any], docCount: Long, data: Map[String, Any])
  extends AggBucket
    with Transformable

case class CompositeAggResult(name: String, buckets: Seq[CompositeBucket], afterKey: Map[String, Any])
  extends BucketAggregation {
  def bucket(key: Map[String, Any]): CompositeBucket = bucketOpt(key).get

  def bucketOpt(key: Map[String, Any]): Option[CompositeBucket] = buckets.find(_.key.equals(key))
}

object CompositeAggResult {
  def apply(name: String, data: Map[String, Any]): CompositeAggResult = CompositeAggResult(
    name,
    data("buckets").asInstanceOf[Seq[Map[String, Any]]].map { map =>
      CompositeBucket(
        map("key").asInstanceOf[Map[String, Any]],
        map("doc_count").toString.toInt,
        map
      )
    },
    if (data.contains("after_key")) data("after_key").asInstanceOf[Map[String, Any]] else Map.empty
  )
}