/*
 * $Id$
 *
 * Copyright (c) 2021 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.df.uba.server.common.elastic.invoker

import com.netapp.df.uba.server.common.elastic.ElasticSearchHandler.createBasicSearchRequest
import com.netapp.df.uba.server.common.elastic.ElasticUtils.{AlertActionResultsIndexReadAlias, DataSourcesIndexReadAlias}
import com.netapp.df.uba.server.common.elastic.aggregations.{TermsAggregationInfo, TopHitsAggregationInfo}
import com.netapp.df.uba.server.common.elastic.utils.ActionUtils.formatErrorMessageFromDeviceError
import com.netapp.df.uba.server.common.elastic.utils.FieldConstants
import com.netapp.df.uba.server.common.elastic.utils.FieldConstants.DatasourceIdFieldName
import com.netapp.df.uba.server.common.elastic.{ElasticSearchHandler, ElasticSettings, ElasticUtils}
import com.netapp.df.uba.server.common.misc.dbmodel.ActionType.{ActionType, CREATE_SNAPSHOT, LIMIT_USER_ACCESS, RESTORE_USER_ACCESS}
import com.netapp.df.uba.server.common.misc.dbmodel._
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.sksamuel.elastic4s.ElasticApi.{existsQuery, multi, must, nestedQuery, termQuery, termsQuery}
import com.sksamuel.elastic4s.http.ElasticClient
import com.sksamuel.elastic4s.http.ElasticDsl.BuildableTermsNoOp
import com.sksamuel.elastic4s.http.search.{MultiSearchResponse, SearchHit}
import com.sksamuel.elastic4s.searches.MultiSearchRequest
import com.sksamuel.elastic4s.searches.queries.{BoolQuery, Query}
import com.sksamuel.elastic4s.searches.sort.{FieldSort, SortOrder}
import spray.json._

import scala.util.{Failure, Success, Try}

/**
  * Base trait represents the last action invoker<br>
  * Subclasses are defined based on the type of last action.<br>
  * Use cases:<br>
  *  - Only alert actions per alertId
  *  - Only user actions per userId
  *  - Mixed scenario where both alert and user actions are combined
  *    to get consolidated list of last action per alertId (where alertId -> userId map is passed)
  */
trait LastActionInvoker extends LogTrait {
  implicit val alertActionsJsonFormats: AlertActionsJsonFormats = new AlertActionsJsonFormats()

  // Alert ids if available
  def alertIds: Set[String]

  // User Ids if available
  def userIds: Set[String]

  // Object invoking the last action query e.g Alert or User
  def actionByObject: ObjectAction.Action

  // id set provided is empty or not
  def isEmpty: Boolean

  // Construct the final result based on the last action query invoker
  def getFinalResultMap(alertIdToResultMap: Map[String, List[ActionTaken]],
                        userIdToResultMap: Map[String, List[ActionTaken]]): Map[String, List[ActionTaken]]

  // Base or top level query to filter actions on
  val baseQuery: Query
  /* Get the action based on descending startTimestamp to get the latest document (most recent action performed)
  Form topHits subAggregation to get only 1 hit which is most recent.*/
  val topHitsAggregationInfo: TopHitsAggregationInfo = {
    TopHitsAggregationInfo(size = 1, sort = List(FieldSort("startTimestamp").order(SortOrder.DESC)))
  }

  // Form Terms (Group by) aggregation on all actionTypes with size not more than possible actionTypes
  def actionTypeTermsAggregationInfo: TermsAggregationInfo = {
    TermsAggregationInfo("actionType", objectToActionTypeMap(actionByObject).size, subAggregationInfoOption = Some(topHitsAggregationInfo))
  }

  // Main aggregation with rest of the sub aggregations
  def mainAggregationInfo: TermsAggregationInfo

  // Final elasticsearch boolean query
  private def constructBoolQuery: BoolQuery = {
    must(baseQuery)
      .withMust(
        termsQuery("actionType", objectToActionTypeMap(actionByObject)),
        existsQuery("endTimestamp") // Not considering the InProgress
      )
      .withNot(nestedQuery("subActions", termQuery("subActions.actionStatus", ActionStatus.InProgress.toString))) // Not considering the InProgress
  }

  // Search request to be queried
  def buildSearchRequest(elasticSettings: ElasticSettings, tenantId: String): MultiSearchRequest = {
    val searchRequest = ElasticSearchHandler.createBasicSearchRequest(constructBoolQuery, AlertActionResultsIndexReadAlias(elasticSettings), tenantId)
      .size(0) // Don't fetch the hits
      .fetchSource(false)

    // Build the final request by appending the aggregation to the search request
    val aggregationSearchRequest = searchRequest
      .aggs(mainAggregationInfo.buildAggregationSnippet())

    multi(aggregationSearchRequest)
  }

  // Extraction logic to be implemented.
  def extractResult(multiSearchResponse: MultiSearchResponse, tenantId: String, elasticSettings: ElasticSettings, elasticClient: ElasticClient): Map[String, List[ActionTaken]]

  def getSnapshotResult(lastActionResultDb: AlertActionResultDb): ActionTaken = {
    // collect all failed subActions
    val failedSubActions = lastActionResultDb
      .subActions
      .filter(_.actionStatus.equals(ActionStatus.Failure))
    // Mode of the alert action
    val lastActionMode = if (lastActionResultDb.isAutomatic) ActionMode.Automatic else ActionMode.Manual
    // some subActions failed
    if (failedSubActions.nonEmpty) {
      // timestamp for last failed subAction
      val lastFailedTimestamp = failedSubActions
        .sortBy(_
          .actionResult
          .get
          .asInstanceOf[VolumeSnapshotDetails]
          .volumeSnapshotResultDb
          .snapshotTimestamp
          .get)(Ordering[Long].reverse)
        .head
        .actionResult
        .get
        .asInstanceOf[VolumeSnapshotDetails]
        .volumeSnapshotResultDb
        .snapshotTimestamp
        .get
      // build error message
      val errorMessage = failedSubActions
        .map(_
          .actionResult
          .get
          .asInstanceOf[VolumeSnapshotDetails]
          .volumeSnapshotResultDb
          .error
          .get) // In case of Failure we definitely have error defined
        .mkString("\n")

      ActionTaken(ActionType.CREATE_SNAPSHOT,
        Some(SnapshotAction(
          lastFailedTimestamp,
          ActionStatus.Failure,
          lastActionMode,
          lastActionResultDb.createdBy,
          Some(errorMessage)
        ))
      )
    }
    else { // Last successful snapshot
      ActionTaken(ActionType.CREATE_SNAPSHOT,
        Some(SnapshotAction(lastActionResultDb.endTimestamp.get,
          ActionStatus.Success,
          lastActionMode,
          lastActionResultDb.createdBy,
          None))
      )
    }
  }

  def getLimitUserResult(lastActionResultDb: AlertActionResultDb, tenantId: String, elasticSettings: ElasticSettings, elasticClient: ElasticClient): ActionTaken = {
    // collect all subActions
    val subActions = lastActionResultDb
      .subActions
    // try to collect one successful subAction
    val successSubActionOption = subActions
      .find(_.actionStatus.equals(ActionStatus.Success))
    // collect all error subActions
    val errorSubActions = subActions
      .filter(_.actionResult
        .get
        .asInstanceOf[UserAccessDetails]
        .userAccessResultDb
        .error
        .isDefined)
    // Mode of the user action
    val lastActionMode = if (lastActionResultDb.isAutomatic) ActionMode.Automatic else ActionMode.Manual
    // last failed action result. If no success, pick error subAction
    val responseDb = successSubActionOption
      .getOrElse(errorSubActions.head)
      .actionResult
      .get
      .asInstanceOf[UserAccessDetails]
      .userAccessResultDb
    // try to get the duration based on expiryMode
    val durationOption = responseDb.expiryMode match {
      case Some(ExpiryMode.Permanent) => Some(responseDb.expiryMode.get.toString)
      case Some(ExpiryMode.TimeBased) => Try(responseDb.duration.get.toString).toOption
      case _ => None
    }
    // some subActions in error
    if (errorSubActions.nonEmpty) {
      // Try and get datasource names by Id map
      val datasourceNamesByIds = getDataSourceNamesByIds(elasticSettings,
        elasticClient,
        errorSubActions.map(_.deviceId.get).toSet,
        tenantId,
        DatasourceIdFieldName) match {
        case Success(datasourceNamesByIdsMap) => datasourceNamesByIdsMap
        case Failure(exception) => throw new Exception("Failed to generate datasources id to name map", exception)
      }
      val deviceErrorDetailList = errorSubActions
        .map(subAction => {
          DeviceErrorDetail(
            datasourceNamesByIds
              .getOrElse(subAction.deviceId.getOrElse(""), FieldConstants.UNKNOWN),
            subAction
              .actionResult
              .get
              .asInstanceOf[UserAccessDetails]
              .userAccessResultDb
              .error
              .get)
        })
      // build error message
      val errorMessage = formatErrorMessageFromDeviceError(deviceErrorDetailList)
        .map(x => s"${x.deviceName}: ${x.error}")
        .mkString("\n")
      // append actionList
      ActionTaken(lastActionResultDb.actionType,
        Some(UserLimitAction(
          lastActionResultDb.endTimestamp.get,
          ActionStatus.Failure,
          lastActionMode,
          subActions.size,
          errorSubActions.size,
          responseDb.limitUserAccessMode,
          responseDb.userIps.getOrElse(Nil),
          responseDb.expiryTimestamp,
          durationOption,
          lastActionResultDb.createdBy,
          Some(errorMessage)
        ))
      )
    }
    else { // no failed subAction
      ActionTaken(lastActionResultDb.actionType,
        Some(UserLimitAction(
          lastActionResultDb.endTimestamp.get,
          ActionStatus.Success,
          lastActionMode,
          subActions.size,
          0,
          responseDb.limitUserAccessMode,
          responseDb.userIps.getOrElse(Nil),
          responseDb.expiryTimestamp,
          durationOption,
          lastActionResultDb.createdBy,
          None
        ))
      )
    }
  }

  /**
    * Actions are mapped to their respective action types
    * AlertAction -> actions invoked on an alert
    * UserAction -> actions invoked on a user
    */
  val objectToActionTypeMap: Map[ObjectAction.Action, List[ActionType]] = {
    Map(
      ObjectAction.AlertAction -> List(CREATE_SNAPSHOT, LIMIT_USER_ACCESS),
      ObjectAction.UserAction -> List(LIMIT_USER_ACCESS, RESTORE_USER_ACCESS)
    )
  }

  /**
    * Returns search hits for datasources for specified datasource ids or datasourceIds
    *
    * @param idFieldName Id field name to be referred in datasourceDb. Possible values - {"id", "datasourceId"}
    */
  def getDataSourceNamesByIds(elasticSettings: ElasticSettings,
                              elasticClient: ElasticClient,
                              dataSourceIds: Set[String],
                              tenantId: String,
                              idFieldName: String = FieldConstants.IdFieldName): Try[Map[String, String]] = {

    if (!List(FieldConstants.IdFieldName, FieldConstants.DatasourceIdFieldName)
      .contains(idFieldName)) {
      val failureMessage = s"Unsupported IdField: $idFieldName passed while creating datasourceId to datasourceName map"
      logErrorT(tenantId, failureMessage)
      Failure(new Exception(failureMessage))
    }
    else {
      val context = s"Getting search hits of datasources by $idFieldName: ${dataSourceIds.mkString(", ")}"
      logInfoT(tenantId, context)
      val searchQuery = termsQuery(idFieldName, dataSourceIds)

      val dataSourceSearchRequest = createBasicSearchRequest(
        searchQuery,
        DataSourcesIndexReadAlias(elasticSettings),
        tenantId)

      /*Returning all the datasources assuming that datasource count will not exceed the max limit of documents elastic search can return*/
      val dataSourceSearchHits = ElasticUtils.searchAllHits(elasticClient, dataSourceSearchRequest, s"Tenant: $tenantId; $context")
      getDataSourceNamesByIdsMap(dataSourceSearchHits, idFieldName)
    }
  }

  private def getDataSourceNamesByIdsMap(dataSourceSearchHitsTry: Try[List[SearchHit]], idFieldName: String): Try[Map[String, String]] = {
    dataSourceSearchHitsTry match {
      case Success(dataSourceSearchHits) =>
        val dataSourceDbList = getDataSourceDbList(dataSourceSearchHits.distinct)
        val dataSourceNameByIdsMap = dataSourceDbList
          .map(dataSourceDb =>
            idFieldName match {
              case FieldConstants.DatasourceIdFieldName => (dataSourceDb.datasourceId, dataSourceDb.nameFromDatasource)
              case FieldConstants.IdFieldName => (dataSourceDb.id, dataSourceDb.nameFromDatasource)
            }
          ).toMap
        Success(dataSourceNameByIdsMap)

      case Failure(ex) =>
        Failure(ex)
    }
  }

  // Datasources
  private def getDataSourceDbList(dataSourceSearchHits: List[SearchHit]): List[DatasourceDb] = {
    implicit val dataSourceJsonFormats: InventoryJsonFormats = new InventoryJsonFormats()
    import dataSourceJsonFormats._

    dataSourceSearchHits
      .map(_.sourceAsString.parseJson.convertTo[DatasourceDb])
  }
}

/**
  * Actions are categorized depending on which forensic object the action is taken on
  */
object ObjectAction extends Enumeration {
  type Action = Value

  val AlertAction: Action = Value("alert_action")
  val UserAction: Action = Value("user_action")
}
