/*
 * $Id$
 *
 * Copyright (c) 2021 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.df.uba.server.common.elastic.invoker

import com.netapp.df.uba.server.common.elastic.ElasticSettings
import com.netapp.df.uba.server.common.elastic.aggregations.TermsAggregationInfo
import com.netapp.df.uba.server.common.misc.dbmodel._
import com.sksamuel.elastic4s.ElasticApi.termsQuery
import com.sksamuel.elastic4s.http.ElasticDsl.BuildableTermsNoOp
import com.sksamuel.elastic4s.http.search.{MultiSearchResponse, SearchResponse, TermsAggResult}
import com.sksamuel.elastic4s.http.{ElasticClient, SourceAsContentBuilder}
import com.sksamuel.elastic4s.searches.queries.Query
import spray.json._


/**
  * Class representing the action query invoker for AlertActions
  *
  * @param alertIdSet Set of alertIds on which last actions are queried
  */
case class LastAlertActionInvoker(alertIdSet: Set[String]) extends LastActionInvoker {
  override def alertIds: Set[String] = alertIdSet

  override def userIds: Set[String] = Set.empty

  override def actionByObject: ObjectAction.Action = ObjectAction.AlertAction

  override def isEmpty: Boolean = alertIdSet.isEmpty

  // Actions are first filtered by alertIds
  override val baseQuery: Query = termsQuery("alertId", alertIds)

  // Group by alertIds to get the last actions for each alert id
  def mainAggregationInfo: TermsAggregationInfo = {
    TermsAggregationInfo("alertId", alertIds.size, subAggregationInfoOption = Some(actionTypeTermsAggregationInfo))
  }

  def getFinalResultMap(alertIdToResultMap: Map[String, List[ActionTaken]],
                        userIdToResultMap: Map[String, List[ActionTaken]]): Map[String, List[ActionTaken]] = alertIdToResultMap

  // Extracting the alert action results
  def extractResult(searchResponse: SearchResponse, tenantId: String, elasticSettings: ElasticSettings, elasticClient: ElasticClient): Map[String, List[ActionTaken]] = {
    import alertActionsJsonFormats._

    val mainAggregationResultOption: Option[TermsAggResult] = mainAggregationInfo
      .extractAggregationResult(tenantId, searchResponse)

    val actionTakenMapByAlertId = if (mainAggregationResultOption.isEmpty) {
      Map.empty[String, List[ActionTaken]]
    } else {
      val mainAggregationResult = mainAggregationResultOption.get

      val mainAggregationBuckets = mainAggregationResult
        .buckets

      mainAggregationBuckets
        .map(alertIdBucket => {
          val actionTypeBuckets = alertIdBucket
            .terms(actionTypeTermsAggregationInfo.aggregationName)
            .buckets

          val alertActionResultDbSorted: List[AlertActionResultDb] = actionTypeBuckets
            .map(actionTypeBucket => {
              // Get the most recent action for each actionType
              val actionResultTopHitSource = actionTypeBucket
                .tophits(topHitsAggregationInfo.aggregationName)
                .hits
                .head
                .source
              val sourceAsString: String = SourceAsContentBuilder(actionResultTopHitSource).string()

              sourceAsString.parseJson.convertTo[AlertActionResultDb]
            })
            .sortBy(_.endTimestamp.get)(Ordering[Long].reverse)
            .toList

          val actionList = alertActionResultDbSorted
            .map(lastActionResultDb => {
              lastActionResultDb.actionType match {
                case ActionType.CREATE_SNAPSHOT =>
                  getSnapshotResult(lastActionResultDb)

                case ActionType.LIMIT_USER_ACCESS =>
                  getLimitUserResult(lastActionResultDb, tenantId, elasticSettings, elasticClient)

                case otherActionType => // Other actionTypes are not implemented yet
                  ActionTaken(otherActionType, None)
              }
            })

          alertIdBucket.key -> actionList
        })
        .toMap
    }
    actionTakenMapByAlertId
  }

  // Extract result from multi search response
  override def extractResult(multiSearchResponse: MultiSearchResponse, tenantId: String, elasticSettings: ElasticSettings, elasticClient: ElasticClient): Map[String, List[ActionTaken]] = {
    extractResult(multiSearchResponse.successes.head, tenantId, elasticSettings, elasticClient)
  }
}
