/*
 * $Id$
 *
 * Copyright (c) 2021 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.df.uba.server.common.elastic.invoker

import com.netapp.df.uba.server.common.elastic.ElasticSettings
import com.netapp.df.uba.server.common.elastic.aggregations.TermsAggregationInfo
import com.netapp.df.uba.server.common.elastic.utils.FieldConstants._
import com.netapp.df.uba.server.common.misc.dbmodel._
import com.sksamuel.elastic4s.ElasticApi.termsQuery
import com.sksamuel.elastic4s.http.ElasticDsl.BuildableTermsNoOp
import com.sksamuel.elastic4s.http.search.{MultiSearchResponse, SearchResponse, TermsAggResult}
import com.sksamuel.elastic4s.http.{ElasticClient, SourceAsContentBuilder}
import com.sksamuel.elastic4s.searches.queries.Query
import spray.json._

import scala.util.{Failure, Success}

/**
  * Class representing the action query invoker for UserActions
  *
  * @param userIdSet Set of userIds on which last actions are queried
  */
case class LastUserActionInvoker(userIdSet: Set[String]) extends LastActionInvoker {
  override def alertIds: Set[String] = Set.empty[String]

  override def userIds: Set[String] = userIdSet

  override def actionByObject: ObjectAction.Action = ObjectAction.UserAction

  override def isEmpty: Boolean = userIds.isEmpty

  // Actions are first filtered by userIds
  override val baseQuery: Query = termsQuery("userId", userIds)

  // Group by userIds to get the last actions for each user id
  def mainAggregationInfo: TermsAggregationInfo = {
    TermsAggregationInfo("userId", userIds.size, subAggregationInfoOption = Some(actionTypeTermsAggregationInfo))
  }

  override def getFinalResultMap(alertIdToResultMap: Map[String, List[ActionTaken]],
                                 userIdToResultMap: Map[String, List[ActionTaken]]): Map[String, List[ActionTaken]] = userIdToResultMap

  // Extracting the last user action results
  def extractResult(searchResponse: SearchResponse, tenantId: String, elasticSettings: ElasticSettings, elasticClient: ElasticClient): Map[String, List[ActionTaken]] = {
    import alertActionsJsonFormats._

    val mainAggregationResultOption: Option[TermsAggResult] = mainAggregationInfo
      .extractAggregationResult(tenantId, searchResponse)

    val actionTakenMapByUserId = if (mainAggregationResultOption.isEmpty) {
      Map.empty[String, List[ActionTaken]]
    } else {
      val mainAggregationResult = mainAggregationResultOption.get

      val mainAggregationBuckets = mainAggregationResult
        .buckets

      mainAggregationBuckets
        .map(userIdBucket => {
          val actionTypeBuckets = userIdBucket
            .terms(actionTypeTermsAggregationInfo.aggregationName)
            .buckets
          // Sort based on endTimestamp to iterate from recent action to old ones
          val userActionResultDbSorted: List[AlertActionResultDb] = actionTypeBuckets
            .map(actionTypeBucket => {
              // Get the most recent action for each actionType
              val actionResultTopHitSource = actionTypeBucket
                .tophits(topHitsAggregationInfo.aggregationName)
                .hits
                .head
                .source
              val sourceAsString: String = SourceAsContentBuilder(actionResultTopHitSource).string()

              sourceAsString.parseJson.convertTo[AlertActionResultDb]
            })
            .sortBy(_.endTimestamp.get)(Ordering[Long].reverse)
            .toList
            .take(1)

          val actionList = userActionResultDbSorted
            .map(lastActionResultDb => {
              lastActionResultDb.actionType match {
                case ActionType.LIMIT_USER_ACCESS | ActionType.RESTORE_USER_ACCESS=>
                  getLimitUserResult(lastActionResultDb, tenantId, elasticSettings, elasticClient)

                case otherActionType => // other actionTypes are not implemented yet
                  ActionTaken(otherActionType, None)
              }
            })

          userIdBucket.key -> actionList
        })
        .toMap
    }
    actionTakenMapByUserId
  }

  // Extract result from multi search response
  override def extractResult(multiSearchResponse: MultiSearchResponse, tenantId: String, elasticSettings: ElasticSettings, elasticClient: ElasticClient): Map[String, List[ActionTaken]] = {
    extractResult(multiSearchResponse.successes.head, tenantId, elasticSettings, elasticClient)
  }
}
