/*
 * $Id$
 *
 *  Copyright (c) 2019 NetApp, Inc.
 *  All rights reserved.
 */
package com.netapp.df.uba.server.common.elastic.model

import com.netapp.df.uba.server.common.elastic.exceptions.ValidationException
import com.netapp.df.uba.server.common.elastic.utils.FieldConstants.MaxSearchIntervalDuration
import com.netapp.df.uba.server.common.misc.utils.GenericUtils

/**
  * Search time range enum. Used in indexing queries and searching
  * queries for time range bucket support.
  */
object SearchTimeRange {

  private final val ThreeHoursString = "THREE_HOURS"
  private final val OneDayString = "ONE_DAY"
  private final val ThreeDaysString = "THREE_DAYS"
  private final val OneWeekString = "ONE_WEEK"
  private final val FourteenDaysString = "FOURTEEN_DAYS"
  private final val ThirtyDaysString = "THIRTY_DAYS"
  private val SupportedTimeRanges = List(ThreeHoursString, OneDayString, ThreeDaysString,
    OneWeekString, ThirtyDaysString)

  val ThreeHours: SearchTimeRange = SearchTimeRange(3L * 60 * 60 * 1000)
  val OneDay: SearchTimeRange = SearchTimeRange(24L * 60 * 60 * 1000)
  val ThreeDays: SearchTimeRange = SearchTimeRange(3L * 24 * 60 * 60 * 1000)
  val OneWeek: SearchTimeRange = SearchTimeRange(7L * 24 * 60 * 60 * 1000)
  val FourteenDays: SearchTimeRange = SearchTimeRange(14L * 24 * 60 * 60 * 1000)
  val ThirtyDays: SearchTimeRange = SearchTimeRange(30L * 24 * 60 * 60 * 1000)

  // This apply will convert from the supported strings to an interval in milliseconds
  def apply(searchTimeRangeString: String): SearchTimeRange = searchTimeRangeString match {
    case ThreeHoursString => ThreeHours
    case OneDayString => OneDay
    case ThreeDaysString => ThreeDays
    case OneWeekString => OneWeek
    case FourteenDaysString => FourteenDays
    case ThirtyDaysString => ThirtyDays
    case _ => throw new ValidationException(s"Unsupported time range given, supported time ranges are [${SupportedTimeRanges.mkString(", ")}]")
  }
}

/**
  * fromTimeDelta is the delta used to calculate the time range in milliseconds.
  * A set of pre-defined deltas will be used like three hours, 1 day etc.
  */
case class SearchTimeRange private(fromTimeDelta: Long) {

  /**
    * Get the from time, given the to time based on the delta defined in this instance
    */
  def getFromTime(toTimeEpoch: Long): Long = getFromTime(toTimeEpoch, 0)

  /**
    * Calculate the fromTime of a time range based on specified 'toTime' and 'fromTimeDeltaOverride'
    * Used to enforce max time range on REST side
    *
    * @param toTimeEpoch           the date to get range for (toTime)
    * @param fromTimeDeltaOverride the time range in milliseconds to override the predefined fromTimeDelta. null means no override
    * @return the 'fromTime' for range ending in 'toTime'
    */
  def getFromTime(toTimeEpoch: Long, fromTimeDeltaOverride: Long): Long = {
    val fromTimeEpoch = if (fromTimeDeltaOverride > 0) fromTimeDeltaOverride else fromTimeDelta

    toTimeEpoch - fromTimeEpoch
  }
}

/**
  * Object for TimeFilter to handle the conversion from query parameters in the API to filter objects
  */
object TimeFilter {
  def apply(fromTimeOption: Option[Long], toTimeOption: Option[Long], searchTimeRangeStringOption: Option[String]): TimeFilter = {

    val currentTimeMillis = GenericUtils.epochMills(GenericUtils.currentTimestamp)

    val maxSearchTimeIntervalInMillis = MaxSearchIntervalDuration.toMillis

    // If both modes i.e searchTimeRange(eg: Last 30days) and timeFilter are defined,
    // direct time in terms of timeFilter with fromTime and toTime are given higher priority
    val timeFilter = (fromTimeOption, toTimeOption) match {
      case (Some(fromTime), Some(toTime)) =>
        if (fromTime <= toTime && toTime - fromTime <= maxSearchTimeIntervalInMillis) TimeFilter(fromTime, toTime)
        else {
          require(requirement = false, "Invalid time filter; fromTime is greater than toTime or timeRange is not within supported range of 32 days.")
          TimeFilter(0L, 0L)
        }

      case (Some(fromTime), None) =>
        if (fromTime <= currentTimeMillis) TimeFilter(fromTime, fromTime + maxSearchTimeIntervalInMillis)
        else {
          require(requirement = false, "Invalid fromTime filter, fromTime cannot be ahead of currentTime")
          TimeFilter(0L, 0L)
        }

      case (None, Some(toTime)) =>
        if (toTime <= currentTimeMillis) TimeFilter(toTime - maxSearchTimeIntervalInMillis, toTime)
        else {
          require(requirement = false, "Invalid toTime filter, toTime cannot be ahead of currentTime")
          TimeFilter(0L, 0L)
        }

      case (None, None) =>
        if (searchTimeRangeStringOption.isDefined) {
          val toTime = currentTimeMillis
          val searchTimeRange = SearchTimeRange(searchTimeRangeStringOption.get)
          val fromTime = searchTimeRange.getFromTime(toTime)
          TimeFilter(fromTime, toTime)
        } else {
          TimeFilter(currentTimeMillis - maxSearchTimeIntervalInMillis, currentTimeMillis)
        }

      case _ =>
        require(requirement = false, "Invalid time filter applied.")
        // This will not be executed when require has already thrown exception
        TimeFilter(0L, 0L)
    }
    timeFilter
  }
}

/**
  * A case class representing the time filter to be applied to the queries or timeseries data
  */
case class TimeFilter(fromTime: Long, toTime: Long)
