/*
 * $Id$
 *
 * Copyright (c) 2021 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.df.uba.server.common.elastic.utils

import com.netapp.df.uba.server.common.elastic.ElasticUtils.AlertActionResultsIndexReadAlias
import com.netapp.df.uba.server.common.elastic._
import com.netapp.df.uba.server.common.elastic.aggregations.TermsAggregationInfo
import com.netapp.df.uba.server.common.misc.dbmodel.ActionType.ActionType
import com.netapp.df.uba.server.common.misc.dbmodel._
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.utils.GenericUtils
import com.sksamuel.elastic4s.http.ElasticClient
import com.sksamuel.elastic4s.http.ElasticDsl.{must, termQuery, _}
import com.sksamuel.elastic4s.http.search.MultiSearchResponse
import com.sksamuel.elastic4s.searches.sort.{FieldSort, SortOrder}

import java.time.Instant
import scala.concurrent.duration.{Duration, DurationInt}
import scala.util.{Failure, Success, Try}

/**
  * Utility to manage action related operations
  */
object ActionUtils extends LogTrait {
  val MaxActivityDuration: Duration = 1.day

  /**
    * A set of IPs fetched from different sources like
    * last limit user access, active alerts of the user and from performed activities
    */
  def getClientIps(elasticSettings: ElasticSettings,
                   elasticClient: ElasticClient,
                   tenantId: String,
                   userId: String,
                   activityInterval: Duration = MaxActivityDuration): Try[Set[String]] = {

    val activityIntervalApplied = if (activityInterval.toMillis > MaxActivityDuration.toMillis) MaxActivityDuration else activityInterval
    // Get last performed action
    val lastActionSearchQuery = must(
      termQuery("userId", userId),
      termQuery("actionType", s"${ActionType.LIMIT_USER_ACCESS}"),
      existsQuery("endTimestamp")
    )
    /*
     * Sample ES query
     GET alert-action-results/_search?size=1
      {
        "query": {
          "bool": {
            "must": [
              {
                "term": {
                  "userId": {
                    "value": "56558FCAEB5E93CCFE49A47D7704C7B1AE2161CFC0421FC9D7C19643FDA7ADF6"
                  }
                }
              },
              {
                "exists" : {
                  "field": "endTimestamp"
                }
              }
            ]
          }
        },
        "_source": ["actionType","subActions"],
        "sort": [
          {
            "startTimestamp": {
              "order": "desc"
            }
          }
        ]
      }
     */
    val lastActionSearchRequest = ElasticSearchHandler
      .createBasicSearchRequest(
        lastActionSearchQuery,
        AlertActionResultsIndexReadAlias(elasticSettings),
        tenantId
      )
      .sortBy(FieldSort("startTimestamp").order(SortOrder.DESC))
      .size(1)

    // Activities filtered by user id
    val activitiesByUserIdSearchQuery = must(
      termQuery("userId", userId),
      rangeQuery("eventTimestamp")
        .gte(Instant.now().toEpochMilli - activityIntervalApplied.toMillis)
    )

    val groupByClientIpAggInfo = TermsAggregationInfo("clientIpAddress", size = 100)

    val auditEventsTargetIndices: List[String] = getAuditEventsTargetIndices(elasticClient,
      elasticSettings,
      activityInterval,
      tenantId)

    val activitiesByUserIdSearchRequest = ElasticSearchHandler
      .createBasicSearchRequest(
        activitiesByUserIdSearchQuery,
        ElasticUtils.LatestIndexAlias(MappingTemplates.INDEX_AUDIT_EVENTS_KEY),
        tenantId,
        sourceIncludeList = List("clientIpAddress"),
        indicesOption = Some(auditEventsTargetIndices)
      )
      .size(0)
      .aggs(groupByClientIpAggInfo.buildAggregationSnippet())

    // Build a multi search query to fetch Ips from multiple sources
    val multiSearchRequest = multi(lastActionSearchRequest, activitiesByUserIdSearchRequest)

    val multiSearchResponseTry = ElasticSearchHandler.runQuery(tenantId, s"Executing multi-search query to get clientIps out of activities and last action for userId $userId", elasticClient, multiSearchRequest)

    multiSearchResponseTry match {
      case Success(multiSearchResponse) if multiSearchResponse.failures.isEmpty =>
        Try(extractClientIps(multiSearchResponse, tenantId, groupByClientIpAggInfo))

      case Success(multiSearchResponse) if multiSearchResponse.failures.nonEmpty =>
        val errorMessage = s"Error while executing multi-search request ${multiSearchRequest.show}. One or more query requests failed while executing elasticsearch multi-search."
        logErrorT(tenantId, errorMessage)
        Failure(new Exception(errorMessage))

      case Failure(ex) =>
        logErrorT(tenantId, s"Failed to execute elasticsearch multi-search request to get clientIps out of activities and last action for userId $userId", ex)
        Failure(ex)
    }
  }

  private def extractClientIps(multiSearchResponse: MultiSearchResponse, tenantId: String, clientIpAggInfo: TermsAggregationInfo): Set[String] = {
    import spray.json._

    implicit val alertActionsJsonFormats: AlertActionsJsonFormats = new AlertActionsJsonFormats
    import alertActionsJsonFormats._

    implicit val reducedActionDbFormatter: RootJsonFormat[ReducedActionDb] = jsonFormat2(ReducedActionDb.apply)
    val successList = multiSearchResponse.successes

    // responses from multi-search response
    val (lastActionResponse, activitiesResponse) = (successList.head, successList(1))

    // Get IPs from last action
    val actionResponseHits = lastActionResponse.hits.hits.toList
    val reducedActionDbList = actionResponseHits.map(_.sourceAsString.parseJson.convertTo[ReducedActionDb])
    val clientIpSetFromAction: Set[String] = if (reducedActionDbList.nonEmpty) {
      reducedActionDbList
        .head
        .subActions
        .head
        .actionResult
        .get
        .asInstanceOf[UserAccessDetails]
        .userAccessResultDb
        .userIps
        .getOrElse(List.empty)
        .toSet
    }
    else Set.empty

    // Get IPs from activity aggregation
    val clientIpAggregationOption = clientIpAggInfo
      .extractAggregationResult(tenantId, activitiesResponse)
    // No result found
    val clientIpSetFromActivities: Set[String] = if (clientIpAggregationOption.isEmpty) {
      Set.empty
    }
    else {
      val clientIpMainAggregation = clientIpAggregationOption.get

      clientIpMainAggregation
        .buckets
        .map(_.key)
        .toSet
    }

    // Combine IPs from all sources and return
    clientIpSetFromAction ++ clientIpSetFromActivities
  }

  /**
    * Find the target audit-events indices for given duration
    *
    * @param elasticClient    Elasticsearch client
    * @param elasticSettings  Elasticsearch settings
    * @param activityInterval Given duration to find target indices
    * @param tenantId         Tenant Id
    * @return List of audit-events indices
    */
  private def getAuditEventsTargetIndices(elasticClient: ElasticClient,
                                          elasticSettings: ElasticSettings,
                                          activityInterval: Duration,
                                          tenantId: String): List[String] = {
    val auditEventsIndexBaseName = elasticSettings.indexMap(MappingTemplates.INDEX_AUDIT_EVENTS_KEY).baseName
    val currentTimeMillis = GenericUtils.epochMills(GenericUtils.currentTimestamp)
    val fromTime = currentTimeMillis - activityInterval.toMillis
    val toTime = currentTimeMillis
    ElasticIndexSelector
      .getListOfIndicesToSearch(elasticClient, auditEventsIndexBaseName, fromTime, toTime) match {
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed to find indices for the specified time range $fromTime to $toTime. Reason: ${ex.getMessage}")
        // Fall back to fetch recent indices based on days
        ElasticIndexSelector.getRecentListOfIndicesToSearch(auditEventsIndexBaseName, activityInterval.toDays.toInt)

      case Success(lisOfIndices) =>
        if (lisOfIndices.nonEmpty) {
          logInfoT(tenantId, s"Searching with indices within time range $fromTime to $toTime: ${lisOfIndices.mkString(", ")}")
          lisOfIndices
        } else {
          /*
            This is the case where the search for indices using _cat/indices was successful,
            but there were no index that matched the time criteria. So, this means search would yield 0 hits.
            However, layers of forensic above cannot handle empty indices. So, latest index is returned here.
            This would be a small search that will target just shards of a single index
           */
          List(ElasticUtils.LatestIndexAlias(auditEventsIndexBaseName))
        }
    }
  }

  /**
    * Transform deviceName to Error map to a more compact form. List is grouped by common error message (unknown devices are exception).<br>
    * Ex: [(d1 -> err1), (d2 -> err1), (d3 -> err2), (UNKNOWN/MISSING -> err3), (UNKNOWN/MISSING -> err4)]<br>
    * The above list of DeviceErrorDetail will be transformed as follows:<br>
    * output: [(d1, d2 -> err1), (d3 -> err2), (UNKNOWN/MISSING -> err3), (UNKNOWN/MISSING -> err4)]
    * @param deviceErrorDetails Input list of deviceErrorDetail object
    * @return Compact list of DeviceErrorDetail
    */
  def formatErrorMessageFromDeviceError(deviceErrorDetails: List[DeviceErrorDetail]): List[DeviceErrorDetail] = {
    val (unknownDeviceList, remainingList) = deviceErrorDetails
      .partition(_.deviceName == FieldConstants.UNKNOWN)
    val remainingProcessedList = remainingList
      .groupBy(_.error)
      .map(entry => {
        DeviceErrorDetail(entry._2.map(_.deviceName).mkString(", "), entry._1)
      })
      .toList

    remainingProcessedList ++ unknownDeviceList
  }

  // A reduced class representing truncated form of actionDb with selective source fields included
  case class ReducedActionDb(actionType: ActionType, subActions: List[SubAction])
}