/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */

package com.netapp.df.uba.server.common.elastic.utils

import com.netapp.df.uba.server.common.elastic.invoker.LastActionInvoker
import com.netapp.df.uba.server.common.elastic.{ElasticSearchHandler, ElasticSearchHandlerTrait, ElasticSettings}
import com.netapp.df.uba.server.common.misc.dbmodel._
import com.sksamuel.elastic4s.http.ElasticClient
import com.sksamuel.elastic4s.http.ElasticDsl.{MultiSearchHandler, RichRequest}
import com.sksamuel.elastic4s.http.search.MultiSearchResponse
import com.sksamuel.elastic4s.searches.MultiSearchRequest

import scala.util.{Failure, Success, Try}

trait CommonAlertQueryUtil {

  /**
    * Get details of actions performed on given forensic object
    *
    * @param elasticSettings   ElasticSettings
    * @param elasticClient     Elastic client
    * @param lastActionInvoker Last action query can be invoked by an Alert or User.
    * @param tenantId          Tenant id
    * @return Map of objectId to last actions on the object
    */
  def getLastPerformedActions(elasticSettings: ElasticSettings,
                              elasticClient: ElasticClient,
                              lastActionInvoker: LastActionInvoker,
                              tenantId: String): Try[Map[String, List[ActionTaken]]] = {

    val totalActionTypes = ActionType.values.size

    // Always check aggregation size is non-zero
    if (totalActionTypes > 0 && !lastActionInvoker.isEmpty) {
      // Build a multi search query based on the last action invoking object
      val lastActionSearchQuery: MultiSearchRequest = lastActionInvoker.buildSearchRequest(elasticSettings, tenantId)

      val responseTry: Try[MultiSearchResponse] = getESHandler
        .runQuery(tenantId, "Get latest actions per actionType", elasticClient, lastActionSearchQuery)

      responseTry match {
        case Success(response) if response.failures.isEmpty =>
          // Try to extract the result if all the queries are successful
          Try(lastActionInvoker.extractResult(response, tenantId, elasticSettings, elasticClient))

        case Success(response) if response.failures.nonEmpty =>
          val errorMessage = s"Error while executing multi-search request ${lastActionSearchQuery.show}. One or more query requests failed while executing elasticsearch multi-search."
          Failure(new Exception(errorMessage))

        case Failure(ex) =>
          Failure(new Exception("One or more query requests failed while executing elasticsearch multi-search.", ex))
      }

    } else Success(Map.empty[String, List[ActionTaken]])
  }

  // Overridden in UTs
  def getESHandler: ElasticSearchHandlerTrait = ElasticSearchHandler
}
