/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.actor

import akka.actor.{Actor, ActorLogging, ActorRef, Props}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.flink.cluster
import com.netapp.df.uba.server.common.flink.cluster.{FlinkApiCheckpointHandler, FlinkApiClusterHandler, FlinkApiJobHandler, FlinkApiSavepointHandler}
import com.netapp.df.uba.server.common.flink.settings.{FlinkCluster, FlinkJobConfig, FlinkJobStateControlSettings, FlinkSettings}
import com.netapp.df.uba.server.common.flink.utils.{CheckOrSavePointWithCreateTime, FlinkJobUtils, NoOpSavepointCallbackHandler, SavepointCallbackHandler}
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model._
import com.netapp.df.uba.server.common.misc.settings.CommonSettings

import java.io.File
import scala.util.{Failure, Success, Try}

/**
  * FlinkJobCreator handles the creation of a single flink job.
  */
object FlinkJobCreator extends LogTrait {
  final val ActorName: String = "flinkJobCreator"

  def props(requestorRef: ActorRef,
            flinkCluster: FlinkCluster,
            commonSettings: CommonSettings,
            flinkSettings: FlinkSettings,
            jobStateControlSettings: FlinkJobStateControlSettings,
            savepointCallbackHandler: SavepointCallbackHandler = new NoOpSavepointCallbackHandler): Props = {
    Props(
      new FlinkJobCreator(
        requestorRef,
        flinkCluster,
        commonSettings,
        flinkSettings,
        jobStateControlSettings,
        savepointCallbackHandler
      )
    )
  }

  // Message to be sent to this actor to manage a flink job, initial setup
  case class SetupFlinkJob(jobConfig: FlinkJobConfig,
                           inputJarFileOption: Option[File] = None,
                           lastSavepointOption: Option[CheckOrSavePointWithCreateTime] = None,
                           allowNonRestoredState: Boolean = false) extends Command

  // Message to be sent to this actor to take a savepoint of the job as an on-demand operation
  case class TakeSavePoint(jobId: String,
                           jobName: String,
                           checkpointDirPath: String) extends Command
}

class FlinkJobCreator(requestorRef: ActorRef,
                     flinkCluster: FlinkCluster,
                     commonSettings: CommonSettings,
                     flinkSettings: FlinkSettings,
                     jobStateControlSettings: FlinkJobStateControlSettings,
                     savepointCallbackHandler: SavepointCallbackHandler = new NoOpSavepointCallbackHandler)
  extends Actor
    with ActorLogging
    with FlinkJobUtils {

  import FlinkJobCreator._

  override def receive: Receive = LoggingReceive {
    case SetupFlinkJob(jobConfig, inputJarFileOption, lastSavepointOption, allowNonRestoredState) =>
      // Create a flink api objects
      val flinkApiJobHandler = createFlinkApiJobHandler()
      val flinkApiClusterHandler = createFlinkApiClusterHandler()
      val flinkApiCheckpointHandler = createFlinkApiCheckpointHandler(flinkApiJobHandler)

      // Lock name
      val lockName = flinkApiJobHandler.lockNameForJob(jobConfig.jobName)

      // Load the jar file
      val jarFileTry: Try[File] = loadJarFileInternal(jobConfig, inputJarFileOption)

      jarFileTry match {
        case Failure(exception) =>
          // Jar file loading failed, abort
          requestorRef ! ActionFailed(jobConfig.jobName, Error(ErrorCode.InternalError, exception.getMessage, jobConfig.jobName))

        case Success(jarFile) =>
          // Setup the job in flink cluster
          val setupJobResponseTry: Try[Boolean] = orchestrateSetupJobInFlinkCluster(
            jobConfig,
            lockName,
            flinkApiJobHandler,
            flinkApiClusterHandler,
            flinkApiCheckpointHandler,
            jobStateControlSettings.deploymentJobControl,
            jarFile,
            commonSettings,
            lastSavepointOption,
            savepointCallbackHandler,
            allowNonRestoredState
          )
          // Return status back to parent
          setupJobResponseTry match {
            case Success(_) =>
              requestorRef ! ActionSucceeded(jobConfig.jobName)

            case Failure(exception) =>
              log.error(s"Job: ${jobConfig.jobName} - Failed to setup the job in the cluster. Reason: ${exception.getMessage}")
              requestorRef ! ActionFailed(jobConfig.jobName, Error(ErrorCode.InternalError, exception.getMessage, jobConfig.jobName))
          }
      }

    case TakeSavePoint(jobId, jobName, checkpointDirPath) =>
      // Create the savepoint handler
      val flinkApiSavepointHandler = createFlinkApiSavepointHandler(jobName, checkpointDirPath)

      // Take a savepoint
      val savepointSucceededTry = flinkApiSavepointHandler.takeSavepoint(jobId)
      savepointSucceededTry match {
        case Success(_) => log.info(s"Job: $jobName($jobId) - Savepoint operation succeeded.")
        case Failure(ex) => log.error(s"Job: $jobName($jobId) - Savepoint operation failed. Reason: ${ex.getMessage}.")
      }
      context stop self

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message that is not handled: " + somethingElse)
  }

  private def loadJarFileInternal(jobConfig: FlinkJobConfig, inputJarFileOption: Option[File]): Try[File] = inputJarFileOption match {
    case Some(inputJarFile) => Success(inputJarFile)
    case None =>
      // If the caller did not send the jar file load it here
      loadJarFile(jobConfig, flinkSettings.jobJarDirectory)
  }

  // Create flink API objects
  protected def createFlinkApiJobHandler(): FlinkApiJobHandler = cluster.FlinkApiJobHandler(flinkCluster)

  protected def createFlinkApiClusterHandler() = FlinkApiClusterHandler(flinkCluster)

  protected def createFlinkApiCheckpointHandler(flinkApiJobHandler: FlinkApiJobHandler): FlinkApiCheckpointHandler = {
    cluster.FlinkApiCheckpointHandler(flinkCluster, flinkApiJobHandler)
  }

  protected def createFlinkApiSavepointHandler(jobName: String, checkpointDirPath: String) = FlinkApiSavepointHandler(
    flinkCluster,
    jobName,
    targetSavepointDirectory = checkpointDirPath,
    savepointCallbackHandler
  )
}