/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.actor

import akka.actor.{Actor, ActorLogging, ActorRef, Props, Timers}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.flink.cluster.FlinkApiJobHandler
import com.netapp.df.uba.server.common.flink.cluster.FlinkApiClusterHandler
import com.netapp.df.uba.server.common.flink.cluster.FlinkApiCheckpointHandler
import com.netapp.df.uba.server.common.flink.cluster.FlinkApiJobHandler._
import com.netapp.df.uba.server.common.flink.cluster.FlinkJob
import com.netapp.df.uba.server.common.flink.settings.FlinkSettings
import com.netapp.df.uba.server.common.flink.settings.FlinkJobStateControlSettings
import com.netapp.df.uba.server.common.flink.settings.FlinkJobConfig
import com.netapp.df.uba.server.common.flink.utils.{CheckOrSavePointWithCreateTime, FlinkJobUtils}
import com.netapp.df.uba.server.common.misc.model._
import com.netapp.df.uba.server.common.misc.settings.CommonSettings
import com.netapp.oci.synch.DistributedSynchronousTask
import com.netapp.oci.synch.DistributedSynchronousTask.TaskWithReturn

import java.io.File
import scala.util.{Failure, Success, Try}

/**
  * FlinkJobMonitor handles the monitoring of a single flink job.
  */
object FlinkJobMonitor {
  final val ActorNamePrefix: String = "flinkJobMonitor-"

  def props(jobConfig: FlinkJobConfig,
            commonSettings: CommonSettings,
            flinkSettings: FlinkSettings,
            jobStateControlSettings: FlinkJobStateControlSettings,
            requestorRef: ActorRef): Props = Props(
    new FlinkJobMonitor(
      jobConfig,
      commonSettings,
      flinkSettings,
      jobStateControlSettings,
      requestorRef))

  // Message to be sent to start job monitoring
  case object StartFlinkJobMonitoring
}

class FlinkJobMonitor(jobConfig: FlinkJobConfig,
                      commonSettings: CommonSettings,
                      flinkSettings: FlinkSettings,
                      jobStateControlSettings: FlinkJobStateControlSettings,
                      requestorRef: ActorRef) extends Actor
  with ActorLogging
  with Timers
  with FlinkJobUtils {

  import FlinkJobMonitor._

  private val monitorJobControl = jobStateControlSettings.monitorJobControl
  private val jobName = jobConfig.jobName

  override def receive: Receive = LoggingReceive {
    case StartFlinkJobMonitoring =>

      // Might come again if this actor was recreated after termination
      if (!timers.isTimerActive(PeriodicTimerKey))
        timers.startPeriodicTimer(PeriodicTimerKey, PeriodicTick, jobConfig.jobCheckInterval)

    case PeriodicTick =>
      monitorJob() match {
        case Success(_) => requestorRef ! ActionSucceeded(jobConfig.jobName)
        case Failure(ex) => requestorRef ! ActionFailed(jobConfig.jobName, Error(ErrorCode.InternalError, ex.getMessage, jobConfig.jobName))
      }

    case somethingElse => log.warning(s"${this.getClass.getSimpleName} actor received a message that is not handled: " + somethingElse)
  }

  /**
    * Utility method to do job monitoring
    */
  private def monitorJob(): Try[Boolean] = {
    // Create flink API objects
    val flinkApiJobHandler = createFlinkApiJobHandler()
    val flinkApiClusterHandler = createFlinkApiClusterHandler()
    val flinkApiCheckpointHandler = createFlinkApiCheckpointHandler(flinkApiJobHandler)

    // Lock name
    val lockName = flinkApiJobHandler.lockNameForJob(jobConfig.jobName)

    try {
      //noinspection ConvertExpressionToSAM
      new DistributedSynchronousTask(lockName).execute[Try[Boolean]](new TaskWithReturn[Try[Boolean]] {
        override def run(): Try[Boolean] = {

          try {
            // Load the job if present using the job id.
            val flinkJobTry: Try[FlinkJob] = flinkApiJobHandler.getFlinkJob(jobName)

            flinkJobTry match {
              case Failure(_) =>
                // Extract the retry count and sleep time between retries for flink cluster invocations
                val flinkClusterApiRetryCount = monitorJobControl.retryCount
                val sleepTimeBetweenRetriesMillis = monitorJobControl.sleepTime.toMillis.toInt

                /*
                  Failed to load job, retry for sometime in failure scenario, it could be because:
                  a) flink cluster is restarting,
                  b) flink HA is trying to reload the job
                 */
                val flinkConnectivityStatusTry = flinkApiClusterHandler.retryFlinkConnectivity(
                  jobName,
                  flinkClusterApiRetryCount,
                  sleepTimeBetweenRetriesMillis
                )

                flinkConnectivityStatusTry match {
                  case Failure(exception) => // Error out immediately by throwing an exception
                    log.error(s"Job: $jobName - Flink connectivity failed for the job. Reason: ${exception.getMessage}")
                    throw exception

                  case Success(_) => log.info(s"Job: $jobName - Flink connectivity successful for the flink job.")
                }

                // Check if the job already exists in the flink cluster, and if yes obtain the status of the same.
                val jobStatusTry = flinkApiJobHandler.retryFlinkJobStatus(
                  jobConfig.jobName,
                  flinkClusterApiRetryCount,
                  sleepTimeBetweenRetriesMillis
                )
                jobStatusTry match {
                  case Failure(exception) =>
                    log.error(s"Job: $jobName - Job does not exist. Setting up a new job. Reason: ${exception.getMessage}")
                    setupJob(flinkApiJobHandler, flinkApiCheckpointHandler)

                  case Success(jobStatusValue) =>
                    log.info(s"Job: $jobName - Job is in valid state $jobStatusValue")
                    Success(true)
                }

              /*
               Valid states from monitoring perspective are:
               RUNNING,SCHEDULED,CREATED RESTARTING CANCELLING SUSPENDING
               FAILING RECONCILING, FINISHED, CANCELLED
               */
              case Success(flinkJob) if flinkApiJobHandler.isJobStateValidForMonitoring(flinkJob.state.getOrElse("")) =>
                //Job is running and in a valid state
                log.info(s"Job: $jobName (${flinkJob.jid}) - Flink job has valid state for monitoring - ${flinkJob.state.getOrElse("")}")
                Success(true)

              /*
                When the job is in "Failed" state, we will check any recent checkpoint files are available for the job. If any check point file is
                available, we will get the recent checkpoint file and load the recent state of the job.
                We achieve the same during startup of the job in deployment phase of SFJM/TJM by overriding helm configuration for savepoint file
               */
              case Success(flinkJob) if jobConfig.jobSavePointEnabled && flinkJob.state.get.equals(JobStateFailed) =>

                val retainedCheckpointFiles: List[CheckOrSavePointWithCreateTime] = getRetainedCheckpointList(
                  jobConfig,
                  flinkApiCheckpointHandler
                )
                setupJob(flinkApiJobHandler, flinkApiCheckpointHandler, retainedCheckpointFiles)

              // Failure case just restart it.
              case Success(flinkJob) =>
                /*
                  Job exists, but not in the right state, usually this happen when the job is in stopped state.
                  Here job might have failed because of external factors like, kafka issue or schema registry issue
                  In this case SFJM/TJM will attempt to reload the job till it is successful.
                 */
                log.error(s"Job: $jobName (${flinkJob.jid}) - Flink job exists with state ${flinkJob.state.getOrElse("")}. Repeating the setup of the job.")
                setupJob(flinkApiJobHandler, flinkApiCheckpointHandler)
            }
          } catch {
            case exception: Exception =>
              log.error(s"Job: $jobName - Exception when checking if the job is running in the cluster. Reason: ${exception.getMessage}")
              Failure(exception)
          }
        }
      })
    } catch {
      case exception: Exception =>
        log.error(s"Job: $jobName - Failed to get distributed synchronous task for $lockName. Reason: ${exception.getMessage}")
        Failure(exception)
    }
  }

  // Method to setup a job
  private def setupJob(flinkApiJobHandler: FlinkApiJobHandler,
                       flinkApiCheckpointHandler: FlinkApiCheckpointHandler,
                       retainedCheckpointFiles: List[CheckOrSavePointWithCreateTime] = Nil): Try[Boolean] = {

    // Load the jar file
    val jarFileTry: Try[File] = loadJarFile(jobConfig, flinkSettings.jobJarDirectory)

    jarFileTry match {
      case Failure(jarFileEx) =>
        log.error(s"Job: $jobName - Jar file failed to load. Reason: ${jarFileEx.getMessage}")
        Failure(jarFileEx)

      case Success(jarFile) => // Setup the job in flink cluster
        setupJobWithOrWithoutCheckpoints(
          flinkApiJobHandler,
          flinkApiCheckpointHandler,
          jobConfig,
          jarFile,
          commonSettings,
          retainedCheckpointFiles
        )
    }
  }

  // Create flink API objects
  protected def createFlinkApiJobHandler(): FlinkApiJobHandler = FlinkApiJobHandler(flinkSettings.cluster)

  protected def createFlinkApiClusterHandler(): FlinkApiClusterHandler = FlinkApiClusterHandler(flinkSettings.cluster)

  protected def createFlinkApiCheckpointHandler(flinkApiJobHandler: FlinkApiJobHandler): FlinkApiCheckpointHandler = FlinkApiCheckpointHandler(flinkSettings.cluster, flinkApiJobHandler)
}