/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.actor

import akka.actor.{Actor, ActorLogging, ActorRef, Props, Terminated}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.flink.actor.FlinkJobCreator.SetupFlinkJob
import com.netapp.df.uba.server.common.flink.actor.FlinkJobMonitor.StartFlinkJobMonitoring
import com.netapp.df.uba.server.common.flink.settings.{FlinkJobConfig, FlinkJobStateControlSettings, FlinkSettings}
import com.netapp.df.uba.server.common.misc.model._
import com.netapp.df.uba.server.common.misc.settings.CommonSettings
import com.netapp.df.uba.server.common.misc.utils.ActorPathUtils

/**
  * Unused class as of now.
  *
  * This was used pre-pipeline model days where there was a
  * single Flink job per type running in a default flink cluster.
  *
  * This actor has code to instantiate the FlinkJobCreator per job
  * and also to run the FlinkJobMonitor for each job type
  */

object FlinkManager {
  final val ActorName: String = "flinkManager"

  def props(commonSettings: CommonSettings,
            flinkSettings: FlinkSettings,
            jobControlSettings: FlinkJobStateControlSettings,
            requestorRef: ActorRef): Props =
    Props(new FlinkManager(commonSettings, flinkSettings, jobControlSettings, requestorRef))

  // A message to send to this actor to start the flink job management
  case object ManageFlinkJobs
  case object ManageFlinkMonitoring

  // Context used for flink job management
  final val ContextFlinkManagement: String = "flinkJobs"
}

class FlinkManager(commonSettings: CommonSettings,
                   flinkSettings: FlinkSettings,
                   jobControlSettings: FlinkJobStateControlSettings,
                   requestorRef: ActorRef) extends Actor with ActorLogging {
  import FlinkManager._

  override def receive: Receive = initialMode

  // Start the job creation actors for flink jobs serially as currently submitting jobs in parallel is failing.
  private def initialMode: Receive = LoggingReceive {
    case ManageFlinkJobs =>

      // Manage all jobs
      if (flinkSettings.flinkJobs.nonEmpty) {
        /*
         Create the job creation actor, this is a pool
         but with only 1 worker since job creation currently does not work in parallel
        */
        val jobCreatorActorRef = createFlinkJobCreatorActor()

        // Watch the actor
        context watch jobCreatorActorRef

        // Send messages to create the first flink job so that they are executed serially
        val firstJobConfig = flinkSettings.flinkJobs.head
        jobCreatorActorRef ! SetupFlinkJob(firstJobConfig)

        // Change context to wait for result of the first job creation
        context.become(createWaitMode(firstJobConfig.jobName, Map.empty, jobCreatorActorRef, flinkSettings.flinkJobs.drop(1)))
      } else {
        // Nothing to do for FlinkManager, so return success to AppManager actor

        // Indicate success of creation of all flink jobs to the parent
        requestorRef ! ActionSucceeded(ContextFlinkManagement)

        // Move to running mode
        log.debug(s"Moving to running mode as all job setup is completed.")
        context.become(runningMode)
      }

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message in initialMode that is not handled: " + somethingElse)
  }

  /**
    * Receive mode where FlinkManager is waiting for flink jobs to be set up
    */
  private def createWaitMode(currentJobName: String,
                             jobResultMap: Map[String, ActionStatus],
                             jobCreatorActorRef: ActorRef,
                             remainingJobs: List[FlinkJobConfig]): Receive = LoggingReceive {

    // Receive the status of every job handler action
    case actionStatus@ActionSucceeded(receivedJobName) =>
      if (receivedJobName != currentJobName) {
        log.warning(s"${this.getClass.getSimpleName} actor received a message in createWaitMode that is not handled: " + actionStatus)
      }
      else {
        // Evaluate the result and decide whether to remain in context or switch
        evaluateResult(
          jobResultMap,
          actionStatus,
          jobCreatorActorRef,
          remainingJobs)
      }

    case actionStatus@ActionFailed(receivedJobName, _) =>
      if (receivedJobName != currentJobName) {
        log.warning(s"${this.getClass.getSimpleName} actor received a message in createWaitMode that is not handled: " + actionStatus)
      } else {
        // Evaluate the result and decide whether to remain in context or switch
        evaluateResult(
          jobResultMap,
          actionStatus,
          jobCreatorActorRef,
          remainingJobs)
      }

    // job handler child has terminated
    case Terminated(jobCreatorChild) if jobCreatorChild.path.name.contains(FlinkJobCreator.ActorName) =>

      val errorMessage = s"FlinkJobCreator actor (${jobCreatorChild.path.toString}) has terminated."
      log.error(errorMessage)

      // Mark as failed for all remaining jobs
      val remainingJobStatusesMap: Map[String, ActionFailed] = flinkSettings.flinkJobs
        .filterNot(job => jobResultMap.contains(job.jobName))
        .map(job => job.jobName -> ActionFailed(job.jobName, Error(ErrorCode.InternalError, errorMessage, job.jobName)))
        .toMap

      val newJobResultMap = jobResultMap ++ remainingJobStatusesMap

      // create response action
      createResponseAction(newJobResultMap)

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message in createWaitMode that is not handled: " + somethingElse)
  }

  // Running mode where status of job management is received
  private def runningMode: Receive = LoggingReceive {

    // Receive the status of every job management periodic case
    case ManageFlinkMonitoring =>
      // Create a monitor actor for every job
      flinkSettings
        .flinkJobs
        .map(job => {
          // Create the job monitor actor
          val jobMonitor = createFlinkJobMonitorActor(job)
          // Watch the actor
          context watch jobMonitor

          // Send the message to setup the job
          jobMonitor ! StartFlinkJobMonitoring
          // Map entry
          job.jobName -> jobMonitor
        })

    // Receive the status of every job monitor periodic activity and log it
    case actionStatus: ActionStatus =>
      actionStatus match {
        case ActionSucceeded(jobName) =>
          log.info(s"Received successful flink job monitoring message for $jobName.")

        case ActionFailed(jobName, error) =>
          log.error(s"Received failure in flink job monitoring for $jobName. ErrorCode: ${error.code}, Message: ${error.message}")
      }

    // Flink job handler actor has terminated
    case Terminated(jobMonitorRef) =>
      val jobName = ActorPathUtils.suffixNameFromPath(jobMonitorRef)

      log.error(s"FlinkJobMonitor actor (${jobMonitorRef.path.toString}) has terminated.")

      // Recreate the flink job monitor actor
      createFlinkJobMonitorActor(flinkSettings.flinkJobs.find(_.jobName == jobName).get)

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message in runningMode that is not handled: " + somethingElse)
  }

  // Utility method
  private def evaluateResult(jobResultMap: Map[String, ActionStatus],
                             actionStatus: ActionStatus,
                             jobCreatorActorRef: ActorRef,
                             remainingJobs: List[FlinkJobConfig]): Unit = {

    actionStatus match {
      case ActionSucceeded(jobName) =>
        log.info(s"Received status as succeeded for $jobName")

        // Add the result to the jobResultMap with key as job name
        val newJobResultMap = jobResultMap ++ Seq(actionStatus.context -> actionStatus)

        // Check if more are to be waited for
        remainingJobs.size - 1 match {
          // If all the jobs are accounted for, check on next steps using createResultAction method
          case -1 =>
            // Stop the job creation actor
            context unwatch jobCreatorActorRef
            context stop jobCreatorActorRef

            // Create appropriate response
            createResponseAction(newJobResultMap)

          // Need to wait for the result of more flink jobs
          case _ =>
            // Give one more job to create to the job creator
            val nextJobConfig = remainingJobs.head
            jobCreatorActorRef ! SetupFlinkJob(nextJobConfig)
            context.become(
              createWaitMode(
                nextJobConfig.jobName,
                newJobResultMap,
                jobCreatorActorRef,
                remainingJobs.drop(1)))
        }

      case ActionFailed(jobName, error) =>

        val failedJobErrorMessage = s"Received status as failed for $jobName. Error: $error."
        log.error(failedJobErrorMessage)

        val newJobResultMapWithJob = jobResultMap ++ Map(jobName -> actionStatus)

        // Mark as failed for all remaining jobs as there is no point in trying the other ones
        val remainingJobsErrorMessage = s"Not creating job since a previous job $jobName has failed."
        val remainingJobStatusesMap: Map[String, ActionFailed] = remainingJobs
          .map(job => job.jobName -> ActionFailed(job.jobName, Error(ErrorCode.InternalError, remainingJobsErrorMessage, job.jobName)))
          .toMap

        val finalJobResultMap = newJobResultMapWithJob ++ remainingJobStatusesMap

        // create response action
        createResponseAction(finalJobResultMap)
    }
  }

  // Utility method
  protected def createResponseAction(jobResultMap: Map[String, ActionStatus]): Unit = {

    // Check if job creation succeeded for all
    if (jobResultMap.values.forall(_.isInstanceOf[ActionSucceeded])) {

      // Indicate success of creation of all flink jobs to the parent
      requestorRef ! ActionSucceeded(ContextFlinkManagement)

      // Move to running mode and handle monitoring of jobs
      self ! ManageFlinkMonitoring
      log.debug(s"Moving to running mode as all job setup is completed.")

      // Move to running mode
      context.become(runningMode)

    } else {
      // Get the list of all errors in job creation
      val allErrors = jobResultMap
        .values
        .filterNot(_.isInstanceOf[ActionSucceeded])
        .map(_.asInstanceOf[ActionFailed].error)
        .toList

      // Create an error object
      val errorObject = Error(ErrorCode.InternalError, "Flink job management failed.", ContextFlinkManagement, Some(allErrors))

      // Indicate to parent about failure and go back to initialMode
      requestorRef ! ActionFailed(ContextFlinkManagement, errorObject)

      context.become(initialMode)
    }
  }

  // actor creation separated out so that unit tests can override it
  protected def createFlinkJobCreatorActor(): ActorRef = {
    context.actorOf(
      FlinkJobCreator.props(
        self,
        flinkSettings.cluster,
        commonSettings,
        flinkSettings,
        jobControlSettings,
      ),
      FlinkJobCreator.ActorName
    )
  }

  protected def createFlinkJobMonitorActor(job: FlinkJobConfig): ActorRef = {
    context.actorOf(
      FlinkJobMonitor.props(
        job,
        commonSettings,
        flinkSettings,
        jobControlSettings,
        self
      ),
      s"${FlinkJobMonitor.ActorNamePrefix}${job.jobName}"
    )
  }
}
