/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.cluster

import akka.http.scaladsl.model.StatusCodes
import com.netapp.df.uba.server.common.flink.cluster.FlinkApiJobHandler.TimeoutValueInMillis
import com.netapp.df.uba.server.common.flink.settings.FlinkCluster
import com.netapp.df.uba.server.common.misc.log.LogTrait
import org.apache.http.client.config.RequestConfig
import org.apache.http.client.methods.HttpGet
import org.apache.http.impl.client.{CloseableHttpClient, HttpClients}
import org.apache.http.util.EntityUtils
import org.apache.http.{Header, HttpHost, HttpRequest}

import java.net.{MalformedURLException, URL}
import scala.util.Try

// A flink exception
case class FlinkProcessingException(message: String) extends Exception(message)

trait FlinkApiBaseUtils
  extends LogTrait
  with FlinkApisJsonFormats {
  /**
    * Utility to create a HTTP client with timeout configurations
    */
  def createHttpClientWithTimeouts(disableRedirects: Boolean = false): CloseableHttpClient = {
    // Some timeout configurations
    val requestConfig = RequestConfig.custom()
      .setConnectionRequestTimeout(TimeoutValueInMillis)
      .setConnectTimeout(TimeoutValueInMillis)
      .setSocketTimeout(TimeoutValueInMillis)
      .build()

    // Create a HTTP client
    val httpClientBuilder = HttpClients.custom()
      .setDefaultRequestConfig(requestConfig)

    if (disableRedirects) httpClientBuilder.disableRedirectHandling

    httpClientBuilder
      .build()
  }

  /**
    * Utility to close the HTTP client
    */
  def closeHttpClient(httpClient: CloseableHttpClient): Unit = {
    try {
      // Close the client
      httpClient.close()
    } catch {
      case closeEx: Exception =>
        // Just log the close issue
        logger.error(s"Failed closing connection. Reason: ${closeEx.getMessage}")
    }
  }

  /**
    * Utility method to make a GET HTTP call and get the response in string format
    */
  def getHttpResponseOfURI(flinkCluster: FlinkCluster,
                           uri: String,
                           jobId: String,
                           jobName: String,
                           operationContext: String): Try[String] = Try {

    val httpClient = createHttpClientWithTimeouts()

    val httpRequest = new HttpGet(uri)
    logger.info(s"URI of GET: ${httpRequest.getURI.toString} for: $jobName with job id :$jobId")

    try {
      getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        jobId,
        jobName,
        operationContext
      )
    } finally {
      closeHttpClient(httpClient)
    }
  }

  /**
    * Utility method to make a GET HTTP call and get the response in string format
    */
  def getHttpResponseOfURI(flinkCluster: FlinkCluster,
                           httpClient: CloseableHttpClient,
                           httpRequest: HttpRequest,
                           jobId: String,
                           jobName: String,
                           context: String): String = {
    logger.info(s"Job: $jobName ($jobId) - $context. URI: ${httpRequest.getRequestLine}")

    try {
      // Execute the GET call
      val httpHost = checkAndGetActiveJobManagerHost(flinkCluster)
      val httpResponse = httpClient.execute(httpHost, httpRequest)

      if (httpResponse.getStatusLine == null || (
          httpResponse.getStatusLine.getStatusCode != StatusCodes.OK.intValue &&
            httpResponse.getStatusLine.getStatusCode != StatusCodes.Accepted.intValue)) {

        if (httpResponse.getEntity != null) EntityUtils.consumeQuietly(httpResponse.getEntity)
        throw FlinkProcessingException(s"statusLine: ${httpResponse.getStatusLine}")

      } else {
        val responseBody = EntityUtils.toString(httpResponse.getEntity, "UTF-8")

        logger.info(s"Job: $jobName ($jobId) - Response while performing '$context': $responseBody")
        if (httpResponse.getEntity != null) EntityUtils.consumeQuietly(httpResponse.getEntity)

        responseBody
      }
    } catch {
      case ex: Exception =>
        val errorMessage = s"Job: $jobName ($jobId) - Failed processing HTTP request while performing '$context'. Reason: ${ex.getMessage}."
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)
    }
  }

  /**
    * Figure out what is an active JobManager host, for HA setups if you query one host you might be redirected (307) to another
    *
    * @return the active JobManager HttpHost
    * @throws FlinkProcessingException on failure to get active JobManager
    */
  protected def checkAndGetActiveJobManagerHost(cluster: FlinkCluster): HttpHost = {
    val givenHttpHost = new HttpHost(
      cluster.host,
      cluster.port,
      cluster.protocol.toString
    )
    // URL of / for GET call
    val httpRequest = new HttpGet("/")

    // disable redirects so we can detect redirected requests in the HTTP client
    val httpClient = createHttpClientWithTimeouts(disableRedirects = true)

    try {
      val httpResponse = httpClient.execute(givenHttpHost, httpRequest)

      // temporarily moved, that's how Flink makes us know about active server
      val httpResponseStringTry = if (httpResponse.getStatusLine == null || httpResponse.getStatusLine.getStatusCode == StatusCodes.TemporaryRedirect.intValue) { // 307
        logger.error(s"Active job manager host $givenHttpHost check returned HTTP status code - 307 Redirect.")

        // Extract the location header
        val locationHeaders: Array[Header] = httpResponse.getHeaders("Location")
        if (locationHeaders == null || locationHeaders.isEmpty) {

          if (httpResponse.getEntity != null) EntityUtils.consumeQuietly(httpResponse.getEntity)
          throw FlinkProcessingException(s"Could not find active JobManager for $givenHttpHost, empty Location header returned.")

        } else {
          val activeJobManagerLocationHeader = locationHeaders.head

          try {
            val activeJobManagerLocationURL = new URL(activeJobManagerLocationHeader.getValue)

            val activeJobManagerHttpHost = new HttpHost(
              activeJobManagerLocationURL.getHost,
              activeJobManagerLocationURL.getPort,
              cluster.protocol.toString
            )
            logger.info(s"Active job manager host after redirection from $givenHttpHost is - $activeJobManagerHttpHost")
            activeJobManagerHttpHost
          } catch {
            case ex: MalformedURLException =>
              if (httpResponse.getEntity != null) EntityUtils.consumeQuietly(httpResponse.getEntity)
              throw FlinkProcessingException(s"Could not find active JobManager from location: ${activeJobManagerLocationHeader.getValue}. Reason: ${ex.getMessage}")
          }
        }
      } else {
        logger.info(s"Active job manager host is the given one: $givenHttpHost")
        givenHttpHost
      }
      EntityUtils.consumeQuietly(httpResponse.getEntity)
      httpResponseStringTry
    } catch {
      case ex: Exception =>
        logger.error(s"Failure in HTTP call to flink cluster to get active job manager. Reason: ${ex.getMessage}")
        throw ex
    } finally {
      // Close the HTTP client
      closeHttpClient(httpClient)
    }
  }

  // Utility to try multiple times
  def resultAfterAttempts[T](waitTries: Int,
                             conditionCheck: (CloseableHttpClient, String) => Try[T],
                             httpClient: CloseableHttpClient,
                             conditionParameter: String,
                             checkForSuccess: Boolean = true): Boolean = {
    // An initial sleep
    Thread.sleep(2000)

    var tryNumber = waitTries
    while (tryNumber > 0 && (if (checkForSuccess) conditionCheck(httpClient, conditionParameter).isSuccess else conditionCheck(httpClient, conditionParameter).isFailure)) {
      Thread.sleep(2000)
      tryNumber = tryNumber - 1
    }
    tryNumber > 0
  }
}
