/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.cluster

import com.netapp.df.uba.server.common.flink.settings.FlinkCluster
import com.netapp.df.uba.server.common.misc.log.LogTrait
import spray.json._

import scala.util.{Failure, Success, Try}

/**
  * Set of methods to deal with APIs on checkpoints on the Flink cluster
  * First three methods are invoked from SFJM, rest are supporting private methods in this class.
  */
case class FlinkApiCheckpointHandler(flinkCluster: FlinkCluster,
                                     flinkApiJobHandler: FlinkApiJobHandler)
  extends LogTrait
    with FlinkApiBaseUtils
    with FlinkApisJsonFormats {

  protected final val EmptyCheckpointResponse = "Empty Checkpoint response"

  /**
    * Method to wait until the next checkpoint of the job finishes.
    * This is to ensure that when SFJM takes a savepoint and redeploys the job,
    * there is no delay or conflicts. In case there are any problems while retrieving
    * checkpoint details, this will throw relevant error
    */
  def waitForNextCheckpointToFinish(jobName: String): Try[Boolean] = Try {

    val sleepTimeBetweenRetriesInMillis = 5000

    // Load the flink job
    val flinkJobTry = flinkApiJobHandler.getFlinkJob(jobName)

    flinkJobTry match {
      case Success(flinkJob) =>
        val jobId = flinkJob.jid

        // Get the checkpoint configuration of the job, i.e is the stateful job having 2min or 5min checkpointing interval
        val checkpointConfigTry = getCheckpointConfigurationOfJob(jobId, jobName)

        checkpointConfigTry match {
          case Success(checkpointConfig) =>
            verifyCheckpointProgressOfTheJob(
              jobId,
              jobName,
              checkpointConfig.interval,
              sleepTimeBetweenRetriesInMillis
            )
          case Failure(ex) => throw ex
        }
      case Failure(ex) =>
        val errorMessage = s"Job: $jobName - Failed to load the job in order to wait for next checkpoint. Reason: ${ex.getMessage}"
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)
    }
  }

  /**
    * Method to get retained checkpoint details for a failed job.
    * Helps in restarting the job using a retained checkpoint of the job
    */
  def getRetainedCheckpointsOfFailedJob(jobName: String): Try[List[(String, Long)]] = Try {

    // Load the flink job
    val flinkJobTry = flinkApiJobHandler.getFlinkJob(jobName)

    flinkJobTry match {
      case Success(flinkJob) =>
        val jobId = flinkJob.jid

        val checkpointsResponseTry = getHttpResponseOfURI(
          flinkCluster,
          uri = s"/jobs/$jobId/checkpoints",
          jobId,
          jobName,
          operationContext = "retrieving checkpoint details of the failure job"
        )

        checkpointsResponseTry match {
          case Success(checkpointsResponse) =>
            processRetainedCheckpointsOfFailedJob(checkpointsResponse, jobId, jobName)

          case Failure(ex) => throw ex
        }
      case Failure(ex) =>
        val errorMessage = s"Job: $jobName - Failed to load the failed job in order to get checkpoint details. Reason: ${ex.getMessage}"
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)
    }
  }

  /**
    * Checks if the age of the checkpoint is within the limits defined
    */
  def isAgeOfTheCheckpointWithinLimit(jobName: String,
                                      checkpointFile: String,
                                      checkpointFileCreationTime: Long,
                                      maxAgeLimitInMillis: Long): Boolean = {

    val ageOfCheckpointFileInMillis: Long = System.currentTimeMillis - checkpointFileCreationTime
    logger.info(s"Job: $jobName - Age of the checkpoint file $checkpointFile is $ageOfCheckpointFileInMillis ms and upper limit of age is $maxAgeLimitInMillis")

    ageOfCheckpointFileInMillis < maxAgeLimitInMillis
  }

  /*-------------------------------------Private methods(Protected for UT)----------------------------*/

  /**
    * Method to get the checkpoint configuration of the job. This only gives information
    * like checkpointing interval for a job and calls the URL
    * /jobs/<jobId>/checkpoints/config
    */
  protected def getCheckpointConfigurationOfJob(jobId: String, jobName: String): Try[CheckPointConfig] = Try {
    // Load the checkpoint configuration
    val checkpointsConfigResponseTry = getHttpResponseOfURI(
      flinkCluster,
      uri = s"/jobs/$jobId/checkpoints/config",
      jobId,
      jobName,
      operationContext = "retrieving checkpoint configuration"
    )
    checkpointsConfigResponseTry match {
      case Success(checkpointsConfigResponse) =>
        checkpointsConfigResponse
          .parseJson
          .convertTo[CheckPointConfig]

      case Failure(ex) => throw ex
    }
  }

  /**
    * Verify the progress of next checkpoint and wait for the same before deploying a job to avoid conflicts
    */
  protected def verifyCheckpointProgressOfTheJob(jobId: String,
                                               jobName: String,
                                               checkpointIntervalInMillis: Int,
                                               sleepTimeBetweenRetriesInMillis: Int): Boolean = {
    // get current checkpoint counter of the job
    val currentCheckpointData: Try[(Int, String)] = getCurrentCheckpointCounterAndStatus(jobId, jobName)

    val previousCheckpointId = currentCheckpointData match {

      case Success((checkpointId, _)) =>
        logger.info(s"Job: $jobName ($jobId) - Received latest checkpoint $checkpointId from flink cluster.")
        //save current check point counter
        checkpointId

      case Failure(ex) =>
        if (ex.getMessage.equals(EmptyCheckpointResponse)) {
          logger.warn(s"Job: $jobName ($jobId) - Checkpoint operation is not yet started, waiting for checkpoint to happen.")
          0
        } else {
          //in case any issue, there is some problem with the system, throw ex
          val errorMessage = s"Job: $jobName ($jobId) - Failed while getting checkpoint details. Reason: ${ex.getMessage}"
          logger.error(errorMessage)
          throw FlinkProcessingException(errorMessage)
        }
    }

    // A maximum of twice the checkpoint interval will be waited for the next checkpoint to complete.
    val maxWaitTimeInMillis = 2 * checkpointIntervalInMillis

    // Counter to keep track of how long the sleep is done.
    var overallSleepTimeInMills = 0

    // keep waiting for next checkpoint
    while (overallSleepTimeInMills < maxWaitTimeInMillis) {

      val checkpointStatus = getCurrentCheckpointCounterAndStatus(jobId, jobName)

      checkpointStatus match {
        case Success((checkpointId, checkpointStatus)) =>

          if (checkpointId == previousCheckpointId) {
            logger.info(s"Job: $jobName ($jobId) - Job is still having the old checkpoint $checkpointId, retrying after $sleepTimeBetweenRetriesInMillis millis for checkpoint progress " +
              s"(Total time waited so far is $overallSleepTimeInMills ms and maximum wait time is ${2 * checkpointIntervalInMillis} ms)")

            Thread.sleep(sleepTimeBetweenRetriesInMillis)
            overallSleepTimeInMills += sleepTimeBetweenRetriesInMillis

          } else {
            logger.info(s"Job: $jobName ($jobId) - New checkpoint id $checkpointId, Waiting for checkpoint to complete.")
            val remainingSleepTime = maxWaitTimeInMillis - overallSleepTimeInMills

            return verifyCheckpointCompletionStatusOfGivenId(checkpointId, checkpointStatus, remainingSleepTime, jobId, jobName)
          }
        case Failure(ex) =>
          if (ex.getMessage.equals(EmptyCheckpointResponse)) {

            logger.warn(s"Job: $jobName ($jobId) - Checkpoint operation is not yet started, waiting for $sleepTimeBetweenRetriesInMillis millis " +
              s"(Total time waited so far is $overallSleepTimeInMills ms and maximum wait time is ${2 * checkpointIntervalInMillis} ms)")

            overallSleepTimeInMills += sleepTimeBetweenRetriesInMillis
            Thread.sleep(sleepTimeBetweenRetriesInMillis)

          } else {
            val errorMessage = s"Job: $jobName ($jobId) - Job has failed with exception ${ex.getMessage} while verifying checkpoint progress of the job"
            logger.error(errorMessage)
            throw FlinkProcessingException(errorMessage)
          }
      }
    }
    val errorMessage = s"Job: $jobName - Failed while verifying checkpoint progress."
    logger.error(errorMessage)
    throw FlinkProcessingException(errorMessage)
  }

  /**
    * Gets the running checkpoint counter of the given job and the status
    */
  protected def getCurrentCheckpointCounterAndStatus(jobId: String, jobName: String): Try[(Int, String)] = Try {
    // Get the checkpoints list of the job
    val checkpointListOfJobStringTry = getHttpResponseOfURI(
      flinkCluster,
      uri = s"/jobs/$jobId/checkpoints",
      jobId,
      jobName,
      operationContext = "retrieving checkpoint information"
    )

    checkpointListOfJobStringTry match {
      case Success(checkpointListOfJobString) =>

        val latestCheckpointOption: Option[CheckPointReqElementNameLatest] = try {

          val latestCheckpointOptionInternal = checkpointListOfJobString
            .parseJson
            .convertTo[CheckPointReqOfJob]
            .latest

          if (latestCheckpointOptionInternal.isEmpty) {
            val errorMessage = s"Job: $jobName ($jobId) - Failed while parsing Json response of the checkpoint. Reason: latest field is missing "
            logger.error(errorMessage)
            return Failure(FlinkProcessingException(EmptyCheckpointResponse))
          }

          if (latestCheckpointOptionInternal.get.completed.isEmpty) {
            val errorMessage = s"Job: $jobName ($jobId) - Failed while parsing Json response of the checkpoint. Reason: completed field is missing "
            logger.error(errorMessage)
            return Failure(FlinkProcessingException(EmptyCheckpointResponse))
          }

          latestCheckpointOptionInternal
        } catch {
          case ex: Exception =>
            val errorMessage = s"Job: $jobName ($jobId) - Failed while parsing Json response of the checkpoint. Reason: ${ex.getMessage}"
            logger.error(errorMessage)
            throw FlinkProcessingException(errorMessage)
        }

        val latestCheckpointReq = latestCheckpointOption.get

        val completedCheckpointDetail: CheckPointDetail = latestCheckpointReq
          .completed
          .get

        val checkpointCounterOption = completedCheckpointDetail.id

        checkpointCounterOption match {
          case Some(checkpointCounter) =>

            // Log the size details of the checkpoint
            logCheckpointStateSizeDetails(jobId, jobName, completedCheckpointDetail)

            val statusOfCompletedCheckpoint: Option[String] = completedCheckpointDetail.status
            if (statusOfCompletedCheckpoint.isEmpty) {
              val errorMessage = s"Job: $jobName ($jobId) - Check point status is empty"
              logger.error(errorMessage)
              throw FlinkProcessingException(errorMessage)
            }

            // Return value
            (checkpointCounter, statusOfCompletedCheckpoint.get)

          case None =>
            val errorMessage = s"Job: $jobName ($jobId) - Problem while parsing check point counter."
            logger.error(errorMessage)
            throw FlinkProcessingException(errorMessage)
        }
      case Failure(ex) => throw ex
    }
  }

  /**
    * Method to wait for remainingSleepTimeMills time for the checkpoint status to change to COMPLETED.
    *
    * During this time this method fails if one of the following conditions are met:
    *
    * a  When checkpoint id got changed
    * b  When the status of checkpoint returned any exception
    * c  When the time elapses and the state is not in COMPLETED
    *
    * This method succeeds as soon as the state changes to COMPLETED
    */
  protected def verifyCheckpointCompletionStatusOfGivenId(checkpointId: Int,
                                                          checkpointStatus: String,
                                                          remainingSleepTimeMills: Int,
                                                          jobId: String,
                                                          jobName: String): Boolean = {
    var localCheckpointStatus = checkpointStatus
    var localSleepTimeMills = 0

    while (!localCheckpointStatus.equals("COMPLETED")) {

      Thread.sleep(1000)
      localSleepTimeMills += 1000

      if (localSleepTimeMills > remainingSleepTimeMills) {
        val errorMessage = s"Job: $jobName ($jobId) - Failed to complete the check point and the status is still $checkpointStatus."
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)
      }

      val currentCheckpointData = getCurrentCheckpointCounterAndStatus(jobId, jobName)

      currentCheckpointData match {
        case Success((currentCheckpointId, currentCheckpointStatusValue)) =>
          localCheckpointStatus = currentCheckpointStatusValue

          if (checkpointId != currentCheckpointId) {
            val errorMessage = s"Job: $jobName ($jobId) - Not completed the previous checkpoint $checkpointId and progressed to next checkpoint $currentCheckpointId."
            logger.error(errorMessage)
            throw FlinkProcessingException(errorMessage)
          }

        case Failure(ex) =>
          val errorMessage = s"Job: $jobName ($jobId) - Failed while checking the progress status of checkpoint $checkpointId. Reason ${ex.getMessage}."
          logger.error(errorMessage)
          throw FlinkProcessingException(errorMessage)
      }
    }
    logger.info(s"Job: $jobName ($jobId) - Completed the progress of checkpoint $checkpointId.")
    true
  }

  /**
    * This message is used for monitoring state size of the stateful jobs
    */
  protected def logCheckpointStateSizeDetails(jobId: String,
                                             jobName: String,
                                             checkpointDetail: CheckPointDetail): Unit = {
    val stateSizeOption = checkpointDetail.`state_size`
    stateSizeOption match {
      case Some(stateSize) => logger.info(s"Job: $jobName ($jobId) - Checkpoint state size (bytes) - $stateSize.")
      case None => logger.error(s"Job: $jobName ($jobId) - Problem while retrieving check point state size details.")
    }
  }

  /**
    * Method processes the job/:jobID/checkpoints API response received from flink cluster after a job is in Failed state
    *
    * Sends back the valid checkpoint file details so that flink cluster can restore the state
    */
  protected def processRetainedCheckpointsOfFailedJob(checkpointsResponseString: String,
                                                    jobId: String,
                                                    jobName: String): List[(String, Long)] = {
    val checkpointsResponse: CheckPointResponseDetailsOfFailureJob = checkpointsResponseString
      .parseJson
      .convertTo[CheckPointResponseDetailsOfFailureJob]

    // Basic validations on the list of checkpoints itself
    if (checkpointsResponse.latest.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - latest field is missing in the response $checkpointsResponseString.")

    if (checkpointsResponse.latest.get.completed.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - latest.completed field is missing in the response $checkpointsResponseString.")

    if (checkpointsResponse.history.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - history field is missing in the response $checkpointsResponseString.")

    // Get the history list
    val historyList = checkpointsResponse
      .history
      .get

    // Filter to a set of valid retained checkpoints
    val listOfValidRetainedCheckpoints: List[CheckPointDetail] = historyList
      .flatMap(checkpointDetail => {

        val checkpointPathAndTimestampTry: Try[(String, Long)] = checkAndExtractPathAndTimestampOfCheckpoint(
          checkpointDetail,
          jobId,
          jobName
        )
        checkpointPathAndTimestampTry match {
          case Success(_) => List(checkpointDetail)

          case Failure(ex) =>
            logger.error(s"Job: $jobName ($jobId) - One of the completed checkpoint response ${checkpointDetail.id.get} failed. Reason: ${ex.getMessage}.")
            Nil
        }
      })

    // We should get the retained check point count at any point of time equals to 3
    if (listOfValidRetainedCheckpoints.isEmpty) {
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - Retained checkpoint count of the job is zero (must be equal to 3, please check production configuration)")
    }

    // Usually latest checkpoint value and first element in the history list are same. In case they are different adding them
    val latestEntryCheckpointDetail: CheckPointDetail = checkpointsResponse.latest.get.completed.get
    val finalListOfValidCheckpoints = if (listOfValidRetainedCheckpoints.head.id.get != latestEntryCheckpointDetail.id.get) {

      logger.warn(s"Job: $jobName ($jobId) - Checkpoint response: Latest completed checkpoint and first checkpoint in history are different: $checkpointsResponseString")
      logger.info(s"Job: $jobName ($jobId) - Adding the checkpoint id of the latest completed entry.")

      val latestIsValidEntryTry: Try[(String, Long)] = checkAndExtractPathAndTimestampOfCheckpoint(
        latestEntryCheckpointDetail,
        jobId,
        jobName
      )
      latestIsValidEntryTry match {
        case Success(_) =>
          // Append the latest to the top of the list
          List(latestEntryCheckpointDetail) ++ listOfValidRetainedCheckpoints

        case Failure(ex) =>
          logger.error(s"Job: $jobName ($jobId) - Latest completed checkpoint response ${latestEntryCheckpointDetail.id.get} failed. Reason: ${ex.getMessage}.")
          listOfValidRetainedCheckpoints
      }
    } else listOfValidRetainedCheckpoints

    // Return only the path and timestamp
    finalListOfValidCheckpoints
      .map(checkpointDetail => (checkpointDetail.external_path.get, checkpointDetail.trigger_timestamp.get))
  }

  /**
    * Method that does basic validation of a checkpoint object
    */
  protected def checkAndExtractPathAndTimestampOfCheckpoint(completedCheckpointDetail: CheckPointDetail,
                                                          jobId: String,
                                                          jobName: String): Try[(String, Long)] = Try {
    // Basic validations of the checkpoint
    if (completedCheckpointDetail.id.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - id field is missing in checkpoint response $completedCheckpointDetail.")

    if (completedCheckpointDetail.status.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - status field is missing in checkpoint response $completedCheckpointDetail.")

    if (completedCheckpointDetail.trigger_timestamp.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - trigger_timestamp field is missing in checkpoint response $completedCheckpointDetail.")

    if (completedCheckpointDetail.external_path.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - external_path field is missing in the checkpoint response $completedCheckpointDetail")

    if (completedCheckpointDetail.discarded.isEmpty)
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - discarded field is missing in the checkpoint response $completedCheckpointDetail.")

    logger.info(s"Job: $jobName ($jobId) - Response received for checkpoint details: $completedCheckpointDetail")

    if (completedCheckpointDetail.discarded.get.equals(true)) {
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - Check point: ${completedCheckpointDetail.external_path.get} is already deleted/discarded by the flink cluster.")
    }
    if (!completedCheckpointDetail.status.get.equals("COMPLETED")) {
      val errorMessage = s"Job: $jobName ($jobId) - Checkpoint status is ${completedCheckpointDetail.status.get} (checkpoint status should be COMPLETED)"
      logger.error(errorMessage)
      throw FlinkProcessingException(errorMessage)
    }
    // Ensure savepoints are not returned as a retained checkpoint
    if (completedCheckpointDetail.is_savepoint.getOrElse(false)) {
      throw FlinkProcessingException(s"Job: $jobName ($jobId) - Check point: ${completedCheckpointDetail.external_path.get} is a savepoint and discarding it as a valid retained checkpoint.")
    }
    (completedCheckpointDetail.external_path.get, completedCheckpointDetail.trigger_timestamp.get)
  }
}