/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.cluster

import com.netapp.df.uba.server.common.flink.settings.FlinkCluster
import com.netapp.df.uba.server.common.misc.log.LogTrait
import org.apache.http.client.methods.HttpGet
import spray.json._

import scala.util.{Failure, Success, Try}

/**
  * A class that handles the cluster related API invocations
  * Example: get jobmanager config, flink connection
  */
case class FlinkApiClusterHandler(flinkCluster: FlinkCluster) extends LogTrait
  with FlinkApisJsonFormats
  with FlinkApiBaseUtils {

  /**
    * Check flink job manager connectivity.
    */
  def checkFlinkConnection(): Try[Boolean] = {
    val responseTry = getHttpResponseOfURI(
      flinkCluster,
      // URL to check for a running jobmanager which runs on the master
      uri = "/jobmanager/config",
      "NA",
      "NA",
      "check flink connection"
    )
    responseTry match {
      case Success(_) => Success(true)
      case Failure(ex) => Failure(ex)
    }
  }

  /**
    * A wrapper over the existing method "checkFlinkConnection" but with connection retries delayed
    * in exponential backoff model until a successful connection to the cluster is made.
    * In case of connection failure to the cluster after maxAttempts returns failure
    */
  def retryFlinkConnectivity(jobName: String,
                             maxAttempts: Int,
                             startSleepTimeBetweenRetriesInMillis: Int): Try[Boolean] = {

    var attemptNumber: Int = 0
    var sleepTimeInMillis: Int = startSleepTimeBetweenRetriesInMillis

    while (attemptNumber < maxAttempts) {
      val connectionStatus = checkFlinkConnection()

      connectionStatus match {
        case Success(_) => return Success(true)

        case Failure(exception) =>
          attemptNumber = attemptNumber + 1
          if (attemptNumber == maxAttempts) {
            logger.error(s"Job: $jobName - Failed to obtain Flink connection after repeatedly trying over $maxAttempts times.")
            return Failure(exception)
          }

          logger.info(s"Job: $jobName - Failed to get Flink connection and waiting for $sleepTimeInMillis milli seconds before retrying.")
          Thread.sleep(sleepTimeInMillis)
          sleepTimeInMillis = sleepTimeInMillis * 2

          logger.info(s"Job: $jobName - Retrying for Flink connection with attempt number $attemptNumber.")
      }
    }
    val errorMessage = s"Job: $jobName - Failed to obtain flink cluster connection."
    logger.error(errorMessage)
    Failure(FlinkProcessingException(errorMessage))
  }

  /**
    * Get the jobmanager's config out for a specific key or for all
    */
  def retrieveJobManagerConfigForKey(keyName: String): Try[String] = {
    val jobManagerConfigMapTry = retrieveJobManagerConfig()

    jobManagerConfigMapTry match {
      case Success(jobManagerConfigMap) =>
        if (jobManagerConfigMap.contains(keyName)) {
          Success(jobManagerConfigMap(keyName))
        } else {
          Failure(throw FlinkProcessingException(s"Failed to find key $keyName in jobmanager config."))
        }

      case Failure(ex) => Failure(ex)
    }
  }

  def retrieveJobManagerConfig(): Try[Map[String, String]] = Try {
    // Create the HTTP client
    val httpClient = createHttpClientWithTimeouts()

    try {
      // Get the jobmanager config
      val httpRequest = new HttpGet(s"/jobmanager/config")

      val jobManagerConfigResponseBody = getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        "NA",
        "JM",
        context = s"jobmanager config"
      )
      return Success(
        jobManagerConfigResponseBody
          .parseJson
          .convertTo[List[JobManagerConfigEntry]]
          .map(configEntry => configEntry.key -> configEntry.value)
          .toMap
      )
    } finally {
      // Close the HTTP client
      closeHttpClient(httpClient)
    }
    Map.empty
  }
}
