/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.cluster

import com.netapp.df.uba.server.common.flink.settings.FlinkCluster
import com.netapp.df.uba.server.common.misc.log.LogTrait
import org.apache.http.HttpEntity
import org.apache.http.client.methods.{HttpDelete, HttpGet, HttpPost}
import org.apache.http.entity.ContentType
import org.apache.http.entity.mime.MultipartEntityBuilder
import org.apache.http.entity.mime.content.FileBody
import org.apache.http.impl.client.{CloseableHttpClient, HttpClients}
import spray.json._

import java.io.File
import java.util.NoSuchElementException
import scala.util.{Failure, Success, Try}

/**
 * A class that handles the jar related API calls on the Flink cluster
 * Add jar, remove jar, get jar id are the main methods
 */
//noinspection DuplicatedCode
case class FlinkApiJarHandler(flinkCluster: FlinkCluster) extends LogTrait
  with FlinkApisJsonFormats
  with FlinkApiBaseUtils {

  /**
   * Add a jar to Flink cluster
   */
  def addJar(jarFile: File, jobName: String, jobJarName: String): Try[String] = Try {

    val jarFileName = jarFile.getName

    // Create a different client that does not have timeouts as addJar takes considerable time
    val httpClientWithoutTimeouts: CloseableHttpClient = HttpClients.createDefault()

    try {
      // Call to upload a jar and append body to upload
      val httpRequest = new HttpPost(s"/jars/upload")
      val binFileBody = new FileBody(jarFile, ContentType.create("application/x-java-archive"), jobJarName)

      val reqEntity: HttpEntity = MultipartEntityBuilder.create()
        .addPart("bin", binFileBody)
        .build()
      httpRequest.setEntity(reqEntity)

      val responseBody = getHttpResponseOfURI(
        flinkCluster,
        httpClientWithoutTimeouts,
        httpRequest,
        "NA",
        jobName,
        context = s"add jar $jarFileName"
      )
      // Parse the response body as JarListInfo
      val addJarResponse = responseBody
        .parseJson
        .convertTo[FlinkAddJarResponse]

      // Return jar id
      addJarResponse
        .filename
        .split("/")
        .last

    } catch {
      case ex: Exception =>
        val errorMessage = s"Job: $jobName - Failed to add jar file $jarFileName. Reason: ${ex.getMessage}"
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)

    } finally {
      closeHttpClient(httpClientWithoutTimeouts)
    }
  }

  /**
   * Utility method to get the id of a jar uploaded to a flink cluster
   */
  def getJarId(httpClient: CloseableHttpClient, jarNameOrId: String): Try[String] = Try {
    // URL to get jar id for the given jar name
    val httpRequest = new HttpGet(s"/jars")

    // Get the response body
    val responseBody = getHttpResponseOfURI(
      flinkCluster,
      httpClient,
      httpRequest,
      "NA",
      jarNameOrId,
      context = s"get jars"
    )

    // Parse the response body as JarListInfo
    val jarListInfo = responseBody
      .parseJson
      .convertTo[FlinkJarListInfo]

    // Find in files element if there is one with the given name
    val jarFileInfoOption: Option[FlinkJarFileInfo] = jarListInfo
      .files
      .filter(_.name.isDefined)
      .find(jarEntry => jarEntry.name.get == jarNameOrId || jarEntry.id.get == jarNameOrId)

    // Get the id out of that jar entry
    jarFileInfoOption match {
      case None => throw new NoSuchElementException(s"Jar: $jarNameOrId - Failed to find jar with the given name.")
      case Some(jarFileInfo) => jarFileInfo.id.get
    }
  }

  /**
   * Remove jar for a flink job in the flink cluster
   */
  def removeJar(jobName: String, jarName: String, maxRetries: Int): Try[Boolean] = {
    val httpClient = createHttpClientWithTimeouts()
    try {
      val jarIdTry = getJarId(httpClient, jarName)
      jarIdTry match {
        case Failure(ex) =>
          logger.error(s"Failed to remove job jar :$jarName.Reason : ${ex.getMessage}")
          Failure(ex)
        case Success(jarId) =>
          removeJar(httpClient, jarId, jobName, jarName, maxRetries)
          logger.info(s"Removed job jar:$jarName.")
          Success(true)
      }
    } finally {
      closeHttpClient(httpClient)
    }
  }

  /**
   * Remove a flink job jar if present in the flink cluster
   * waitTries number of waits to ensure jar has been removed
   */
  def removeJar(httpClient: CloseableHttpClient,
                jarId: String,
                jobName: String,
                jobJarName: String,
                maxRetries: Int): Try[Boolean] = Try {

    // HTTP Delete call to remove a jar is based on id
    val httpRequest = new HttpDelete(s"/jars/$jarId")

    // Get the response body
    getHttpResponseOfURI(
      flinkCluster,
      httpClient,
      httpRequest,
      "NA",
      jobName,
      context = s"remove jar $jobJarName ($jarId)"
    )
    // Try as per the configuration and return the result.
    if (resultAfterAttempts[String](maxRetries, getJarId, httpClient, jarId)) true
    else throw FlinkProcessingException(s"Job: $jobName - Remove jar $jobJarName ($jarId) - Failed even after multiple attempts.")
  }
}
