/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.cluster

import com.netapp.df.uba.server.common.flink.settings.{FlinkCluster, FlinkJobConfig}
import com.netapp.df.uba.server.common.flink.utils.{NoOpSavepointCallbackHandler, SavepointCallbackHandler}
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.settings.CommonSettings
import org.apache.http.client.methods.{HttpGet, HttpPatch, HttpPost}
import org.apache.http.entity.{ContentType, StringEntity}
import org.apache.http.impl.client.CloseableHttpClient
import spray.json._

import java.io.File
import java.net.SocketTimeoutException
import java.util.NoSuchElementException
import scala.util.{Failure, Success, Try}

/**
 * FlinkApiJobHandler provides a set of utility methods to manage flink jobs in the flink cluster.
 */
object FlinkApiJobHandler {

  // Various job state strings that come from Flink
  final val JobStateInitializing = "INITIALIZING"
  final val JobStateRunning = "RUNNING"
  final val JobStateScheduled = "SCHEDULED"
  final val JobStateCreated = "CREATED"
  final val JobStateRestarting = "RESTARTING"
  final val JobStateFailing = "FAILING"
  final val JobStateCancelling = "CANCELLING"
  final val JobStateReconciling = "RECONCILING"
  final val JobStateSuspending = "SUSPENDING"
  final val JobStateSuspended = "SUSPENDED"
  final val JobStateFailed = "FAILED"
  final val JobStateCancelled = "CANCELED"
  final val JobStateFinished = "FINISHED"

  // Timeout for socket, connection request - 100 seconds
  final val TimeoutValueInMillis: Int = 100000

  // Lock namespace for flink job, note: this is appended to /oci/locks used in DistributedSynchronousTask taken from OCI_ commons
  final val FlinkJobLockNamespace = "flinkJob/"
}

//noinspection DuplicatedCode
case class FlinkApiJobHandler(flinkCluster: FlinkCluster)
  extends LogTrait
    with FlinkApiBaseUtils {

  import FlinkApiJobHandler._

  // Lock name used to synchronize the job management across micro-services includes a namespace of flinkJob/
  def lockNameForJob(jobName: String) = s"$FlinkJobLockNamespace${jobName.replace(' ', '_')}"

  // Gets the additional parameters not explicitly defined in the job
  def commonJobParameters(commonSettings: CommonSettings): Map[String, String] = Map(
    "microservice.common.environment" -> commonSettings.environment,
    "microservice.common.loglevel" -> commonSettings.loglevel,
    "microservice.common.external-base-path" -> commonSettings.externalBasePath
  )

  /**
   * Gets the job details for the given job name
   * 2 versions one with HTTP client and another without it
   */
  def getFlinkJob(jobName: String): Try[FlinkJob] = {
    // Create the HTTP client
    val httpClient = createHttpClientWithTimeouts()

    try {
      // Return the job id
      getFlinkJob(httpClient, jobName)
    } finally {
      // Close the HTTP client
      closeHttpClient(httpClient)
    }
  }

  /**
   * Gets full details of the job from Flink cluster
   */
  def getFlinkJob(httpClient: CloseableHttpClient, jobName: String): Try[FlinkJob] = Try {

    val httpRequest = new HttpGet(s"/jobs/overview")

    val jobsOverviewResponseBody = getHttpResponseOfURI(
      flinkCluster,
      httpClient,
      httpRequest,
      "NA",
      jobName,
      context = s"job details"
    )
    // Get the response body and parse as FlinkJobs
    val flinkJobs: List[FlinkJob] = jobsOverviewResponseBody
      .parseJson
      .convertTo[FlinkJobs]
      .jobs

    logger.info(s"Job: $jobName - Received the response from Flink cluster for jobs. " +
      s"Count: ${flinkJobs.size}, jobNames and statuses: ${flinkJobs.map(job => s"${job.name} -> ${job.state}").mkString(", ")}")

    val jobListWithMatchingName: List[FlinkJob] = flinkJobs
      .filter(_.name == jobName)

    if (jobListWithMatchingName.isEmpty) {
      val errorMessage = s"Job: $jobName - Failed to find job with the given name."
      logger.error(errorMessage)
      throw new NoSuchElementException(errorMessage)
    }

    // Get list of running jobs with the same name. end-time of the RUNNING job is always -1
    val detailsOfRunningJobs = jobListWithMatchingName
      .filter(_.`end-time`.get == -1)

    /*
      If there is more than one job running in prod cluster due to some manual intervention reason,
      clean it up, the latest one is cleaned up in the supporting method
     */
    if (detailsOfRunningJobs.size > 1) {
      logger.error(s"Job: $jobName - More than one running instances of the job are present. Need to stop duplicate jobs.")

      try {
        val jobRetainedPostStopOfDuplicates: FlinkJob = retainOldestAndStopDuplicateJobs(httpClient, detailsOfRunningJobs)
        return Success(jobRetainedPostStopOfDuplicates)
      }
      catch {
        case ex: Exception =>
          val errorMessage = s"Job: $jobName - Stopping duplicate jobs failed. Reason: ${ex.getMessage}"
          logger.error(errorMessage)
          throw FlinkProcessingException(errorMessage)
      }
    }

    // Case when there is a single running job
    if (detailsOfRunningJobs.nonEmpty) {
      logger.info(s"Job: $jobName - Returning the single running job: ${detailsOfRunningJobs.head.jid}")
      detailsOfRunningJobs.head

    } else {
      // No running instance of the job is present. Return the job which is recently stopped
      val latestTerminatedJob = jobListWithMatchingName
        .maxBy(_.`end-time`.get)

      logger.info(s"Job: $jobName - Returning a terminated job: ${latestTerminatedJob.jid}")

      // Return this job
      latestTerminatedJob
    }
  }

  /**
   * A wrapper over the existing method "getFlinkJob" but with job status retries delayed
   * in exponential backoff mode until a successful job status is available.
   * In case of failure to get job status after maxAttempts, this method throws exception.
   *
   * This is because in HA mode it takes some time to load job and bring it to running state.
   * Otherwise SFJM/TJM and Flink HA try to load the job at the same time and cluster goes to inconsistent state
   */
  def retryFlinkJobStatus(jobName: String, maxAttempts: Int, startSleepTimeBetweenRetriesInMillis: Int): Try[String] = Try {

    var attemptNumber = 0
    var sleepTimeInMillis = startSleepTimeBetweenRetriesInMillis

    while (attemptNumber < maxAttempts) {

      logger.info(s"Job: $jobName - Trying for flink job status with attempt number ${attemptNumber + 1}")

      val flinkJobTry = getFlinkJob(jobName)

      flinkJobTry match {
        case Success(job) => return Success(job.state.get)

        case Failure(ex) =>
          logger.info(s"Job: $jobName - Failed to obtain job status and waiting for $sleepTimeInMillis ms before retrying.")

          attemptNumber = attemptNumber + 1
          if (attemptNumber == maxAttempts) {
            logger.error(s"Job: $jobName - Failed to obtain job status after repeatedly trying for $maxAttempts times. Reason: ${ex.getMessage}.")
            throw ex
          }

          Thread.sleep(sleepTimeInMillis)
          sleepTimeInMillis = sleepTimeInMillis * 2
      }
    }
    logger.error(s"Job: $jobName - Failed to obtain the job status")
    throw FlinkProcessingException(s"Job: $jobName - Failed to obtain the job status")
  }

  /**
   * Utility to check if job is considered as in good state
   * In general the following strings can come
   * "enum" : [ "CREATED", "RUNNING", "FAILING", "FAILED", "CANCELLING", "CANCELED", "FINISHED", "RESTARTING", "SUSPENDING", "SUSPENDED", "RECONCILING" ]
   */
  def isJobStateValidForMonitoring(jobState: String): Boolean = jobState match {
    /*
     States for which we can consider it as valid.
     SFJM should not attempt to bring up the new job during the monitoring phase
     for these states
     */
    case JobStateRunning |
         JobStateScheduled |
         JobStateInitializing |
         JobStateCreated |
         JobStateRestarting |
         JobStateCancelling |
         JobStateSuspending |
         JobStateFailing |
         JobStateReconciling |
         JobStateFinished |
         JobStateCancelled => true

    case _ => false
  }

  /**
   * setupJob handles the setting up of the flink job in the flink cluster.
   * Requires cluster settings and job configuration
   *
   * additionalParameters are ones that are common to all jobs in their respective reference.conf's
   *
   * Reset job scenario:
   *   - remove jar
   *   - add jar
   *   - kill current job
   *   - add new job
   *     Non-reset job scenario:
   *     - add jar if it does not exist
   *     - add new job
   */
  def setupJob(jobConfig: FlinkJobConfig,
               jobJar: File,
               additionalParameters: Map[String, String],
               savepointCallbackHandler: SavepointCallbackHandler = new NoOpSavepointCallbackHandler,
               allowNonRestoredState: Boolean = false): Try[Boolean] = Try {

    val jobName = jobConfig.jobName
    val jobJarName = jobConfig.jobJarName

    // basic jar file validations
    if (!jobJar.exists()) throw FlinkProcessingException(s"Job: $jobName - Jar file ${jobJar.getName} does not exist.")
    if (!jobJar.isFile) throw FlinkProcessingException(s"Job: $jobName - Jar file ${jobJar.getName} is not a file.")

    // Create a savepoint manager object if job needs it
    val flinkApiSavepointHandlerOption = if (jobConfig.jobSavePointEnabled) {
      Some(createFlinkApiSavepointHandler(jobConfig, savepointCallbackHandler))
    } else None

    // Create a jar API handler
    val flinkApiJarHandler = createFlinkApiJarHandler()

    // Create a HTTP client
    val httpClient = createHttpClientWithTimeouts()

    try {
      // First get the jar id
      val jarIdTry1 = flinkApiJarHandler.getJarId(httpClient, jobJarName)
      jarIdTry1 match {

        case Success(jarId) =>
          logger.info(s"Job: $jobName - Jar present for $jobJarName with id $jarId.")

          // jar already present and job has to be reset, so remove the jar file
          if (jobConfig.jobReset) {
            logger.info(s"Job: $jobName - Removing the jar $jobJarName as the call is to reset the job.")
            flinkApiJarHandler.removeJar(
              httpClient,
              jarId,
              jobName,
              jobJarName,
              jobConfig.jobLaunchMaxRetries
            )
          }
        case Failure(_) =>
          logger.info(s"Job: $jobName - Jar not present for $jobJarName.")
        // Nothing to do here
      }

      // Get the jar id again
      val jarIdTry2 = flinkApiJarHandler.getJarId(httpClient, jobJarName)
      val updatedJarIdTry = jarIdTry2 match {

        case Success(jarId) =>
          // jar already present and hence nothing to be done
          logger.info(s"Job: $jobName - Jar id $jarId present for $jobJarName.")
          Success(jarId)

        case Failure(_) =>
          logger.info(s"Job: $jobName - Adding jar $jobJarName.")
          // Add the jar now, handles both reset and non-reset cases
          flinkApiJarHandler.addJar(jobJar, jobName, jobJarName)
      }
      logger.info(s"Job: $jobName - setupJob - updatedJarId: $updatedJarIdTry, jobJarName: $jobJarName")

      updatedJarIdTry match {
        case Success(updatedJarId) =>

          // Jar added successfully - Continue to handle job based on the jar
          if (jobConfig.jobReset) {
            logger.info(s"Job: $jobName - Stop the job as job has to be reset.")
            stopJob(httpClient, jobConfig, flinkApiSavepointHandlerOption)
          }
          logger.info(s"Job: $jobName - Executed stopJob and now executing getJob")

          // Try to get the job id
          val flinkJobTry = getFlinkJob(httpClient, jobName)
          flinkJobTry match {
            /*
              Stateful jobs: After executing stop, wait for job to go from RUNNING to FINISHED for stateful jobs as it can take some time
              Stateless jobs: RUNNING to CANCELED state for stateless jobs
             */
            case Success(job) if job.state.isDefined && job.state.get.equals(JobStateRunning) =>

              val jobStatusChangeTry: Try[Boolean] = if (jobConfig.jobSavePointEnabled) {
                waitUntilJobChangesGivenState(
                  jobName,
                  JobStateRunning,
                  JobStateFinished,
                  httpClient
                )
              } else {
                waitUntilJobChangesGivenState(
                  jobName,
                  JobStateRunning,
                  JobStateCancelled,
                  httpClient
                )
              }
              jobStatusChangeTry match {
                case Success(_) =>
                  logger.info(s"Job: $jobName - Adding the job after the state got changed from $JobStateRunning to $JobStateFinished or $JobStateCancelled")
                  addJob(httpClient, jobConfig, additionalParameters, updatedJarId, flinkApiSavepointHandlerOption, allowNonRestoredState)
                  true

                case Failure(ex) =>
                  logger.error(s"Job: $jobName - Job failed while checking the state transition from $JobStateRunning to $JobStateFinished or $JobStateCancelled")
                  throw ex
              }

            // DFPS-4525
            case Success(job) if job.state.isDefined && job.state.get.equals(JobStateCancelling) =>
              val jobStatusChangeTry = waitUntilJobChangesGivenState(
                jobName,
                JobStateCancelling,
                JobStateCancelled,
                httpClient
              )
              jobStatusChangeTry match {
                case Success(_) =>
                  logger.info(s"Job: $jobName - Adding the job after the state got changed from $JobStateCancelling to $JobStateCancelled")
                  addJob(httpClient, jobConfig, additionalParameters, updatedJarId, flinkApiSavepointHandlerOption, allowNonRestoredState)
                  true

                case Failure(ex) =>
                  logger.error(s"Job: $jobName - Failed while checking the state transition from $JobStateCancelling to $JobStateCancelled")
                  throw ex
              }

            case Success(job) if job.state.isDefined && !isJobReadyForDeployment(job.state.get) =>
              logger.info(s"Job: $jobName - Not adding the job since job id ${job.jid} exists for the job and its state is not one valid for deployment: ${job.state.get}.")
              true

            case Success(job) =>
              // Add the job as job exists but not with the required state
              logger.info(s"Job: $jobName - Adding the job since the state of the job is: ${job.state.getOrElse("NotDefined")}.")
              addJob(httpClient, jobConfig, additionalParameters, updatedJarId, flinkApiSavepointHandlerOption, allowNonRestoredState)
              true

            case Failure(_) =>
              // Add the job as job does not exist
              logger.info(s"Job: $jobName - Adding the job as it does not exist.")
              addJob(httpClient, jobConfig, additionalParameters, updatedJarId, flinkApiSavepointHandlerOption, allowNonRestoredState)
              true
          }
        case Failure(ex) =>
          logger.error(s"Job: $jobName - Failed to add/get jar id for $jobJarName. Reason: ${ex.getMessage}")
          throw ex
      }
    } finally {
      logger.info(s"Job: $jobName - Call closeHttpClient in setupJob.")
      // close the HTTP client
      closeHttpClient(httpClient)
    }
  }

  /*-------------------------------------Private methods(Protected for UT)----------------------------*/

  /**
   * Method to stop duplicate jobs in Flink cluster, ie more than one job with the same name in running status
   * Logic: Keeps the oldest job
   * Returns the retained job
   *
   * Called from getFlinkJob
   */
  protected def retainOldestAndStopDuplicateJobs(httpClient: CloseableHttpClient, duplicateJobList: List[FlinkJob]): FlinkJob = {

    // Sort the jobs such that the oldest job is at the head
    val sortedJobs = duplicateJobList
      .sortBy(_.`start-time`.get)
    val oldestJob = sortedJobs.head

    val duplicateJobsToStop = sortedJobs.drop(1)

    duplicateJobsToStop.foreach(duplicateJob => {
      val cancelStatusTry = cancelJob(httpClient, duplicateJob.jid, duplicateJob.name)
      logger.info(s"Job: ${duplicateJob.name} (${duplicateJob.jid}) - Duplicate job cancel status: $cancelStatusTry")
    })
    oldestJob
  }

  /**
   * Cancel a running job in the flink cluster
   */
  protected def cancelJob(name: String): Try[Boolean] = {
    val httpClient = createHttpClientWithTimeouts()
    try {
      val flinkJobTry = getFlinkJob(httpClient, name)
      flinkJobTry match {
        case Success(flinkJob) =>
          cancelJob(httpClient, flinkJob.jid, flinkJob.name)
          logger.info(s"Job: $name - Cancelled.")
          Success(true)
        case Failure(ex) =>
          logger.error(s"Job: $name - Failed to cancel the flink job. Reason: ${ex.getMessage}")
          Failure(ex)
      }
    } finally {
      closeHttpClient(httpClient)
    }
  }

  /**
   * Cancel a running job in the flink cluster
   */
  protected def cancelJob(httpClient: CloseableHttpClient, jobId: String, jobName: String): Try[Boolean] = Try {

    val httpRequest = new HttpPatch(s"/jobs/$jobId?mode=cancel")

    // Execute the patch call
    getHttpResponseOfURI(
      flinkCluster,
      httpClient,
      httpRequest,
      jobId,
      jobName,
      context = s"Cancel job PATCH call"
    )
    true
  }


  /**
   * Stops job in the flink cluster
   */
  protected def stopJob(httpClient: CloseableHttpClient,
                        jobConfig: FlinkJobConfig,
                        flinkApiSavepointHandlerOption: Option[FlinkApiSavepointHandler] = None): Try[Boolean] = Try {
    val jobName = jobConfig.jobName

    // Get hold of the job id for the job name
    val flinkJobTry = getFlinkJob(httpClient, jobName)

    flinkJobTry match {
      // Case where job exists
      case Success(flinkJob) if flinkJob.state.isDefined && isJobReadyForStopping(flinkJob.state.get) =>

        val jobId = flinkJob.jid
        if (jobConfig.jobSavePointEnabled) {

          logger.info(s"Job: $jobName ($jobId) - Stopping the stateful job.")
          return stopJobWithSavepoint(httpClient, jobName, jobId, flinkApiSavepointHandlerOption.get)
        } else {

          logger.info(s"Job: $jobName ($jobId) - Cancelling the stateless job.")
          return cancelJob(httpClient, flinkJob.jid, flinkJob.name)
        }
      case _ =>
        logger.info(s"Job: $jobName - Job does not exist or does not have valid state for stopping.")
        true
    }
  }

  // Utility to check if job is considered as in valid state such that we can stop the job without any issue
  protected def isJobReadyForStopping(jobState: String): Boolean = jobState match {

    // States for which we can consider it as valid states for stopping the job
    case JobStateRunning |
         JobStateScheduled |
         JobStateInitializing |
         JobStateCreated |
         JobStateRestarting |
         JobStateCancelling |
         JobStateSuspending |
         JobStateFailing |
         JobStateReconciling => true

    /*
      Other states include FINISHED, CANCELED, SUSPENDED, FAILED
      In these states, job is already stopped and hence no need to stop.
     */
    case _ => false
  }

  /**
   * Invokes the stop operation with a savepoint
   * Only stateful jobs are going to use the state while uploading the new jar
   */
  protected def stopJobWithSavepoint(httpClient: CloseableHttpClient,
                                     jobName: String,
                                     jobId: String,
                                     flinkApiSavepointHandler: FlinkApiSavepointHandler): Try[Boolean] = Try {
    // Save job state with savepoint request before stopping the job. Even if savepoint fails, job reset happens currently.
    val stopJobStatusTry = flinkApiSavepointHandler.stopJobWithSavepoint(httpClient, jobId)
    stopJobStatusTry match {
      case Success(_) =>
        logger.info(s"Job: $jobName ($jobId) - Stop operation with savepoint is successful.")
        true

      case Failure(ex) =>
        val errorMessage = s"Job: $jobName ($jobId) - Stop operation with savepoint failed. Reason: ${ex.getMessage}"
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)
    }
  }

  /**
   * Method to wait until the job state changes from currentJobState to expectedJobState
   * with configured retry count and wait time
   */
  protected def waitUntilJobChangesGivenState(jobName: String,
                                              currentJobState: String,
                                              expectedJobState: String,
                                              httpClient: CloseableHttpClient,
                                              maxAttempts: Int = 20,
                                              sleepTimeBetweenRetriesInMillis: Int = 4000): Try[Boolean] = Try {
    var attemptNumber = 0

    // Wait for upper limit number of attempts of 40*2 = 80 seconds
    while (attemptNumber < maxAttempts) {
      val flinkJobTry = getFlinkJob(httpClient, jobName)

      flinkJobTry match {
        case Success(job) if job.state.get.equals(currentJobState) =>
          logger.info(s"Job: $jobName - Attempt number ${attemptNumber + 1}, Job is still in the $currentJobState.")

          attemptNumber = attemptNumber + 1
          if (attemptNumber == maxAttempts) {
            val errorMessage = s"Job: $jobName - Job stuck in :$currentJobState state and has not changed to expected state: $expectedJobState state, after retrying $maxAttempts times."
            logger.error(errorMessage)
            throw FlinkProcessingException(errorMessage)
          }
          Thread.sleep(sleepTimeBetweenRetriesInMillis)

        case Success(job) if job.state.get.equals(expectedJobState) =>
          // State transition as expected
          return Success(true)

        case Success(job) =>
          val errorMessage = s"Job: $jobName - Has switched to wrong state ${job.state} from the state $currentJobState (expected job state is $expectedJobState)"
          logger.error(errorMessage)
          throw FlinkProcessingException(errorMessage)

        case Failure(ex) => throw ex
      }
    }
    throw FlinkProcessingException(s"Job: $jobName - Job stuck in :$currentJobState state and has not changed into expected state: $expectedJobState state, after retrying $attemptNumber times")
  }

  /**
   * Add a job to Flink
   */
  protected def addJob(httpClient: CloseableHttpClient,
                       jobConfig: FlinkJobConfig,
                       additionalParameters: Map[String, String],
                       jarId: String,
                       flinkApiSavepointHandlerOption: Option[FlinkApiSavepointHandler] = None,
                       allowNonRestoredState: Boolean = false,
                       maxAttempts: Int = 8,
                       sleepTimeBetweenRetriesInMillis: Int = 1000): Unit = {
    val jobName = jobConfig.jobName

    // Program arguments for the flink job
    val flinkJobArguments = if (jobConfig.jobParameters.nonEmpty || additionalParameters.nonEmpty) {
      // If at least one parameter is present add the program-args parameter
      val allParameters = jobConfig.jobParameters ++ additionalParameters

      allParameters
        .flatMap(entry => List(s"--${entry._1}", s"${entry._2}")) // One item for key and another for value
        .mkString(" ")
    } else ""

    logger.info(s"Job: $jobName parameters: $flinkJobArguments")

    val jarRunJsonString = getJarRunJsonString(jobConfig, flinkJobArguments, flinkApiSavepointHandlerOption)

    try {
      /*
       Launch a job using jar id

       Refer to https://ci.apache.org/projects/flink/flink-docs-release-1.9/monitoring/rest_api.html
       on allowNonRestoredState - Boolean value that specifies whether the job submission should be
       rejected if the savepoint contains state that cannot be mapped back to the job.
       */
      val httpRequest = new HttpPost(s"/jars/$jarId/run?allowNonRestoredState=$allowNonRestoredState")
      httpRequest.setEntity(new StringEntity(jarRunJsonString, ContentType.APPLICATION_JSON))

      getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        "NA",
        jobName,
        context = "add job"
      )
      // Check the post savepoint loading operation -- Process savepoint only if the savepoint file is present
      if (jobConfig.jobSavePointEnabled && flinkApiSavepointHandlerOption.get.getSavepointFilePathOption.isDefined) {
        val postJobSavepointResponseTry = waitForJobStateToRestoreFromSavepoint(
          httpClient,
          jobConfig,
          flinkApiSavepointHandlerOption.get
        )
        postJobSavepointResponseTry match {
          case Failure(ex) => throw ex
          case Success(_) => // Do nothing
        }
      }
      // Return value
      if (!resultAfterAttempts(jobConfig.jobLaunchMaxRetries, getFlinkJob, httpClient, jobName, checkForSuccess = false)) {
        throw FlinkProcessingException(s"Job: $jobName - Failed to add job even after multiple attempts.")
      }
    } catch {
      case ex: SocketTimeoutException =>

        logger.error(s"Job: $jobName - Timeout error while adding job. Reason: ${ex.getMessage}. Waiting for job to be in running state")

        val jobRunningStatusTry = waitForJobToBeRunningStatus(jobName, maxAttempts, sleepTimeBetweenRetriesInMillis)
        jobRunningStatusTry match {
          case Success(true) =>
            logger.info(s"Job: $jobName - Successfully deployed.")

          case Failure(ex1) =>
            throw FlinkProcessingException(s"Job: $jobName - Failed to add job. Reason: ${ex1.getMessage}.")

          case _ => //Shouldn't reach here ideally
            throw FlinkProcessingException(s"Job: $jobName - Failed to add job.")
        }
      case ex: Exception =>
        val errorMessage = s"Job: $jobName - Failed to add job. Reason: ${ex.getMessage}."
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)
    }
  }

  /**
   * Create the JSON String used in the body of RunJar method based on the availability of the savepoint file
   * Takes into account the savepoint path required if the job had a savepoint
   */
  protected def getJarRunJsonString(jobConfig: FlinkJobConfig,
                                    programArgs: String,
                                    flinkApiSavepointHandlerOption: Option[FlinkApiSavepointHandler] = None): String = {
    val jobName = jobConfig.jobName
    /*
      Savepoint can be used when starting the job only if savepoint file is created during the JobReset operation and
      savepoint is enabled. This ensures that the flink job starts reading the state available from savepoint file
    */
    flinkApiSavepointHandlerOption match {

      case Some(flinkApiSavepointHandler)
        if flinkApiSavepointHandler.isSavepointAvailable(jobConfig.jobSavePointEnabled, jobConfig.jobExternalSavePointFileOption) =>
        // Always choose the savepoint target directory which is defined
        val savepointFileOption: Option[String] = if (jobConfig.jobExternalSavePointFileOption.isDefined) {

          logger.info(s"Job: $jobName - Currently using the save point file, ${jobConfig.jobExternalSavePointFileOption} which is defined externally.")
          jobConfig.jobExternalSavePointFileOption

        } else {
          logger.info(s"Job: $jobName - Currently using the save point file ${flinkApiSavepointHandler.getSavepointFilePathOption}, which is generated internally")
          flinkApiSavepointHandler.getSavepointFilePathOption
        }
        val jarRunSavePointRequestBody = RunJarWithSavePoint(
          programArgs,
          jobConfig.jobMainClass,
          savepointFileOption.get
        )
        val jarRunJsonString = jarRunSavePointRequestBody
          .toJson
          .prettyPrint

        logger.info(s"Job: $jobName - Jar run JSON string: $jarRunJsonString")
        jarRunJsonString

      case _ =>
        if (flinkApiSavepointHandlerOption.isDefined) logger.info(s"Job: $jobName - Savepoint file is not present.")

        // Resort back to the plain RunJar without savepoint JSON string
        val jarRunRequestBody = RunJar(
          programArgs,
          jobConfig.jobMainClass
        )
        val jarRunJsonString = jarRunRequestBody
          .toJson
          .prettyPrint

        logger.info(s"Job: $jobName - Jar run JSON string: $jarRunJsonString")
        jarRunJsonString
    }
  }

  /**
   * This method waits for a job to go to Running state.
   * Typically used in cases where a job is submitted for execution and there is a need to wait for its running status.
   */
  protected def waitForJobToBeRunningStatus(jobName: String, maxAttempts: Int, startSleepTimeBetweenRetriesInMillis: Int): Try[Boolean] = Try {

    if (maxAttempts <= 0) {
      val errorMessage = s"Job: $jobName - Not waiting for the job to be running. Reason: maxAttempts is <= 0."
      logger.error(errorMessage)
      throw FlinkProcessingException(errorMessage)
    }
    val flinkApiClusterHandler = FlinkApiClusterHandler(flinkCluster)
    val flinkConnectivityStatusTry = flinkApiClusterHandler.retryFlinkConnectivity(
      jobName,
      maxAttempts,
      startSleepTimeBetweenRetriesInMillis
    )
    flinkConnectivityStatusTry match {
      case Failure(ex) =>
        logger.error(s"Job: $jobName - Flink connectivity failed for the job when waiting for the job to be in Running status.")
        throw ex

      case Success(_) =>
        logger.info(s"Job: $jobName - Flink connectivity is successful for the job when waiting for the job to be in Running status.")
    }

    var attemptNumber = 0
    var sleepTimeInMillis = startSleepTimeBetweenRetriesInMillis

    while (attemptNumber < maxAttempts) {

      logger.info(s"Job: $jobName - Waiting for job to be in running state with attempt number ${attemptNumber + 1}")

      val jobStatusTry = retryFlinkJobStatus(jobName, maxAttempts, startSleepTimeBetweenRetriesInMillis)

      jobStatusTry match {
        case Success(jobStatus) if JobStateRunning == jobStatus => // Return immediately if the status is fine.
          return Success(true)

        case _ =>
          if (jobStatusTry.isSuccess) logger.info(s"Job: $jobName - Job is still in ${jobStatusTry.get} status, waiting for $sleepTimeInMillis ms before retrying.")
          else logger.info(s"Job: $jobName - Job is not yet in $JobStateRunning status, waiting for $sleepTimeInMillis ms before retrying.")

          attemptNumber = attemptNumber + 1
          if (attemptNumber >= maxAttempts) {

            val errorMessage = s"Job: $jobName - Job is not in $JobStateRunning status even after repeatedly trying over $maxAttempts times."
            logger.error(errorMessage)
            throw FlinkProcessingException(errorMessage)
          }

          Thread.sleep(sleepTimeInMillis)
          sleepTimeInMillis = sleepTimeInMillis * 2
      }
    }
    throw FlinkProcessingException(s"Job: $jobName - Job is not in $JobStateRunning status even after repeatedly trying over $maxAttempts times.")
  }

  /**
   * Method that handles the post activities after a job starts with savepoint
   * This includes disposal of savepoint
   */
  protected def waitForJobStateToRestoreFromSavepoint(httpClient: CloseableHttpClient,
                                                      jobConfig: FlinkJobConfig,
                                                      flinkApiSavepointHandler: FlinkApiSavepointHandler): Try[Boolean] = Try {
    val jobName = jobConfig.jobName
    val flinkJobTry = getFlinkJob(jobName)

    flinkJobTry match {
      case Success(flinkJob) =>

        val jobId = flinkJob.jid

        // If a job is in canceled state, throw exception immediately
        if (flinkJob.state.get.equals("CANCELED")) throw FlinkProcessingException(s"Job: $jobName ($jobId) - Job is in canceled state and is not running.")

        // Check the progress of the savepoint operation
        val savepointTryString = flinkApiSavepointHandler.getCheckpointStatus(httpClient, flinkJob.jid)

        savepointTryString match {
          case Success(_) =>
            logger.info(s"Job: $jobName ($jobId) - Successfully restored the state after the savepoint operation.")

          case Failure(ex) =>
            val errorMessage = s"Job: $jobName ($jobId) - Job has failed to restore state after savepoint operation. Reason: ${ex.getMessage}."
            logger.error(errorMessage)
            throw FlinkProcessingException(errorMessage)
        }

        // Clear the savepoint file state
        flinkApiSavepointHandler.clearSavepointFilePath()

        // All good
        true

      case Failure(ex) =>
        val errorMessage = s"Job: $jobName - Unable to get the job details while processing savepoint operation. Reason: $ex, ${ex.getMessage}."
        logger.error(errorMessage)
        throw FlinkProcessingException(errorMessage)
    }
  }

  // Utility to check if job is considered in valid state so that setupJob can start deploying the job into remote flink cluster
  protected def isJobReadyForDeployment(jobState: String): Boolean = jobState match {

    // States for which we can consider it as valid for deploying a new job, mostly immediately after stopping
    case JobStateFailed |
         JobStateFinished |
         JobStateSuspended |
         JobStateCancelled => true

    // ALl other cases - Not ready for deployment
    case _ => false
  }

  // Code separated out so that unit tests can over-ride it
  protected def createFlinkApiSavepointHandler(jobConfig: FlinkJobConfig,
                                               savepointCallbackHandler: SavepointCallbackHandler) = {
    FlinkApiSavepointHandler(
      flinkCluster,
      jobConfig.jobName,
      jobConfig.jobSavePointTargetDir,
      savepointCallbackHandler
    )
  }

  protected def createFlinkApiJarHandler(): FlinkApiJarHandler = FlinkApiJarHandler(flinkCluster)
}