/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.cluster

import akka.http.scaladsl.marshallers.sprayjson.SprayJsonSupport
import spray.json.{DefaultJsonProtocol, RootJsonFormat}

/**
  * Case classes used as JSON responses in replies from flink cluster for
  * a. /jars
  * b. /jobs/overview
  * c. /savepoint API calls
  * d. /checkpoint related API calls
  */
case class FlinkJarEntryInfo(name: Option[String],
                             description: Option[String])

case class FlinkJarFileInfo(id: Option[String],
                            name: Option[String],
                            uploaded: Int,
                            entry: List[FlinkJarEntryInfo])

case class FlinkJarListInfo(address: Option[String],
                            files: List[FlinkJarFileInfo])

case class FlinkAddJarResponse(filename: String,
                               status: String)

// There are more fields in a flink job, but not needed for the logic
case class FlinkJob(jid: String,
                    name: String,
                    state: Option[String],
                    `start-time`: Option[Long],
                    `end-time`: Option[Long],
                    duration: Option[Long])

case class FlinkJobs(jobs: List[FlinkJob])

// To run a jar
case class RunJar(programArgs: String,
                  entryClass: String)

case class RunJarWithSavePoint(programArgs: String,
                               entryClass: String,
                               savepointPath: String)

//Savepoint response
case class SavePointResponse(`request-id`: String)

case class SavePointRequestObject(`target-directory`: String,
                                  `cancel-job`: Boolean)

//savepoint status and location response
case class SavePointOperationStatus(id: Option[String])

case class SavePointOperationLocation(location: Option[String])

case class SavePointStatusResponse(status: Option[SavePointOperationStatus],
                                   operation: Option[SavePointOperationLocation])

//Savepoint disposal
case class SavePointDisposalRequest(`savepoint-path`: String)

//Savepoint disposal response
case class SavePointDisposalStatus(`request-id`: String)

//Check point details
case class CheckPointDetail(id: Option[Int],
                            status: Option[String],
                            `state_size`: Option[Int],
                            trigger_timestamp: Option[Long] = None,
                            external_path: Option[String] = None,
                            discarded: Option[Boolean] = None,
                            is_savepoint: Option[Boolean] = None)

case class CheckPointReqElementNameLatest(completed: Option[CheckPointDetail])

case class CheckPointReqOfJob(latest: Option[CheckPointReqElementNameLatest])

//Check point configuration details
case class CheckPointConfig(mode: String,
                            interval: Int,
                            timeout: Long)

//Failure Checkpoint details
case class CheckPointRestoredDetailsOfFailureJob(id: Option[Int],
                                                 restore_timestamp: Option[Long],
                                                 is_savepoint: Option[Boolean],
                                                 external_path: Option[String])

case class CheckPointLatestDetailsOfFailureJob(completed: Option[CheckPointDetail],
                                               restored: Option[CheckPointRestoredDetailsOfFailureJob])

case class CheckPointResponseDetailsOfFailureJob(latest: Option[CheckPointLatestDetailsOfFailureJob],
                                                 history: Option[List[CheckPointDetail]])

case class JobManagerConfigEntry(key: String, value: String)
case class JobManagerConfigResponse(configEntries: List[JobManagerConfigEntry])

//Flink stop job request params
case class StopJobRequest(targetDirectory: String,
                          drain: Boolean)
/**
  * FlinkResponseJsonFormats provides the implicit format conversion for the flink responses
  */
trait FlinkApisJsonFormats extends SprayJsonSupport with DefaultJsonProtocol {
  implicit val jarEntryInfoFormatter: RootJsonFormat[FlinkJarEntryInfo] = jsonFormat2(FlinkJarEntryInfo.apply)
  implicit val jarFileInfoFormatter: RootJsonFormat[FlinkJarFileInfo] = jsonFormat4(FlinkJarFileInfo.apply)
  implicit val jarListInfoFormatter: RootJsonFormat[FlinkJarListInfo] = jsonFormat2(FlinkJarListInfo.apply)
  implicit val jarAddResponseFormatter: RootJsonFormat[FlinkAddJarResponse] = jsonFormat2(FlinkAddJarResponse.apply)

  implicit val jobFormatter: RootJsonFormat[FlinkJob] = jsonFormat6(FlinkJob.apply)
  implicit val jobsFormatter: RootJsonFormat[FlinkJobs] = jsonFormat1(FlinkJobs.apply)
  implicit val jarRunFormatter: RootJsonFormat[RunJar] = jsonFormat2(RunJar.apply)
  implicit val jarRunWithSavePointFormatter: RootJsonFormat[RunJarWithSavePoint] = jsonFormat3(RunJarWithSavePoint.apply)

  implicit val savePointResponseFormatter: RootJsonFormat[SavePointResponse] = jsonFormat1(SavePointResponse.apply)
  implicit val savePointRequestFormatter: RootJsonFormat[SavePointRequestObject] = jsonFormat2(SavePointRequestObject.apply)

  implicit val savePointOperationStatusFormatter: RootJsonFormat[SavePointOperationStatus] = jsonFormat1(SavePointOperationStatus.apply)
  implicit val savePointOperationLocationFormatter: RootJsonFormat[SavePointOperationLocation] = jsonFormat1(SavePointOperationLocation.apply)
  implicit val savePointStatusResponseFormatter: RootJsonFormat[SavePointStatusResponse] = jsonFormat2(SavePointStatusResponse.apply)

  implicit val SavePointDisposalRequestFormatter: RootJsonFormat[SavePointDisposalRequest] = jsonFormat1(SavePointDisposalRequest.apply)
  implicit val SavePointDisposalStatusFormatter: RootJsonFormat[SavePointDisposalStatus] = jsonFormat1(SavePointDisposalStatus.apply)

  implicit val completedAllFieldsFormatter: RootJsonFormat[CheckPointDetail] = jsonFormat7(CheckPointDetail.apply)
  implicit val latestAllFormatter: RootJsonFormat[CheckPointReqElementNameLatest] = jsonFormat1(CheckPointReqElementNameLatest.apply)
  implicit val latestCheckPointOfJobTestFormatter: RootJsonFormat[CheckPointReqOfJob] = jsonFormat1(CheckPointReqOfJob.apply)

  implicit val checkPointConfigFormatter: RootJsonFormat[CheckPointConfig] = jsonFormat3(CheckPointConfig.apply)

  implicit val checkPointRestoredDetailsOfFailureJobFormatter: RootJsonFormat[CheckPointRestoredDetailsOfFailureJob] = jsonFormat4(CheckPointRestoredDetailsOfFailureJob.apply)

  implicit val stopJobRequestFormatter: RootJsonFormat[StopJobRequest] = jsonFormat2(StopJobRequest.apply)
  implicit val checkPointLatestDetailsOfFailureJobFormatter: RootJsonFormat[CheckPointLatestDetailsOfFailureJob] = jsonFormat2(CheckPointLatestDetailsOfFailureJob.apply)
  implicit val checkPointReponseDetailsOfFailureJobFormatter: RootJsonFormat[CheckPointResponseDetailsOfFailureJob] = jsonFormat2(CheckPointResponseDetailsOfFailureJob.apply)

  implicit val jobManagerConfigEntryFormatter: RootJsonFormat[JobManagerConfigEntry] = jsonFormat2(JobManagerConfigEntry.apply)
  implicit val jobManagerConfigResponseFormatter: RootJsonFormat[JobManagerConfigResponse] = jsonFormat1(JobManagerConfigResponse.apply)
}