/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.cluster

import com.netapp.df.uba.server.common.flink.settings.FlinkCluster
import com.netapp.df.uba.server.common.flink.utils.{NoOpSavepointCallbackHandler, SavepointCallbackHandler}
import com.netapp.df.uba.server.common.misc.log.LogTrait
import org.apache.http.client.methods.{HttpGet, HttpPost}
import org.apache.http.entity.{ContentType, StringEntity}
import org.apache.http.impl.client.CloseableHttpClient
import spray.json._

import scala.util.{Failure, Success, Try}

/**
  * FlinkApiSavepointHandler class handles savepoint creation and disposal of the savepoint file operation
  */
//noinspection DuplicatedCode
case class FlinkApiSavepointHandler(flinkCluster: FlinkCluster,
                                    jobName: String,
                                    targetSavepointDirectory: String,
                                    savepointCallbackHandler: SavepointCallbackHandler = new NoOpSavepointCallbackHandler)
  extends LogTrait
    with FlinkApisJsonFormats
    with FlinkApiBaseUtils {

  // Variable to store savepoint file
  private var savepointFilePathOption: Option[String] = None

  // Get/Set methods for savepoint file
  def getSavepointFilePathOption: Option[String] = savepointFilePathOption

  def clearSavepointFilePath(): Unit = savepointFilePathOption = None

  /**
    * Checks if the savepoint file is available for the job
    */
  def isSavepointAvailable(jobSavePointEnabled: Boolean,
                           externalSavepointFileOption: Option[String]): Boolean = if (jobSavePointEnabled && (savepointFilePathOption.isDefined || externalSavepointFileOption.isDefined)) true else false

  /**
    * Stop the job with savepoint for stateful jobs. For all other jobs, savepoint is ignored
    */
  def stopJobWithSavepoint(httpClient: CloseableHttpClient,
                           jobId: String): Try[Boolean] = Try {
    // Body for savepoint
    val httpRequest = new HttpPost(s"jobs/$jobId/stop")
    val stopJobRequest = StopJobRequest(targetSavepointDirectory, drain = true)
    httpRequest.setEntity(new StringEntity(stopJobRequest.toJson.toString, ContentType.APPLICATION_JSON))

    // Issue stop command and receive the response
    val stopJobResponseString = getHttpResponseOfURI(
      flinkCluster,
      httpClient,
      httpRequest,
      jobId,
      jobName,
      context = "stop job with savepoint"
    )
    return postSavepointProcessing(
      httpClient,
      jobId,
      stopJobResponseString
    )
  }

  /**
    * Issue savepoint request to the remote cluster for the given job id
    */
  def takeSavepoint(jobId: String): Try[Boolean] = Try {
    // URI to invoke a savepoint on the job
    val httpRequest = new HttpPost(s"jobs/$jobId/savepoints")

    // Body for savepoint
    val savepointRequestObject = SavePointRequestObject(targetSavepointDirectory, `cancel-job` = false)
    httpRequest.setEntity(new StringEntity(savepointRequestObject.toJson.toString, ContentType.APPLICATION_JSON))

    // Create a HTTP client
    val httpClient = createHttpClientWithTimeouts()

    try {
      // Extract the response body from the response
      val savepointResponseString = getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        jobId,
        jobName,
        context = "savepoint request"
      )

      // postSavepointProcessing takes care of calling savepointCallbackHandler's triggered/succeeded/failed
      return postSavepointProcessing(
        httpClient,
        jobId,
        savepointResponseString
      )
    } finally {
      logger.info(s"Job: $jobName - Call closeHttpClient in takeSavepoint.")
      // close the HTTP client
      closeHttpClient(httpClient)
    }
  }

  /**
    * This method very the progress of the Checkpoint status after the savepoint is loaded
    */
  def getCheckpointStatus(httpClient: CloseableHttpClient,
                          jobId: String,
                          maxAttempts: Int = 30,
                          sleepTimeBetweenRetriesInMillis: Int = 4000): Try[String] = Try {

    val httpRequest = new HttpGet(s"/jobs/$jobId/checkpoints")

    // Increasing the waiting time of job status after loading with savepoint to 120 seconds
    var attemptNumber = maxAttempts
    while (attemptNumber > 0) {
      val checkpointsResponseString = getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        jobId,
        jobName,
        context = s"get checkpoints status"
      )
      /*
        Let's keep polling for sometime for checkpoint progress. So that we give equal opportunity for both the cases
        case 1 => savepoint without data, this usually happens when there is no data in system flowing still savepoint file is created
        case 2 => savepoint with data, this usually happens when there is data flowing in system
       */
      if (verifyCheckpointProgress(checkpointsResponseString)) return Success("Savepoint operation is successful")
      else {
        attemptNumber -= 1
        logger.info(s"Job: $jobName ($jobId) - Retrying.... Savepoint progress check...(retry count ${maxAttempts - attemptNumber}")
        Thread.sleep(sleepTimeBetweenRetriesInMillis)
      }
    }
    throw FlinkProcessingException(s"Job: $jobName ($jobId) - Flink Savepoint failed for the job.")
  }

  /*-------------------------------------Private methods(Protected for UT)----------------------------*/

  /**
    * Internal method to handle post processing after savepoint operation
    * 1. Stop job with savepoint
    * 2. Explicit call to take savepoint
    */
  protected def postSavepointProcessing(httpClient: CloseableHttpClient,
                                        jobId: String,
                                        savepointResponseString: String): Try[Boolean] = {
    val savepointResponse: SavePointResponse = try {
      savepointResponseString
        .parseJson
        .convertTo[SavePointResponse]
    } catch {
      case ex: Exception =>
        logger.error(s"Job: $jobName ($jobId) - Savepoint operation failed to parse savepoint response. Reason: ${ex.getMessage}.")
        return Failure(ex)
    }

    Try {
      /*
       Callback in case the caller wants to register the savepoint request
       Once savepointTriggered is invoked, either succeeded or failed should be mandatorily invoked on savepointCallbackHandler
       */
      val savepointRequestId = savepointResponse.`request-id`
      savepointCallbackHandler.savepointTriggered(savepointRequestId)

      // SavePointResponse above gives only the request-id, to get the actual savepoint-path,one more API needs to be invoked
      val savepointStatusResponseTry: Try[SavePointStatusResponse] = retrieveSavepointStatusResponse(
        httpClient,
        savepointRequestId,
        jobId,
        jobName
      )
      savepointStatusResponseTry match {
        case Success(savepointStatusResponse) =>
          val savepointPathOption = if (savepointStatusResponse.operation.isDefined) savepointStatusResponse.operation.get.location else None
          savepointPathOption match {
            case Some(savepointPath) =>
              logger.info(s"Job: $jobName ($jobId) - Savepoint operation output file is $savepointPath.")
              savepointCallbackHandler.savepointSucceeded(savepointRequestId, savepointPath)
              extractSavepointPath(savepointPath)
              true

            case None =>
              savepointCallbackHandler.savepointFailed(savepointRequestId)
              throw FlinkProcessingException(s"Job: $jobName ($jobId) - Savepoint path file is not defined.")
          }

        case Failure(ex) =>
          logger.error(s"Job: $jobName - Savepoint operation for the job failed. Reason: ${ex.getMessage}")
          savepointCallbackHandler.savepointFailed(savepointRequestId)
          throw ex
      }
    }
  }

  /**
    * Save the savepoint in a local variable
    * Savepoint file syntax given by flink starts with "file:/", then strip it to receive actual file name
    */
  protected def extractSavepointPath(savepointPath: String): Unit = {
    if (savepointPath.isEmpty) throw FlinkProcessingException(s"Job: $jobName - Savepoint file is empty")
    if (savepointPath.contains("file:/")) savepointFilePathOption = Some(savepointPath.substring("file:/".length - 1))
    else savepointFilePathOption = Some(savepointPath)
  }

  /**
    * This method executes the savepoint request operation and returns the savepoint path
    */
  protected def retrieveSavepointStatusResponse(httpClient: CloseableHttpClient,
                                                savepointRequestId: String,
                                                jobId: String,
                                                jobName: String,
                                                maxAttempts: Int = 90,
                                                sleepTimeBetweenRetriesInMillis: Int = 2000): Try[SavePointStatusResponse] = Try {
    // get call on a specific savepoint operation
    val httpRequest = new HttpGet(s"/jobs/$jobId/savepoints/$savepointRequestId")

    //Increasing the attempt count and wait time so that SFJM max waits for 120 seconds for save point progression
    var attemptNumber = maxAttempts

    // Execute the GET call
    while (attemptNumber > 0) {
      val savepointStatusResponseString = getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        jobId,
        jobName,
        context = s"retrieve savepoint operation progress"
      )
      val savepointStatusResponse = savepointStatusResponseString
        .parseJson
        .convertTo[SavePointStatusResponse]

      val savepointStatusIdOption = savepointStatusResponse.status match {
        case Some(statusValue) => statusValue.id
        case None => None
      }
      if (savepointStatusIdOption.isDefined) {
        // Check savepoint status, if not completed, retry
        if (!savepointStatusIdOption.get.equals("COMPLETED")) {

          logger.info(s"Job: $jobName ($jobId) - Savepoint status is ${savepointStatusIdOption.get}, again retrying (attempt number $attemptNumber)")
          Thread.sleep(sleepTimeBetweenRetriesInMillis)
          attemptNumber -= 1

        } else {
          logger.info(s"Job: $jobName ($jobId) - Savepoint operation successfully finished with savepointRequestId: $savepointRequestId")
          return Success(savepointStatusResponse)
        }
      } else {
        logger.info(s"Job: $jobName ($jobId) - Savepoint status is not defined, again retrying (attempt number $attemptNumber)")
        Thread.sleep(sleepTimeBetweenRetriesInMillis)
        attemptNumber -= 1
      }
    }
    val errorMessage = s"Job: $jobName ($jobId) - Failed to get savepoint status after attempting multiple times with savepointRequestId: $savepointRequestId"
    logger.error(errorMessage)
    throw FlinkProcessingException(errorMessage)
  }

  /**
    * Checks the restoration status of the previous savepoint which is loaded after supplying the savepoint file
    */
  protected def verifyCheckpointProgress(checkpointStatusResponse: String): Boolean = {
    try {
      val mainCheckpointResponseFields = checkpointStatusResponse
        .parseJson
        .asJsObject

      val counts: JsValue = mainCheckpointResponseFields
        .getFields("counts")
        .toList
        .head

      val countsAsJsonObject = counts
        .prettyPrint
        .parseJson
        .asJsObject

      val restored: Seq[JsValue] = countsAsJsonObject
        .getFields("restored")

      val restoredStatus = restored
        .head
        .toString
        .toInt

      val checkpointProgressCount = countsAsJsonObject
        .getFields("total")
        .head
        .toString
        .toInt

      val failedCount = countsAsJsonObject
        .getFields("failed")
        .head
        .toString
        .toInt

      // If any failed count is there we should return immediately
      if (failedCount > 0) return false

      /*
        When savepoint is restored, restoration status is set to 1.
        Savepoint progress is indicated by the checkpointProgressCount.
       */
      if (restoredStatus == 1 && checkpointProgressCount >= 0) true
      else false
    } catch {
      case ex: Exception =>
        logger.error(s"verifyCheckpointProgress exception. Reason: ${ex.getMessage}")
        false
    }
  }

  /**
    * Disposes the savepoint file created during the savepoint operation
    */
  //noinspection ScalaUnusedSymbol
  def disposeSavepoint(savepointFile: String): Try[Boolean] = Try {
    val httpClient = createHttpClientWithTimeouts()

    // POST request for disposing a savepoint
    val httpRequest = new HttpPost("/savepoint-disposal")
    val savepointDisposalCommandValue = SavePointDisposalRequest(savepointFile)
    httpRequest.setEntity(new StringEntity(savepointDisposalCommandValue.toJson.toString, ContentType.APPLICATION_JSON))

    try {
      val savepointDisposalStatusString = getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        jobId = "NA",
        jobName,
        context = s"dispose savepoint"
      )
      // Request to dispose a savepoint only returns a request id.
      val savepointDisposalStatus = savepointDisposalStatusString
        .parseJson
        .convertTo[SavePointDisposalStatus]

      /*
       Get the actual status of disposal using this request-id
       Maximum wait here is 10seconds
       */
      val isSavepointDisposedTry = checkIfSavepointIsDisposed(
        httpClient,
        jobName,
        savepointFile,
        savepointDisposalStatus.`request-id`
      )
      isSavepointDisposedTry match {
        case Success(_) =>
          // delete savepoint file successful
          logger.info(s"Job: $jobName - Savepoint file $savepointFile deleted successfully for the job.")
          true

        case Failure(ex) =>
          throw FlinkProcessingException(s"Job: $jobName - Savepoint file $savepointFile failed to delete with the exception ${ex.getMessage}.")
      }
    } catch {
      case ex: Exception => throw ex
    } finally {
      closeHttpClient(httpClient)
    }
  }

  /**
    * Get the status of savepoint disposal request of the savepoint file
    * created during the savepoint operation
    */
  protected def checkIfSavepointIsDisposed(httpClient: CloseableHttpClient,
                                          jobName: String,
                                          savepointFile: String,
                                          disposalStatusRequestId: String,
                                          maxAttempts: Int = 5,
                                          sleepTimeBetweenRetriesInMillis: Int = 2000): Try[Boolean] = Try {
    var attemptNumber = maxAttempts
    while (attemptNumber > 0) {
      val httpRequest = new HttpGet(s"/savepoint-disposal/$disposalStatusRequestId")

      val savepointDisposalStatus = getHttpResponseOfURI(
        flinkCluster,
        httpClient,
        httpRequest,
        jobId = "NA",
        jobName,
        context = s"savepoint disposal status for $savepointFile - $disposalStatusRequestId"
      )
      try {
        /*
        Sample API responses:
          Success Response: {\"status\":{\"id\":\"COMPLETED\"},\"operation\":{}}"
          Failure Response: "{\"status\":{\"id\":\"COMPLETED\"},\"operation\":{\"failure-cause\":{\"class\":\"java.util.concurrent.CompletionException\"}}}"
         */
        val statusValue: Seq[JsValue] = savepointDisposalStatus
          .parseJson
          .asJsObject
          .getFields("status")

        val operationId: Seq[JsValue] = statusValue
          .head
          .asJsObject
          .getFields("id")

        val operationIdValue: String = operationId.head.convertTo[String]
        // Marking as failure in two scenarios, when status is completed and when there is wrong savepoint file
        if (operationIdValue.equals("COMPLETED")) {

          val postSavepointOperationJsValue: Seq[JsValue] = savepointDisposalStatus
            .parseJson
            .asJsObject
            .getFields("operation")

          val postSavepointOperationStatusMap: Map[String, JsValue] = postSavepointOperationJsValue
            .head
            .asJsObject
            .fields

          // Case of wrong savepoint file
          if (postSavepointOperationStatusMap.contains("failure-cause")) throw FlinkProcessingException(s"Savepoint disposal failed because of the wrong savepoint file present.")
          else logger.info(s"Job: $jobName - Savepoint disposal of $savepointFile successful.")

          // Return value
          return Success(true)
        } else {
          //This is the case of In complete operation
          attemptNumber -= 1
          Thread.sleep(sleepTimeBetweenRetriesInMillis)
        }
      } catch {
        case ex: Exception =>
          logger.error(s"Job: $jobName - checkIfSavepointIsDisposed exception for $savepointFile. Reason: ${ex.getMessage}")
          //This is the case of In complete operation
          attemptNumber -= 1
          Thread.sleep(sleepTimeBetweenRetriesInMillis)
      }
    }
    throw FlinkProcessingException(s"Failed processing flink savepoint disposal status for $savepointFile of job after multiple attempts.")
  }
}