/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.settings

import com.netapp.df.uba.server.common.flink.settings.FlinkProtocol.FlinkProtocol
import com.netapp.df.uba.server.common.misc.utils.SettingUtils
import com.netapp.df.uba.server.common.misc.utils.SettingUtils._
import com.typesafe.config.{Config, ConfigException}

import scala.collection.JavaConverters._
import scala.concurrent.duration._
import scala.language.implicitConversions

/**
  * Enumeration representing the protocol used to communicate with Flink Cluster.
  */
object FlinkProtocol extends Enumeration {
  type FlinkProtocol = Value

  val Http: FlinkProtocol = Value("http")
  val Https: FlinkProtocol = Value("https")
}

/**
  * FlinkJobConfig represents the settings for one flink job to be submitted to the flink cluster.
  */
object FlinkJobConfig {
  def apply(config: Config): FlinkJobConfig = {

    /*
     A flink job needs a set of parameters to operate on.
     They will be sent as parameters. In the Flink job, they will get set as system properties
     or environment variables before loading the typesafe config of the job.
      */
    val parameterMap: Map[String, String] = if (config.hasPath("parameters")) {
      config.getConfig("parameters").entrySet()
        .asScala
        .map(item => item.getKey -> item.getValue.unwrapped().toString)
        .toMap
    } else Map.empty

    // Job name
    val jobName = SettingUtils.mandatoryConfigValue(config, "job-name")

    // Save point settings. Mandate the savepoint directory if savepoint is enabled
    val savepointEnabled = config.getBoolean("job-savepoint-enabled")
    val savepointDirectory = if (savepointEnabled) SettingUtils.mandatoryConfigValue(config, "job-savepoint-dir") else ""

    //This will be supplied through helm or by program, by default value defined in configuration is None
    val externalSavePointFile = if (savepointEnabled) {
        val savepointFile = config.getString("job-external-savepoint-file")

        if (savepointFile.equals("None")) None
        else if (savepointFile.isEmpty) None
        else Some(savepointFile)

      } else None

    val inputTopicNameKeys = SettingUtils
      .getNonEmptyValueStringSet(config, "input-topic-name-keys")
      .toList

    if (inputTopicNameKeys.isEmpty)
      throw new ConfigException.BadValue("input-topic-name-keys", s"At least one input topic should be specified for the job $jobName.")

    FlinkJobConfig(
      jobName,
      SettingUtils.mandatoryConfigValue(config, "job-jar-name"),
      SettingUtils.mandatoryConfigValue(config, "job-main-class"),
      SettingUtils.validMinInt(config, "job-launch-max-retries", 1),
      config.getBoolean("job-reset"),
      SettingUtils.validMinDuration(config, "job-check-interval"),
      parameterMap,
      SettingUtils.validMinDuration(config, "job-checkpoint-file-max-age"),
      inputTopicNameKeys,
      config.getBoolean("is-topic-owner"),
      config.getInt("pipeline.restart-strategy.fixed-delay.attempts"),
      config.getDuration("pipeline.restart-strategy.fixed-delay.delay"),
      externalSavePointFile,
      savepointEnabled,
      savepointDirectory
    )
  }
}

case class FlinkJobConfig(jobName: String,
                          jobJarName: String,
                          jobMainClass: String,
                          jobLaunchMaxRetries: Int,
                          jobReset: Boolean,
                          jobCheckInterval: FiniteDuration,
                          jobParameters: Map[String, String],
                          checkpointFileMaxAge: FiniteDuration,
                          inputTopicNameKeys: List[String],
                          isJobTopicOwner: Boolean,
                          pipelineRestartStrategyFixedDelayAttempts: Int,
                          pipelineRestartStrategyFixedDelayDelay: FiniteDuration,
                          jobExternalSavePointFileOption: Option[String] = None,
                          jobSavePointEnabled: Boolean = false,
                          jobSavePointTargetDir: String = ""
                         )

// Represents the settings to connect to flink cluster
object FlinkCluster {
  def apply(config: Config): FlinkCluster = FlinkCluster(
    SettingUtils.mandatoryConfigValue(config, "host"),
    SettingUtils.validMinInt(config, "port", 1),
    FlinkProtocol.withName(SettingUtils.mandatoryConfigValue(config, "protocol"))
  )
}
case class FlinkCluster(host: String,
                        port: Int,
                        protocol: FlinkProtocol)

/**
  * FlinkJobStateControl represents the sleep-time and retry-count when setting up a job
  * or during monitoring if applicable
  */
object FlinkJobStateControl {
  def apply(config: Config): FlinkJobStateControl = FlinkJobStateControl(
    config.getInt("retry-count"),
    config.getDuration("sleep-time")
  )
}
case class FlinkJobStateControl(retryCount: Int, sleepTime: FiniteDuration)

/**
  * FlinkJobStateControlSettings is a collection of settings representing how long job manager's
  * have to wait before checking the flink job status and how many times it has to retry for
  * getting valid job status before setting up a new job.
  */
object FlinkJobStateControlSettings {

  def apply(config: Config): FlinkJobStateControlSettings = {
    val deployConfig = FlinkJobStateControl(config.getConfig("deploy"))
    val monitorConfig = if (config.hasPath("monitor")) {
      FlinkJobStateControl(config.getConfig("monitor"))
    } else FlinkJobStateControl(8, 2500.millis)  // Monitoring is currently not applicable for pipeline jobs
    FlinkJobStateControlSettings(deployConfig, monitorConfig)
  }
}
case class FlinkJobStateControlSettings(deploymentJobControl: FlinkJobStateControl,
                                        monitorJobControl: FlinkJobStateControl)
/**
  * Overall flink settings. Contains a set of jobs to be managed by the micro-service
  * and the job configuration for each of the job.
  */
object FlinkSettings {
  def apply(config: Config, jobKeys: List[String] = Nil): FlinkSettings = {
    val flinkJobs = jobKeys
      .map(jobKey => FlinkJobConfig(config.getConfig(s"jobs.$jobKey")))

    FlinkSettings(
      FlinkCluster(config),
      SettingUtils.mandatoryConfigValue(config, "jar-directory").stripSuffix("/"),
      flinkJobs
    )
  }
}

case class FlinkSettings(cluster: FlinkCluster,
                         jobJarDirectory: String,
                         flinkJobs: List[FlinkJobConfig])
