/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.flink.utils

import com.netapp.df.uba.server.common.flink.cluster.{FlinkApiCheckpointHandler, FlinkApiClusterHandler, FlinkApiJobHandler}
import com.netapp.df.uba.server.common.flink.settings.{FlinkJobConfig, FlinkJobStateControl}
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.settings.CommonSettings
import com.netapp.oci.synch.DistributedSynchronousTask
import com.netapp.oci.synch.DistributedSynchronousTask.TaskWithReturn
import FlinkApiJobHandler._
import java.io.File
import scala.util.{Failure, Success, Try}

/**
  * A case class representing a checkpoint/savepoint location and its age
  */
case class CheckOrSavePointWithCreateTime(location: String, createTime: Long)

/**
  * A set of utility methods used during job creation
  */
trait FlinkJobUtils extends LogTrait {
  /*
    Utility method to check for loading the jar file, checking if it exists and if it is a file and returning the File object.
   */
  def loadJarFile(jobConfig: FlinkJobConfig,
                  jarDirectory: String): Try[File] = Try {
    val jarFile = new File(s"$jarDirectory/${jobConfig.jobJarName}")
    logger.info(s"Job: ${jobConfig.jobName} - Loading the jar file with path ${jarFile.getPath}")

    if (!jarFile.exists()) throw new Exception(s"Job: ${jobConfig.jobName} - File with path ${jarFile.getCanonicalPath} does not exist.")
    else if (!jarFile.isFile) throw new Exception(s"Job: ${jobConfig.jobName} - Path ${jarFile.getCanonicalPath} does not correspond to a file.")
    else jarFile
  }

  /**
    * Detailed orchestration of setting up a job in Flink cluster
    */
  def orchestrateSetupJobInFlinkCluster(jobConfig: FlinkJobConfig,
                                        lockName: String,
                                        flinkApiJobHandler: FlinkApiJobHandler,
                                        flinkApiClusterHandler: FlinkApiClusterHandler,
                                        flinkApiCheckpointHandler: FlinkApiCheckpointHandler,
                                        deploymentJobControl: FlinkJobStateControl,
                                        jarFile: File,
                                        commonSettings: CommonSettings,
                                        lastSavepointOption: Option[CheckOrSavePointWithCreateTime] = None,
                                        savepointCallbackHandler: SavepointCallbackHandler = new NoOpSavepointCallbackHandler(),
                                        allowNonRestoredState: Boolean = false): Try[Boolean] = {
    try {
      // noinspection ConvertExpressionToSAM
      new DistributedSynchronousTask(lockName).execute[Try[Boolean]](new TaskWithReturn[Try[Boolean]] {
        override def run(): Try[Boolean] = {
          try {

            // Extract the retry count and sleep time between retries for flink cluster invocations
            val flinkClusterApiRetryCount = deploymentJobControl.retryCount
            val sleepTimeBetweenRetriesMillis = deploymentJobControl.sleepTime.toMillis.toInt

            // Ensure connectivity to Flink cluster
            val flinkConnectivityStatusTry = flinkApiClusterHandler.retryFlinkConnectivity(
              jobConfig.jobName,
              flinkClusterApiRetryCount,
              sleepTimeBetweenRetriesMillis
            )

            flinkConnectivityStatusTry match {
              case Failure(exception) =>
                // Error out immediately by throwing an exception
                logger.error(s"Job: ${jobConfig.jobName} - Flink connectivity failed for the job. Reason: ${exception.getMessage}")
                throw exception

              case Success(_) =>
                logger.info(s"Job: ${jobConfig.jobName} - Flink connectivity successful for the flink job.")
            }

            // Check if the job already exists in the flink cluster, and if yes obtain the status of the same.
            val jobStatusTry = flinkApiJobHandler.retryFlinkJobStatus(
              jobConfig.jobName,
              flinkClusterApiRetryCount,
              sleepTimeBetweenRetriesMillis
            )

            val retainedCheckPoints: List[CheckOrSavePointWithCreateTime] = jobStatusTry match {
              case Failure(_) =>
                logger.info(s"Job: ${jobConfig.jobName} - Job does not exist. Setting up a new job.")
                Nil // No old checkpoints anyway

              /**
                * Job exists, now different statuses have to be handled appropriately.
                * As explained in the document:
                * https://ci.apache.org/projects/flink/flink-docs-release-1.5/api/java/org/apache/flink/runtime/execution/ExecutionState.html#RECONCILING
                * A new job is setup only from RUNNING state and terminal states FINISHED, CANCELED, SUSPENDED and FAILED.
                */
              case Success(jobStatus) if jobStatus == JobStateRunning =>
                /*
                  Job status is running.
                  - For savepoint/checkpoint enabled jobs, a wait is done for the next checkpoint to finish
                 */
                if (jobConfig.jobSavePointEnabled) {

                  val checkpointStatusTry = flinkApiCheckpointHandler.waitForNextCheckpointToFinish(jobConfig.jobName)

                  checkpointStatusTry match {
                    case Failure(exception) =>
                      logger.error(s"Job: ${jobConfig.jobName} - Error when waiting for checkpoint to finish. Reason: ${exception.getMessage}.")
                      throw exception

                    case Success(_) =>
                      logger.info(s"Job: ${jobConfig.jobName} - Successful in checkpoint progress for the job.")
                      Nil
                  }
                } else Nil

              case Success(jobStatus) if jobStatus == JobStateFailed =>
                /*
                  Job status is failed.
                  - For savepoint/checkpoint enabled jobs, a wait is done for the next checkpoint to finish
                  - For the rest of them, no retained checkpoints
                 */
                if (jobConfig.jobSavePointEnabled) getRetainedCheckpointList(jobConfig, flinkApiCheckpointHandler)
                else Nil

              case Success(jobStatus) if jobStatus == JobStateFinished || jobStatus == JobStateCancelled || jobStatus == JobStateSuspended =>
                logger.info(s"Job: ${jobConfig.jobName} - Job is in $jobStatus state. Setting up new job.")
                Nil // No retained checkpoints in this case

              case Success(jobStatus) =>
                /*
                  All other cases for job status:
                    FAILING,RESTARTING..etc will come out and retry until job is in either failed or running or finished state
                 */
                throw new Exception(s"Job: ${jobConfig.jobName} - Job is in $jobStatus state. Cannot setup new job as this can lead to duplicate jobs.")
            }
            // Setup the job in flink cluster
            setupJobWithOrWithoutCheckpoints(
              flinkApiJobHandler,
              flinkApiCheckpointHandler,
              jobConfig,
              jarFile,
              commonSettings,
              retainedCheckPoints,
              lastSavepointOption,
              savepointCallbackHandler,
              allowNonRestoredState
            )
          } catch {
            case exception: Exception => Failure(exception)
          }
        }
      })
    } catch {
      case exception: Exception =>
        logger.error(s"Job: ${jobConfig.jobName} - Failed to get distributed synchronous task for $lockName. Reason: ${exception.getMessage}")
        Failure(exception)
    }
  }

  /**
    * Utility to get list of retained checkpoints for a job with failed status case
    */
  def getRetainedCheckpointList(jobConfig: FlinkJobConfig,
                                flinkApiCheckpointHandler: FlinkApiCheckpointHandler): List[CheckOrSavePointWithCreateTime] = {

    val retainedCheckPointListOfFailedJobTry: Try[List[(String, Long)]] =
      flinkApiCheckpointHandler.getRetainedCheckpointsOfFailedJob(jobConfig.jobName)

    retainedCheckPointListOfFailedJobTry match {
      /*
        There is some problem in retrieving the checkpoint file details from flink cluster
        Setup new job in this case. Otherwise SFJM will go through an infinite loop until job state disappears from flink cluster
       */
      case Failure(exception) =>
        logger.error(s"Job: ${jobConfig.jobName} - Failed when retrieving checkpoint list of the failed job. " +
          s"Setting up the new job without any checkpoint file. Reason: ${exception.getMessage}")

        Nil

      case Success(retainedCheckPointListOfFailedJob) =>
        /*
        Response contain a list and each element of list has 2 elements.
        1) Checkpoint file details
        2) Creation time of the checkpoint file
         */
        retainedCheckPointListOfFailedJob
          .map(retainedCheckpoint => CheckOrSavePointWithCreateTime(retainedCheckpoint._1, retainedCheckpoint._2))
    }
  }

  /**
    * Setup of a job using retained checkpoints if available, else setup as a fresh job
    */
  def setupJobWithOrWithoutCheckpoints(flinkApiJobHandler: FlinkApiJobHandler,
                                       flinkApiCheckpointHandler: FlinkApiCheckpointHandler,
                                       jobConfig: FlinkJobConfig,
                                       jarFile: File,
                                       commonSettings: CommonSettings,
                                       retainedCheckPointFiles: List[CheckOrSavePointWithCreateTime],
                                       lastSavepointOption: Option[CheckOrSavePointWithCreateTime] = None,
                                       savepointCallbackHandler: SavepointCallbackHandler = new NoOpSavepointCallbackHandler,
                                       allowNonRestoredState: Boolean = false): Try[Boolean] = {

    val setupJobCheckpointList: List[Option[String]] = computeCheckpointListToUse(
      jobConfig,
      flinkApiCheckpointHandler,
      retainedCheckPointFiles,
      lastSavepointOption
    )
    setupJobCheckpointList
      .map(checkpointFileOption => {
        /**
          * Define another JobConfig object with the checkpoint file path and by enabling the savepoint flag
          * NOTE: The way of running the job using the checkpoint file and the savepoint file is one and same
          */
        val updatedJobConfig = jobConfig.copy(jobExternalSavePointFileOption = checkpointFileOption)

        logger.info(s"Job: ${jobConfig.jobName} - setupJobWithOrWithoutCheckpoints with updatedJobConfig: $updatedJobConfig.")
        val setupStatusTry: Try[Boolean] = setupJobInternal(
          flinkApiJobHandler,
          updatedJobConfig,
          jarFile,
          commonSettings,
          savepointCallbackHandler,
          allowNonRestoredState
        )
        if (setupStatusTry.isSuccess) return setupStatusTry // break out if the job was setup successfully.
        setupStatusTry
      })
      .last // Give the result as the final attempt in case of all failures.
  }

  protected def computeCheckpointListToUse(jobConfig: FlinkJobConfig,
                                           flinkApiCheckpointHandler: FlinkApiCheckpointHandler,
                                           retainedCheckPointFiles: List[CheckOrSavePointWithCreateTime],
                                           lastSavepointOption: Option[CheckOrSavePointWithCreateTime] = None,
                                          ): List[Option[String]] = {
    // Get the external savepoint file if configured explicitly for this job's reset
    val externalSavepointFileOptionInJobConfig = jobConfig.jobExternalSavePointFileOption

    /*
      List of checkpoints to consider for start of the job depends on the following conditions:
      1. If external savepoint is configured, use only that, do not fall back to retained checkpoints or savepoints
      2. If last savepoint (comes in pipeline model) and a set of retained checkpoints are present try
      with each one of them in order such that the latest one is tried first - only if it within age
      3. If none of the above works, then the job gets started without old state
     */
    if (externalSavepointFileOptionInJobConfig.isDefined) {
      logger.info(s"Job: ${jobConfig.jobName} - computeCheckpointListToUse - External savepoint provided for the job: ${externalSavepointFileOptionInJobConfig.get}.")
      List(externalSavepointFileOptionInJobConfig)
    }
    else {
      val lastSavepointInList: List[CheckOrSavePointWithCreateTime] = if (lastSavepointOption.isEmpty) Nil else {
        logger.info(s"Job: ${jobConfig.jobName} - Job has a last savepoint: ${lastSavepointOption.get}")
        List(lastSavepointOption.get)
      }
      logger.info(s"Job: ${jobConfig.jobName} - computeCheckpointListToUse - retainedCheckPointFiles.size(): ${retainedCheckPointFiles.size}.")

      val fullListOfCheckpointsToTry = (retainedCheckPointFiles ++ lastSavepointInList)
        .filter(checkOrSavepoint => {
          /*
            Check that the age of the checkpoint file (current time - creation time) under the configured limits specific to the flink job.
            If it is well under the limits then use the checkpoint file to load the old state. Otherwise ignore the check point file
           */
          val isAgeWithinLimit = flinkApiCheckpointHandler.isAgeOfTheCheckpointWithinLimit(
            jobConfig.jobName,
            checkOrSavepoint.location,
            checkOrSavepoint.createTime,
            jobConfig.checkpointFileMaxAge.toMillis
          )
          logger.info(s"Job: ${jobConfig.jobName} - Checkpoint/Savepoint file ${checkOrSavepoint.location} with creation time ${checkOrSavepoint.createTime} - isAgeWithinLimit: $isAgeWithinLimit.")
          isAgeWithinLimit
        })
        .sortBy(_.createTime)(Ordering[Long].reverse)
        .map(entry => Some(entry.location))

      logger.info(s"Job: ${jobConfig.jobName} - computeCheckpointListToUse - " +
        s"fullListOfCheckpointsToTry.size(): ${fullListOfCheckpointsToTry.size}, items: ${fullListOfCheckpointsToTry.map(entry => entry.getOrElse("None")).mkString(", ")}.")
      /*
       Append the option of setting up the job without checkpoint as an additional option
       To represent the case of setup without any checkpoint or savepoint i.e no old state
       */
      fullListOfCheckpointsToTry ++ List(None)
    }
  }

  /**
    * Internal method to call the setupJob for the job
    */
  private def setupJobInternal(flinkApiJobHandler: FlinkApiJobHandler,
                               jobConfig: FlinkJobConfig,
                               jarFile: File,
                               commonSettings: CommonSettings,
                               savepointCallbackHandler: SavepointCallbackHandler,
                               allowNonRestoredState: Boolean): Try[Boolean] = {

    logger.info(s"Job: ${jobConfig.jobName} - setupJobInternal to setup the job.")

    val setupContext = if (jobConfig.jobExternalSavePointFileOption.isDefined) {
      s" with the checkpoint/savepoint file: ${jobConfig.jobExternalSavePointFileOption.get}"
    } else ""

    // Setup the job
    val setupJobResponseTry = flinkApiJobHandler.setupJob(
      jobConfig,
      jarFile,
      flinkApiJobHandler.commonJobParameters(commonSettings),
      savepointCallbackHandler,
      allowNonRestoredState
    )

    setupJobResponseTry match {
      case Success(_) =>
        // Able to setup the job, so log and return immediately.
        logger.info(s"Job: ${jobConfig.jobName} - Successfully setup the job$setupContext.")
        Success(true)

      case Failure(exception) =>
        logger.error(s"Job: ${jobConfig.jobName} - Failed to setup job$setupContext. Reason: ${exception.getMessage}")
        Failure(exception)
    }
  }
}