/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.kafka

import akka.actor.Status.Failure
import akka.actor.{Actor, ActorLogging, ActorRef, Props}
import akka.http.scaladsl.Http
import akka.http.scaladsl.model.{HttpRequest, HttpResponse, StatusCodes}
import akka.pattern.pipe
import akka.stream.{ActorMaterializer, ActorMaterializerSettings}
import com.netapp.df.uba.server.common.misc.model._

import scala.concurrent.ExecutionContext
import scala.util.Try

/**
  * KafkaInfraChecker is the actor that is responsible for checking connectivity to kafka related
  * components. Includes the check of kafka brokers and schema registry.
  */
object KafkaInfraChecker {
  final val ActorName: String = "kafkaInfraChecker"
  def props(kafkaSettings: KafkaSettings,
            requestorRef: ActorRef): Props = Props(new KafkaInfraChecker(kafkaSettings, requestorRef))

  final val ContextKafkaInfraCheck: String = "kafkaInfraCheck"
}

class KafkaInfraChecker(kafkaSettings: KafkaSettings, requestorRef: ActorRef) extends Actor with ActorLogging {

  import KafkaInfraChecker._

  // Implicits required for http client call made for schema registry checks
  final implicit val materializer: ActorMaterializer = ActorMaterializer(ActorMaterializerSettings(context.system))
  final implicit val executionContext: ExecutionContext = context.system.dispatcher

  override def preStart(): Unit = {
    /*
    Check validity of schema registry, i.e the URL provided in the configuration
    Schema registry URL is generally http://<ip>:8081
    To really confirm that it is schema registry and not any HTTP server, lets append /subjects to the URL
      */
    val schemaRegistrySubjectsUrl = kafkaSettings.schemaRegistryUrl + "/subjects"

    // Fire the GET call for http://schema_registry_host:port/subjects. Response will be received by the actor as messages
    checkSchemaRegistryUrlCheck(schemaRegistrySubjectsUrl)
  }

  // Since preStart checks for schema registry first, that will be the initial mode and the checks return as messages to the actor due to pipe
  override def receive: Receive = schemaRegistryCheckMode

  private def schemaRegistryCheckMode: Receive = {

    // Schema registry URL check
    case schemaRegistryResponse@HttpResponse(StatusCodes.OK, _, _, _) =>
      log.info(s"Schema registry check for URL ${kafkaSettings.schemaRegistryUrl} succeeded.")
      schemaRegistryResponse.discardEntityBytes()

      // Now that schema registry check passed, a check on Kafka broker list is done.
      self ! CheckInfra
      context.become(kafkaCheckMode)

    case schemaRegistryResponse @ HttpResponse(code, _, _, _) =>
      schemaRegistryResponse.discardEntityBytes()

      val errorMessage = s"Schema registry check failed, response code: $code"
      log.error(errorMessage)

      // Send action failed message back to parent
      requestorRef ! ActionFailed(ContextKafkaInfraCheck, Error(ErrorCode.ConnectionNotEstablished, errorMessage, kafkaSettings.schemaRegistryUrl))

    case Failure(schemaRegistryFailureException) =>
      val errorMessage = s"Schema registry check failed with exception message as: ${schemaRegistryFailureException.getMessage}"
      log.error(errorMessage)

      // Send action failed message back to parent
      requestorRef ! ActionFailed(ContextKafkaInfraCheck, Error(ErrorCode.ConnectionNotEstablished, errorMessage, kafkaSettings.schemaRegistryUrl))

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message in schemaRegistryCheckMode that is not handled: " + somethingElse)
  }

  // Receive case for kafka broker check
  private def kafkaCheckMode: Receive = {

    // Message to check for kafka brokers
    case CheckInfra =>
      checkKafkaConnection() match {
        case scala.util.Success(_) =>
          log.info(s"Kafka broker connection check for broker list ${kafkaSettings.bootstrapServers} succeeded.")

          // Send action done to parent
          requestorRef ! ActionSucceeded(ContextKafkaInfraCheck)

        // Kafka broker check failed
        case scala.util.Failure(ex) =>
          val errorMessage = s"Kafka broker connection check failed with message: ${ex.getMessage}"
          log.error(errorMessage)

          // Send action done to parent
          requestorRef ! ActionFailed(ContextKafkaInfraCheck, Error(ErrorCode.ConnectionNotEstablished, errorMessage, kafkaSettings.bootstrapServers))
      }

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message in kafkaCheckMode that is not handled: " + somethingElse)
  }

  // Code to keep the some code separate so that unit tests can override it
  protected def checkSchemaRegistryUrlCheck(schemaRegistrySubjectsUrl: String): Unit = {
    // Fire the GET call for http://schema_registry_host:port/subjects. Response will be received by the actor as messages
    Http(context.system).singleRequest(HttpRequest(uri = schemaRegistrySubjectsUrl)).pipeTo(self)
  }

  // Code to check for kafka connection, kept separate so that unit tests can override it
  protected def checkKafkaConnection(): Try[Boolean] =
    KafkaUtils.checkBrokerConnectivity(kafkaSettings.bootstrapServers, kafkaSettings.adminClientRequestTimeout)
}