/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.kafka

import com.netapp.df.uba.server.common.misc.utils.SettingUtils
import com.typesafe.config.{Config, ConfigException}

import scala.collection.JavaConverters._
import scala.concurrent.duration.{FiniteDuration, _}

/**
  * A marker trait for case classes that represent the topics to which data is published to
  * from the micro-service or the case class representing the topics from which data
  * is consumed by the micro-service
  */
trait TopicTrait {
  def name: String
}

/**
  * Topic just includes the name and does not include anything suggestive of
  * producer checks or consumer checks
  */
object InfoTopic {
  def apply(config: Config): InfoTopic = new InfoTopic(
    SettingUtils.mandatoryConfigValue(config, "name")
  )
}

case class InfoTopic(name: String) extends TopicTrait

/**
  * ProducerTopic represents the configuration of a topic
  * to which the micro-service is publishing data.
  * It contains all the configuration options of the topic
  *
  * A complete check is done to ensure the configurarion of the topic is fine
  */
object ProducerTopic {

  def apply(config: Config): ProducerTopic = {

    /*
     Create a map of all config entries under the topic configuration which is not replciation
     factor and partition count and name

     events: {
        name: "auditevents.raw"
        partitions: 20
        replication-factor: 2

        cleanup.policy: delete
        compression.type: snappy
        retention.ms: 4 hours
      }

      For the above entry, it will create a map with three entries
      */
    val configMap: Map[String, String] = config.entrySet()
      .asScala
      .filterNot(item => item.getKey == "partitions" ||
                          item.getKey == "replication-factor" ||
                          item.getKey == "name" ||
                          item.getKey == "is-pipeline-topic")
      .map(item => item.getKey -> item.getValue.unwrapped().toString)
      .toMap

    // All except 2 are pipeline topics, hence marking default as true
    val isPipelineTopic = if (config.hasPath("is-pipeline-topic")) config.getBoolean("is-pipeline-topic") else true

    new ProducerTopic(
      SettingUtils.mandatoryConfigValue(config, "name"),
      SettingUtils.validMinInt(config, "partitions", 1),
      SettingUtils.validMinInt(config, "replication-factor", 1).asInstanceOf[Short],
      configMap,
      isPipelineTopic
    )
  }
}

case class ProducerTopic(name: String,
                         partitionCount: Int,
                         replicationFactor: Short,
                         config: Map[String, String],
                         isPipelineTopic: Boolean = true) extends TopicTrait
/**
  * ConsumerTopic represents a topic from which data is consumed by the micro-service.
  * Only an existence check will be done for these topics
  */
object ConsumerTopic {
  def apply(config: Config): ConsumerTopic = ConsumerTopic(
    SettingUtils.mandatoryConfigValue(config, "name")
  )
}

case class ConsumerTopic(name: String) extends TopicTrait

/**
  * Companion object for KafkaSettings to load from config object
  */
object KafkaSettings {

  def apply(config: Config,
            publishingTopics: List[String] = Nil,
            subscribingTopics: List[String] = Nil,
            infoTopics: List[String] = Nil): KafkaSettings = {

    // For each of the topic to which the micro-service is publishing, create the ProducerTopic with the full configuration
    val producersMap: Map[String, TopicTrait] = if (config.hasPath("topic.producer")) {
      // Get the main producer configuration
      val producerTopicConfigs = config.getConfig("topic.producer")

      publishingTopics
        .map(producerTopicKey => producerTopicKey -> ProducerTopic(producerTopicConfigs.getConfig(producerTopicKey)))
        .toMap
    } else Map.empty

    // For each of the topic to which the micro-service is subscribing, create the ConsumerTopic with the name
    val consumersMap: Map[String, TopicTrait] = if (config.hasPath("topic.consumer")) {
      // Get the main consumer configuration
      val consumerTopicConfigs = config.getConfig("topic.consumer")

      subscribingTopics
        .map(consumerTopicKey => consumerTopicKey -> ConsumerTopic(consumerTopicConfigs.getConfig(consumerTopicKey)))
        .toMap
    } else Map.empty

    // For each of the topic to which the info is required, create the InfoTopic with the name
    val infosMap: Map[String, TopicTrait] = if (config.hasPath("topic.info")) {
      // Get the main info configuration
      val infoTopicConfigs = config.getConfig("topic.info")

      infoTopics
        .map(infoTopicKey => infoTopicKey -> InfoTopic(infoTopicConfigs.getConfig(infoTopicKey)))
        .toMap
    } else Map.empty

    // Full map is a union of the two
    val topicMap: Map[String, TopicTrait] = producersMap ++ consumersMap ++ infosMap

    // Check the mandatory existence of schema registry
    val schemaRegistry: String = SettingUtils.mandatoryConfigValue(config, s"schema.registry.url").stripSuffix("/")
    if (schemaRegistry.isEmpty) throw new ConfigException.BadValue(s"schema.registry.url", "Value cannot be empty.")

    KafkaSettings(
      SettingUtils.mandatoryConfigValue(config, s"bootstrap.servers"),
      SettingUtils.mandatoryConfigValue(config, s"zookeeper.servers"),
      SettingUtils.mandatoryConfigValue(config, s"schema.registry.url").stripSuffix("/"),
      SettingUtils.validMinDuration(config, s"admin.client.request.timeout.ms", 1.milliseconds),
      topicMap,
      config.getString(s"key.subject.name.strategy"),
      config.getString(s"value.subject.name.strategy")
    )
  }
}

/**
  * ReceiverSettings contains kafka custom settings needed in other parts
  * than creating kafka producer, like schema registry URL needed to send avro schema + data as
  * produced data
  */
case class KafkaSettings(bootstrapServers: String,
                         zookeeperServers: String,
                         schemaRegistryUrl: String,
                         adminClientRequestTimeout: FiniteDuration,
                         topicMap: Map[String, TopicTrait],
                         keySubjectNameStrategy: String,
                         valueSubjectNameStrategy: String) {

  // Before invoking all the topic, index and flink manager's that deal with distributed locks, set the system property
  def setZookeeperServerSystemProperty(): Unit = System.setProperty("DOCKER_ZOOKEEPER_SERVERS", zookeeperServers)
}