/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.kafka

import com.netapp.df.uba.server.common.misc.utils.SettingUtils
import com.typesafe.config.{Config, ConfigException}

/**
  * Companion object for KafkaStreamsSettings to load from config object
  * to be shared across different applications.
  */
object KafkaStreamsSettings {

  def apply(config: Config): KafkaStreamsSettings = {

    // Check state dir name should not have a / character
    val stateDir = config.getString("state.dir")
    if (stateDir.contains("/")) throw new ConfigException.BadValue("state.dir", "Kafka streams state directory name should not have a / character.")

    KafkaStreamsSettings(
      config.getString("application.id"),
      SettingUtils.configValueWithEmptyAsNone(config, "cache.max.bytes.buffering"),
      SettingUtils.configValueWithEmptyAsNone(config, "commit.interval.ms"),
      config.getInt("num.standby.replicas"),
      config.getInt("replication.factor"),
      config.getString("acks"),
      config.getBoolean("reset.app"),
      config.getString("auto.offset.reset"),
      config.getInt("num.stream.threads"),
      stateDir,
      SettingUtils.configValueWithEmptyOrMissingAsNone(config, "state.cleanup.delay.ms"),
      // Refer https://cwiki.apache.org/confluence/display/KAFKA/KIP-91+Provide+Intuitive+User+Timeouts+in+The+Producer
      // From streams 2.3 onwards, delivery.timeout.ms gives a guaranteed upper bound on when a record will either get
      // sent, fail or expire from the point when producer.send returns.
      // delivery.timeout.ms >= linger.ms + retry.backoff.ms + request.timeout.ms
      // Its default is 2 mins(compared to earlier request.timeout.ms of 30s in streams 1.1.0), so we shouldn't ideally
      // need to change this now. But just making these configurable in case we need to tune any of these to avoid any
      // timeout errors in streams application
      SettingUtils.configValueWithEmptyOrMissingAsNone(config, "delivery.timeout.ms"),
      SettingUtils.configValueWithEmptyOrMissingAsNone(config, "producer.request.timeout.ms"),
      SettingUtils.configValueWithEmptyOrMissingAsNone(config, "linger.ms"),
      SettingUtils.configValueWithEmptyOrMissingAsNone(config, "batch.size"),
      SettingUtils.configValueWithEmptyOrMissingAsNone(config, "max.task.idle.ms")
    )
  }
}

/**
  * StreamsCommonSettings contains common settings for a stream application
  * like application id, and other options meant for kafka streams application.
  * Reference: https://docs.confluent.io/current/streams/developer-guide/config-streams.html
  */
case class KafkaStreamsSettings(applicationId: String,
                                cacheMaxBytesBufferingOption: Option[String],
                                commitIntervalMsOption: Option[String],
                                standbyReplicas: Int,
                                replicationFactorForStreamStateStores: Int,
                                producerAcks: String,
                                resetStreamApp: Boolean,
                                autoOffsetReset: String,
                                numStreamThreads: Int,
                                stateDir: String,
                                stateCleanupDelayMsOption: Option[String],
                                deliveryTimeoutMsOption: Option[String],
                                producerRequestTimeoutMsOption: Option[String],
                                producerLingerMsOption: Option[String],
                                producerBatchSizeOption: Option[String],
                                maxTaskIdleMsOption: Option[String] = None
                               )

