/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.kafka

import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.oci.synch.DistributedSynchronousTask
import com.netapp.oci.synch.DistributedSynchronousTask.TaskWithReturn
import io.confluent.kafka.serializers.{AbstractKafkaAvroSerDeConfig, KafkaAvroDeserializer, KafkaAvroSerializer}
import org.apache.kafka.clients.admin._
import org.apache.kafka.clients.consumer.{ConsumerConfig, KafkaConsumer, OffsetAndMetadata}
import org.apache.kafka.common.config.ConfigResource
import org.apache.kafka.common.errors.InvalidReplicationFactorException
import org.apache.kafka.common.serialization.StringDeserializer
import org.apache.kafka.common.{KafkaException, TopicPartition}

import java.util.concurrent.TimeUnit
import java.util.{Properties, UUID}
import java.{lang, util}
import scala.collection.JavaConverters._
import scala.concurrent.ExecutionException
import scala.concurrent.duration.FiniteDuration
import scala.util.{Failure, Success, Try}

// To hold keys of topic block in reference.conf
object KafkaConsts {
  final val TOPIC_AUDITEVENTS_RAW_KEY = "auditevents-raw"
  final val TOPIC_AUDITEVENTS_ENRICHED_KEY = "auditevents-enriched"
  final val TOPIC_AUDITEVENTS_RESOLVED_KEY = "auditevents-resolved"
  final val TOPIC_AUDITEVENTS_COUNTED_KEY = "auditevents-counted"

  final val TOPIC_COMMUNITIES_DETECTED_KEY = "communities-detected"
  final val TOPIC_COMMUNITIES_RESOLVED_KEY = "communities-resolved"

  final val TOPIC_COMMUNITY_DETECTION_RUNS_KEY = "community-detection-runs"
  final val TOPIC_COMMUNITY_RESOLUTION_RUNS_KEY = "community-resolution-runs"

  final val TOPIC_USERS_RAW_KEY = "users-raw"
  final val TOPIC_INVENTORY_RAW_KEY = "inventory-raw"

  final val TOPIC_OPSRATE_METRICS_KEY = "opsrate-metrics"

  final val TOPIC_ALERTS_RAW_KEY = "alerts-raw"
  final val TOPIC_ALERTS_PROCESSED_KEY = "alerts-processed"
  final val TOPIC_ALERTS_CLOUDEVENTS_KEY = "alerts-cloudevents"

  final val TOPIC_ALERT_ACTION_RESULTS_KEY = "alert-action-results"

  final val TOPIC_TRAFFIC_RAW_KEY = "traffic-raw"
  final val TOPIC_TRAFFIC_ENRICHED_KEY = "traffic-enriched"

  final val TOPIC_T_TEST_KEY = "t-test-metrics"

  final val TOPIC_AGENT_COLLECTOR_HEALTH_CLOUDEVENTS_KEY = "agent-collector-health-cloudevents"

  final val TOPIC_ONTAP_ARW_SUSPECT_DETAILS_KEY = "ontap-arw-suspect-details"
}

/**
  * A utility object for kafka admin management
  * Includes methods for checking connectivity, topic listing, config listing, topic creation and so on
  */
//noinspection ScalaDeprecation
object KafkaUtils extends LogTrait {

  // Lock namespace for topics, note: this is appended to /oci/locks used in DistributedSynchronousTask taken from OCI_ commons
  private val TopicLockNamespace = "topic/"

  /**
    * Create a kafka admin client with a setting for request timeout for all admin requests
    */
  def createAdminClient(bootstrapServers: String, adminClientRequestTimeout: FiniteDuration): Try[AdminClient] = {
    // Check for empty configuration
    if (bootstrapServers.isEmpty) Failure[AdminClient](new Exception("Bootstrap server list is empty."))
    else {
      // Properties for the admin client
      val adminClientProps = new Properties()
      adminClientProps.setProperty(AdminClientConfig.BOOTSTRAP_SERVERS_CONFIG, bootstrapServers)
      adminClientProps.setProperty(AdminClientConfig.REQUEST_TIMEOUT_MS_CONFIG, adminClientRequestTimeout.toMillis.toString)

      // Admin client fails to get created in some cases like empty bootstrap servers
      try {
        Success(AdminClient.create(adminClientProps))
      } catch {
        case ex: KafkaException if ex.getCause != null =>
          logger.error(s"Exception when creating kafka admin client. Reason: ${ex.getMessage}, Cause: ${ex.getCause.getMessage}")
          Failure(ex)

        case ex: Exception =>
          // Can never happen as per code of AdminClient.create
          logger.error(s"Exception when creating kafka admin client. Reason: ${ex.getMessage}")
          Failure(ex)
      }
    }
  }

  /**
   * Get the consumer group lag.
   *
   * @param groupId        consumer group Id
   * @param kafkaServerUrl kafka server url
   * @param topicIds       topics of this consumer group for which lag is needed. If empty, all the topics of this consumer group are
   *                       fetched.
   * @param timeoutMs      kafka client query timeout
   * @return list of all the partitions this consumer group is subscribed to, with their lag information
   *
   *         Steps to get the lag:
   *         1. Get the current offset for all the topic partitions for this consumer group.
   *         2. Create a dummy consumer, with same set of partitions assigned as for this consumer group.
   *         3. Get the end offsets for each of the partitions using this dummy consumer.
   *         4. Calculate lag for each partition as 3-1.
   *         5. Return list of all the partitions in response, with their offset and lag information.
   *         6. Finally total consumer group lag can be calculated by sum of all the individual partition lags returned by this method.
   *
   *         NOTE:Kafka 2.5 and later has a new API added KafkaAdminClient - getLatestOffsets(), avoiding need to create dummy consumer
   *         to fetch latest offsets for all topics/partitions. Need to refactor this method on moving to Kafka 2.5.
   */
  def getConsumerGroupLag(groupId: String, kafkaServerUrl: String, topicIds: List[String], timeoutMs: Int): Try[List[PartitionOffsetAndLag]] = {
    val adminClientTry = createAdminClient(kafkaServerUrl, FiniteDuration(timeoutMs, TimeUnit.MILLISECONDS))
    val dummyConsumer = getDummyConsumer(kafkaServerUrl)
    val partitionOffsetsAndLagTry: Try[List[PartitionOffsetAndLag]] = adminClientTry match {
      case Success(adminClient) => getConsumerGroupLag(groupId, topicIds, adminClient, dummyConsumer, timeoutMs)

      case Failure(ex) => logger.error(s"Failed to get admin client, Error: ${ex.getMessage}")
        Failure(ex)
    }

    try {
      dummyConsumer.close()
    } catch {
      case ex: Exception => logger.error(s"Failed to close dummy consumer. Exception:${ex.getMessage}")
    }

    // Close the admin client
    closeAdminClient(adminClientTry)
    partitionOffsetsAndLagTry
  }

  /**
    * Check for broker connectivity.
    * Includes a timeout to list topics from the kafka brokers brokers.
    * A very short time here leads to "Timed out waiting for node assignment"
    */
  def checkBrokerConnectivity(bootstrapServers: String, adminClientRequestTimeout: FiniteDuration): Try[Boolean] = {

    val kafkaAdminClientTry: Try[AdminClient] = KafkaUtils.createAdminClient(bootstrapServers, adminClientRequestTimeout)

    kafkaAdminClientTry match {
      case Success(kafkaAdminClient) =>
        // Try a sample API like listing topics to verify connectivity
        try {
          val describeClusterResult = kafkaAdminClient.describeCluster()
          describeClusterResult.clusterId().get()

          // Close the admin client
          closeAdminClient(kafkaAdminClientTry)

          Success(true)
        } catch {
          case ex: Exception =>
            logger.error(s"Kafka connectivity check failed. Reason: ${ex.getMessage}")
            Failure(ex)
        }

      // Scenario where admin client was never created
      case Failure(ex) =>
        // Already logged case
        Failure[Boolean](ex)
    }
  }

  /**
    * Gets the result of a single topic describe output
    */
  def topicDescribe(adminClient: AdminClient, topicName: String, timeoutMs: Int = 15000): Try[TopicDescription] = Try {
    val topicDescribeResult = adminClient.describeTopics(
      List(topicName).asJava,
      new DescribeTopicsOptions().timeoutMs(timeoutMs)
    )
    val topicDescription = topicDescribeResult.all().get()
    topicDescription.get(topicName)
  }

  /**
    * Check the presence of a topic by calling list topics API
    */
  def topicExists(adminClient: AdminClient, topicName: String): Try[Boolean] = Try {
    adminClient
      .listTopics()
      .names()
      .get()
      .contains(topicName)
  }

  /**
    * A way to increase the partitions of a topic
    */
  def increasePartitions(adminClient: AdminClient, topicName: String, partitionCount: Int): Try[Boolean] = Try {
    val partitionMap = new java.util.HashMap[String, NewPartitions]
    partitionMap.put(topicName, NewPartitions.increaseTo(partitionCount))
    adminClient
      .createPartitions(partitionMap)
    true
  }

  /**
    * Gets the result of config of a topic.
    * Note: This method returns success even in case of a non-existant topic
    */
  def topicConfig(adminClient: AdminClient, topicName: String): Try[Config] = try {
    val configResource = new ConfigResource(ConfigResource.Type.TOPIC, topicName)

    val topicConfigResult = adminClient.describeConfigs(List(configResource).asJava).all().get()

    Success(topicConfigResult.get(configResource))

  } catch {
    case ex: ExecutionException =>
      logger.error(s"Topic config get for $topicName failed. Reason: ${ex.getMessage}")
      Failure(ex)
  }

  /**
    * Check for config mismatch of a topic against what is expected and return the set of config that have mismatched
    */
  def topicConfigCompare(config: Config, expectedConfigItems: Map[String, String]): Map[String, String] = {
    expectedConfigItems
      .filter(expectedConfig => {
        val actualConfigEntryOption = Option(config.get(expectedConfig._1))

        actualConfigEntryOption.isDefined &&
          actualConfigEntryOption.get.value() != expectedConfig._2
      })
      .map(expectedConfig => {
        val actualConfigEntryOption = Option(config.get(expectedConfig._1))

        if (actualConfigEntryOption.isEmpty) expectedConfig._1 -> "null"
        else expectedConfig._1 -> actualConfigEntryOption.get.value()
      })
  }

  /**
    * Alter configuration of a topic
    */
  def alterTopicConfig(adminClient: AdminClient, topicDetails: ProducerTopic): Try[Boolean] = {
    // Create a config resource of topic type and given topic name
    val configResource = new ConfigResource(ConfigResource.Type.TOPIC, topicDetails.name)
    val configResourceMap: java.util.Map[ConfigResource, Config] = new java.util.HashMap

    // Create a config out of the configuration entries provided for ProducerTopic
    val configEntries = topicDetails
      .config
      .map(entry => new ConfigEntry(entry._1, entry._2))
      .toList
      .asJava

    configResourceMap.put(configResource, new Config(configEntries))

    // Invoke the alter configs API call
    try {
      //noinspection ScalaDeprecation
      adminClient.alterConfigs(configResourceMap).all().get()

      Success(true)

    } catch {
      case ex: Exception =>
        logger.error(s"Altering topic config for ${topicDetails.name} failed. Reason: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Utility method to handle a topic that the micro-service is publishing data to
    */
  def handleProducerTopic(producerTopic: ProducerTopic, adminClient: AdminClient, zookeeperServers: String): Try[Boolean] = {

    // Name of the topic
    val topicName = producerTopic.name

    // Lock name used to synchrnoize the topic management across micro-services includes a namespace of topic/
    val lockName = s"$TopicLockNamespace$topicName"

    // Necessary for distributed lock to work
    System.setProperty("DOCKER_ZOOKEEPER_SERVERS", zookeeperServers)

    try {
      //noinspection ConvertExpressionToSAM
      new DistributedSynchronousTask(lockName).execute[Try[Boolean]](new TaskWithReturn[Try[Boolean]] {

        override def run(): Try[Boolean] = {

          try {
            // Check the topic exists or not using describe.
            // Describe is used instead of list so that partition and replication factor count is also obtained.
            val topicDescribeTry = KafkaUtils.topicDescribe(adminClient, topicName)

            topicDescribeTry match {

              case Success(topicDescription) =>
                /*
                 Check topic partition count and replication factor
                  */
                val partitions = topicDescription.partitions()

                // Get the actual partition and replication factor size
                val actualPartitionSize = partitions.size()
                val actualReplicationFactorSize = if (partitions.isEmpty) 0
                else partitions.get(0).replicas().size()

                // Warning for mismatch of partition
                if (actualPartitionSize != producerTopic.partitionCount) {
                  logger.warn(s"Topic name: $topicName, Partition count mismatch. Actual: $actualPartitionSize, Expected: ${producerTopic.partitionCount}")

                  if (producerTopic.partitionCount < actualPartitionSize) logger.warn(s"Topic name: $topicName. Expected partition count (${producerTopic.partitionCount}) is lower than the current value($actualPartitionSize).")
                  else increasePartitions(adminClient, topicName, producerTopic.partitionCount)
                }

                // Warning for mismatch of replication factor
                if (actualReplicationFactorSize != producerTopic.replicationFactor) {
                  logger.warn(s"Topic name: $topicName, Replication factor mismatch. Actual: $actualReplicationFactorSize, Expected: ${producerTopic.replicationFactor}")
                }

                // Check if configuration matches as expected
                val topicConfigTry = KafkaUtils.topicConfig(adminClient, producerTopic.name)

                topicConfigTry match {
                  case Success(config) =>

                    // Get the comparison of expected and actual config of the topic
                    val configMisMatchResult = KafkaUtils.topicConfigCompare(config, producerTopic.config)

                    if (configMisMatchResult.isEmpty) {
                      logger.info(s"Topic name: $topicName, configuration matched.")
                      Success(true)

                    } else {
                      logger.warn(s"Topic name: $topicName, Actual configuration not same as expected - $configMisMatchResult")
                      KafkaUtils.alterTopicConfig(adminClient, producerTopic)
                    }

                  case Failure(ex) =>
                    logger.error(s"Topic name: $topicName, failed to get configuration of the topic. Reason: ${ex.getMessage}")
                    Failure(ex)
                }

              case Failure(ex) =>

                logger.info(s"Topic name: $topicName, topic not present (Exception: ${ex.getMessage}, hence creating it.")
                // Failed to get topic description which means topic not present, create it
                val topicCreatedTry = KafkaUtils.createTopic(adminClient, producerTopic)

                // If the creation was successful, alter the configuration to match what is required
                topicCreatedTry match {

                  case Success(_) =>
                    // Set the required configuration
                    val configAlterTry = alterTopicConfig(adminClient, producerTopic)

                    configAlterTry match {
                      case Success(_) =>
                        logger.info(s"Successfully set the configuration of the topic $topicName.")
                        configAlterTry

                      case Failure(configEx) =>
                        logger.error(s"Topic name: $topicName, failed to configure the topic. Reason: ${configEx.getMessage}")
                        Failure(configEx)
                    }
                  case Failure(createEx) =>
                    logger.error(s"Topic name: $topicName, failed to create the topic. Reason: ${createEx.getMessage}")
                    Failure(createEx)
                }
            }
          } catch {
            case ex: Exception =>
              logger.error(s"Topic name: ${producerTopic.name}, Exception occured: ${ex.getMessage}")
              Failure(ex)
          }
        }
      })
    } catch {
      case ex: Exception =>
        logger.error(s"Topic producer failed to get distributed synchronous task for $lockName. Reason: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Utility method to handle a topic that the micro-service is subscribing to
    */
  def handleConsumerTopic(consumerTopic: ConsumerTopic, adminClient: AdminClient, zookeeperServers: String): Try[Boolean] = {
    // Name of the topic
    val topicName = consumerTopic.name

    // Lock name used to synchronize the topic management across micro-services includes a namespace of topic/
    val lockName = s"$TopicLockNamespace$topicName"

    // Necessary for distributed lock to work
    System.setProperty("DOCKER_ZOOKEEPER_SERVERS", zookeeperServers)

    try {
      //noinspection ConvertExpressionToSAM
      new DistributedSynchronousTask(lockName).execute[Try[Boolean]](new TaskWithReturn[Try[Boolean]] {
        override def run(): Try[Boolean] = {

          // Check if the topic is present in the kafka server
          KafkaUtils.topicExists(adminClient, topicName) match {

            case Success(true) =>
              Success(true)

            case Success(false) =>
              logger.error(s"Topic name: $topicName to consume from does not exist.")
              Failure(new Exception(s"Topic name: $topicName to consume from does not exist."))

            case Failure(ex) =>
              // Failed to get topic description which means topic not present, mark as failure when its a consumer topic
              logger.error(s"Topic name: $topicName, Failure to load topic description: ${ex.getMessage}")
              Failure(ex)
          }
        }
      })
    } catch {
      case ex: Exception =>
        logger.error(s"Topic consumer failed to get distributed synchronous task for $lockName. Reason: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Create a topic and set the configuration as provided
    */
  //noinspection NoTailRecursionAnnotation
  def createTopic(adminClient: AdminClient, topicDetails: ProducerTopic): Try[Boolean] = {
    // Create a new topic object
    val newTopic = new NewTopic(topicDetails.name,
      topicDetails.partitionCount,
      topicDetails.replicationFactor)

    try {
      // Create the topic and wait for the response
      adminClient.createTopics(List(newTopic).asJava).all().get()
      Success(true)

    } catch {
      case ex: ExecutionException =>
        ex.getCause match {
          case irex: InvalidReplicationFactorException =>
            if (topicDetails.replicationFactor > 1) {

              /*
              Most of the times in dev/qa environments, the number of brokers might not be sufficient to support
              > 1 replication factor. To counter that issue, reduce the replication factor and proceed.
              We do not expect this to happen in production environments.
               */
              val newReplicationFactor = (topicDetails.replicationFactor - 1).asInstanceOf[Short]

              logger.warn(s"Unable to create topic ${topicDetails.name} with replication factor ${topicDetails.replicationFactor}, not enough brokers. Retrying with replication factor: $newReplicationFactor")

              createTopic(adminClient, ProducerTopic(
                topicDetails.name,
                topicDetails.partitionCount,
                newReplicationFactor,
                topicDetails.config)
              )
            } else {
              logger.error(s"Failed to create topic ${topicDetails.name} with InvalidReplicationFactorException. Reason: ${irex.getMessage}")
              Failure(irex)
            }

          case _ =>
            // for all other causes of ExecutionException, throw the issue out
            logger.error(s"Failed to create topic ${topicDetails.name} with ExecutionException. Reason: ${ex.getMessage}.")
            Failure(ex)
        }
      case ex: Exception =>
        // Other exceptions
        logger.error(s"Failed to create topic ${topicDetails.name}. Reason: ${ex.getMessage}.")
        Failure(ex)
    }
  }

  /**
   * Closes an admin client
   */
  def closeAdminClient(adminClientTry: Try[AdminClient]): Unit = {
    adminClientTry match {
      case Success(adminClient) =>
        // Ignore the close exception if any
        try adminClient.close(10, TimeUnit.MILLISECONDS) catch {
          case ex: Exception =>
            logger.debug(s"Failed to close kafka admin client after connectivity check. Reason: ${ex.getMessage}")
        }
      case Failure(ex) =>
        logger.warn(s"Admin client is not created due to ${ex.getMessage}. Hence, no close.")
    }
  }

  /**
   * Get the consumer group lag.
   *
   * @param groupId        consumer group Id
   * @param topicIds       topics of this consumer group for which lag is needed. If empty, all the topics of this consumer group are
   *                       fetched.
   * @param adminClient    Kafka admin client to use. Caller to ensure cleanup after use.
   * @param dummyConsumer  Kafka dummy consumer to use to determine lag. Caller to ensure cleanup after use.
   * @param timeoutMs      kafka client query timeout
   * @return list of all the partitions this consumer group is subscribed to, with their lag information
   *
   *         Steps to get the lag:
   *         1. Get the current offset for all the topic partitions for this consumer group.
   *         2. Create a dummy consumer, with same set of partitions assigned as for this consumer group.
   *         3. Get the end offsets for each of the partitions using this dummy consumer.
   *         4. Calculate lag for each partition as 3-1.
   *         5. Return list of all the partitions in response, with their offset and lag information.
   *         6. Finally total consumer group lag can be calculated by sum of all the individual partition lags returned by this method.
   *
   *         NOTE:Kafka 2.5 and later has a new API added KafkaAdminClient - getLatestOffsets(), avoiding need to create dummy consumer
   *         to fetch latest offsets for all topics/partitions. Need to refactor this method on moving to Kafka 2.5.
   */
  def getConsumerGroupLag(groupId: String, topicIds: List[String], adminClient: AdminClient, dummyConsumer: KafkaConsumer[String, String], timeoutMs: Int = 10000): Try[List[PartitionOffsetAndLag]] = {
    if (null == dummyConsumer || null == adminClient) {
      return Failure(new IllegalArgumentException("Missing kafka consumer or admin client"))
    }

    //Since the method dynamically assigns and unassigns topic partitions to the dummy consumer, need to synchronize this
    //to avoid multiple callers sharing same consumer instance
    dummyConsumer.synchronized {
      val partitionOffsetsAndLagTry: Try[List[PartitionOffsetAndLag]] = Try {
        val consumerGroupOffsets: util.Map[TopicPartition, OffsetAndMetadata] = adminClient
          .listConsumerGroupOffsets(groupId, new ListConsumerGroupOffsetsOptions().timeoutMs(timeoutMs))
          .partitionsToOffsetAndMetadata
          .get

        dummyConsumer.assign(consumerGroupOffsets.keySet)

        val topicEndOffsets: util.Map[TopicPartition, lang.Long] = dummyConsumer.endOffsets(consumerGroupOffsets.keySet, java.time.Duration.ofMillis(timeoutMs))

        //Unsubscribe once end offsets are determined.
        dummyConsumer.assign(util.Collections.emptySet())
        consumerGroupOffsets
          .entrySet()
          .asScala
          .filter(entry =>
            topicIds.isEmpty || topicIds.contains(entry.getKey.topic()))
          .map(entry => {
            val topicPartition = entry.getKey
            val endOffset = topicEndOffsets.get(topicPartition)
            val currentOffset = entry.getValue.offset()
            val topicLag = endOffset - currentOffset
            PartitionOffsetAndLag(topicPartition.topic(), topicPartition.partition(), currentOffset, endOffset, if (topicLag < 0) 0 else topicLag)
          }).toList
      }

      partitionOffsetsAndLagTry
    }
  }

  def getDummyConsumer(kafkaServerUrl: String): KafkaConsumer[String, String] = {
    val kafkaConsumerProperties: Properties = new Properties
    kafkaConsumerProperties.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, kafkaServerUrl)
    kafkaConsumerProperties.put(ConsumerConfig.CLIENT_ID_CONFIG, "consumer-" + UUID.randomUUID.toString)
    kafkaConsumerProperties.put(ConsumerConfig.GROUP_ID_CONFIG, "dummy-consumer-for-lag-detection")
    kafkaConsumerProperties.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, classOf[StringDeserializer])
    kafkaConsumerProperties.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, classOf[StringDeserializer])
    new KafkaConsumer[String, String](kafkaConsumerProperties)
  }

  /**
    * Utility method to create a kafka avro serializer as value stored in the topic is in avro format
    * Serializer needs schema registry URL in its config for its operation
    * Along with that key and value naming strategy is as per configuration settings
    * Refer to https://www.confluent.io/blog/put-several-event-types-kafka-topic/
    */
  def createKafkaAvroSerializer(schemaRegistryUrl: String,
                                keySubjectNameStrategy: String = "io.confluent.kafka.serializers.subject.RecordNameStrategy",
                                valueSubjectNameStrategy: String = "io.confluent.kafka.serializers.subject.RecordNameStrategy",
                                isKey: Boolean = false
                               ): KafkaAvroSerializer = {
    val kafkaAvroSerializer = new KafkaAvroSerializer()

    val serializerProperties: java.util.Map[String, String] = new java.util.HashMap[String, String]()
    serializerProperties.put(AbstractKafkaAvroSerDeConfig.SCHEMA_REGISTRY_URL_CONFIG, schemaRegistryUrl)
    serializerProperties.put(AbstractKafkaAvroSerDeConfig.KEY_SUBJECT_NAME_STRATEGY, keySubjectNameStrategy)
    serializerProperties.put(AbstractKafkaAvroSerDeConfig.VALUE_SUBJECT_NAME_STRATEGY, valueSubjectNameStrategy)

    kafkaAvroSerializer.configure(serializerProperties, isKey) // false for record values
    kafkaAvroSerializer
  }

  /**
    * Creation of kafka avro deserializer
    */
  def createKafkaAvroDeserializer(schemaRegistryUrl: String,
                                  keySubjectNameStrategy: String = "io.confluent.kafka.serializers.subject.RecordNameStrategy",
                                  valueSubjectNameStrategy: String = "io.confluent.kafka.serializers.subject.RecordNameStrategy",
                                  isKey: Boolean = false): KafkaAvroDeserializer = {
    val kafkaAvroDeserializer = new KafkaAvroDeserializer()

    val deserializerProperties: java.util.Map[String, String] = new java.util.HashMap[String, String]()
    deserializerProperties.put(AbstractKafkaAvroSerDeConfig.SCHEMA_REGISTRY_URL_CONFIG, schemaRegistryUrl)
    deserializerProperties.put(AbstractKafkaAvroSerDeConfig.KEY_SUBJECT_NAME_STRATEGY, keySubjectNameStrategy)
    deserializerProperties.put(AbstractKafkaAvroSerDeConfig.VALUE_SUBJECT_NAME_STRATEGY, valueSubjectNameStrategy)

    kafkaAvroDeserializer.configure(deserializerProperties, isKey)  // false for record values
    kafkaAvroDeserializer
  }
}

case class PartitionOffsetAndLag(topic: String, partition: Int, curOffset: Long, endOffset: Long, lag: Long)
