/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.kafka

import java.util.Properties

import org.apache.kafka.clients.consumer.ConsumerConfig
import org.apache.kafka.clients.producer.ProducerConfig
import org.apache.kafka.common.serialization.Serdes
import org.apache.kafka.streams.StreamsConfig
import org.apache.kafka.streams.errors.LogAndContinueExceptionHandler

object StreamsSettingUtils {
  /**
    * Utility method to create kafka streams configuration as Properties object
    */
  def createStreamsConfigPropertiesObject(kafkaSettings: KafkaSettings, streamSettings: KafkaStreamsSettings, basePath: String): Properties = {

    val configProperties = new Properties

    // Add all properties set in reference.conf/application.conf under section <microservicename>.kafka.streams.config.common

    // Mandatory ones
    configProperties.put(StreamsConfig.APPLICATION_ID_CONFIG, streamSettings.applicationId)
    configProperties.put(StreamsConfig.BOOTSTRAP_SERVERS_CONFIG, kafkaSettings.bootstrapServers)

    if (streamSettings.cacheMaxBytesBufferingOption.isDefined) configProperties.put(StreamsConfig.CACHE_MAX_BYTES_BUFFERING_CONFIG, streamSettings.cacheMaxBytesBufferingOption.get)
    if (streamSettings.commitIntervalMsOption.isDefined) configProperties.put(StreamsConfig.COMMIT_INTERVAL_MS_CONFIG, streamSettings.commitIntervalMsOption.get)

    // Number of standby replicas. Standby replicas are shadow copies of local state stores.
    configProperties.put(StreamsConfig.NUM_STANDBY_REPLICAS_CONFIG, streamSettings.standbyReplicas.toString)

    /*
      Specifies the replication factor of internal topics that Kafka Streams creates when local states are used
      or a stream is repartitioned for aggregation. Replication is important for fault tolerance.
      Without replication even a single broker failure may prevent progress of the stream processing application.
      It is recommended to use a similar replication factor as source topics. It is currently configured as 2 mostly.
     */
    configProperties.put(StreamsConfig.REPLICATION_FACTOR_CONFIG, streamSettings.replicationFactorForStreamStateStores.toString)

    // A producer specific configurationn, meant for resiliency
    configProperties.put(StreamsConfig.producerPrefix(ProducerConfig.ACKS_CONFIG), streamSettings.producerAcks)

    // Specify default serializer and deserializer whenever data is materialized, read/write from topics, state stores
    configProperties.put(StreamsConfig.DEFAULT_KEY_SERDE_CLASS_CONFIG, Serdes.String().getClass)
    configProperties.put(StreamsConfig.DEFAULT_VALUE_SERDE_CLASS_CONFIG, Serdes.String().getClass)

    /*
    Default deserialization exception handler allows you to manage record exceptions that fail to deserialize.
    This can be caused by corrupt data, incorrect serialization logic, or unhandled record types.
    Default is LogAndFailExceptionHandler which is changed here to LogAndContinueExceptionHandler as it is not necessary to fail
    if one record fails to deserialize for whatever reason.
     */
    configProperties.put(StreamsConfig.DEFAULT_DESERIALIZATION_EXCEPTION_HANDLER_CLASS_CONFIG, classOf[LogAndContinueExceptionHandler].getName)

    /*
    default production exception handler allows you to manage exceptions triggered when trying to interact with a broker such as attempting to produce a record that is too large.
    By default, Kafka provides and uses the DefaultProductionExceptionHandler that always fails when these exceptions occur.
    Note: DEFAULT_PRODUCTION_EXCEPTION_HANDLER_CLASS_CONFIG in StreamsConfig is marked private by mistake
    Change it to IgnoreRecordTooLargeException
     */
    configProperties.put("default.production.exception.handler", classOf[IgnoreRecordTooLargeException].getName)

    /*
     Read the topic from the very beginning if no previous consumer offsets are found for this app.
     Resetting an app will set any existing consumer offsets to zero,
     so setting this config combined with resetting will cause the application to re-process all the input data in the topic.
      */
    configProperties.put(StreamsConfig.consumerPrefix(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG), streamSettings.autoOffsetReset)
    configProperties.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, streamSettings.autoOffsetReset)

    /*
    This specifies the number of stream threads in an instance of the Kafka Streams application.
    The stream processing code runs in these thread.
     */
    configProperties.put(StreamsConfig.NUM_STREAM_THREADS_CONFIG, streamSettings.numStreamThreads.toString)

    /*
    The state directory. Kafka Streams persists local states under the state directory.
    Each application has a subdirectory on its hosting machine that is located under the state directory.
    The name of the subdirectory is the application ID. The state stores associated with the application are created under this subdirectory.
     */
    configProperties.put(StreamsConfig.STATE_DIR_CONFIG, s"$basePath/${streamSettings.stateDir}")

    if (streamSettings.stateCleanupDelayMsOption.isDefined) configProperties.put(StreamsConfig.STATE_CLEANUP_DELAY_MS_CONFIG, streamSettings.stateCleanupDelayMsOption.get)

    // delivery.timeout.ms is a new config in streams 2.3. Since commons is still using older streams 1.1.0, using
    // hardcoded config for now
    if (streamSettings.deliveryTimeoutMsOption.isDefined) configProperties.put("delivery.timeout.ms", streamSettings.deliveryTimeoutMsOption.get)

    // request.timeout.ms is a common property name for consumer and producer. Hence need to add prefix to distinguish.
    if (streamSettings.producerRequestTimeoutMsOption.isDefined) configProperties.put(StreamsConfig.producerPrefix(StreamsConfig.REQUEST_TIMEOUT_MS_CONFIG), streamSettings.producerRequestTimeoutMsOption.get)

    if (streamSettings.producerLingerMsOption.isDefined) configProperties.put(ProducerConfig.LINGER_MS_CONFIG, streamSettings.producerLingerMsOption.get)

    if (streamSettings.producerBatchSizeOption.isDefined) configProperties.put(ProducerConfig.BATCH_SIZE_CONFIG, streamSettings.producerBatchSizeOption.get)

    /*
      Maximum amount of time a stream task will stay idle when not all of its partition buffers contain records,
      to avoid potential out-of-order record processing across multiple input streams
      stream-table join leads to race conditions at times. Setting this option to other than default value (0) will help
      those situations.
      Reference: https://stackoverflow.com/questions/54412041/kafka-kstream-ktable-join-race-condition

      This option came post 2.2.0, common refers to library 1.1.0 and hence cannot use a constant to set it.
      The actual streams applications RDP, CD, CR use 2.3.0 so it will be fine.
     */
    if (streamSettings.maxTaskIdleMsOption.isDefined) configProperties.put("max.task.idle.ms", streamSettings.maxTaskIdleMsOption.get)

    /*
    Following are the ones which are not configured as the defaults are fine
      processing.guarantee        = at_least_once default. Can be either at-least-once (default) or exactly-once.
      client.id                   = client id used for kafka producer and consumer running as part of streams app
      commit.interval.ms          = 30000 milliseconds (at-least-once) / 100 milliseconds (exactly-once)
      default.timestamp.extractor = Will be message timestamp which will be provided correctly from producer
                                    Default extractor is FailOnInvalidTimestamp. This extractor retrieves built-in timestamps
                                    that are automatically embedded into Kafka messages by the Kafka producer client.
                                    service-agent-data-receiver extracts the timestamp from audit-events and marks it as message timestamp.
                                    log.message.timestamp.type is CreateTime by default
     */
    
    configProperties
  }
}
