/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.kafka

import akka.actor.{Actor, ActorLogging, ActorRef, Props}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.kafka.KafkaUtils._
import com.netapp.df.uba.server.common.misc.model._
import org.apache.kafka.clients.admin.AdminClient

import scala.util.{Failure, Success, Try}

/**
  * Topic Manager actor handles the topic creation, checks, modifications for a micro-service.
  */
object TopicManager {
  final val ActorName: String = "topicManager"

  def props(kafkaSettings: KafkaSettings, requestorRef: ActorRef): Props = Props(new TopicManager(kafkaSettings, requestorRef))

  // A message to send to this actor to start the topic management
  case object ManageTopics extends Command

  // Context used for action on topics
  final val ContextTopicManagement: String = "topics"
}

class TopicManager(kafkaSettings: KafkaSettings, requestorRef: ActorRef) extends Actor with ActorLogging {

  import TopicManager._

  // Create the admin client for kafka
  private val adminClientTry: Try[AdminClient] = KafkaUtils.createAdminClient(
    kafkaSettings.bootstrapServers,
    kafkaSettings.adminClientRequestTimeout
  )

  // Close the admin client
  override def postStop(): Unit = KafkaUtils.closeAdminClient(adminClientTry)

  // Receive method for TopicManager
  override def receive: Receive = LoggingReceive {

    // Handles a single message and returns responses back to the sender
    case ManageTopics =>

      // Check if admin client was created in preStart or not
      adminClientTry match {

        case Failure(ex) =>
          log.warning(s"Failed to get admin client with bootstrap servers ${kafkaSettings.bootstrapServers}. Reason: ${ex.getMessage}")

          // Send failed case
          requestorRef ! ActionFailed(ContextTopicManagement, Error(ErrorCode.ServiceNotRunning, ex.getMessage, kafkaSettings.bootstrapServers))

        case Success(adminClient) =>
          /*
            For each of the topics asked for
              - Producer Topics - Create it, if absent, check config if present and alter if required
              - Consumer Topics - Check for existence

            Return a consolidated failure result for any of the topics that fails the above checks
            */
          val failedResults: Map[String, Try[Boolean]] = kafkaSettings.topicMap.values
            .map {
              case producerTopic: ProducerTopic if !producerTopic.isPipelineTopic =>
                // For each topic that is producer based, create it if it does not exist
                producerTopic.name -> handleProducerTopic(producerTopic, adminClient, kafkaSettings.zookeeperServers)

              case consumerTopic: ConsumerTopic =>
                // For each topic that is to be consumed, check for presence
                consumerTopic.name -> handleConsumerTopic(consumerTopic, adminClient, kafkaSettings.zookeeperServers)

              case actuallyConsumerTopic: ProducerTopic if actuallyConsumerTopic.isPipelineTopic =>
                /*
                  Special case: In pipeline model, the topic is created by the consumer job
                  where as in the older model, the topic was created by the producer.
                  To handle this scenario, this flag is there which says to create only for pipeline model.
                  If it is true, then skip topic checks in SFJM topic management during bootstrap, and simply mark
                  it as success.
                 */
                actuallyConsumerTopic.name -> Success(true)

              case infoTopic: InfoTopic =>
                // Nothing to be done, mark it as success
                infoTopic.name -> Success(true)
            }
            .filterNot(_._2.isSuccess) // Only include failed results
            .toMap

          failedResults.size match {
            case 0 =>
              // Send the action succeeded message back to the micro-service
              requestorRef ! ActionSucceeded(ContextTopicManagement)

            case _ =>
              // Create an overall list of errors that takes into account all errors.
              val allErrors: List[Error] = failedResults
                .map(entry => (entry._1, entry._2.asInstanceOf[Failure[Boolean]]))
                .map(entry => Error(ErrorCode.InternalError, entry._2.exception.getMessage, entry._1))
                .toList

              // Send the action failed event back to the micro-service
              requestorRef ! ActionFailed(ContextTopicManagement,
                Error(ErrorCode.InternalError, "Failed in managing topics", ContextTopicManagement, Some(allErrors)))
          }
      }

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message that is not handled: " + somethingElse)
  }
}