/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.actor

import akka.actor.{Actor, ActorLogging, Props}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.misc.model.ServiceName.ServiceName
import com.netapp.df.uba.server.common.misc.model._

/**
  * StateManager tracks state of various services within the micro-service.
  */
object StateManager {
  final val ActorName: String = "stateManager"
  final val ActorPath: String = s"/user/appManager/$ActorName"

  def props(internalServiceList: List[ServiceName]): Props = Props(new StateManager(internalServiceList))
}

class StateManager(internalServiceList: List[ServiceName]) extends Actor with ActorLogging {

  /**
    * receive is implemented by sending it the initial state with Initializing for LdapConnectionChecker
    */
  override def receive: Receive = {
    val serviceMap: Map[ServiceName, ServiceStatus] = internalServiceList
      .map(serviceItem => serviceItem -> Initializing)
      .toMap

    stateReceive(serviceMap)
  }

  /**
    * Maintains on an on-going basis the service state map for each service
    * Updates the internal service state map whenever a new service event message is received by this actor.
    */
  private def stateReceive(serviceStateMap: Map[ServiceName, ServiceStatus]): Receive = LoggingReceive {

    case ServiceEventMessage(serviceName, state) =>
      val newServiceStateMap: Map[ServiceName, ServiceStatus] = updateStateMap(serviceName, state, serviceStateMap)

      /*
        For every state transition, a check is done on the overall component state or state of the micro-service.
       */
      val componentState = calculateComponentState(newServiceStateMap)
      componentState match {
        /*
         If the overall component state has reached an error, send the ComponentInError message back to sender
         Its upto the sender (a specific micro-service) to decide what to do with the error state
         Could be retry, could be full exit or some cleanup and then exit.
          */
        case ComponentError(reason) =>
          sender() ! ComponentInError(reason)

        case _ =>
          // Nothing to do
      }
      context.become(stateReceive(newServiceStateMap))

    case GetStateDetails => sender() ! StateDetails(calculateComponentState(serviceStateMap), serviceStateMap)

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message that is not handled: " + somethingElse)
  }

  /**
    * Add the new state(serviceState) entry for service(serviceName) in the state map
    */
  private def updateStateMap(serviceName: ServiceName,
                             serviceState: ServiceStatus,
                             currentServiceStateMap: Map[ServiceName, ServiceStatus]): Map[ServiceName, ServiceStatus] = {
    /*
     Add the new service->state entry to the map as long as the key is present.
     Ideally, only keys present in map should come to this actor. So add a warning if some other key comes
      */
    val newServiceStateMap = if (currentServiceStateMap.contains(serviceName)) currentServiceStateMap + (serviceName -> serviceState)
    else {
      log.warning(s"Unsupported entry received by StateManager. ServiceName: $serviceName, ServiceState: $serviceState.")
      currentServiceStateMap
    }

    log.info(s"Updated service state map in StateManager is: $newServiceStateMap")
    newServiceStateMap
  }

  /**
  Calculate the current component state based on values of all internal services in the map.
    Starting with : ConnectorInitializing

    Any one item has Failed -> Failed
    Any one item is Initializing -> Initializing
    Any one item is Stopping -> Stopping
    All items are Stopped -> Stopped
    Running
    */
  private def calculateComponentState(serviceStateMap: Map[ServiceName, ServiceStatus]): ComponentState = {
    if (serviceStateMap.exists(_._2.isInstanceOf[Failed])) {
      if (serviceStateMap.size > 1) ComponentError("Error state. " +
          serviceStateMap
            .filter(_._2.isInstanceOf[Failed])
            .map(entry => s"Service name: ${entry._1.toString}. Reason for failure: ${entry._2.asInstanceOf[Failed].reason}")
            .toList
            .mkString(", "))
      else ComponentError(serviceStateMap
            .filter(_._2.isInstanceOf[Failed])
            .map(entry => entry._2.asInstanceOf[Failed].reason)
            .toList
            .mkString(", "))
    }
    else if (serviceStateMap.exists(_._2 == Initializing)) ComponentInitializing
    else if (serviceStateMap.exists(_._2 == Stopping)) ComponentStopping
    else if (serviceStateMap.forall(service => service._2 == Stopped || service._2 == NotEnabled)) ComponentStopped
    else ComponentRunning
  }
}