/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.api

import akka.http.scaladsl.model.HttpRequest
import akka.http.scaladsl.model.StatusCodes.Unauthorized
import akka.http.scaladsl.server
import akka.http.scaladsl.server.Directives.{complete, optionalCookie, optionalHeaderValueByName}
import akka.http.scaladsl.server.{Directive, Directive1, Route}
import com.netapp.df.uba.server.common.misc.apikey.ApiKey
import com.netapp.df.uba.server.common.misc.interceptor.InterceptorPublicApiKeyValidator
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model.Role.Role
import com.netapp.df.uba.server.common.misc.model._
import com.netapp.df.uba.server.common.misc.utils.RbUtils
import com.netapp.oci.jwt.{ExpiredTokenException, InvalidTokenException, JWTAuthData, JWTUtils}

import java.util
import java.util.{Calendar, Collections, ResourceBundle}
import scala.collection.JavaConverters._
import scala.util.{Failure, Success, Try}

object AuthDirectives {
  // Environment where the JWT token validation will be bypassed
  val BypassVaultAuthEnvironment = "novault"
}

/**
  * A set of common authentication directives to be used across micro-services
  * that are scala based and expose REST APIs in SaaS layer
  */
trait AuthDirectives extends LogTrait with JsonFormats {

  // A map of empty cookies
  private val emptyCookiesMap: util.Map[String, String] = Collections.emptyMap()

  // A set of empty headers
  private val emptyHeadersMap: util.Map[String, String] = Collections.emptyMap()

  // Resource bundle
  private val resourceBundle: ResourceBundle = ResourceBundle.getBundle("ExtMessageBundleMisc")

  /**
    * A custom directive for checking if caller has valid JWT token and to derive the tenant context out of it.
    * The JWT token can come in three ways:
    * <ol>
    * <li>Custom header - PublicApi Request</li>
    * <li>Cookie - Mostly used in UI</li>
    * <li>Header - Direct API call and PublicApi Requests</li>
    * </ol>
    *
    * @param environment  Deployment environment
    * @param httpRequest  Http request
    * @param allowedRoles Set of allowed roles
    * @return A tuple of {@link TenantContext}, {@link JwtTokenContext}, Option of {@link ApiKey}
    */
  def authenticateAndExtractTenant(environment: String,
                                   httpRequest: Option[HttpRequest] = None,
                                   allowedRoles: Set[Role] = Set(Role.Admin)): Directive[(TenantContext, JwtTokenContext, Option[ApiKey])] =

    server.Directive { inner ⇒
      val interceptorPublicApiKeyValidator = getInterceptorPublicApiKeyValidator(httpRequest)
      val publicApiKeyStatusTry = interceptorPublicApiKeyValidator.getCloudSecureJwtToken

      publicApiKeyStatusTry match {
        case Success(_) =>

          //We process Authorization part only if the public API Key is present otherwise we will handle the
          //authorization part to default processing defined below
          val publicApiRequestStatusTry = interceptorPublicApiKeyValidator.validatePublicApiRequest

          publicApiRequestStatusTry match {

            case Success((jwtAuthData, apiKey)) =>
              val tenantContext = TenantContext(jwtAuthData.getTenant)
              logger.debug(s"Tenant: ${tenantContext.tenantId}; JWT token verified successfully.")

              // Pass Jwt token context information
              val jwtTokenContext = JwtTokenContext(jwtAuthData)
              inner(Tuple3(tenantContext, jwtTokenContext, Some(apiKey)))

            case Failure(exception) =>
              logger.error("Failure when validating the token", exception)
              completeWithTokenExceptionMessage(exception)
          }

        case Failure(_) =>
          // Check if there is a cookie named as JWT
          optionalCookie(JWTUtils.JWT_COOKIE_ID) {
            case Some(cookie) =>

              // Create a cookie map
              val cookies = Map(cookie.name -> cookie.value).asJava

              // Authenticate using JWT
              doJWTAuthentication(cookies, emptyHeadersMap, environment, allowedRoles) match {
                // Successful validation
                case Success(jwtAuthData) =>
                  jwtAuthData.getClaims
                  // Pass tenant information to inner routes
                  val tenantContext = TenantContext(jwtAuthData.getTenant)
                  logger.debug(s"Tenant: ${tenantContext.tenantId}; JWT token in cookie verified successfully.")

                  // Pass Jwt token context information
                  val jwtTokenContext = JwtTokenContext(jwtAuthData)

                  inner(Tuple3(tenantContext, jwtTokenContext, None))

                // Authentication failed
                case Failure(exception) if exception.isInstanceOf[InvalidTokenException] || exception.isInstanceOf[ExpiredTokenException] =>
                  if (logger.isDebugEnabled) exception.printStackTrace()
                  completeWithTokenExceptionMessage(exception)

                case Failure(exception: AuthorizationException) =>
                  if (logger.isDebugEnabled) exception.printStackTrace()
                  completeWithUnAuthorizedMessage

                case Failure(exception) =>
                  exception.printStackTrace()
                  completeWithJwtInvalidMessage
              }

            // No cookie, check header
            case None =>
              // Check if there is a header with the given JwtHeaderName
              optionalHeaderValueByName(JWTUtils.JWT_HEADER_NAME) {
                case Some(jwtToken) =>

                  // Create a header map
                  val headers = Map(JWTUtils.JWT_HEADER_NAME -> jwtToken).asJava

                  // Authenticate using JWT
                  doJWTAuthentication(emptyCookiesMap, headers, environment, allowedRoles) match {

                    // Successful validation
                    case Success(jwtAuthData) =>
                      //Pass tenant information to inner routes
                      val tenantContext = TenantContext(jwtAuthData.getTenant)
                      logger.debug(s"Tenant: ${tenantContext.tenantId}; JWT token in header verified successfully.")

                      // Pass Jwt token context information
                      val jwtTokenContext = JwtTokenContext(jwtAuthData)

                      inner(Tuple3(tenantContext, jwtTokenContext, None))

                    // Authentication failed
                    case Failure(exception) if exception.isInstanceOf[InvalidTokenException] || exception.isInstanceOf[ExpiredTokenException] =>
                      if (logger.isDebugEnabled) exception.printStackTrace()
                      completeWithTokenExceptionMessage(exception)

                    case Failure(exception: AuthorizationException) =>
                      if (logger.isDebugEnabled) exception.printStackTrace()
                      completeWithUnAuthorizedMessage

                    case Failure(exception) =>
                      exception.printStackTrace()
                      completeWithJwtInvalidMessage
                  }

                // No cookie or header
                case None =>
                  completeWithJwtInvalidMessage
              }
          }
      }
    }

  /**
    * Custom directive responsible to validate the category of apiKey to endpoint in case it come from a custom header.
    * This directive is ONLY applied for public api endpoints.
    *
    * @param jwtTokenContext         Token context
    * @param httpRequest             Http Request
    * @param restMethodDocumentation RestMethodDocumentation wraps the endpoint categories
    * @param apiKey                  ApiKey passed to validate allowed categories
    * @return Directive with boolean flag
    */
  def validateCategory(jwtTokenContext: JwtTokenContext, httpRequest: HttpRequest,
                       restMethodDocumentation: RestMethodDocumentation, apiKey: Option[ApiKey]): Directive1[Boolean] =
    Directive { inner =>
      if (apiKey.isDefined) {
        val interceptorPublicApiKeyValidator = getInterceptorPublicApiKeyValidator(Some(httpRequest))
        /*
        * In case of public API endpoints, we need to know the the category it belongs to,
        * to properly validate the apiKey against the category.
        * RestMethodDocumentation object which has the category info, is passed along with the request for further validation here.
        */
        interceptorPublicApiKeyValidator.validateApiKeyCategory(jwtTokenContext, restMethodDocumentation, apiKey.get) match {
          case Success(_) =>
            logger.debug(s"Tenant: ${jwtTokenContext.tenantId}; ApiKey category validated successfully")
            inner(Tuple1(true))

          case Failure(ex) =>
            completeWithTokenExceptionMessage(ex)
        }
      }
      else
        inner(Tuple1(true)) // No validation required since JWT is passed via cookie or auth header
    }

  /**
    * Utility method to complete a REST API request with ErrorCode of invalid credentials
    */
  private def completeWithJwtInvalidMessage: Route = {
    val errorMessage: String = RbUtils.rbMessage(resourceBundle, RbUtils.RBKEY_API_JWT_TOKEN_INVALID)
    complete(Unauthorized -> Error(ErrorCode.InvalidCredentials, errorMessage, "JWT Token"))
  }

  /**
    * Utility method to complete a REST API request with ErrorCode of unauthorized user
    */
  private def completeWithTokenExceptionMessage(exception: Throwable): Route = {
    if (exception.isInstanceOf[ExpiredTokenException]) complete(Unauthorized -> Error(ErrorCode.TokenExpired, exception.getMessage, "JWT Token"))
    else complete(Unauthorized -> Error(ErrorCode.InvalidToken, exception.getMessage, "JWT Token"))
  }

  /**
    * Utility method to complete a REST API request with ErrorCode of unauthorized user
    */
  private def completeWithUnAuthorizedMessage: Route = {
    val errorMessage: String = RbUtils.rbMessage(resourceBundle, RbUtils.RBKEY_API_USER_NOT_AUTHORIZED)
    complete(Unauthorized -> Error(ErrorCode.NotAuthorized, errorMessage, "JWT Token"))
  }

  /**
    * Internal method to do the actual JWT authentication using the JWT library
    */
  private def doJWTAuthentication(cookiesMap: util.Map[String, String],
                                  headersMap: util.Map[String, String],
                                  environment: String,
                                  allowedRoles: Set[Role]): Try[JWTAuthData] = {
    if (environment == AuthDirectives.BypassVaultAuthEnvironment) {
      logger.debug("Creating JWTAuthData to bypass authentication for dev environment.")

      val calendar = Calendar.getInstance()
      calendar.set(Calendar.YEAR, calendar.get(Calendar.YEAR) + 1)

      Try(
        new JWTAuthData(
          "tenant1",
          "dev_user",
          calendar.getTime,
          "admin", "user", "guest"
        )
      )
    } else {
      logger.debug("Calling vault service to validate JWT token...")

      // Proper JWT authentication using the secret from the vault
      val jwtAuthDataTry = Try(JWTUtils.fromJWT(cookiesMap, headersMap))

      // Validate the result from vault on token
      jwtAuthDataTry match {
        // Case when jwt is valid, do the role checks
        case Success(jwtAuthData) if jwtAuthData != null =>

          // Role check for APIs
          val rolesInJwtToken: Set[String] = jwtAuthData.getRoles.asScala.toSet
          val allowedRolesSet: Set[String] = allowedRoles.map(_.toString)

          // If allowed roles and roles in JWT token has at least one element in common, then it is allowed.
          if (rolesInJwtToken.intersect(allowedRolesSet).nonEmpty) jwtAuthDataTry
          else Try(throw new AuthorizationException("Not authorized."))

        case Success(null) =>
          Try(throw new Exception("Invalid JWT."))

        case Failure(_) =>
          // Failure in token validation. Return the failure.
          jwtAuthDataTry
      }
    }
  }

  private case class AuthorizationException(message: String, cause: Throwable) extends Exception(message) {
    def this(message: String) = this(message, null: Throwable)

    override def getCause: Throwable = cause
  }

  // Used for UT override
  def getInterceptorPublicApiKeyValidator(httpRequest: Option[HttpRequest]): InterceptorPublicApiKeyValidator =
    InterceptorPublicApiKeyValidator(httpRequest)
}