/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.api

import java.util.UUID

import akka.http.scaladsl.marshallers.sprayjson.SprayJsonSupport
import com.netapp.df.uba.server.common.misc.dbmodel.AuditEventType.AuditEventType
import com.netapp.df.uba.server.common.misc.dbmodel.CaseSensitiveNature.CaseSensitiveNature
import com.netapp.df.uba.server.common.misc.dbmodel.ConnectorType.ConnectorType
import com.netapp.df.uba.server.common.misc.dbmodel.DataVolumeState.DataVolumeState
import com.netapp.df.uba.server.common.misc.dbmodel.DataVolumeType.DataVolumeType
import com.netapp.df.uba.server.common.misc.dbmodel.MetadataEventType.MetadataEventType
import com.netapp.df.uba.server.common.misc.dbmodel.MetadataObjectType.MetadataObjectType
import com.netapp.df.uba.server.common.misc.dbmodel.Protocol.Protocol
import com.netapp.df.uba.server.common.misc.dbmodel.ProtocolDb.ProtocolDb
import com.netapp.df.uba.server.common.misc.dbmodel.UserDirectoryLocation.UserDirectoryLocation
import com.netapp.df.uba.server.common.misc.dbmodel.UserIdType.UserIdType
import com.netapp.df.uba.server.common.misc.dbmodel._
import com.netapp.df.uba.server.common.misc.model.ErrorCode.ErrorCode
import com.netapp.df.uba.server.common.misc.model.ServiceName.ServiceName
import com.netapp.df.uba.server.common.misc.model.{ComponentStateApi, Error, ErrorCode, MessageResponse, Property, ServiceApi, ServiceName}
import spray.json.{DefaultJsonProtocol, DeserializationException, JsString, JsValue, JsonFormat, RootJsonFormat}

trait JsonFormats extends SprayJsonSupport with DefaultJsonProtocol {

  /**
    * Implicit JSON formatting for UUIds
    */
  implicit object UUIDJsonFormat extends RootJsonFormat[UUID] {
    override def read(json: JsValue): UUID = UUID.fromString(json.convertTo[String])

    override def write(uuid: UUID) = JsString(uuid.toString)
  }

  /**
    * Computes json format for a scala enumeration
    */
  def jsonEnumFormatter[T <: Enumeration](enumeration: T): JsonFormat[T#Value] = new JsonFormat[T#Value] {
    def write(obj: T#Value) = JsString(obj.toString)

    def read(json: JsValue): T#Value = json match {
      case JsString(txt) => enumeration.withName(txt)
      case other => throw DeserializationException(s"Expected a value from Enumeration $enumeration instead of $other")
    }
  }

  // Actual implicits to convert domain specific model classes

  // Commonly used ones for error and status and property
  implicit val errorCodeEnumFormatter: JsonFormat[ErrorCode] = jsonEnumFormatter(ErrorCode)
  implicit val messageResponseFormatter: RootJsonFormat[MessageResponse] = jsonFormat1(MessageResponse.apply)

  // Refer to https://github.com/spray/spray-json#jsonformats-for-recursive-types since Error is a recursive object
  implicit val errorFormatter: RootJsonFormat[Error] = rootFormat(lazyFormat(jsonFormat(Error, "code", "message", "target", "details")))

  implicit val serviceNameEnumFormatter: JsonFormat[ServiceName] = jsonEnumFormatter(ServiceName)
  implicit val serviceApiFormatter: RootJsonFormat[ServiceApi] = jsonFormat3(ServiceApi.apply)
  implicit val componentStateApiFormatter: RootJsonFormat[ComponentStateApi] = jsonFormat4(ComponentStateApi.apply)

  implicit val propertyFormatter: RootJsonFormat[Property] = jsonFormat2(Property.apply)

  /*
  Enum formatters used for enums in audit events, datasources, data volumes and user directories.
  Since they are used by agent-data-receiver as well as elasticsearch-objects-store, they are defined here.
   */
  implicit val auditEventTypeEnumFormatter: JsonFormat[AuditEventType] = jsonEnumFormatter(AuditEventType)
  implicit val userIdTypeEnumFormatter: JsonFormat[UserIdType] = jsonEnumFormatter(UserIdType)

  implicit val metadataEventTypeEnumFormatter: JsonFormat[MetadataEventType] = jsonEnumFormatter(MetadataEventType)
  implicit val metadataObjectTypeEnumFormatter: JsonFormat[MetadataObjectType] = jsonEnumFormatter(MetadataObjectType)

  implicit val connectorTypeEnumFormatter: JsonFormat[ConnectorType] = jsonEnumFormatter(ConnectorType)

  implicit val protocolEnumFormatter: JsonFormat[Protocol] = jsonEnumFormatter(Protocol)
  implicit val protocolFormatter: RootJsonFormat[ProtocolInfo] = jsonFormat3(ProtocolInfo.apply)

  implicit val protocolDbEnumFormatter: JsonFormat[ProtocolDb] = jsonEnumFormatter(ProtocolDb)
  implicit val protocolDbFormatter: RootJsonFormat[ProtocolInfoDb] = jsonFormat3(ProtocolInfoDb.apply)

  implicit val dataVolumeTypeEnumFormatter: JsonFormat[DataVolumeType] = jsonEnumFormatter(DataVolumeType)
  implicit val caseSensitiveEnumFormatter: JsonFormat[CaseSensitiveNature] = jsonEnumFormatter(CaseSensitiveNature)
  implicit val dataVolumeStateEnumFormatter: JsonFormat[DataVolumeState] = jsonEnumFormatter(DataVolumeState)

  implicit val userDirectoryLocationEnumFormatter: JsonFormat[UserDirectoryLocation] = jsonEnumFormatter(UserDirectoryLocation)
}
