/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.dbmodel

import com.netapp.df.uba.server.common.misc.dbmodel.ActionStatus.ActionStatus
import com.netapp.df.uba.server.common.misc.dbmodel.ActionType.ActionType
import com.netapp.df.uba.server.common.misc.dbmodel.ExpiryMode.ExpiryMode
import com.netapp.df.uba.server.common.misc.dbmodel.LimitUserAccessMode.LimitUserAccessMode
import com.netapp.df.uba.server.common.misc.dbmodel.UserBlockType.UserBlockType
import com.netapp.df.uba.server.common.misc.model.{UserAccessResult, VolumeSnapshotResult}

/**
  * Class representing the database model of AlertActionResult in Elasticsearch
  *
  * alertId - Always filled for "Create Snapshot" action, optional for "User restriction"
  * as that action can be initiated in alert or user context<br>
  * userId - Always filled for "User restriction" action and empty for "Create Snapshot" action
  */
case class AlertActionResultDb(tenantId: String,
                               actionId: String,
                               actionType: ActionType,
                               alertId: Option[String],
                               userId: Option[String],
                               isAutomatic: Boolean,
                               startTimestamp: Long, // set when an alert action is created
                               endTimestamp: Option[Long] = None, // set when action is either success or failure
                               createdBy: Option[String] = None, // only applicable for manual actions, ie when isAutomatic is false
                               subActions: List[SubAction])

/**
  * Class representing the SubAction for AlertActionResultDb
  */
case class SubAction(deviceId: Option[String] = None,
                     volumeId: Option[String] = None,
                     policyId: Option[String] = None,
                     policyName: Option[String] = None,
                     actionStatus: ActionStatus,
                     actionResult: Option[ResponseDb])


object UserAccessResultDb {
  def apply(userAccessResult: UserAccessResult, userBlockType: UserBlockType): UserAccessResultDb = {
    UserAccessResultDb(userAccessResult.userIps,
      userAccessResult.sid,
      Some(userBlockType),
      userAccessResult.limitUserAccessMode,
      userAccessResult.expiryMode,
      userAccessResult.expiryTimestamp,
      userAccessResult.duration,
      userAccessResult.error)
  }
}

object VolumeSnapshotResultDb {
  def apply(volumeSnapshotResult: VolumeSnapshotResult): VolumeSnapshotResultDb = {
    VolumeSnapshotResultDb(volumeSnapshotResult.snapshotName,
      volumeSnapshotResult.snapshotTimestamp,
      volumeSnapshotResult.error)

  }
}

case class VolumeSnapshotResultDb(snapshotName: Option[String] = None,
                                  snapshotTimestamp: Option[Long] = None,
                                  error: Option[String] = None)

case class UserAccessResultDb(userIps: Option[List[String]],
                              sid: Option[String],
                              userBlockType: Option[UserBlockType],
                              limitUserAccessMode: Option[LimitUserAccessMode],
                              expiryMode: Option[ExpiryMode] = None,
                              expiryTimestamp: Option[Long] = None,
                              duration: Option[Long] = None,
                              error: Option[String] = None)

/**
  * Enum to specify underlying userBlockType mechanism for a given UserAccess subAction.<br>
  * SubActions are added for each datasource and different userBlockType are possible for every datasource.
  */
object UserBlockType extends Enumeration {
  type UserBlockType = Value

  val IPBased: UserBlockType = Value("IPBased")
  val SidBased: UserBlockType = Value("SidBased")
  val IPAndSidBased: UserBlockType = Value("IPAndSidBased")
}

case class VolumeSnapshotDetails(volumeSnapshotResultDb: VolumeSnapshotResultDb) extends ResponseDb

case class UserAccessDetails(userAccessResultDb: UserAccessResultDb) extends ResponseDb

trait ResponseDb

/**
  * Enumeration representing the alert action statuses
  */
object ActionStatus extends Enumeration {
  type ActionStatus = Value

  val InProgress: ActionStatus = Value("InProgress")
  val Success: ActionStatus = Value("Success")
  val Failure: ActionStatus = Value("Failure")
}

/**
  * This mode handles only the user restriction mode
  * read only means user is restricted for read only
  * block means user is fully restricted for both read and write .
  */
object LimitUserAccessMode extends Enumeration {
  type LimitUserAccessMode = Value

  val READ_ONLY: LimitUserAccessMode = Value("readOnly")
  val BLOCK: LimitUserAccessMode = Value("block")
}

/**
  * Expiry Mode enum
  * User restriction can expire based on Time or it may not expire in that
  * case it will be permanent.
  */
object ExpiryMode extends Enumeration {
  type ExpiryMode = Value

  val TimeBased: ExpiryMode = Value("timeBased")
  val Permanent: ExpiryMode = Value("permanent")
}