/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.dbmodel

import com.netapp.df.uba.server.common.misc.dbmodel.AlertSeverity.AlertSeverity
import com.netapp.df.uba.server.common.misc.dbmodel.AlertStatus.AlertStatus
import com.netapp.df.uba.server.common.misc.dbmodel.AlertTrigger.AlertTrigger
import com.netapp.df.uba.server.common.misc.dbmodel.AlertType.AlertType
import com.netapp.df.uba.server.common.misc.dbmodel.TrustScore.TrustScore


/**
  * Enumeration representing the alert severities
  */
object AlertSeverity extends Enumeration {
  type AlertSeverity = Value

  val Ok: AlertSeverity = Value("ok")
  val Low: AlertSeverity = Value("low")
  val Warning: AlertSeverity = Value("warning")
  val Critical: AlertSeverity = Value("critical")
}

/**
  * Enumeration representing the alert statuses
  */
object AlertStatus extends Enumeration {
  type AlertStatus = Value

  val New: AlertStatus = Value("New")
  val InProgress: AlertStatus = Value("InProgress")
  val Resolved: AlertStatus = Value("Resolved")
  val Dismissed: AlertStatus = Value("Dismissed")
}

/**
  * Enumeration representing the alert type
  */
object AlertType extends Enumeration {
  type AlertType = Value

  val UserActivityRate: AlertType = Value("User Activity Rate")
  val CommunityActivityRate: AlertType = Value("Community Activity Rate")
  val UserVersusCommunityActivityRate: AlertType = Value("User vs. Community Activity Rate")
  val UserJoinedManyCommunities: AlertType = Value("User Joined Many Communities")
  val RansomwareAttack: AlertType = Value("Ransomware Attack")
  val Sabotage: AlertType = Value("Sabotage")
  val DataDestructionFileDeletionAttack: AlertType = Value("Data Destruction - File Deletion")
}

/**
  * Enumeration representing the anomaly detection algorithm being used
  */
object AnomalyDetectionAlgorithmName extends Enumeration {
  type AnomalyDetectionAlgorithmName = Value
  val Red : AnomalyDetectionAlgorithmName  = Value("RED")
  val TTest : AnomalyDetectionAlgorithmName = Value("tTest")
}

/**
 * Transition represents a change in alert type, status, severity or timestamp at a specific point in time
 */

case class Transition(status: AlertStatus, severity: AlertSeverity, alertTimestamp: Long, alertType: Option[AlertType] = None)

/**
  * Affected datavolume for an alert for snapshot or blocking user purpose.
  *
  * @param dataSourceId DatasourceId
  * @param dataVolumeId DatavolumeId
  * @param clientIps    clientIps
  */
case class AlertDataVolume(dataSourceId: String, dataVolumeId: String, clientIps: Option[String], volumeEncryptionTimestamp: Option[Long] = None)

/**
 * Enumeration representing the alert trigger. An alert e.g. could be an outcome of an anomaly detection or an
 * external detection like via ontap.
 */
object AlertTrigger extends Enumeration {
  type AlertTrigger = Value

  val Ontap: AlertTrigger = Value("Ontap")
  val Anomaly: AlertTrigger = Value("Anomaly")
}

/**
  * Enumeration representing the alert trust score
  */
object TrustScore extends Enumeration {
  type TrustScore = Value

  val High: TrustScore = Value("High")
  val Medium: TrustScore = Value("Medium")
  val Low: TrustScore = Value("Low")
  val NA: TrustScore = Value("NA")
}

/**
  * Enumeration representing the unknown user alert error scenarios.
  * There could be some cases wherein user may not be identified for an alert. E.g. for Ontap ARW attacks.
  * Need to correctly identify error in such cases.
  */
object UnknownUserAlertErrorCode extends Enumeration {
  type UnknownUserAlertErrorCode = Value

  val UnmonitoredSvm: UnknownUserAlertErrorCode = Value("UnmonitoredSvm")
  val ExcludedVolume: UnknownUserAlertErrorCode = Value("ExcludedVolume")
  val SvmCollectorError: UnknownUserAlertErrorCode = Value("SvmCollectorError")
  val Other: UnknownUserAlertErrorCode = Value("Other")
}

/**
  * Class representing the database model of the alert object in Elasticsearch
  */
case class AlertDb(tenantId: String,
                   id: String,
                   displayId: String,
                   userId: Option[String],
                   userName: Option[String],
                   communityId: Option[String],
                   communityName: Option[String],
                   synopsis: String,
                   attributes: Map[String, String],
                   alertTimestamp: Long,
                   alertType: AlertType,
                   severity: AlertSeverity,
                   status: AlertStatus,
                   note: String,
                   transitions: List[Transition],
                   changePercentage: Double,
                   userDomain: String,
                   isFalsePositive: Option[Boolean] = None,
                   patternStartTimestamp: Option[Long] = None,
                   patternEndTimestamp: Option[Long] = None,
                   dataVolumes: Option[List[AlertDataVolume]] = None,
                   alertTrigger: Option[AlertTrigger] = None,
                   trustScore: Option[TrustScore] = None
                  ) extends ObjWithIdAndLastReported {

  // Override the lastReportedTimestamp used in the trait ObjWithIdAndLastReported
  override def lastReportedTimestamp: Long = alertTimestamp
}

/**
  * A set of constants used in alert implementations
  * both in flink jobs and forensic APIs
  */
object AlertConstants {
  // Common ones
  final val MetricId = "metricId"
  final val ChangePercentage = "changePercentage"

  // Specific to alert type of abnormal access type
  final val TypicalAccessRate = "typicalAccessRate"
  final val AbnormalAccessRate = "abnormalAccessRate"
  final val MetricValue = "metricValue"
  final val AnomalousMetrics = "anomalousMetrics"
  // Anomaly Detection Algorithm
  final val AnomalyDetectionAlgorithm = "adAlgorithm"
  final val AlertVisible = "alertVisible"

  // Specific to UserJoinedManyCommunities alert type
  final val CommunitiesPreviousValue = "communitiesPreviousValue"
  final val CommunitiesCurrentValue = "communitiesCurrentValue"
  final val PreviousMetricTimestamp = "previousMetricTimestamp"

  // Specific to ransomware alert type
  final val RansomwareConfidenceScore = "ransomwareConfidenceScore"
  final val RansomwarePriority = "ransomwarePriority"
  final val RansomwareThresholdMode = "ransomwareThresholdMode"
  final val RansomwareThresholdValue = "ransomwareThresholdValue"

  final val RansomwareTotalEventCount = "ransomwareTotalEventCount"
  final val RansomwareEventCountAfterFiltering = "ransomwareEventCountAfterFiltering"
  final val RansomwareTotalEntityCount = "ransomwareTotalEntityCount"
  final val RansomwareEntityCountAfterFiltering = "ransomwareEntityCountAfterFiltering"

  final val RansomwareDetectedEntityPercentage = "ransomwareDetectedEntityPercentage"
  final val RansomwareDetectedEntityCount = "ransomwareDetectedEntityCount"
  final val RansomwareDetectedExtensions = "ransomwareDetectedExtensions"
  final val RansomwareDetectedSubPatternType = "ransomwareDetectedSubPatternType"

  // Specific to ontap ransomware detector
  final val OntapRansomwareConfidenceScore = "ontapRansomwareConfidenceScore"
  final val OntapRansomwarePriority = "ontapRansomwarePriority"
  final val OntapRansomwareThresholdMode = "ontapRansomwareThresholdMode"
  final val OntapRansomwareThresholdValue = "ontapRansomwareThresholdValue"
  final val OntapRansomwareSuspectEventCount = "ontapRansomwareSuspectEventCount"
  final val OntapRansomwareSuspectEventCountAfterFiltering = "ontapRansomwareSuspectEventCountAfterFiltering"
  final val OntapRansomwareSuspectEntityCount = "ontapRansomwareSuspectEntityCount"
  final val OntapRansomwareSuspectEntityCountAfterFiltering = "ontapRansomwareSuspectEntityCountAfterFiltering"
  final val OntapRansomwareFalsePositiveEntityPercentage = "ontapRansomwareFalsePositiveEntityPercentage"
  final val OntapRansomwareFalsePositiveEntityCount = "ontapRansomwareFalsePositiveEntityCount"
  final val OntapRansomwareFalsePositiveEventPercentage = "ontapRansomwareFalsePositiveEventPercentage"
  final val OntapRansomwareFalsePositiveEventCount = "ontapRansomwareFalsePositiveEventCount"

  // Specific to false positive detector
  final val FalsePositiveConfidenceScore = "falsePositiveConfidenceScore"
  final val FalsePositivePriority = "falsePositivePriority"
  final val FalsePositiveThresholdMode = "falsePositiveThresholdMode"
  final val FalsePositiveThresholdValue = "falsePositiveThresholdValue"

  final val FalsePositiveTotalEventCount = "falsePositiveTotalEventCount"
  final val FalsePositiveEventCountAfterFiltering = "falsePositiveEventCountAfterFiltering"
  final val FalsePositiveTotalEntityCount = "falsePositiveTotalEntityCount"
  final val FalsePositiveEntityCountAfterFiltering = "falsePositiveEntityCountAfterFiltering"

  final val FalsePositiveDetectedEventPercentage = "falsePositiveDetectedEventPercentage"
  final val FalsePositiveDetectedEventCount = "falsePositiveDetectedEventCount"
  final val FalsePositiveDetectedEntityPercentage = "falsePositiveDetectedEntityPercentage"
  final val FalsePositiveDetectedEntityCount = "falsePositiveDetectedEntityCount"
  final val FalsePositiveAlertTrigger = "falsePositiveAlertTrigger"

  // Specific to dataDestruction pattern detector
  final val DataDestructionConfidenceScore = "dataDestructionConfidenceScore"
  final val DataDestructionPriority = "dataDestructionPriority"
  final val DataDestructionThresholdMode = "dataDestructionThresholdMode"
  final val DataDestructionThresholdValue = "dataDestructionThresholdValue"
  final val DataDestructionMinimumDeletedEntitiesThreshold = "dataDestructionMinimumDeletedEntitiesThreshold"

  final val DataDestructionTotalEventCount = "dataDestructionTotalEventCount"
  final val DataDestructionEventCountAfterFiltering = "dataDestructionEventCountAfterFiltering"
  final val DataDestructionTotalEntityCount = "dataDestructionTotalEntityCount"
  final val DataDestructionEntityCountAfterFiltering = "dataDestructionEntityCountAfterFiltering"

  final val DataDestructionDetectedEventPercentage = "dataDestructionDetectedEventPercentage"
  final val DataDestructionDetectedEventCount = "dataDestructionDetectedEventCount"
  final val DataDestructionDetectedEntityPercentage = "dataDestructionDetectedEntityPercentage"
  final val DataDestructionDetectedEntityCount = "dataDestructionDetectedEntityCount"

  //Specific to low-confidence alert detector
  final val LowConfidencePatternConfidenceScore = "lowConfidencePatternConfidenceScore"
  final val LowConfidencePatternPriority = "lowConfidencePatternPriority"
  final val LowConfidencePatternDetected = "lowConfidencePatternDetected"

  //Specific to Unknown user alerts
  final val UserNotDetectedError = "userNotDetectedError"
}