/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.dbmodel

import com.netapp.df.uba.server.common.misc.dbmodel.ActionMode.ActionMode
import com.netapp.df.uba.server.common.misc.dbmodel.ActionStatus.ActionStatus
import com.netapp.df.uba.server.common.misc.dbmodel.ActionType.ActionType
import com.netapp.df.uba.server.common.misc.dbmodel.AlertType.AlertType
import com.netapp.df.uba.server.common.misc.dbmodel.LimitUserAccessMode.LimitUserAccessMode
import com.netapp.df.uba.server.common.misc.dbmodel.PolicyType.PolicyType

/**
  * Enumeration representing the alert policy type
  */
object PolicyType extends Enumeration {
  type PolicyType = Value

  val Warning: PolicyType = Value("warning")
  val Attack: PolicyType = Value("attack")
}

/**
  * Class representing the device object on which policy is defined
  *
  * @param name          Device name
  * @param datasource_id datasource Id
  * @param paths         list of datavolume paths
  */
case class PolicyDevice(name: String,
                        datasource_id: String,
                        paths: Option[List[String]])

/**
  * Enumeration representing the alert action types
  */
object ActionType extends Enumeration {
  type ActionType = Value
  val CREATE_SNAPSHOT: ActionType = Value("create_snapshot")
  val LIMIT_USER_ACCESS: ActionType = Value("limit_user_access")
  val RESTORE_USER_ACCESS: ActionType = Value("restore_user_access")
  val NOTIFY_USER: ActionType = Value("notify_user")
  val NOTIFY_MGR: ActionType = Value("notify_manager")
  val NOTIFY_ADMINS: ActionType = Value("notify_admins")
}
/**
  * Class Representing Alert Policy
  */
case class Policy(policyId: String, policyName: String)

/**
  * Class Representing Feature flag
  *
  * @param name    Name of the feature flag
  * @param enabled Is enabled or not
  */
case class FeatureFlag(name: String, enabled: Boolean)

/**
  * Class representing the model of the alert action parameter object.
  * Parameter name is mandatory. Have kept parameter values as optional
  * for now.
  */

trait Parameter

/**
  * There can be Parameter which are List of String
  *
  * @param parameterName  Parameter name
  * @param parameterValue Value as list of string
  */
case class ListActionParameter(parameterName: String,
                               parameterValue: Option[List[String]] = None) extends Parameter

/**
  * There can be Parameter which are StringActionParameter
  *
  * @param parameterName  Parameter name
  * @param parameterValue Value as String
  */
case class StringActionParameter(parameterName: String,
                                 parameterValue: Option[String] = None) extends Parameter

/**
  * Class representing the model of the alert action object.
  * Action type is mandatory, but parameters are optional depending on action type
  */
case class PolicyAction(actionType: ActionType,
                        parameters: Option[List[Parameter]] = None)

case class AlertPolicy(uuid: String,
                       name: String,
                       is_active: Boolean,
                       policy_type: PolicyType,
                       alert_types: List[AlertType],
                       actions: List[PolicyAction],
                       devices: Option[List[PolicyDevice]] = None)

// Snapshot action detail for a given volume
case class SnapshotDetails(timestamp: Long,
                           mode: ActionMode,
                           status: ActionStatus,
                           snapshotName: Option[String],
                           error: Option[String])

/*
 * ActionTaken object represents set of last actions taken on a particular alert.
 * Supports multiple actionTypes with corresponding details
 */
case class ActionTaken(actionType: ActionType,
                       lastActionDetails: Option[ActionBase]) extends LastAction

case class LastSuccessfulActionTaken(actionType: ActionType,
                                     lastSuccessfulActionDetails: Option[ActionBase]) extends LastAction

// Latest action detail for ActionType: CREATE_SNAPSHOT
case class SnapshotAction(timestamp: Long,
                          status: ActionStatus,
                          mode: ActionMode,
                          createdBy: Option[String],
                          error: Option[String]
                         ) extends ActionBase

// Latest action detail for ActionType: LIMIT_USER_ACCESS and RESTORE_USER_ACCESS
case class UserLimitAction(timestamp: Long,
                           status: ActionStatus,
                           mode: ActionMode,
                           totalCount: Int,
                           failureCount: Int,
                           restrictionMode: Option[LimitUserAccessMode],
                           userIps: List[String],
                           expiryTimestamp: Option[Long],
                           duration: Option[String],
                           createdBy: Option[String],
                           error: Option[String]
                          ) extends ActionBase

// Device names to error detail
case class DeviceErrorDetail(deviceName: String, error: String)

// ENUMs for Alert action modes
object ActionMode extends Enumeration {
  type ActionMode = Value

  val Automatic: ActionMode = Value("Automatic")
  val Manual: ActionMode = Value("Manual")
}

// Base trait for all actions
trait ActionBase {
  def timestamp: Long

  def status: ActionStatus

  def mode: ActionMode

  def createdBy: Option[String]

  def error: Option[String]
}

trait LastAction

