/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.dbmodel

import com.netapp.df.uba.server.common.misc.api.JsonFormats
import com.netapp.df.uba.server.common.misc.dbmodel.ActionMode.ActionMode
import com.netapp.df.uba.server.common.misc.dbmodel.ActionStatus.ActionStatus
import com.netapp.df.uba.server.common.misc.dbmodel.ActionType.ActionType
import com.netapp.df.uba.server.common.misc.dbmodel.AlertSeverity.AlertSeverity
import com.netapp.df.uba.server.common.misc.dbmodel.AlertStatus.AlertStatus
import com.netapp.df.uba.server.common.misc.dbmodel.AlertTrigger.AlertTrigger
import com.netapp.df.uba.server.common.misc.dbmodel.AlertType.AlertType
import com.netapp.df.uba.server.common.misc.dbmodel.AuditEventType.AuditEventType
import com.netapp.df.uba.server.common.misc.dbmodel.TrustScore.TrustScore
import com.netapp.df.uba.server.common.misc.dbmodel.ExpiryMode.ExpiryMode
import com.netapp.df.uba.server.common.misc.dbmodel.LimitUserAccessMode.LimitUserAccessMode
import com.netapp.df.uba.server.common.misc.dbmodel.MetadataObjectType.MetadataObjectType
import com.netapp.df.uba.server.common.misc.dbmodel.PolicyType.PolicyType
import com.netapp.df.uba.server.common.misc.dbmodel.UserBlockType.UserBlockType
import com.netapp.df.uba.server.common.misc.model.{ActionResult, Response, UserAccessResponse, UserAccessResult, VolumeSnapshotResponse, VolumeSnapshotResult}
import spray.json._

/**
  * CommunityJsonFormats is to provide implicit conversions from the object to json format using spray
  */
class CommunityJsonFormats extends JsonFormats {
  implicit val communityDbFormatter: RootJsonFormat[CommunityDb] = jsonFormat11(CommunityDb.apply)
  implicit val communityMappingDbFormatter: RootJsonFormat[CommunityMappingDb] = jsonFormat10(CommunityMappingDb.apply)
}

/**
  * InventoryJsonFormats is to provide implicit conversions from the object to json format using spray
  */
class InventoryJsonFormats extends JsonFormats {

  // All inventory data db model object formatters
  implicit val datasourceDbFormatter: RootJsonFormat[DatasourceDb] = jsonFormat12(DatasourceDb.apply)
  implicit val dataVolumeDbFormatter: RootJsonFormat[DataVolumeDb] = jsonFormat13(DataVolumeDb.apply)
  implicit val userDirectoryDbFormatter: RootJsonFormat[UserDirectoryDb] = jsonFormat8(UserDirectoryDb.apply)
  implicit val shareDbFormatter: RootJsonFormat[ShareDb] = jsonFormat10(ShareDb.apply)
  implicit val exportDbFormatter: RootJsonFormat[ExportDb] = jsonFormat8(ExportDb.apply)

  // inventory objects
  implicit val inventoryObjectDbFormatter: RootJsonFormat[InventoryObjectDb] = new RootJsonFormat[InventoryObjectDb] {
    override def read(json: JsValue): InventoryObjectDb = {
      val fields = json.asJsObject.fields

      if (fields.contains("nameFromDatasource")) json.convertTo[DatasourceDb]
      else if (fields.contains("dataVolumeType")) json.convertTo[DataVolumeDb]
      else if (fields.contains("cifsServer")) json.convertTo[ShareDb]
      else if (fields.contains("objectType")) json.convertTo[ExportDb]
      else json.convertTo[UserDirectoryDb]
    }

    override def write(obj: InventoryObjectDb): JsValue = {
      obj match {
        case msg: DatasourceDb => msg.toJson
        case msg: DataVolumeDb => msg.toJson
        case msg: ShareDb => msg.toJson
        case msg: ExportDb => msg.toJson
        case msg: UserDirectoryDb => msg.toJson
      }
    }
  }
}

/**
  * EventsJsonFormats is to provide implicit conversions from the object to json format using spray.
  * Explicit read/write is provided as spray JSON supports maximum 22 fields only and here there are 23 fields.
  */
class EventsJsonFormats extends JsonFormats {
  implicit val auditEventDbFormatter: RootJsonFormat[AuditEventDb] = new RootJsonFormat[AuditEventDb] {
    override def read(json: JsValue): AuditEventDb = {

      // All the fields
      val fields = json.asJsObject.fields

      AuditEventDb(
        fromField[String](json, "tenantId"),
        fromField[String](json, "id"),
        fromField[String](json, "datasourceId"),
        fromField[Long](json, "eventTimestamp"),
        fromField[AuditEventType](json, "eventType"),
        fromField[MetadataObjectType](json, "objectType"),
        fromField[String](json, "canonicalPath"),
        fromField[String](json, "displayPath"),
        fromField[String](json, "entityName"),
        fromField[String](json, "objectExtension"),
        if (fields.contains("originalCanonicalPath")) fields("originalCanonicalPath").convertTo[Option[String]] else None,
        if (fields.contains("originalDisplayPath")) fields("originalDisplayPath").convertTo[Option[String]] else None,
        fromField[String](json, "userDisplayName"),
        if (fields.contains("mountPath")) fields("mountPath").convertTo[Option[String]] else None,
        fromField[String](json, "clientIpAddress"),
        if (!fields.contains("clientCoordinates") || fields.contains("clientCoordinates").toString.isEmpty) None
        else fields("clientCoordinates").convertTo[Option[String]],
        fromField[Long](json, "sizeBytes"),
        fromField[String](json, "dataVolumeId"),
        fromField[String](json, "userId"),
        if (fields.contains("rawUserId")) fields("rawUserId").convertTo[Option[String]] else None,
        fromField[Option[ProtocolInfoDb]](json, "protocol"),
        fromField[String](json, "entityId"),
        if (fields.contains("originalEntityId")) fields("originalEntityId").convertTo[Option[String]] else None,
        if (fields.contains("justCreatedEntityEvent")) fields("justCreatedEntityEvent").convertTo[Option[Boolean]] else Some(false),
        if (fields.contains("isTemporary")) fields("isTemporary").convertTo[Option[Boolean]] else Some(false),
        if (fields.contains("isHomeDirectoryEvent")) fields("isHomeDirectoryEvent").convertTo[Option[Boolean]] else Some(false)
      )
    }

    override def write(obj: AuditEventDb): JsObject = {
      val jsValues: List[(String, JsValue)] = List(
        ("tenantId", StringJsonFormat.write(obj.tenantId)),
        ("id", JsString(obj.id)),
        ("datasourceId", JsString(obj.datasourceId)),
        ("eventTimestamp", JsNumber(obj.eventTimestamp)),
        ("eventType", auditEventTypeEnumFormatter.write(obj.eventType)),
        ("objectType", metadataObjectTypeEnumFormatter.write(obj.objectType)),
        ("canonicalPath", JsString(obj.canonicalPath)),
        ("displayPath", JsString(obj.displayPath)),
        ("entityName", JsString(obj.entityName)),
        ("objectExtension", JsString(obj.objectExtension)),
        if (obj.originalCanonicalPath.isDefined) ("originalCanonicalPath", JsString(obj.originalCanonicalPath.get)) else ("originalCanonicalPath", JsNull),
        if (obj.originalDisplayPath.isDefined) ("originalDisplayPath", JsString(obj.originalDisplayPath.get)) else ("originalDisplayPath", JsNull),
        ("userDisplayName", JsString(obj.userDisplayName)),
        if (obj.mountPath.isDefined) ("mountPath", JsString(obj.mountPath.get)) else ("mountPath", JsNull),
        ("clientIpAddress", JsString(obj.clientIpAddress)),
        if (obj.clientCoordinates.isEmpty || obj.clientCoordinates.get.isEmpty)
          ("clientCoordinates", JsNull)
        else ("clientCoordinates", JsString(obj.clientCoordinates.get)),
        ("sizeBytes", JsNumber(obj.sizeBytes)),
        ("dataVolumeId", JsString(obj.dataVolumeId)),
        ("userId", JsString(obj.userId)),
        if (obj.rawUserId.isDefined) ("rawUserId", JsString(obj.rawUserId.get)) else ("rawUserId", JsNull),
        if (obj.protocol.isDefined) ("protocol", protocolDbFormatter.write(obj.protocol.get)) else ("protocol", JsNull),
        ("entityId", JsString(obj.entityId)),
        if (obj.originalEntityId.isDefined) ("originalEntityId", JsString(obj.originalEntityId.get)) else ("originalEntityId", JsNull),
        if (obj.justCreatedEntityEvent.isDefined) ("justCreatedEntityEvent", JsBoolean(obj.justCreatedEntityEvent.get)) else ("justCreatedEntityEvent", JsNull),
        if (obj.isTemporary.isDefined) ("isTemporary", JsBoolean(obj.isTemporary.get)) else ("isTemporary", JsNull),
        if (obj.isHomeDirectoryEvent.isDefined) ("isHomeDirectoryEvent", JsBoolean(obj.isHomeDirectoryEvent.get)) else ("isHomeDirectoryEvent", JsNull)
      )
      val nonNullValues: List[(String, JsValue)] = jsValues
        .filterNot(_._2 == JsNull)

      JsObject(nonNullValues: _*)
    }
  }
}

/**
  * UserJsonFormats is to provide implicit conversions from the object to json format using spray
  */
class UserJsonFormats extends JsonFormats {
  implicit val userDbFormatter: RootJsonFormat[UserDb] = jsonFormat13(UserDb.apply)
}

/**
  * OpsRateMetricJsonFormats is to provide implicit conversions from the object to json format using spray
  */
class OpsRateMetricJsonFormats extends JsonFormats {
  implicit val opsRateMetricDbFormatter: RootJsonFormat[OpsRateMetricDb] = jsonFormat10(OpsRateMetricDb.apply)
}

/**
  * UserCommunityCountMetricJsonFormats is to provide implicit conversions from the object to json format using spray
  */
class UserCommunityCountMetricJsonFormats extends JsonFormats {
  implicit val userCommunityCountMetricDbFormatter: RootJsonFormat[UserCommunityCountMetricDb] = jsonFormat5(UserCommunityCountMetricDb.apply)
}

/**
 * AlertJsonFormats is to provide implicit conversions from the object to json format using spray for alert objects.
 * Explicit read/write is provided as spray JSON supports maximum 22 fields only and here there are 23 fields.
 */
class AlertJsonFormats extends JsonFormats {
  implicit val alertTypeEnumFormatter: JsonFormat[AlertType] = jsonEnumFormatter(AlertType)
  implicit val alertStatusEnumFormatter: JsonFormat[AlertStatus] = jsonEnumFormatter(AlertStatus)
  implicit val alertSeverityEnumFormatter: JsonFormat[AlertSeverity] = jsonEnumFormatter(AlertSeverity)
  implicit val trustScoreEnumFormatter: JsonFormat[TrustScore] = jsonEnumFormatter(TrustScore)
  implicit val transitionFormatter: RootJsonFormat[Transition] = jsonFormat4(Transition.apply)
  implicit val dataVolumeFormatter: RootJsonFormat[AlertDataVolume] = jsonFormat4(AlertDataVolume.apply)
  implicit val alertTriggerEnumFormatter: JsonFormat[AlertTrigger] = jsonEnumFormatter(AlertTrigger)

  implicit val alertDbFormatter: RootJsonFormat[AlertDb] = new RootJsonFormat[AlertDb] {
    override def read(json: JsValue): AlertDb = {

      // All the fields
      val fields = json.asJsObject.fields

      AlertDb(
        fromField[String](json, "tenantId"),
        fromField[String](json, "id"),
        fromField[String](json, "displayId"),
        if (fields.contains("userId")) fields("userId").convertTo[Option[String]] else None,
        if (fields.contains("userName")) fields("userName").convertTo[Option[String]] else None,
        if (fields.contains("communityId")) fields("communityId").convertTo[Option[String]] else None,
        if (fields.contains("communityName")) fields("communityName").convertTo[Option[String]] else None,
        fromField[String](json, "synopsis"),
        fromField[Map[String, String]](json, "attributes"),
        fromField[Long](json, "alertTimestamp"),
        fromField[AlertType](json, "alertType"),
        fromField[AlertSeverity](json, "severity"),
        fromField[AlertStatus](json, "status"),
        fromField[String](json, "note"),
        fromField[List[Transition]](json, "transitions"),
        fromField[Double](json, "changePercentage"),
        fromField[String](json, "userDomain"),
        if (fields.contains("isFalsePositive")) fields("isFalsePositive").convertTo[Option[Boolean]] else None,
        if (fields.contains("patternStartTimestamp")) fields("patternStartTimestamp").convertTo[Option[Long]] else None,
        if (fields.contains("patternEndTimestamp")) fields("patternEndTimestamp").convertTo[Option[Long]] else None,
        if (fields.contains("dataVolumes")) fields("dataVolumes").convertTo[Option[List[AlertDataVolume]]] else None,
        if (fields.contains("alertTrigger")) fields("alertTrigger").convertTo[Option[AlertTrigger]] else None,
        if (fields.contains("trustScore")) fields("trustScore").convertTo[Option[TrustScore]] else None
      )
    }

    override def write(obj: AlertDb): JsObject = {
      val jsValues: List[(String, JsValue)] = List(
        ("tenantId", StringJsonFormat.write(obj.tenantId)),
        ("id", JsString(obj.id)),
        ("displayId", JsString(obj.displayId)),
        if (obj.userId.isDefined) ("userId", JsString(obj.userId.get)) else ("userId", JsNull),
        if (obj.userName.isDefined) ("userName", JsString(obj.userName.get)) else ("userName", JsNull),
        if (obj.communityId.isDefined) ("communityId", JsString(obj.communityId.get)) else ("communityId", JsNull),
        if (obj.communityName.isDefined) ("communityName", JsString(obj.communityName.get)) else ("communityName", JsNull),
        ("synopsis", JsString(obj.synopsis)),
        ("attributes", obj.attributes.toJson),
        ("alertTimestamp", JsNumber(obj.alertTimestamp)),
        ("alertType", alertTypeEnumFormatter.write(obj.alertType)),
        ("severity", alertSeverityEnumFormatter.write(obj.severity)),
        ("status", alertStatusEnumFormatter.write(obj.status)),
        ("note", JsString(obj.note)),
        ("transitions", JsArray(obj.transitions.map(transition => transitionFormatter.write(transition)).toVector)),
        ("changePercentage", JsNumber(obj.changePercentage)),
        ("userDomain", JsString(obj.userDomain)),
        if (obj.isFalsePositive.isDefined) ("isFalsePositive", JsBoolean(obj.isFalsePositive.get)) else ("isFalsePositive", JsNull),
        if (obj.patternStartTimestamp.isDefined) ("patternStartTimestamp", JsNumber(obj.patternStartTimestamp.get)) else ("patternStartTimestamp", JsNull),
        if (obj.patternEndTimestamp.isDefined) ("patternEndTimestamp", JsNumber(obj.patternEndTimestamp.get)) else ("patternEndTimestamp", JsNull),
        if (obj.dataVolumes.isDefined) ("dataVolumes", JsArray(obj.dataVolumes.get.map(dataVolume => dataVolumeFormatter.write(dataVolume)).toVector)) else ("dataVolumes", JsNull),
        if (obj.alertTrigger.isDefined) ("alertTrigger", alertTriggerEnumFormatter.write(obj.alertTrigger.get)) else ("alertTrigger", JsNull),
        if (obj.trustScore.isDefined) ("trustScore", trustScoreEnumFormatter.write(obj.trustScore.get)) else ("trustScore", JsNull)
      )
      val nonNullValues: List[(String, JsValue)] = jsValues
        .filterNot(_._2 == JsNull)

      JsObject(nonNullValues: _*)
    }
  }
}

class AlertPolicyJsonFormats extends AlertJsonFormats {

  implicit val actionTypeEnumFormatter: JsonFormat[ActionType] = jsonEnumFormatter(ActionType)
  implicit val actionParameterFormatter: RootJsonFormat[ListActionParameter] = jsonFormat2(ListActionParameter.apply)
  implicit val policyActionParameterFormatter: RootJsonFormat[StringActionParameter] = jsonFormat2(StringActionParameter.apply)

  implicit val parameterFormatter: RootJsonFormat[Parameter] = new RootJsonFormat[Parameter] {
    override def write(obj: Parameter): JsValue = {

      obj match {
        case msg: StringActionParameter => msg.toJson
        case msg: ListActionParameter => msg.toJson
      }
    }

    override def read(json: JsValue): Parameter = {
      val fields = json.asJsObject.fields
      // Extra double quotes to be removed
      val fieldValues = fields.map(u => u._2.toString().replaceAll("^\"|\"$", "")).toList
      if (fieldValues.contains("mode") || fieldValues.contains("durationInHours")) {
        json.convertTo[StringActionParameter]
      }
      else json.convertTo[ListActionParameter]
    }
  }

  //type is a reserved character, hence need to explicitly specify field name to convert the "type" json field
  //to actionType
  implicit val policyActionFormatter: RootJsonFormat[PolicyAction] = jsonFormat(PolicyAction.apply, "type", "parameters")

  implicit val policyTypeEnumFormatter: JsonFormat[PolicyType] = jsonEnumFormatter(PolicyType)
  implicit val policyDeviceFormatter: RootJsonFormat[PolicyDevice] = jsonFormat3(PolicyDevice.apply)

  implicit val alertPolicyFormatter: RootJsonFormat[AlertPolicy] = jsonFormat7(AlertPolicy.apply)
  implicit val policyFormatter: RootJsonFormat[Policy] = jsonFormat2(Policy.apply)
}

class AlertActionsJsonFormats extends AlertPolicyJsonFormats {
  implicit val datasourceActionsFormatter: RootJsonFormat[DatasourceAlertAction] = jsonFormat3(DatasourceAlertAction.apply)
  implicit val actionStatusEnumFormatter: JsonFormat[ActionStatus] = jsonEnumFormatter(ActionStatus)
  implicit val actionModeFormatter: JsonFormat[ActionMode] = jsonEnumFormatter(ActionMode)
  implicit val limitUserAccessModeEnumFormatter: JsonFormat[LimitUserAccessMode] = jsonEnumFormatter(LimitUserAccessMode)
  implicit val expiryModeEnumFormatter: JsonFormat[ExpiryMode] = jsonEnumFormatter(ExpiryMode)
  implicit val userBlockTypeFormatter: JsonFormat[UserBlockType] = jsonEnumFormatter(UserBlockType)
  implicit val userAccessResultFormatter: RootJsonFormat[UserAccessResult] = jsonFormat12(UserAccessResult.apply)
  implicit val volumeSnapshotResultFormatter: RootJsonFormat[VolumeSnapshotResult] = jsonFormat8(VolumeSnapshotResult.apply)
  implicit val userAccessResultDbFormatter: RootJsonFormat[UserAccessResultDb] = jsonFormat8(UserAccessResultDb.apply)
  implicit val volumeSnapshotResultDbFormatter: RootJsonFormat[VolumeSnapshotResultDb] = jsonFormat3(VolumeSnapshotResultDb.apply)
  implicit val volumeSnapshotDetailsFormatter: RootJsonFormat[VolumeSnapshotDetails] = jsonFormat1(VolumeSnapshotDetails.apply)
  implicit val userAccessDetailsFormatter: RootJsonFormat[UserAccessDetails] = jsonFormat1(UserAccessDetails.apply)

  implicit val responseDbFormatter: RootJsonFormat[Option[ResponseDb]] = new RootJsonFormat[Option[ResponseDb]] {
    override def write(obj: Option[ResponseDb]): JsValue = {

      obj.getOrElse(None) match {
        case msg: VolumeSnapshotDetails => msg.toJson
        case msg: UserAccessDetails => msg.toJson
        case None => JsObject.empty
      }
    }

    override def read(json: JsValue): Option[ResponseDb] = {
      val fields = json.asJsObject.fields
      if (fields.contains("volumeSnapshotResultDb")) json.convertTo[Option[VolumeSnapshotDetails]]
      else if (fields.contains("userAccessResultDb")) json.convertTo[Option[UserAccessDetails]]
      else None
    }
  }

  implicit val subActionsFormatter: RootJsonFormat[SubAction] = jsonFormat6(SubAction.apply)
  implicit val alertActionResultDbFormatter: RootJsonFormat[AlertActionResultDb] = jsonFormat10(AlertActionResultDb.apply)

  // Used in config
  implicit val volumeSnapshotResponseFormatter: RootJsonFormat[VolumeSnapshotResponse] = jsonFormat1(VolumeSnapshotResponse.apply)
  implicit val userAccessResponseFormatter: RootJsonFormat[UserAccessResponse] = jsonFormat1(UserAccessResponse.apply)
  implicit val responseFormatter: RootJsonFormat[Response] = new RootJsonFormat[Response] {
    override def write(obj: Response): JsValue = {
      obj match {
        case msg: VolumeSnapshotResponse => msg.toJson
        case msg: UserAccessResponse => msg.toJson
      }
    }

    override def read(json: JsValue): Response = {
      val fields = json.asJsObject.fields
      if (fields.contains("volumeSnapshotResult")) json.convertTo[VolumeSnapshotResponse]
      else json.convertTo[UserAccessResponse]
    }
  }

  implicit val alertActionResultFormatter: RootJsonFormat[ActionResult] = jsonFormat8(ActionResult.apply)

  implicit val alertActionObjectFormatter: RootJsonFormat[ActionBase] = new RootJsonFormat[ActionBase] {
    override def read(json: JsValue): ActionBase = {
      val fields = json.asJsObject.fields
      if (fields.contains("userIps")) json.convertTo[UserLimitAction]
      else json.convertTo[SnapshotAction] // NOTE: with new supported ActionTypes we may need changes here
    }

    override def write(obj: ActionBase): JsValue = {
      obj match {
        case msg: SnapshotAction => msg.toJson
        case msg: UserLimitAction => msg.toJson
      }
    }
  }

  implicit val snapshotActionFormatter: RootJsonFormat[SnapshotAction] = jsonFormat5(SnapshotAction.apply)
  implicit val userRestrictActionFormatter: RootJsonFormat[UserLimitAction] = jsonFormat11(UserLimitAction.apply)

  implicit val lastActionObjectFormatter: RootJsonFormat[LastAction] = new RootJsonFormat[LastAction] {
    override def read(json: JsValue): LastAction = {
      val fields = json.asJsObject.fields
      if (fields.contains("lastActionDetails")) json.convertTo[ActionTaken]
      else json.convertTo[LastSuccessfulActionTaken] // NOTE: with new supported ActionTypes we may need changes here
    }

    override def write(obj: LastAction): JsValue = {
      obj match {
        case msg: ActionTaken => msg.toJson
        case msg: LastSuccessfulActionTaken => msg.toJson
      }
    }
  }

  implicit val actionTakenFormatter: RootJsonFormat[ActionTaken] = jsonFormat2(ActionTaken.apply)
  implicit val lastSuccessfulActionFormatter: RootJsonFormat[LastSuccessfulActionTaken] = jsonFormat2(LastSuccessfulActionTaken.apply)

  implicit val deviceErrorDetailFormatter: RootJsonFormat[DeviceErrorDetail] = jsonFormat2(DeviceErrorDetail.apply)
}

/**
  * StateObjectDbJsonFormats is to provide implicit conversions from the object to json format using spray for state objects of Flink stored in Elastic
  */
class StateObjectDbJsonFormats extends JsonFormats {
  implicit val stateDbFormatter: JsonFormat[StateObjectDb] = jsonFormat3(StateObjectDb.apply)
}

class FeatureFlagJsonFormats extends JsonFormats {
  implicit val featureFlagFormatter: JsonFormat[FeatureFlag] = jsonFormat2(FeatureFlag.apply)
}
