/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.http

import java.security.KeyStore
import java.security.cert.X509Certificate

import org.apache.http.client.config.RequestConfig
import org.apache.http.conn.ssl.{NoopHostnameVerifier, SSLConnectionSocketFactory}
import org.apache.http.impl.client.{CloseableHttpClient, HttpClients}
import org.apache.http.ssl.SSLContextBuilder

import scala.util.{Failure, Success, Try}

object HttpClientSingleton {
  // Timeout for socket, connection request
  final val TimeoutValueInMillis: Int = 100000
  // Lock to synchronize
  private val lock = new Object()
  @volatile private var httpClientTry: Option[Try[CloseableHttpClient]] = None
  @volatile private var httpClientWithoutRedirectsTry: Option[Try[CloseableHttpClient]] = None

  /**
   * Http client singleton creator for most general usecases. Its recommended to create a single
   * client instance throughout application.
   * https://hc.apache.org/httpcomponents-client-ga/tutorial/html/fundamentals.html#d5e213
   *
   * @return Try[CloseableHttpClient]
   */
  def getHttpClient: Try[CloseableHttpClient] = {
    var localHttpClientTry: Option[Try[CloseableHttpClient]] = httpClientTry
    if (localHttpClientTry.isEmpty || !localHttpClientTry.get.isSuccess) {
      lock.synchronized {
        localHttpClientTry = httpClientTry

        //Need to use double locking pattern to ensure only one client instance gets created. Reference
        // https://en.wikipedia.org/wiki/Double-checked_locking. Local copying ensures volatile access is minimized
        // to improve reads.
        if (localHttpClientTry.isEmpty) {
          httpClientTry = Some(createHttpClientTry())
          localHttpClientTry = httpClientTry
        }
      }
    }
    localHttpClientTry.get
  }

  /**
   * Http client singleton creator with disabled automatic redirect handling.
   * This allows caller to explicitly handle 307 Redirect response.
   *
   * @return Try[CloseableHttpClient]
   */
  def getHttpClientWithoutRedirects: Try[CloseableHttpClient] = {
    var localHttpClientTry: Option[Try[CloseableHttpClient]] = httpClientWithoutRedirectsTry
    if (localHttpClientTry.isEmpty || !localHttpClientTry.get.isSuccess) {
      lock.synchronized {
        localHttpClientTry = httpClientWithoutRedirectsTry

        //Need to use double locking pattern to ensure only one client instance gets created. Reference
        // https://en.wikipedia.org/wiki/Double-checked_locking. Local copying ensures volatile access is minimized
        // to improve reads.
        if (localHttpClientTry.isEmpty) {
          httpClientWithoutRedirectsTry = Some(createHttpClientTry(true))
          localHttpClientTry = httpClientWithoutRedirectsTry
        }
      }
    }
    localHttpClientTry.get
  }

  private def createHttpClientTry(disableRedirects: Boolean = false): Try[CloseableHttpClient] = try {
    val builder = new SSLContextBuilder
    builder.loadTrustMaterial(null.asInstanceOf[KeyStore], (chain: Array[X509Certificate], authType: String) => true)
    val sslSf = new SSLConnectionSocketFactory(builder.build, new NoopHostnameVerifier)

    // Some timeout configurations
    val requestConfig = RequestConfig.custom()
      .setConnectionRequestTimeout(TimeoutValueInMillis)
      .setConnectTimeout(TimeoutValueInMillis)
      .setSocketTimeout(TimeoutValueInMillis)
      .build()

    // Create a HTTP client
    val httpClientBuilder = HttpClients.custom()
      .setSSLSocketFactory(sslSf)
      .setDefaultRequestConfig(requestConfig)
      .setSSLHostnameVerifier(new NoopHostnameVerifier)

    if (disableRedirects) httpClientBuilder.disableRedirectHandling

    val httpClient = httpClientBuilder.build()
    Success(httpClient)
  } catch {
    case ex: Exception =>
      Failure(ex)
  }
}
