/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.interceptor

import akka.http.scaladsl.model._
import com.auth0.jwt.JWT
import com.auth0.jwt.interfaces.{Claim, DecodedJWT}
import com.netapp.ci.apikey.ApiKeyPermission
import com.netapp.df.uba.server.common.misc.apikey.model.CategoryAndApplications
import com.netapp.df.uba.server.common.misc.apikey.utils.SecurityUtils
import com.netapp.df.uba.server.common.misc.apikey.{ApiKey, ApiKeyCategory}
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model.{ACCESS_TYPE, JwtTokenContext, RestMethodDocumentation}
import com.netapp.oci.jwt.{JWTAuthData, JWTUtils}

import java.time.Clock
import java.util
import java.util.Date
import scala.collection.JavaConverters._
import scala.util.{Failure, Success, Try}


/**
  * Validator class for the public APIs
  * Validates the API key retrieved from http header of the HTTP request
  * by using the api key value  retrieved from distributed read through cache
  */
case class InterceptorPublicApiKeyValidator(httpRequest: Option[HttpRequest]) extends LogTrait {

  private val validPublicApiEndpoints: Set[String] = Set(
    "/rest/v1/cloudsecure/activities",
    "/rest/v1/cloudsecure/alerts",
    "/rest/v1/traffic/records"
  )

  def getCloudSecureJwtToken: Try[String] = {

    if (httpRequest.isEmpty) {
      return Failure(new Exception("http request is empty"))
    }
    val cloudSecureHeader: Seq[HttpHeader] = httpRequest.get.headers.filter(_.name().toLowerCase == InterceptorApiCacheService.publicApiHeaderName)

    if (cloudSecureHeader.isEmpty) {
      if (isValidPublicApiEndPoint) {
        logger.warn(s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: ${InterceptorApiCacheService.publicApiHeaderName} header is not present in the request")
      }
      return Failure(new Exception("Invalid public api header"))
    }
    Success(cloudSecureHeader.head.value())
  }

  /**
    * This method first calls fromJWT method and receives the decoded key
    * In case any exception comes then it tries to retrieve the key by skipping the signature validation
    */
  def getJwtAuthData(jwtToken: String): JWTAuthData = {

    var jWTAuthData: Option[JWTAuthData] = None

    try {
      jWTAuthData = Some(JWTUtils.fromJWT(jwtToken))
    } catch {
      case _: Exception =>
        logger.info(s"Public API token signature verification issue with vault, trying to decode token without validating signature")
        //Try to get the token by ignoring the signature verification
        jWTAuthData = getJwtClaimsWithoutSignatureValidation(jwtToken)
    }

    if (jWTAuthData.isEmpty) {
      throw new RuntimeException("Invalid token")
    }
    jWTAuthData.get
  }

  /**
    * Below code is same as in JWTUtils.Java after skipping signature verification.
    * Following is the reason for skipping the signature validation:
    *
    * 1. API Access key has maximum validity of 1 year. Where as Signature of the key expires in short interval of time, 3 months duration.
    * 2. If any customer created a key of 1 year, has to invalidate and create new one.
    * 3. API Access key still can't be broken with respect to security even the signature is not validated.
    *
    * Reason is: multiple levels of Authorization steps.
    */
  def getJwtClaimsWithoutSignatureValidation(token: String): Option[JWTAuthData] = {

    val jwt: DecodedJWT = JWT.decode(token)

    var tenant: Option[String] = None
    var login: Option[String] = None
    var roles: Option[Array[String]] = None
    val extraClaims: util.HashMap[String, String] = new util.HashMap[String, String]

    for (claimsEntrySet <- jwt.getClaims.entrySet.asScala) {
      val claimName: String = claimsEntrySet.getKey
      val claim: Claim = claimsEntrySet.getValue
      logger.debug("JWTUtils::fromJWT claimName: >>" + claimName + "<<")
      claimName match {
        case JWTAuthData.CLAIM_TENANT =>
          tenant = Some(claim.asString)

        case JWTAuthData.CLAIM_LOGIN =>
          login = Some(claim.asString)
        case JWTAuthData.CLAIM_ROLES =>
          roles = Some(claim.asArray(classOf[String]))
        case _ =>
          extraClaims.put(claimName, claim.asString)
      }
    }
    if (login.isEmpty) return None

    val expire: Date = jwt.getExpiresAt
    val info: JWTAuthData = new JWTAuthData(tenant.get, login.get, expire, extraClaims, roles.toString)
    logger.debug("Parsed JWT access token <" + token + "> and get " + info)
    Some(info)
  }

  /**
    * Method implemented as described in the file:com.netapp.oci.springboot.commons.security.JwtAuthenticationFilter
    */
  protected def validatePublicApiKey(apiKeyJwtTokenFromPublicApiResponse: String): Try[(JWTAuthData, ApiKey)] = {

    // ApiKeyCache service must be initialized by the service which is using the public api otherwise throw exception here
    if (InterceptorApiCacheService.getApiKeyCacheService.isEmpty) {
      return Failure(new Exception("Insufficient Authentication"))
    }

    //Retrieve the apiKeyCacheService object
    val apiKeyCacheService = InterceptorApiCacheService.getApiKeyCacheService.get
    try {
      //Decrypt the JWT token retrieved from public api http request
      val jwtAuthData: JWTAuthData = getJwtAuthData(apiKeyJwtTokenFromPublicApiResponse)
      logger.debug(s"Successfully validated the JWT method")
      val tenantId = jwtAuthData.getTenant
      val apiKeyId = jwtAuthData.getLogin
      if (tenantId == null || apiKeyId == null) {

        if (tenantId == null) {
          logger.error(s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed Reason: tenantId is null")
        } else {
          logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed Reason: apiKeyId is null")
        }
        return Failure(new Exception("Invalid API key"))
      }

      //Retrieve the set of JWT tokens available for a tenant from the distributed read through cache
      val tenantApiKeys = apiKeyCacheService.getTenantApiKeys(tenantId)

      if (tenantApiKeys == null) {
        logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed Reason: tenantApiKeys is null")
        return Failure(new Exception("Invalid API key"))
      }

      //Retrieve the apiKey corresponding to the current public api request
      val apiKey = tenantApiKeys.getApiKey(apiKeyId)

      //Check the current status of the apiKey and take action accordingly
      if (apiKey == null || !apiKey.isEnabled || apiKey.isRevoked) {
        if (apiKey == null) {
          logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKey is null")
        } else if (!apiKey.isEnabled) {
          logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKey is disabled")
        } else if (apiKey.isRevoked) {
          logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKey is Revoked")
        }
        return Failure(new Exception("Invalid API key"))
      }

      //Validate the signature of the apiKey from cache with the public api
      val jwtTokenSignature = ApiKey.generateJwtTokenSignature(apiKeyJwtTokenFromPublicApiResponse)

      if (!jwtTokenSignature.equals(apiKey.getJwtTokenSignature)) {
        logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKey signature mismatch")
        return Failure(new Exception("Invalid API key signature"))
      }

      //Validate the expire time
      if (apiKey.getExpireTime < Clock.systemUTC().millis()) {
        logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKey is expired")
        return Failure(new Exception("API key expired"))
      }

      //validate the READ/WRITE permissions
      val apiKeyPermission = ApiKeyPermission.fromPermissionString(apiKey.getPermission)

      if (apiKeyPermission == null ||
        apiKeyPermission == ApiKeyPermission.READ && !"GET".equals(httpRequest.get.method.value.toUpperCase()) ||
        apiKeyPermission == ApiKeyPermission.WRITE && "GET".equals(httpRequest.get.method.value.toUpperCase())) {

        if (apiKeyPermission == null) {
          logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKeyPermission is null")
        } else if (apiKeyPermission == ApiKeyPermission.READ) {
          logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKeyPermission is not valid READ request")
        } else if (apiKeyPermission == ApiKeyPermission.WRITE) {
          logErrorT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: apiKeyPermission is not valid WRITE request")
        }
        return Failure(new Exception("API key has insufficient permissions"))
      }
      logInfoT(tenantId, s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Success")
      Success(jwtAuthData, apiKey)
    } catch {
      case ex: Exception =>
        logger.error(s"Public API Request: ${httpRequest.get.getUri().path()}, Status: Failed, Reason: ${ex.printStackTrace()}")
        Failure(ex)
    }
  }

  def validateApiKeyCategory(jwtTokenContext: JwtTokenContext, restMethodDocumentation: RestMethodDocumentation, apiKey: ApiKey): Try[Boolean] = {
    val claims = jwtTokenContext.claims
    val apiKeyType = if (claims.contains("api") && claims("api").equalsIgnoreCase("true")) ApiKeyType.CUSTOM
    else ApiKeyType.GENERIC

    apiKeyType match {
      case ApiKeyType.GENERIC =>
        Success(true)

      case ApiKeyType.CUSTOM =>
        val apiKeyCategoriesAnnotated = restMethodDocumentation.categories

        if (apiKeyCategoriesAnnotated.isEmpty)
          return Failure(new Exception("Method is not allowed to be accessed by API key"))

        // API token can access only public APIs
        if (restMethodDocumentation.access != ACCESS_TYPE.PUBLIC)
          return Failure(new Exception("Method is not allowed to be accessed by API key"))
        Try {
          val apiKeyCategories: util.List[CategoryAndApplications] = SecurityUtils.getApikeyCategories(apiKey)
          apiKeyCategories.forEach(
            categoryAndApplication => {
              val apiKeyCategory = ApiKeyCategory.fromCategoryString(categoryAndApplication.getId)
              if (apiKeyCategory != null) {
                apiKeyCategoriesAnnotated.foreach(
                  currAnnotatedCategory =>
                    if (apiKeyCategory == currAnnotatedCategory) {
                      // TODO: Check for feature flag as well
                      return Success(true)
                    }
                )
              }
            }
          )
          throw new Exception("ApiKey category requirement not met")
        } match {
          case Success(_) => Success(true)
          case Failure(ex) =>
            logger.error("Failed to validate the apiKey category")
            Failure(new Exception("Method is not allowed to be accessed by API key", ex))
        }

      case somethingElse =>
        logger.error(s"Invalid ApiKeyType $somethingElse")
        Failure(new Exception("Method is not allowed to be accessed by API key"))
    }

  }

  private object ApiKeyType extends Enumeration {
    type ApiKeyType = Value
    val GENERIC, CUSTOM = Value
  }

  protected def isValidPublicApiEndPoint: Boolean = {
    validPublicApiEndpoints.contains(httpRequest.get.getUri().getPathString)
  }

  def validatePublicApiRequest: Try[(JWTAuthData, ApiKey)] = {
    val tokenTry = getCloudSecureJwtToken
    tokenTry match {
      case Success(token) =>
        if (!isValidPublicApiEndPoint) Failure(new Exception("Invalid public api endpoint"))
        else validatePublicApiKey(token)

      case Failure(exception) =>
        Failure(exception)
    }
  }
}