
/*
 * $Id$
 *
 * Copyright (c) 2019 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.model

import com.netapp.df.uba.server.common.misc.dbmodel.AlertSeverity.AlertSeverity
import com.netapp.df.uba.server.common.misc.dbmodel.AlertStatus.AlertStatus
import com.netapp.df.uba.server.common.misc.dbmodel.AlertType.AlertType
import com.netapp.df.uba.server.common.misc.dbmodel._
import spray.json.RootJsonFormat

import scala.collection.mutable.ListBuffer

/**
  * AlertCloudEvent represents the data sent in the alerts.cloudevents topic
  * that will be consumed by the notifications service. The format is going to be exactly same
  * as how notifications service will consume. It contains an eventType and all data related to the
  * alert under the field 'data'
  */
object AlertCloudEvent {

  final val EventTypeAlerts = "CS_ALERT_STATUS"
  final val EventTypeRansomwareAlerts = "CS_RANSOMWARE_ALERT_STATUS"
  final val EventTypeDataDestructionAlerts = "CS_DATA_DESTRUCTION_ALERT_STATUS"
  final val TenantExtension = "tenantId"
  final val KeyExtension = "key"
  final val ActionTakenDisplayStringMap = Map(
    "create_snapshot" -> "Snapshot Taken",
    "limit_user_access" -> "Access Blocked",
    "restore_user_access" -> "Access Restored")

  private def getErrorMsgForLimitUser(userLimitAction: UserLimitAction) = {
    val successCount = userLimitAction.totalCount - userLimitAction.failureCount
    (userLimitAction.restrictionMode.get, successCount) match {
      case (LimitUserAccessMode.BLOCK, 0) => "Block action failed on all SVMs"
      case (LimitUserAccessMode.BLOCK, _) => s"Access blocked on $successCount SVMs"
      case _ => ""
    }
  }

  def apply(alertDb: AlertDb): AlertCloudEvent = {

    val alertData: AlertData =
      AlertData(
        alertDb.tenantId,
        alertDb.id,
        alertDb.displayId,
        alertDb.alertType,
        alertDb.severity,
        alertDb.synopsis,
        alertDb.userName.getOrElse(""),
        alertDb.communityName.getOrElse(""),
        alertDb.userId.getOrElse(""),
        alertDb.communityId.getOrElse(""),
        alertDb.alertTimestamp,
        alertDb.status,
        alertDb.note,
        alertDb.changePercentage,
        alertDb.attributes,
        List.empty)

    val cloudEventAlertType = alertData.`type` match {
      case AlertType.RansomwareAttack => AlertCloudEvent.EventTypeRansomwareAlerts
      case AlertType.DataDestructionFileDeletionAttack => AlertCloudEvent.EventTypeDataDestructionAlerts
      case _ => AlertCloudEvent.EventTypeAlerts
    }
    AlertCloudEvent(
      cloudEventAlertType,
      alertData,
      Map(
        TenantExtension -> alertData.tenantId,
        KeyExtension -> alertData.tenantId)
    )
  }

  def getActionsTakenAsString(actionsTaken: List[ActionTaken]): List[String] = {
    val actionsTakenBuffer: ListBuffer[String] = new ListBuffer[String]

    actionsTaken.map(action => action.actionType match {
      case ActionType.CREATE_SNAPSHOT =>
        if (action.lastActionDetails.isEmpty) { // This should NOT happen as this is only for InProgress case
          actionsTakenBuffer += "None"
        } else if (action.lastActionDetails.get.asInstanceOf[SnapshotAction].error.isEmpty) { // if no error, use actionType
          actionsTakenBuffer += ActionTakenDisplayStringMap(action.actionType.toString)
        } else { // if error, use error
          actionsTakenBuffer += action.lastActionDetails.get.asInstanceOf[SnapshotAction].error.get
        }
      case ActionType.LIMIT_USER_ACCESS =>
        if (action.lastActionDetails.isEmpty) { // This should NOT happen as this is only for InProgress case
          actionsTakenBuffer += "None"
        } else if (action.lastActionDetails.get.asInstanceOf[UserLimitAction].error.isEmpty) { // if no error, use actionType
          actionsTakenBuffer += ActionTakenDisplayStringMap(action.actionType.toString)
        } else { // if error, use error
          actionsTakenBuffer += getErrorMsgForLimitUser(action.lastActionDetails.get.asInstanceOf[UserLimitAction])
        }
      case _ => // handle later
    })

    actionsTakenBuffer.toList
  }

  // Create AlertCloudEvent object with extensions.
  // Extensions is needed for the notifications service to pick up metadata about events like key/tenant etc
  def apply(alertData: AlertData): AlertCloudEvent = {
    val cloudEventAlertType = alertData.`type` match {
      case AlertType.RansomwareAttack => AlertCloudEvent.EventTypeRansomwareAlerts
      case AlertType.DataDestructionFileDeletionAttack => AlertCloudEvent.EventTypeDataDestructionAlerts
      case _ => AlertCloudEvent.EventTypeAlerts
    }
    AlertCloudEvent(
      cloudEventAlertType,
      alertData,
      Map(
        TenantExtension -> alertData.tenantId,
        KeyExtension -> alertData.tenantId)
    )
  }

  def apply(alertDb: AlertDb, actionsTaken: List[ActionTaken]): AlertCloudEvent = {
    val actionsTakenAsString: List[String] = getActionsTakenAsString(actionsTaken)
    val alertData: AlertData =
      AlertData(
        alertDb.tenantId,
        alertDb.id,
        alertDb.displayId,
        alertDb.alertType,
        alertDb.severity,
        alertDb.synopsis,
        alertDb.userName.getOrElse(""),
        alertDb.communityName.getOrElse(""),
        alertDb.userId.getOrElse(""),
        alertDb.communityId.getOrElse(""),
        alertDb.alertTimestamp,
        alertDb.status,
        alertDb.note,
        alertDb.changePercentage,
        alertDb.attributes,
        actionsTakenAsString)

    AlertCloudEvent(alertData)
  }

}

/* Using scala classes in java, issues in deserialization
https://stackoverflow.com/questions/42739919/deserialize-scala-deserialized-object-in-java
https://medium.com/nick-tune-tech-strategy-blog/strict-json-deserialization-with-enum-validation-b6861f11371c
actionTaken should ideally be used as List[ActionTaken], but because of deserialization issues, keeping it List[String] for now.
*/

// Data part in the cloud event object
case class AlertData(tenantId: String,
                     id: String,
                     displayId: String,
                     `type`: AlertType,
                     severity: AlertSeverity,
                     synopsis: String,
                     userName: String,
                     communityName: String,
                     userId: String,
                     communityId: String,
                     alertTimestamp: Long,
                     status: AlertStatus,
                     note: String,
                     changePercentage: Double,
                     attributes: Map[String, String],
                     actionsTaken: List[String])

case class AlertCloudEvent(eventType: String = AlertCloudEvent.EventTypeAlerts, data: AlertData, extensions: Map[String, String])

/**
  * AlertActionsJsonFormats is to provide implicit conversions from the object to json format using spray for alert objects
  */
class AlertCloudEventJsonFormats extends AlertActionsJsonFormats {
  implicit val alertDataFormatter: RootJsonFormat[AlertData] = jsonFormat16(AlertData.apply)
  implicit val alertCloudEventFormatter: RootJsonFormat[AlertCloudEvent] = jsonFormat3(AlertCloudEvent.apply)
}
