/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.rest

import com.netapp.df.uba.server.common.misc.dbmodel.AlertType.AlertType
import com.netapp.df.uba.server.common.misc.dbmodel._
import org.apache.http.client.methods._
import org.apache.http.client.utils.URIBuilder
import org.apache.http.entity.ContentType
import org.apache.http.message.BasicHeader
import org.apache.http.{HttpHeaders, HttpHost}
import spray.json.DefaultJsonProtocol.BooleanJsonFormat
import spray.json._

import scala.collection.mutable.ListBuffer
import scala.util.{Failure, Success, Try}

/**
  * A REST library for config management API usage.
  */
class ConfigMgmtRestClient extends RestUtils {

  //TODO: See if there is a better way to refer these from via some environment. Right now commons doesnt have its own helm, so
  // forced to hardcode.
  private val ConfigMgmtServiceName = "cs-service-config-mgmt.cloudsecure.svc.cluster.local"

  private val ConfigMgmtServiceNameEnv = "CONFIG_MGMT_HOST_NAME"
  private val ConfigMgmtServicePortEnv = "CONFIG_MGMT_PORT"
  private val ConfigMgmtServiceSchemeEnv = "CONFIG_MGMT_SCHEME"

  private val serviceNameToPortMap = Map(ConfigMgmtServiceName -> "8080")

  /**
    * Get Feature Flags
    *
    * @param tenantId tenant id
    * @return
    */
  def getFeatureFlag(tenantId: String): Try[List[FeatureFlag]] = {
    val uriBuilder: URIBuilder = new URIBuilder().setPath("/rest/v1/feature-flags") // set base path

    val uri = uriBuilder.build().toString

    val httpResponseBodyTry = execute(Option(tenantId), HttpGet.METHOD_NAME, uri, getHttpHost)
    httpResponseBodyTry match {
      case Success(httpResponseBody) =>
        implicit val featureFlagJsonFormats: FeatureFlagJsonFormats = new FeatureFlagJsonFormats
        import featureFlagJsonFormats._
        // parseJson should always be enclosed in Try to catch any exception it throws
        Try(httpResponseBody.parseJson.convertTo[List[FeatureFlag]])
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed retrieving feature flags: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Fetch UUID's of Data collectors from Config
    *
    * @param tenantId tenant id
    * @return
    */
  def getDataCollectors(tenantId: String): Try[List[String]] = {
    val uriBuilder: URIBuilder = new URIBuilder().setPath("/rest/v1/collectors") // set base path
    uriBuilder.addParameter("collectorCategory", "DATA") // add params for Only Data Collectors

    val uri = uriBuilder.build().toString

    val httpResponseBodyTry = execute(Option(tenantId), HttpGet.METHOD_NAME, uri, getHttpHost)

    val datasourceUUIDs = ListBuffer[String]()

    httpResponseBodyTry match {
      case Success(httpResponseBody) =>
        Try {
          // Only extract datasource uuid
          httpResponseBody.parseJson.asInstanceOf[JsArray].elements.foreach(jsValue => {
            // Remove extra quotes
            datasourceUUIDs.append(jsValue.asJsObject.fields("uuid").toString().replaceAll("^\"|\"$", ""))
          })
          datasourceUUIDs.toList
        }
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed fetching data collectors: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Fetch details of a specific data-collector from Config mgmt service
    *
    * @param tenantId tenant id
    * @param collectorUuid UUID of the collector to fetch details for.
    * @return details of the collector in json format or Failure
    */
  def getDataCollector(tenantId: String, collectorUuid: String): Try[JsObject] = {
    val uriBuilder: URIBuilder = new URIBuilder().setPath(s"/rest/v1/collectors/$collectorUuid")
    val uri = uriBuilder.build().toString

    val httpResponseBodyTry = execute(Option(tenantId), HttpGet.METHOD_NAME, uri, getHttpHost)
    httpResponseBodyTry match {
      case Success(httpResponseBody) =>
        Try {
          httpResponseBody.parseJson.asJsObject
        }
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed fetching details of data collector $collectorUuid: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Fetch Global User Access Config
    *
    * @param tenantId tenant id
    * @return
    */
  def isUserAccessConfigEnabled(tenantId: String): Try[Boolean] = {
    val uriBuilder: URIBuilder = new URIBuilder().setPath("/rest/v1/configuration") // set base path

    val uri = uriBuilder.build().toString

    val httpResponseBodyTry = execute(Option(tenantId), HttpGet.METHOD_NAME, uri, getHttpHost)

    httpResponseBodyTry match {
      case Success(httpResponseBody) =>
        Try {
          // Only extract user access config enabled property.
          httpResponseBody.parseJson.asJsObject.fields("user-access-config").asJsObject.fields("is-enabled").convertTo[Boolean]
        }
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed fetching user access global config: ${ex.getMessage}")
        Failure(ex)
    }

  }

  /**
    * Get Alert policies for this tenant, alert type and devices
    *
    * @param tenantId  TenantId for which policies need to be retrieved
    * @param alertType Optional alert type filter for which policies need to be retrieved
    * @param devices   Optional devices filter for which policies need to be retrieved
    * @param isActive  Optional flag to retrieve active or inactive policies. By default only active policies are fetched.
    * @return List of policies or Failure if error occurs
    */
  def getAlertPolicies(tenantId: String, alertType: Option[AlertType] = None, devices: Option[List[String]] = None, isActive: Boolean = true): Try[List[AlertPolicy]] = {
    val uriBuilder: URIBuilder = new URIBuilder().setPath("/rest/v1/alert-policies") // set base path
    uriBuilder.addParameter("is_active", isActive.toString) // add params as needed
    if (alertType.isDefined) uriBuilder.addParameter("alert_types", alertType.get.toString)
    if (devices.isDefined && devices.get.nonEmpty) uriBuilder.addParameter("devices", devices.get.mkString(","))

    val uri = uriBuilder.build().toString

    val httpResponseBodyTry = execute(Option(tenantId), HttpGet.METHOD_NAME, uri, getHttpHost)
    httpResponseBodyTry match {
      case Success(httpResponseBody) =>
        implicit val alertPolicyJsonFormats: AlertPolicyJsonFormats = new AlertPolicyJsonFormats
        import alertPolicyJsonFormats._
        // parseJson should always be enclosed in Try to catch any exception it throws
        Try(httpResponseBody.parseJson.convertTo[List[AlertPolicy]])
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed retrieving alert policies: ${ex.getMessage}")
        Failure(ex)
    }
  }

  /**
    * Trigger an alert action (like Create snapshot, Notify admin etc.).
    *
    * @param tenantId         TenantId for which to trigger action
    * @param alertId          AlertId which is causing this action to be triggered
    * @param userId           userId associated with the alert
    * @param reportedIdOption reportedId of the user if present
    * @param isAutomatic      Whether this is an automated action or manual. All UI triggered actions are manual(false), while for backend
    *                         jobs which call this API, this is automatic(true).
    * @param snapshotActions  All the snapshot actions to be triggered
    * @param userLimitActions All the user limit/restore actions to be triggered
    * @return Alert actionId if request is successful, else Failure
    */
  def triggerAlertAction(tenantId: String, alertId: String, userId: String, reportedIdOption: Option[String], alertType: AlertType, isAutomatic: Boolean, snapshotActions: List[DatasourceAlertAction] = List.empty, userLimitActions: List[DatasourceAlertAction] = List.empty): Try[String] = {
    val alertActionsUri: String = s"/rest/v1/alert-actions"

    implicit val alertActionsJsonFormats: AlertActionsJsonFormats = new AlertActionsJsonFormats
    import alertActionsJsonFormats._

    var alertActions = JsArray()
    if (snapshotActions.nonEmpty) {
      alertActions = JsArray(JsObject("type" -> JsString(ActionType.CREATE_SNAPSHOT.toString),
        "instances" -> JsArray(snapshotActions.map(_.toJson).toVector)))
    }

    if (userLimitActions.nonEmpty) {
      alertActions = JsArray(alertActions.elements :+
        JsObject("type" -> JsString(ActionType.LIMIT_USER_ACCESS.toString),
          "instances" -> JsArray(userLimitActions.map(_.toJson).toVector)))
    }

    val actionPayload = JsObject("alert_id" -> JsString(alertId),
      "alert_type" -> JsString(alertType.toString),
      "is_automatic" -> JsBoolean(isAutomatic),
      "user_id" -> JsString(userId),
      if (reportedIdOption.isDefined) "reportedId" -> JsString(reportedIdOption.get) else "reportedId" -> JsNull,
      "actions" -> alertActions
    )

    val httpResponseBodyTry = execute(
      Option(tenantId),
      HttpPost.METHOD_NAME,
      alertActionsUri,
      getHttpHost,
      Some(actionPayload.prettyPrint),
      Array(new BasicHeader(HttpHeaders.CONTENT_TYPE, ContentType.APPLICATION_JSON.getMimeType))
    )
    httpResponseBodyTry match {
      case Success(httpResponseBody) =>
        Try {
          httpResponseBody.parseJson.asJsObject.fields("action_id").convertTo[String]
        }
      case Failure(ex) =>
        logErrorT(tenantId, s"Failed triggering alert action: ${ex.getMessage}")
        Failure(ex)
    }
  }

  private def getHttpHost: HttpHost = {
    //NOTE: Only Contract based UTs are supposed to override these.
    new HttpHost(
      System.getProperty(ConfigMgmtServiceNameEnv, ConfigMgmtServiceName),
      System.getProperty(ConfigMgmtServicePortEnv, serviceNameToPortMap(ConfigMgmtServiceName)).toInt,
      System.getProperty(ConfigMgmtServiceSchemeEnv, "https")
    )
  }
}