/*
 * $Id$
 *
 * Copyright (c) 2021 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.rest

import com.netapp.df.uba.server.common.misc.api.AuthDirectives
import com.netapp.df.uba.server.common.misc.http.HttpClientSingleton
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.oci.central.vault.{CentralVaultClientAccess, CentralVaultClientException}
import com.netapp.oci.jwt.{JWTAuthData, JWTUtils}
import org.apache.http.client.methods._
import org.apache.http.entity.StringEntity
import org.apache.http.util.EntityUtils
import org.apache.http.{Header, HttpHeaders, HttpHost, HttpStatus}

import java.util.Calendar
import scala.util.{Failure, Success, Try}

trait RestUtils extends LogTrait {
  // A dummy CS user for generating token
  private val ContrailServiceUser = "cs-user"

  // A dummy tenant Id for generating token with tenant_admin privileges
  protected val DefaultTenantIdForJwt = "default-tenant"

  //JWT token expiry
  private val JwtTokenExpiryMins = 5

  //Right now, vault service is set explicitly via helm for some of the SAAS services like forensics. Setting it by default here, so that
  //we don't need to explicitly set it in helm for each and every SAAS service. There is no helm template for commons, so forced to hardcode
  //here.
  private val VaultServiceName = "cs-central-vault.cloudsecure.svc.cluster.local"

  protected def execute(tenantIdOption: Option[String],
                      method: String,
                      uri: String,
                      httpHost: HttpHost,
                      requestPayload: Option[String] = None, httpHeaders: Array[Header] = Array.empty): Try[String] = {

    // Tenant id to be used for JWT and log messages
    val (tenantId, jwtRole) = tenantIdOption match {
      case None => (DefaultTenantIdForJwt, "tenant_admin")
      case Some(id) => (id, "admin")
    }

    val httpClientTry = HttpClientSingleton.getHttpClient
    httpClientTry match {

      case Success(httpClient) =>
        logInfoT(tenantId, s"Generating JWT token with $jwtRole privileges")
        val jwtTokenTry: Try[String] = getJwtToken(tenantId, jwtRole)

        jwtTokenTry match {
          case Success(jwtToken) =>
            Try {
              val httpRequest: HttpRequestBase =
                if (HttpDelete.METHOD_NAME == method) new HttpDelete(uri)
                else if (HttpGet.METHOD_NAME == method) new HttpGet(uri)
                else if (HttpPost.METHOD_NAME == method) {
                  val httpPost: HttpPost = new HttpPost(uri)
                  httpPost.setEntity(new StringEntity(requestPayload.getOrElse("")))
                  httpPost
                } else if (HttpPut.METHOD_NAME == method) {
                  val httpPut: HttpPut = new HttpPut(uri)
                  httpPut.setEntity(new StringEntity(requestPayload.getOrElse("")))
                  httpPut
                } else throw ContrailRestException(s"Unsupported HTTP request method $method for $uri", null)

              if (!httpHeaders.isEmpty) httpRequest.setHeaders(httpHeaders)
              httpRequest.setHeader(HttpHeaders.AUTHORIZATION, s"Bearer $jwtToken")

              logInfoT(tenantId, s"HTTP $method, ${httpRequest.getURI.toString}")

              val httpResponse = httpClient.execute(httpHost, httpRequest)
              val httpResponseTry: Try[String] = if (httpResponse.getStatusLine != null
                && httpResponse.getStatusLine.getStatusCode >= HttpStatus.SC_OK
                && httpResponse.getStatusLine.getStatusCode < HttpStatus.SC_MULTIPLE_CHOICES) {

                // Extract the response body from the response
                val responseBody = EntityUtils.toString(httpResponse.getEntity)
                logInfoT(tenantId, s"Received the following response: $responseBody")
                Success(responseBody)

              } else {
                val errorMessage = s"Failed HTTP response. statusLine: ${httpResponse.getStatusLine}"
                logErrorT(tenantId, errorMessage)
                logErrorT(tenantId, s"Failed response: ${EntityUtils.toString(httpResponse.getEntity)}")
                Failure(ContrailRestException(errorMessage, null))
              }
              EntityUtils.consumeQuietly(httpResponse.getEntity)
              httpResponseTry.get
            }
          case Failure(ex) =>
            Failure(ex)
        }

      case Failure(ex) =>
        logErrorT(tenantId, s"Failed to get http client. Reason: ${ex.getMessage}")
        Failure(ContrailRestException(s"Failed to get http client. Reason: ${ex.getMessage}", ex.getCause))
    }
  }

  private def getJwtToken(tenantId: String, roles: String*): Try[String] = {
    val environment = Option(System.getenv("DEPLOY_ENV"))
    if (environment.getOrElse("") == AuthDirectives.BypassVaultAuthEnvironment) {
      logDebugT(tenantId, "Creating JWTAuthData to bypass authentication for dev environment.")

      val calendar = Calendar.getInstance()
      calendar.set(Calendar.YEAR, calendar.get(Calendar.YEAR) + 1)

      Try {
        val jwtAuthData = new JWTAuthData(
          "tenant1",
          "dev_user",
          calendar.getTime,
          "admin", "user", "guest")
        JWTUtils.toJWT(jwtAuthData)
      }
    } else {
      val jwtSecretTry: Try[String] = {
        val secret = {
          //To avoid vault access for UTs
          val secretEnv = System.getenv("JWT_SECRET")
          if (secretEnv != null) secretEnv else System.getProperty("JWT_SECRET")
        }
        if ((secret != null) && secret.nonEmpty) Success(secret) else {
          logInfoT(tenantId, "Fetching JWT key from CS central vault")
          System.setProperty("VAULT_HOST_NAME", VaultServiceName)
          try Success(CentralVaultClientAccess.getInstance.getJWTKey)
          catch {
            case e: CentralVaultClientException =>
              logErrorT(tenantId, "Exception retrieving JWT key from central vault.")
              Failure(ContrailRestException("Error retrieving JWT key from central vault", e))
          }
        }
      }
      jwtSecretTry match {
        case Success(jwtSecret) =>
          val calendar: Calendar = Calendar.getInstance()
          calendar.add(Calendar.MINUTE, JwtTokenExpiryMins)
          val jwtAuthData = new JWTAuthData(tenantId, ContrailServiceUser, calendar.getTime, roles: _*)
          Success(JWTUtils.toJWT(jwtAuthData, jwtSecret))
        case Failure(ex) =>
          logErrorT(tenantId, "Failed to generate JWT token...")
          Failure(ex)
      }
    }
  }
}

case class ContrailRestException(message: String, throwable: Throwable) extends Exception(message, throwable)