/*
 * $Id$
 *
 * Copyright (c) 2020 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.settings

import com.netapp.df.uba.server.common.misc.settings.FilterMode.FilterMode
import com.netapp.df.uba.server.common.misc.utils.SettingUtils
import com.typesafe.config.{Config, ConfigException}

// An enum to represent the type of filtering to be done
object FilterMode extends Enumeration {
  type FilterMode = Value

  // Used by ADR micro-service
  val IncludeMode: FilterMode = Value("include")
  val ExcludeMode: FilterMode = Value("exclude")
  val NoFilterMode: FilterMode = Value("none")
}

/**
  * In different micro-services, at times, there is a need to include or exclude certain objects
  * and the following provides a generic setting to do the same.
  *
  * Example: Prevent CD from running for specific tenants
  * Example: Drop events of a specific collector in ADR
  */
object FilterObjectSettings {
  def apply(config: Config): FilterObjectSettings = {
    val filterMode = FilterMode.withName(config.getString("filter-mode"))
    val objectsToFilter = SettingUtils.getNonEmptyValueStringSet(config, "objects-to-filter")
    filterMode match {
      case FilterMode.IncludeMode | FilterMode.ExcludeMode if objectsToFilter.isEmpty =>
        throw new ConfigException.BadValue(s"objects-to-filter", "Objects to filter can not be empty when filter-mode is 'include' or 'exclude'.")
      case FilterMode.NoFilterMode if objectsToFilter.nonEmpty =>
        throw new ConfigException.BadValue(s"objects-to-filter", "Objects to filter should be empty when filter-mode is 'none'.")
      case _ => // All fine
    }
    FilterObjectSettings(
      filterMode,
      objectsToFilter
    )
  }
}

case class FilterObjectSettings(
                       filterMode: FilterMode,
                       objectsToFilter: Set[String]
                       ) {
  // A method that can be used to know whether a specific object should be accepted for processing or not
  def isObjectToBeAccepted(objectValue: String): Boolean = {
    filterMode match {
      case FilterMode.NoFilterMode =>
        true // no filtering, so accept

      case FilterMode.ExcludeMode =>
        if (objectsToFilter.contains(objectValue)) false
        else true

      case FilterMode.IncludeMode =>
        if (objectsToFilter.contains(objectValue)) true
        else false
    }
  }
}