/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.utils

import akka.actor.{Actor, ActorRef, ReceiveTimeout}
import akka.event.LoggingReceive
import akka.util.Timeout

import scala.concurrent.ExecutionContext
import scala.concurrent.duration._

/**
  * A trait to indicate that the actor intends to send ask messages to other actors
  */
trait ExecContextAndTimeOutImplicits extends Actor {
  // Implicit definition of execution context
  implicit val executionContext: ExecutionContext = context.system.dispatcher

  // Implicit definition of timeout required in ask pattern calls
  protected implicit val timeout: Timeout = 60.seconds
}

/**
  * A trait to aggregate responses from multiple actors and return an aggregated response back to original sender
  */
trait AppActorUser extends Actor {

  // Utility receive method to add with orElse to main receive method of an actor
  def timedOutReceive: Receive = LoggingReceive {
    case ReceiveTimeout =>
      context stop self
  }

  // Ask data from an actor given an ActorRef by sending it a message
  def askDataFromActor(actorRef: ActorRef, message: Any): Unit = {
    actorRef ! message
  }

  // Ask data from an actor given an actor path by sending it a message
  def askDataFromActor(actorPath: String, message: Any): Unit = {
    context.actorSelection(actorPath) ! message
  }

  // Actually send response to the original sender and stop the worker actor
  def sendResponse(originalSender: ActorRef, messageToSend: Any): Unit = {
    // Send the message to route handler
    originalSender ! messageToSend

    // Work done, stop itself
    context stop self
  }
}

/**
  * A trait to aggregate responses from multiple actors and return an aggregated response back to original sender
  */
trait AppAggregator extends AppActorUser {

  // Internal state to maintain count of the number of actors from which responses has to be waited for.
  private var repliesToWaitFor: Int = 0

  // Ask data from an actor given an ActorRef by sending it a message
  override def askDataFromActor(actorRef: ActorRef, message: Any): Unit = {
    super.askDataFromActor(actorRef, message)
    repliesToWaitFor += 1
  }

  // Ask data from an actor given an actor path by sending it a message
  override def askDataFromActor(actorPath: String, message: Any): Unit = {
    super.askDataFromActor(actorPath, message)
    repliesToWaitFor += 1
  }

  // Check if responses are all received and if yes, send it to original sender
  def checkAndSendResponse(originalSender: ActorRef): Unit = {
    repliesToWaitFor -= 1

    if (repliesToWaitFor == 0) {
      sendResponse(originalSender, aggregatedResponse)
    }
  }

  // To be overridden to get an aggregated response
  protected def aggregatedResponse: Any
}