/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.utils

import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model.{Property, ServiceApi}
import com.netapp.df.uba.server.common.misc.settings.CommonSettings.CommonSettingPrefix
import com.typesafe.config.{Config, ConfigFactory, ConfigValueFactory}
import org.slf4j.event.Level

import java.text.SimpleDateFormat
import java.time.format.DateTimeFormatter
import java.time.{Instant, ZoneId, ZonedDateTime}
import java.util.{Date, TimeZone}
import scala.util.{Failure, Success, Try}

object GenericUtils extends LogTrait {

  /**
    * Utility method to get the current timestamp in the API compatible format
    * References based on which this was decided
    * - Microsoft REST API guidelines recommends the following in this section:
    * https://github.com/Microsoft/api-guidelines/blob/vNext/Guidelines.md#113-json-serialization-of-dates-and-times
    * "The Web has coalesced around the ECMAScript subset of ISO 8601 date formats (ISO 8601)"
    *
    * ECMA guidelines is stated here:
    * http://www.ecma-international.org/ecma-262/5.1/#sec-15.9.1.15
    *
    * Swagger API specification guidelines also recommends the same thing when the format used is dateTime with type as string
    * https://swagger.io/docs/specification/data-models/data-types/
    * https://tools.ietf.org/html/rfc3339#section-5.6
    */
  def currentTimestamp: String = ZonedDateTime.now(ZoneId.of("UTC")).format(DateTimeFormatter.ISO_INSTANT)

  /**
    * Checks for supported log levels
    */
  def checkLoglevel(loglevel: String): Boolean = loglevel match {
    case "ERROR" | "WARN" | "INFO" | "DEBUG" => true
    case _ => false
  }

  /**
    * An unfortunate bug in akka logging - https://github.com/akka/akka/issues/16745
    *   - slf4j log level is WARN but akka supports log level as WARNING leading to inconsistencies
    *     Supported log levels are DEBUG, INFO, WARN and ERROR only
    *     akka.loglevel and akka.stdout-loglevel in reference.conf refer to ${datasource.loglevel}
    *     Even before actor system is initialized with the config, the config is adjusted here to change WARN in slf4j to WARNING for akka
    *     Also, support for only valid levels is tested here.
    */
  def logAdjustedConfig: Config = {
    val currentConfig: Config = ConfigFactory.load()
    val adjustedConfig: Try[Config] = logAdjustedConfig(currentConfig)

    adjustedConfig match {
      case Success(config) => config

      case Failure(ex) =>
        System.err.println(ex)
        System.exit(1)
        currentConfig
    }
  }

  def logAdjustedConfig(currentConfig: Config): Try[Config] = {

    // Get the loglevel setting
    val microServiceLogLevel = currentConfig.getString(s"$CommonSettingPrefix.loglevel")

    // Check for validity of loglevels to the supported list
    if (!GenericUtils.checkLoglevel(microServiceLogLevel)) {
      Failure(new Exception(s"Unsupported log level in configuration - $microServiceLogLevel. Valid levels are: ERROR, WARN, INFO, DEBUG"))
    } else {
      // Define the akka log level as same as component log level, but change to WARNING if WARN is used in slf4j
      val akkaLogLevel = if (microServiceLogLevel == Level.WARN.toString) "WARNING" else microServiceLogLevel

      // Create the new config with correct log levels for akka.loglevel and akka.stdout-loglevel
      Success(currentConfig.withValue("akka.loglevel", ConfigValueFactory.fromAnyRef(akkaLogLevel))
        .withValue("akka.stdout-loglevel", ConfigValueFactory.fromAnyRef(akkaLogLevel)))
    }
  }

  /**
    * Timestamp string conversion to epochmillis
    */
  def epochMills(timeString: String): Long = {
    // Convert the timestamp reported by the datasource for reported timestamp into epochMilli format as expected by Elasticsearch
    val isoInstantUserReportedTimestamp = DateTimeFormatter.ISO_INSTANT.parse(timeString)
    Instant.from(isoInstantUserReportedTimestamp).toEpochMilli
  }

  /**
    * Timestamp epoch millis to string conversion
    */
  def timeString(epochMillis: Long): String = {
    val formatter = DateTimeFormatter.ISO_INSTANT.withZone(ZoneId.of("UTC"))
    val dateTime = Instant.ofEpochMilli(epochMillis).atZone(ZoneId.of("UTC"))
    dateTime.format(formatter)
  }

  /**
    * @param timestamp Timestamp to be converted to MMM d yyyy HH:mm:ss zzz date format
    * @param timezone  Timezone provided for date conversion
    * @return Date in string format
    */
  def getDateWithTimeZone(timestamp: Long, timezone: String): String = {
    val simpleDateFormat = new SimpleDateFormat("MMM d yyyy HH:mm:ss zzz")
    simpleDateFormat.setTimeZone(TimeZone.getTimeZone(timezone))
    val dateFromTimestamp = new Date(timestamp)
    simpleDateFormat.format(dateFromTimestamp)
  }

  /**
    * Utility for conversion of optional list of properties to a map
    */
  def attributesMap(properties: Option[List[Property]]): Map[String, String] = properties match {
    case None => Map.empty
    case Some(propertyList) => attributesMap(propertyList)
  }

  /**
    * Utility for conversion of list of properties to a map
    */
  def attributesMap(properties: List[Property]): Map[String, String] = properties
    .map(property => property.name -> property.value)
    .toMap

  /**
    * Utility for conversion of serviceApi to a map
    */
  def servicesMap(services: List[ServiceApi]): Map[String, String] = services
    .map(service => service.name -> service.status)
    .toMap

  /**
    * Utility for log base 2
    */
  def log2(value: Double): Double = {
    value match {
      case 0 => 0
      case _ => scala.math.log10(value) / scala.math.log10(2)
    }
  }

  /**
    * Utility to log time taken for some action
    */
  def execAndLogTime[R](block: => R, context: String): R = {
    val start = System.currentTimeMillis()
    val result = block
    val end = System.currentTimeMillis()
    logger.info(s"Elapsed time executing '$context': " + (end - start) + "ms.")
    result
  }

  /**
    * Utility to retry multiple times
    *
    * @param attempts               : Number of times to run the code meant for retry
    * @param context                : A context to help identify the log messages
    * @param sleepInBetweenInMillis : Sleep time between retries, 0 for no sleep in between
    * @param isExponential          : If exponential is true, then after each attempt, the sleep time is doubled
    * @param exceptionCheckForRetry : Retry is done only if the thrown exception passes the check in this function
    * @param functionToRetry        : Actual code meant for retry
    */
  @scala.annotation.tailrec
  def retry[T](attempts: Int,
               context: String,
               sleepInBetweenInMillis: Int = 0,
               isExponential: Boolean = false,
               exceptionCheckForRetry: Exception => Boolean = _ => true
              )(functionToRetry: => T): T = {
    try {
      functionToRetry
    } catch {
      case ex: Exception =>
        // Retry only if the exception passes the check sent by the caller
        if (exceptionCheckForRetry(ex)) {
          if (attempts > 1) {
            logger.error(s"$context - Failed. Reason: ${ex.getMessage};${ex.getCause}; Retrying, remaining attempts ${attempts - 1}")
            // Sleep between retries if applicable
            if (sleepInBetweenInMillis > 0) Thread.sleep(sleepInBetweenInMillis)

            // Compute the new sleep value based on whether isExponential is true or false
            val newSleepValueInMillis: Int = if (isExponential) sleepInBetweenInMillis * 2 else sleepInBetweenInMillis

            // Retry the code
            retry(attempts - 1, context, newSleepValueInMillis, isExponential, exceptionCheckForRetry)(functionToRetry)

          } else {
            logger.error(s"$context - Failed after all retries. Reason: ${ex.getMessage}.")
            throw ex
          }
        } else {
          logger.error(s"$context - Exception not meant for retry attempt. Reason: ${ex.getMessage}.")
          throw ex
        }
    }
  }

  /**
    * Find the user domain based on directoryId or rawUserId. Where rawUserId get preference, if provided.<br>
    * directoryId is in the form ex- ldap:{domain}<br>
    * rawUserId is in the form ex- ldap:{domain}:{sid}
    *
    * @param userDirectoryId User's directory id
    * @param rawUserIdOption User's raw id
    * @return Domain to which user belongs
    */
  def getDomain(userDirectoryId: String, rawUserIdOption: Option[String] = None): String = {
    val domainSource = if (rawUserIdOption.isDefined) rawUserIdOption.get else userDirectoryId
    domainSource match {
      case _ if domainSource.equals("ldap:default") || domainSource.startsWith("ldap:default:") => ""
      case _ if domainSource.startsWith("ldap:") => domainSource.split(":").lift(1).getOrElse("").toLowerCase
      case _ => ""
    }
  }

  /**
    * Get the user's SID from its reportedId.
    *
    * @param reportedId User's raw id generated in the form of ldap:{domain}:{SID}
    *                   or local:{datasourceId}:{SID} for ONTAP local users
    * @return User's SID
    */
  def getSidFromReportedId(reportedId: String): String = {
    reportedId match {
      case _ if reportedId.startsWith("ldap:") || reportedId.startsWith("local:") => reportedId.split(":").lift(2).getOrElse("").toUpperCase
      case _ => ""
    }
  }
}