/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.utils

import com.netapp.df.uba.server.common.misc.dbmodel.MetadataObjectType
import com.netapp.df.uba.server.common.misc.dbmodel.MetadataObjectType.MetadataObjectType
import com.netapp.df.uba.server.common.misc.settings.{EventPathSettings, PathConfigMode}

/**
  * A set of utility methods related to how entity path is handled
  * Stripping certain levels of paths, extension extraction are part of this.
  */
object PathUtils {
  /**
    * Stripped or modified path given a file path. Stripping of levels are done as explained
    * in reference.conf
    */
  def getExcludedDirPath(eventPathSettings: EventPathSettings, canonicalPath: String, mountPath: String): String = {
    eventPathSettings.pathConfigMode match {

      // Current behavior, do not manipulate anything that was sent from collectors
      case PathConfigMode.NoChange =>
        canonicalPath

      // Up-level to mount path level which is equivalent to saying all users who access files in a data volume/mount-path are in the community
      case PathConfigMode.MountPath =>
        mountPath

      // Up-level to some levels of directory in the path of the file.
      case PathConfigMode.Directory =>

        // Check exclude level
        val excludeLevel: Int = eventPathSettings.excludeDirectoryLevel

        // Ensure the path does not have / at the end for ease of manipulation below
        val canonicalPathInEvent: String = canonicalPath
          .stripSuffix("/")

        /*
         Split the canonical and display path based on /
         and filter out the empty ones
         Example: /data1/dir1/file1 will return data1, dir1, file1
          */
        val splitPathCanonicalComponents: List[String] = canonicalPathInEvent
          .split("/")
          .toList
          .filter(_.nonEmpty)

        splitPathCanonicalComponents.size match {
          case n if excludeLevel >= n =>
            /*
             If the number of components in the path is lesser than or equal to configured exclude levels
             then, just consider /, which is actually equivalent to mount path or data volume level event
              */
            "/"

          case _ =>
            /*
            In the other case, remove components from the right as per exclude level and construct back the
            canonical and display path using / as separator.
             */
            splitPathCanonicalComponents
              .dropRight(excludeLevel)
              .mkString("/", "/", "")
        }
    }
  }

  /**
    * Utility method to get name and extension from the display path
    */
  def getLastPartAfterDelimiter(input: String, delimiter: String, defaultValue: String): String = {
    if (input.isEmpty) defaultValue
    else {
      val splitComponents: List[String] = input.split(delimiter).toList
      splitComponents match {
        case Nil =>
          defaultValue

        case components => components.last
      }
    }
  }

  /**
    * Utility method to get extension from a file name
    * Update based on DFPS-2767.
    *
    * Extension computation examples:
    * fileName.2019.hello.xls   --> xls (Use last extension component only)
    * fileName.pdf.exe          --> exe (Used in attachments/malware to dupe folks, exe is the actual extension, use last component)
    * fileName.2019.hello.XLS   --> xls (Conversion of upper to lower case)
    * fileName.log.1            --> log (Log file conventions: Generic removal of last set of extensions based on it being a number)
    * fileName.log.2019.1       --> log (Log file conventions: Generic removal of last set of extensions based on it being a number)
    * fileName.pdf.             --> Default value: file
    */
  def getExtensionComponent(input: String, objectType: MetadataObjectType): String = {

    def isAllDigits(str: String) = str forall Character.isDigit

    objectType match {
      // In case of directories, irrespective of directory name which can include dots, keep extension as the string "directory"
      case MetadataObjectType.Directory => objectType.toString

      // Other cases: File, symbolicLink, stream etc..
      case _ =>
        val firstIndexOfDot: Int = input.indexOf('.')
        firstIndexOfDot match {
          case num if num == input.length - 1 => objectType.toString  // . at the end of the string

          case num if num < 0 => objectType.toString  // no .

          case _ =>   // Dots present case
            // Get all extensions
            val extensionComponents = input
              .split("\\.")
              .drop(1)  // remove the actual entity name
              .map(_.toLowerCase()) // convert to lower case
              .reverse  // start from the end of the string

            extensionComponents.length match {
              case 1 => // Means a single extension whether it is alphanumeric or only numeric
                extensionComponents.head

              case _ => // Means multiple extensions, select the right one as per the logic related to discarding numbers
                extensionComponents
                  .find(!isAllDigits(_)) // Find the first non-digit-only extension
                  .getOrElse(if (extensionComponents.nonEmpty) extensionComponents.head else objectType.toString) // If all are numeric, then give up, give the last number
            }
        }
    }
  }
}
