/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.utils

import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.typesafe.config.{Config, ConfigException, ConfigObject}

import scala.collection.JavaConverters._
import scala.concurrent.duration._
import scala.language.implicitConversions

/**
  * A set of utility methods to load the configuration
  */
object SettingUtils extends LogTrait {
  /**
    * Utility method to convert java Duration's that are used in config file
    * into Scala's FiniteDuration objects
    */
  implicit def asFiniteDuration(d: java.time.Duration): FiniteDuration =
    scala.concurrent.duration.Duration.fromNanos(d.toNanos)

  /**
    * Utility method to load settings of a specific type from the config such that any config exception is
    * caught, logged appropriately, and the exception is thrown for upper layers
    */
  def loadSettings[T](configApply: Config => T, config: Config): T =
    try {
      configApply(config)
    } catch {
      case cex: ConfigException =>
        logger.error(s"ConfigException thrown when loading the configuration of the component. Reason: ${cex.getMessage}}")
        throw new RuntimeException(s"Component configuration has an error. Specific reason: ${cex.getMessage}", cex)
    }

  /**
    * Utility method to get config value for a key where empty or null returns None
    * and a correct value returns Some(String)
    */
  def configValueWithEmptyAsNone(config: Config, configKey: String): Option[String] = {
    if (config.hasPathOrNull(configKey)) {
      if (config.getIsNull(configKey)) None
      else if (config.getString(configKey).isEmpty) None
      else Some(config.getString(configKey))
    } else {
      // Let it throw the Config.MissingException
      Some(config.getString(configKey))
    }
  }

  /**
    * Utility method to get config value for a key where empty or missing or null returns None
    * and a correct value returns Some(String)
    */
  def configValueWithEmptyOrMissingAsNone(config: Config, configKey: String): Option[String] = {
    if (config.hasPathOrNull(configKey)) {
      if (config.getIsNull(configKey)) None
      else if (config.getString(configKey).isEmpty) None
      else Some(config.getString(configKey))
    } else {
      None
    }
  }

  /**
    * Utility method to get config value for a key where empty or null returns None
    * and a correct value returns Some(String)
    */
  def configValueWithEmptyAsNoneInt(config: Config, configKey: String): Option[Int] = {
    if (config.hasPathOrNull(configKey)) {
      if (config.getIsNull(configKey)) None
      else if (config.getString(configKey).isEmpty) None
      else Some(config.getInt(configKey))
    } else {
      // Let it throw the Config.MissingException
      Some(config.getInt(configKey))
    }
  }

  /**
   * Utility method to get config value for a long key where empty or missing or null returns None
   * and a correct value returns Some(Long)
   */
  def configValueWithEmptyOrMissingAsNoneLong(config: Config, configKey: String): Option[Long] = {
    if (config.hasPathOrNull(configKey)) {
      if (config.getIsNull(configKey)) None
      else if (config.getString(configKey).isEmpty) None
      else Some(config.getLong(configKey))
    } else {
      None
    }
  }

  /**
   * Utility method to get config value for a double key where empty or missing or null returns None
   * and a correct value returns Some(Double)
   */
  def configValueWithEmptyOrMissingAsNoneDouble(config: Config, configKey: String): Option[Double] = {
    if (config.hasPathOrNull(configKey)) {
      if (config.getIsNull(configKey)) None
      else if (config.getString(configKey).isEmpty) None
      else Some(config.getDouble(configKey))
    } else {
      None
    }
  }

  /**
   * Utility method to get config value for a boolean key where empty or missing or null returns None
   * and a correct value returns Some(Boolean)
   */
  def configValueWithEmptyOrMissingAsNoneBoolean(config: Config, configKey: String): Option[Boolean] = {
    if (config.hasPathOrNull(configKey)) {
      if (config.getIsNull(configKey)) None
      else if (config.getString(configKey).isEmpty) None
      else Some(config.getBoolean(configKey))
    } else {
      None
    }
  }

  /**
   * Utility method to get config value for a key where empty or null results in BadValue exception
   * and a correct value returns the string
   */
  def mandatoryConfigValue(config: Config, configKey: String): String = {
    val configValueOption: Option[String] = configValueWithEmptyAsNone(config, configKey)
    configValueOption match {
      case None =>
        throw new ConfigException.BadValue(configKey, "Value cannot be empty.")
      case Some(configValue) if configValue.trim().isEmpty =>
        throw new ConfigException.BadValue(configKey, "Value cannot be empty.")
      case Some(configValue) =>
        configValue
    }
  }

  /**
    * Utility method to check if the int value for a key is within the approved range
    */
  def validIntConfig(config: Config, key: String, check: Int => Boolean, message: String): Int = {
    val configValue: Int = config.getInt(key)
    if (!check(configValue)) throw new ConfigException.BadValue(key, message)
    else configValue
  }

  /**
    * Utility method to check if the int value for a key is within the approved range
    */
  def validMinInt(config: Config, key: String, minValue: Int): Int = validIntConfig(config,
      key,
      {value => value >= minValue},
      s"Value should be at least $minValue")

  /**
    * Utility method to check if the int value for a key is within a specific range
    */
  def validIntRange(config: Config, key: String, minValueIncluded: Int, maxValueIncluded: Int): Int = validIntConfig(config,
      key,
      {value => value >= minValueIncluded && value <= maxValueIncluded},
      s"Value should be in the range of $minValueIncluded..$maxValueIncluded")

  /**
    * Utility method to check if the int value for a key is either 0 (indicative of not specifying) or above a min value
    */
  def validMinIntWithDefault(config: Config, key: String, minValue: Int): Int = validIntConfig(config,
      key,
      {value => value == 0 || value >= minValue},
      s"Value should be at least $minValue or 0")


  /**
    * Utility method to check if the duration value for a key is within the approved range
    */
  def validDurationConfig(config: Config, key: String, check: FiniteDuration => Boolean, message: String): FiniteDuration = {
    val configValue: FiniteDuration = config.getDuration(key)
    if (!check(configValue)) throw new ConfigException.BadValue(key, message)
    else configValue
  }

  /**
    * Utility method to check for all intervals as >= minDuration
    */
  def validMinDuration(config: Config, key: String, minDuration: FiniteDuration = 2.minutes): FiniteDuration = {
    validDurationConfig(config,
      key,
      {dur => dur >= minDuration },
      s"Interval times should be greater than or equal to $minDuration")
  }

  /**
    * Utility method to get a list of Config objects in scala
    */
  def toConfigList(javaList: java.util.List[_ <: ConfigObject]) :  List[Config] = javaList
      .asScala
      .map(_.toConfig)
      .toList

  /**
    * Utility method to get a set of strings with none of them as empty
    */
  def getNonEmptyValueStringSet(config: Config, listKeyName: String): Set[String] = {
    val setOfValues = config
      .getStringList(listKeyName)
      .asScala
      .toSet

    if (setOfValues.exists(_.isEmpty)) throw new ConfigException.BadValue(listKeyName, "Value cannot be empty.")
    setOfValues
  }
}