/*
 * $Id$
 *
 * Copyright (c) 2018 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.misc.utils

import com.netapp.df.uba.server.common.misc.model.{ComponentError, ComponentStateApi, Error, ErrorCode, Failed, ServiceApi, StateDetails}

object StateUtils {

  def getComponentStateApi(stateDetails: StateDetails, componentName: String): ComponentStateApi = {
    // Determine the error state of the overall component.
    val errorObject = StateUtils.errorObject(stateDetails, componentName)

    // General set of statuses of the various internal services
    val serviceDetails = StateUtils.serviceDetails(stateDetails)

    // Construct the overall response to be returned
    ComponentStateApi(componentName,
      stateDetails.componentState.stateString,
      serviceDetails,
      errorObject)
  }

  /**
    * Determine the error state of the overall Component. Rules for the same are:
    * - If Component has failed due to other reasons, bubble up the error of individual services and present it in Error details.
    * - If overall Component is any state other than Error, then there will be no error list.
    */
  private def errorObject(stateDetails: StateDetails, componentName: String): Option[Error] = {
    stateDetails.componentState match {

      case ComponentError(reason) =>

        val overallErrorCode = ErrorCode.InternalError

        // Compute the detailed errors of each of the internal services.
        val servicesInError = stateDetails.serviceStatusMap
          .filter(entry => entry._2.isInstanceOf[Failed])
          .map(entry => entry._1 -> entry._2.asInstanceOf[Failed])
          .map(entry => {
            val serviceName = entry._1
            val failedState = entry._2

            Error(failedState.errorCode, failedState.reason, serviceName.toString)
          })
          .toList

        Some(Error(overallErrorCode, reason, componentName, Some(servicesInError)))

      case _ => None
    }
  }

  /**
    * General set of statuses of the various internal services
    */
  def serviceDetails(stateDetails: StateDetails): List[ServiceApi] = {
      stateDetails.serviceStatusMap
        .map(entry => ServiceApi(entry._1.toString, entry._2.statusString))
        .toList
  }
}
