/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.pipeline.job

import akka.actor.{Actor, ActorLogging, ActorRef, Props, Stash, Terminated}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.flink.actor.FlinkJobCreator
import com.netapp.df.uba.server.common.flink.actor.FlinkJobCreator.{SetupFlinkJob, TakeSavePoint}
import com.netapp.df.uba.server.common.flink.cluster.FlinkApiJobHandler
import com.netapp.df.uba.server.common.flink.settings.{FlinkCluster, FlinkJobConfig, FlinkSettings}
import com.netapp.df.uba.server.common.flink.utils.{CheckOrSavePointWithCreateTime, FlinkJobUtils, SavepointCallbackHandler}
import com.netapp.df.uba.server.common.kafka.{KafkaSettings, ProducerTopic}
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model.ErrorCode._
import com.netapp.df.uba.server.common.misc.model._
import com.netapp.df.uba.server.common.misc.settings.CommonSettings
import com.netapp.df.uba.server.common.misc.utils.GenericUtils
import com.netapp.df.uba.server.common.pipeline.client.PipelineClient
import com.netapp.df.uba.server.common.pipeline.data._
import com.netapp.df.uba.server.common.pipeline.job.JobSetupContext.JobSetupContext

import java.io.File
import scala.util.{Failure, Success, Try}

/**
  * PipelineActor handles the creation/reset of a single pipeline flink job.
  */
object PipelineActor extends LogTrait {
  final val ActorName: String = "pipelineActor"
  final val ActorPath: String = s"/user/appManager/$ActorName"

  def props(commonSettings: CommonSettings,
            managerSettings: PipelineJobManagerSettings,
            jobManagerMicroServiceName: String): Props =
    Props(new PipelineActor(commonSettings, managerSettings, jobManagerMicroServiceName))

  // Main message to create a pipeline job
  case class CreatePipeline(pipeline: Pipeline) extends Command

  // Main message to reset a pipeline job
  case class ResetPipeline(pipelineId: String, resetPipelineInfo: ResetPipelineInfo) extends Command

  // Main message to create a savepoint for a pipeline job
  case class CreateSavepoint(pipelineId: String) extends Command
}

/**
  * An enumeration just to distinguish the two modes in which jobs are setup in pipeline model.
  * Either first time creation or a reset of the pipeline.
  */
object JobSetupContext extends Enumeration {
  type JobSetupContext = Value

  val CreatePipelineContext: JobSetupContext = Value("create")
  val ResetPipelineContext: JobSetupContext = Value("reset")
}

//noinspection SameParameterValue
class PipelineActor(commonSettings: CommonSettings,
                    managerSettings: PipelineJobManagerSettings,
                    jobManagerMicroServiceName: String) extends Actor
  with ActorLogging
  with FlinkJobUtils
  with PipelineJobUtils
  with Stash {

  import PipelineActor._

  // private settings fields for ease
  private val defaultKafkaSettings: KafkaSettings = managerSettings.kafkaSettings
  private val flinkSettings: FlinkSettings = managerSettings.flinkSettings

  // pipeline client creation
  private val pipelineClient = new PipelineClient(managerSettings.servicePipelinesUrl)

  override def receive: Receive = mainApiCallMode

  protected def takeSavepoint(pipelineId: String): Either[Error, SavePoint] = {

    // First check if the pipeline id is valid by loading the pipeline
    val pipelineTry = Try(getPipeline(pipelineId, showJobs = true))
    if (pipelineTry.isFailure) return buildErrorObjectForSavePoint(pipelineId, pipelineTry.failed.get, "input pipeline id not found", NotFound, jobManagerMicroServiceName)

    val pipeline = pipelineTry.get
    logger.info(s"Pipeline: $pipelineId - Pipeline found for the given id.")

    // Check and retrieve the base job configuration for the job
    val defaultJobConfigTry = checkAndGetDefaultJobConfig(pipeline, flinkSettings)

    if (defaultJobConfigTry.isFailure) {
      return buildErrorObjectForSavePoint(
        pipeline,
        new Exception(s"Pipeline: ${pipelineString(pipeline)} - Savepoint operation for pipeline failed. Reason: Failed to get default job configuration. Reason: ${defaultJobConfigTry.failed.get.getMessage}"),
        context = "retrieval of basic job configuration",
        InternalError,
        jobManagerMicroServiceName
      )
    }
    val defaultJobConfig = defaultJobConfigTry.get
    logger.info(s"Pipeline: ${pipelineString(pipeline)} - Retrieved the basic job configuration of the pipeline, jobName: ${defaultJobConfig.jobName}, jobSavePointEnabled: ${defaultJobConfig.jobSavePointEnabled}.")

    if (!defaultJobConfig.jobSavePointEnabled) {
      return buildErrorObjectForSavePoint(
        pipeline,
        new Exception(s"Pipeline: ${pipelineString(pipeline)} - Savepoint operation is not supported for stateless pipelines."),
        context = "savepoint not allowed for stateless pipelines",
        BadRequest,
        jobManagerMicroServiceName
      )
    }

    // Check if pipeline status is running or not
    if (pipeline.getPipelineStatus != PipelineStatus.RUNNING) {
      return buildErrorObjectForSavePoint(
        pipeline,
        new Exception(s"Pipeline: ${pipelineString(pipeline)} - Pipeline is in invalid status '${pipeline.getPipelineStatus}' to take savepoint."),
        context = "pipeline status is not running",
        InternalError,
        jobManagerMicroServiceName
      )
    }
    // Check if the job exists for the pipeline
    if (pipeline.getJobs.isEmpty) {
      return buildErrorObjectForSavePoint(
        pipeline,
        new Exception(s"Pipeline: ${pipelineString(pipeline)} - No job is associated with the pipeline."),
        context = "job does not exist for pipeline",
        NotFound,
        jobManagerMicroServiceName
      )
    }
    // Check if the job is in RUNNING status or not, savepoint can be taken only if job is running
    val pipelineJob = pipeline.getJobs.get(0)
    if (pipelineJob.getState != FlinkApiJobHandler.JobStateRunning) {
      return buildErrorObjectForSavePoint(
        pipeline,
        new Exception(s"Pipeline: ${pipelineString(pipeline)} - Job ${pipelineJob.getId} is in ${pipelineJob.getState} state."),
        context = "job state is not running",
        InternalError,
        jobManagerMicroServiceName
      )
    }
    // Get the flink cluster and checkpoint/savepoint directory
    val flinkCluster = createFlinkCluster(pipeline)
    val checkpointDirPathTry = getFlinkClusterCheckpointDirPath(flinkCluster)
    if (checkpointDirPathTry.isFailure) {
      return buildErrorObjectForSavePoint(
        pipeline,
        checkpointDirPathTry.failed.get,
        context = "retrieve jobmanager config for checkpoint dir path",
        InternalError,
        jobManagerMicroServiceName
      )
    }
    val checkpointDirPath = checkpointDirPathTry.get
    logger.info(s"Pipeline: $pipeline - Retrieved checkpoint dir path as $checkpointDirPath")

    // the callback handler for savepoint operations
    val savepointCallbackHandler = createSavepointCallbackHandler(pipeline)

    /*
     In order to make on-demand savepoint operation as async operation,
     it is moved to FlinkJobCreator and invoked here as TakeSavePoint
     */
    val flinkJobCreatorRef = createFlinkJobActor(
      flinkCluster,
      pipeline,
      savepointCallbackHandler,
      addUnique = true
    )

    logger.info(s"Pipeline: ${pipelineString(pipeline)} - Sending message to FlinkJobCreator to take a savepoint.")
    flinkJobCreatorRef ! TakeSavePoint(
      pipelineJob.getId,
      pipelineJob.getName,
      checkpointDirPath
    )
    /*
      takeSavePoint is a REST API call whose return value is either the savepoint object or the error response.
      Waiting for a maximum of 10 seconds here to get an appropriate response
     */
    Try { // Return value need not be checked as it is handled in the subsequent if/else
      GenericUtils.retry(10, "savepoint trigger", sleepInBetweenInMillis = 1000) {
        if (savepointCallbackHandler.getSavepointOption.isEmpty && savepointCallbackHandler.getTriggerFailureExceptionOption.isEmpty) {
          throw new Exception(s"Pipeline: ${pipelineString(pipeline)} - Failed to trigger savepoint on the pipeline.")
        }
      }
    }
    if (savepointCallbackHandler.getSavepointOption.isDefined) {
      Right(savepointCallbackHandler.getSavepointOption.get)

    } else if (savepointCallbackHandler.getTriggerFailureExceptionOption.isDefined) {
      buildErrorObjectForSavePoint(
        pipeline,
        savepointCallbackHandler.getTriggerFailureExceptionOption.get,
        context = "savepoint trigger operation",
        InternalError,
        jobManagerMicroServiceName
      )
    } else {
      buildErrorObjectForSavePoint(
        pipeline,
        new Exception("Did not get savepoint status on time"),
        context = "savepoint trigger operation",
        InternalError,
        jobManagerMicroServiceName
      )
    }
  }

  protected def getPipeline(pipelineId: String,
                            showJobs: Boolean = false): Pipeline = pipelineClient.getPipelineWithShowJobs(pipelineId, showJobs)

  // Create a Flink Job creator actor
  protected def createFlinkJobActor(flinkCluster: FlinkCluster,
                                    pipeline: Pipeline,
                                    savepointCallbackHandler: SavepointCallbackHandler,
                                    addUnique: Boolean = false): ActorRef = {

    val flinkJobCreatorActorName = if (addUnique) {
      // Add a random number since multiple actors might get created for taking savepoint
      val randomNumber = (new scala.util.Random).nextInt(100000)
      s"${FlinkJobCreator.ActorName}-${pipeline.getId}-$randomNumber"

    } else s"${FlinkJobCreator.ActorName}-${pipeline.getId}"

    logger.info(s"Pipeline: ${pipelineString(pipeline)} - Creating FlinkJobCreator actor with name as $flinkJobCreatorActorName.")
    context.actorOf(
      FlinkJobCreator.props(
        self,
        flinkCluster,
        commonSettings,
        managerSettings.flinkSettings,
        managerSettings.jobControlSettings,
        savepointCallbackHandler
      ),
      flinkJobCreatorActorName
    )
  }

  protected def createSavepointCallbackHandler(pipeline: Pipeline): PipelineSavepointCallbackHandler =
    new PipelineSavepointCallbackHandler(pipeline, pipelineClient)

  private def mainApiCallMode: Receive = LoggingReceive {
    // Create a new pipeline job
    case CreatePipeline(inputPipeline) =>
      val creationResponse = createPipeline(inputPipeline)
      creationResponse match {
        case Left(error: Error) => sender ! error
        case Right(pipeline: Pipeline) => sender ! pipeline
      }

    // Reset a  pipeline job
    case ResetPipeline(pipelineId, resetPipelineInfo) =>
      val resetResponse = resetPipeline(pipelineId, resetPipelineInfo)
      resetResponse match {
        case Left(error: Error) => sender ! error
        case Right(pipeline: Pipeline) => sender ! pipeline
      }

    // Take a savepoint
    case CreateSavepoint(pipelineId) =>
      val takeSavePointResponse = takeSavepoint(pipelineId)
      takeSavePointResponse match {
        case Left(error: Error) => sender ! error
        case Right(savepoint: SavePoint) => sender ! savepoint
      }

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor in mainApiCallMode received a message that is not handled: " + somethingElse)
  }

  private def waitForJobSetupMode(pipeline: Pipeline, flinkJobCreatorRef: ActorRef, jobSetupContext: JobSetupContext): Receive = LoggingReceive {

    case CreatePipeline(_) | ResetPipeline(_, _) | CreateSavepoint(_) =>
      stash()

    case ActionSucceeded(jobName) =>
      logger.info(s"Pipeline: ${pipelineString(pipeline)} - Successfully setup the pipeline job with jobName: $jobName.")

      // Updating the pipeline status to RUNNING, and also make it assignable if asked.
      pipeline.setPipelineStatus(PipelineStatus.RUNNING)
      updatePipelineProperties(pipelineClient, pipeline)

      jobSetupContext match {
        case JobSetupContext.CreatePipelineContext if pipeline.getAssignedTenants != null && !pipeline.getAssignedTenants.isEmpty =>
          /*
            Now that pipeline is up and running, assign all the tenants.
            This has to be done only in case the pipeline job is being setup.
           */
          assignTenants(pipelineClient, pipeline) match {
            case Success(_) => logger.info(s"Pipeline: ${pipelineString(pipeline)} -  Tenant assignment successful")
            case Failure(ex) => logger.error(s"Pipeline: ${pipelineString(pipeline)} - Failed to assign tenants. Error: ${ex.getMessage}")
          }

        case _ =>
        // Nothing to do here, tenant assignment is not done again.
        // Note: Pipeline object does not have the assigned tenant also as the GET done is not called with showTenants=true
      }

      // Unwatch and kill the flink job creator actor
      context unwatch flinkJobCreatorRef
      context stop flinkJobCreatorRef

      // If any messages to create/reset job had come, it is put back to mailbox now
      unstashAll()
      context.become(mainApiCallMode)

    case ActionFailed(jobName, error) =>
      logger.info(s"Pipeline: ${pipelineString(pipeline)} - Failed to setup the pipeline job with jobName: $jobName. Error: $error")

      // Updating the pipeline status to RESET_FAILED
      pipeline.setPipelineStatus(PipelineStatus.RESET_FAILED)

      jobSetupContext match {
        case JobSetupContext.CreatePipelineContext =>
          //Since job creation failed, mark the pipeline as un-assignable to avoid any tenant assignment to it.
          pipeline.setIsAssignable(false)

        case JobSetupContext.ResetPipelineContext =>
        // Nothing to do here. In case of RESET failure, need to step in and fix the Pipeline failure to get it back running.
      }

      updatePipelineProperties(pipelineClient, pipeline)

      // Unwatch and kill the flink job creator actor
      context unwatch flinkJobCreatorRef
      context stop flinkJobCreatorRef

      // If any messages to create/reset job had come, it is put back to mailbox now
      unstashAll()
      context.become(mainApiCallMode)

    case Terminated(flinkJobCreatorRef) =>
      log.error(s"FlinkJobCreator actor (${flinkJobCreatorRef.path.toString}) has terminated.")

      // Updating the pipeline status to RESET_FAILED
      pipeline.setPipelineStatus(PipelineStatus.RESET_FAILED)
      updatePipelineProperties(pipelineClient, pipeline)

      // If any messages to create/reset job had come, it is put back to mailbox now
      unstashAll()
      context.become(mainApiCallMode)

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor in waitForJobSetupMode received a message that is not handled: " + somethingElse)
  }

  // Actual method to start pipeline creation
  private def createPipeline(inputPipeline: Pipeline): Either[Error, Pipeline] = {

    // Check basic parameters sent in the pipeline reference
    val basicCheckTry = checkBasicPipelineParameters(inputPipeline)
    if (basicCheckTry.isFailure) return buildErrorObject(inputPipeline, basicCheckTry.failed.get, "basic validation error", BadRequest, jobManagerMicroServiceName)
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Basic validation completed")

    // Check and retrieve the base job configuration for the job
    val defaultJobConfigTry = checkAndGetDefaultJobConfig(inputPipeline, flinkSettings)
    if (defaultJobConfigTry.isFailure) return buildErrorObject(inputPipeline, defaultJobConfigTry.failed.get, "retrieval of basic job configuration", InternalError, jobManagerMicroServiceName)
    val defaultJobConfig = defaultJobConfigTry.get
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Retrieved the basic job configuration of the pipeline.")

    // Ensure topic owner settings are fine
    val topicOwnerCheckTry = validateNonTopicOwnerConfig(inputPipeline, defaultJobConfig, defaultKafkaSettings)
    if (topicOwnerCheckTry.isFailure) return buildErrorObject(inputPipeline, topicOwnerCheckTry.failed.get, "topic owner checks", BadRequest, jobManagerMicroServiceName)
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Checked topic owner based settings of the pipeline.")

    // Check and retrieve the default topic configuration meant for this job
    val defaultTopicConfigTry = checkAndGetDefaultTopicConfig(defaultJobConfig, defaultKafkaSettings)
    if (defaultTopicConfigTry.isFailure) return buildErrorObject(inputPipeline, defaultTopicConfigTry.failed.get, "retrieval of default topic configuration", InternalError, jobManagerMicroServiceName)
    val defaultTopicConfig = defaultTopicConfigTry.get
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Retrieved the topic(s) configuration of the pipeline.")

    // Ensure jar is available for the job
    val jobJarFileTry: Try[File] = loadJarFile(defaultJobConfig, flinkSettings.jobJarDirectory)
    if (jobJarFileTry.isFailure) return buildErrorObject(inputPipeline, jobJarFileTry.failed.get, "load jar file for pipeline", InternalError, jobManagerMicroServiceName)
    val jobJarFile: File = jobJarFileTry.get
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Jar file ${jobJarFile.getName} available for the pipeline.")

    /*
      Now all basic validations that do not change anything in the system are done.
      Next, changes like pipeline creation in database by service-pipelines, topic creation, job creation is done.
     */

    // Construct the pipeline to be used in body in POST request to service-pipelines
    val pipelineRequestTry = buildPipelineObjectForCreation(
      inputPipeline,
      defaultJobConfig,
      defaultTopicConfig
    )
    if (pipelineRequestTry.isFailure) return buildErrorObject(inputPipeline, pipelineRequestTry.failed.get, "create updated pipeline object", InternalError, jobManagerMicroServiceName)
    val pipelineRequest = pipelineRequestTry.get
    logger.info(s"Pipeline: $pipelineRequestTry - Pipeline object created to send to service-pipelines.")

    // Invoke the API to create pipeline on service-pipelines.
    val pipelineResponseTry = createPipeline(pipelineClient, pipelineRequest)
    if (pipelineResponseTry.isFailure) return buildErrorObject(inputPipeline, pipelineResponseTry.failed.get, "create call to service-pipelines", InternalError, jobManagerMicroServiceName)
    val pipelineResponse = pipelineResponseTry.get
    logger.info(s"Pipeline: $pipelineResponse - Pipeline object created by service-pipelines.")

    // Call the internal setup method
    pipelineSetupInternal(
      JobSetupContext.CreatePipelineContext,
      pipelineResponse,
      defaultTopicConfig,
      defaultJobConfig,
      jobJarFile,
      inputPipelineForCreateOption = Some(inputPipeline)
    )
  }

  // Actual method to reset pipeline
  private def resetPipeline(pipelineId: String, resetPipelineInfo: ResetPipelineInfo): Either[Error, Pipeline] = {

    // First check if the pipeline id is valid by loading the pipeline
    val pipelineTry = Try(getPipeline(pipelineId, showJobs = true))
    if (pipelineTry.isFailure) return buildErrorObject(pipelineId, pipelineTry.failed.get, "input pipeline id not found", NotFound, jobManagerMicroServiceName)
    val inputPipeline = pipelineTry.get
    logger.info(s"Pipeline: $pipelineId - Pipeline found for the given id.")

    // Load the default job config of the pipeline type
    // Check and retrieve the base job configuration for the job
    val defaultJobConfigTry = checkAndGetDefaultJobConfig(inputPipeline, flinkSettings)
    if (defaultJobConfigTry.isFailure) return buildErrorObject(inputPipeline, defaultJobConfigTry.failed.get, "retrieval of basic job configuration", InternalError, jobManagerMicroServiceName)
    val defaultJobConfig = defaultJobConfigTry.get
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Retrieved the basic job configuration of the pipeline.")

    // Basic checks of reset parameters
    val savepointForResetRequestTry = checkAndGetSavePointForResetRequest(inputPipeline, pipelineClient, resetPipelineInfo, defaultJobConfig)
    if (savepointForResetRequestTry.isFailure) return buildErrorObject(inputPipeline, savepointForResetRequestTry.failed.get, "basic validation error for reset request", BadRequest, jobManagerMicroServiceName)
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Basic validation of reset request completed.")

    // Check and retrieve the default topic configuration meant for this job
    val defaultTopicConfigTry = checkAndGetDefaultTopicConfig(defaultJobConfig, defaultKafkaSettings)
    if (defaultTopicConfigTry.isFailure) return buildErrorObject(inputPipeline, defaultTopicConfigTry.failed.get, "retrieval of default topic configuration", InternalError, jobManagerMicroServiceName)
    val defaultTopicConfig = defaultTopicConfigTry.get
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Retrieved the topic configuration of the pipeline.")

    // Ensure jar is available for the job
    val jobJarFileTry: Try[File] = loadJarFile(defaultJobConfig, flinkSettings.jobJarDirectory)
    if (jobJarFileTry.isFailure) return buildErrorObject(inputPipeline, jobJarFileTry.failed.get, "load jar file for pipeline", InternalError, jobManagerMicroServiceName)
    val jobJarFile: File = jobJarFileTry.get
    logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Jar file ${jobJarFile.getName} available for the pipeline.")

    // Update the pipeline object with values that need to be copied from the ResetPipelineInfo changes
    if (resetPipelineInfo.getParallelism != null) inputPipeline.setParallelism(resetPipelineInfo.getParallelism)
    if (resetPipelineInfo.getAdditionalJobParameters != null) inputPipeline.setAdditionalJobParameters(resetPipelineInfo.getAdditionalJobParameters)

    // Call the internal setup method
    pipelineSetupInternal(
      JobSetupContext.ResetPipelineContext,
      inputPipeline,
      defaultTopicConfig,
      defaultJobConfig,
      jobJarFile,
      savepointForResetRequestTry.get.externalSavepointOption,
      savepointForResetRequestTry.get.lastSavepointOption,
      resetPipelineInfo.isAllowNonRestoredState
    )
  }

  /**
    * Common internal method used by creation and reset of a pipeline
    */
  private def pipelineSetupInternal(jobSetupContext: JobSetupContext,
                                    pipeline: Pipeline,
                                    defaultTopicConfig: Map[PipelineTopicType, ProducerTopic],
                                    defaultJobConfig: FlinkJobConfig,
                                    jobJarFile: File,
                                    externalSavepointPathOption: Option[String] = None,
                                    lastSavepointOption: Option[CheckOrSavePointWithCreateTime] = None,
                                    allowNonRestoredState: Boolean = false,
                                    inputPipelineForCreateOption: Option[Pipeline] = None): Either[Error, Pipeline] = {
    /*
     Topic creation in 99% of the cases will be done in create_pipeline flow.
     In case the topic is deleted and the job failed and a reset is being called, this will ensure
     the topic is re-created and also checks the configuration parameters of the topic like partitions etc.
     */
    val topicCreationResponseTry = createTopicForPipeline(
      pipeline,
      defaultTopicConfig,
      defaultKafkaSettings
    )

    if (topicCreationResponseTry.isFailure) {
      if (jobSetupContext == JobSetupContext.CreatePipelineContext) {
        // Updating the pipeline status to RESET_FAILED to have consistency in status of failed pipelines.
        pipeline.setPipelineStatus(PipelineStatus.RESET_FAILED)
        updatePipelineProperties(pipelineClient, pipeline)


        // Delete the pipeline since topic creation failed., ignore errors are continue
        deletePipeline(pipelineClient, pipeline)
      }
      return buildErrorObject(pipeline, topicCreationResponseTry.failed.get, s"create topic ${pipeline.getTopicName}", ErrorCode.InternalError, jobManagerMicroServiceName)
    }

    if (pipeline.getTopicNames != null && !pipeline.getTopicNames.isEmpty) {
      logger.info(s"Pipeline: ${pipelineString(pipeline)} - Pipeline topic(s) is ${pipeline.getTopicNames.values()}.")
    } else {
      logger.info(s"Pipeline: ${pipelineString(pipeline)} - Pipeline topic is ${pipeline.getTopicName}.")
    }

    /*
     Get the checkpoints.dir.name irrespective of job type as of now
     Reason: Allows to take savepoints of jobs if needed on stateless jobs too.
     */
    val flinkCluster = createFlinkCluster(pipeline)
    val checkpointDirPathTry = getFlinkClusterCheckpointDirPath(flinkCluster)
    if (checkpointDirPathTry.isFailure) return buildErrorObject(pipeline, checkpointDirPathTry.failed.get, "retrieve jobmanager config for checkpoint dir path", InternalError, jobManagerMicroServiceName)
    val checkpointDirPath = checkpointDirPathTry.get
    logger.info(s"Pipeline: ${pipelineString(pipeline)} - Retrieved checkpoint dir path as $checkpointDirPath")

    // Create the actual job config meant for the job
    val actualJobConfig = buildPipelineJobConfig(
      pipeline,
      defaultJobConfig,
      defaultKafkaSettings,
      managerSettings.servicePipelinesUrl,
      managerSettings.vaultServiceName,
      managerSettings.serviceTenantsUrl,
      managerSettings.serviceConfigMgmtUrl,
      managerSettings.serviceDbsUrl,
      checkpointDirPath,
      externalSavepointPathOption
    )
    logger.info(s"Pipeline: ${pipelineString(pipeline)} - Actual job configuration for the pipeline job created - $actualJobConfig.")

    /*
     Setup the job in flink cluster. This is done using FlinkJobCreator
     actor so that it is done asynchronously.
     The pipelineResponse got from service-pipelines is returned as return value
     */
    val flinkJobCreatorRef = createFlinkJobActor(
      flinkCluster,
      pipeline,
      new PipelineSavepointCallbackHandler(
        pipeline,
        pipelineClient
      )
    )

    /*
     Change status of the pipeline to resetting and update database
     In case of reset, the inputPipeline is already updated with information that
     came from ResetPipelineInfo
     */
    pipeline.setPipelineStatus(PipelineStatus.RESETTING)
    updatePipelineProperties(pipelineClient, pipeline)

    // Changes to be done post creation i.e once it is marked running only.
    if (jobSetupContext == JobSetupContext.CreatePipelineContext) {
      //Since service-pipelines does not assign the tenants on pipeline creation, copy the assigned tenants from input pipeline
      //to perform assignment once pipeline is running
      pipeline.setAssignedTenants(inputPipelineForCreateOption.get.getAssignedTenants)

      //Make the pipeline open for assignment, once it is up and running.
      pipeline.setIsAssignable(inputPipelineForCreateOption.get.getIsAssignable)
    }

    logger.info(s"Pipeline: ${pipelineString(pipeline)} - Sending message to FlinkJobCreator to setup the job.")
    flinkJobCreatorRef ! SetupFlinkJob(actualJobConfig, Some(jobJarFile), lastSavepointOption, allowNonRestoredState)

    // Watch the flink job creator ref
    context watch flinkJobCreatorRef

    // Change the context so as to wait for the job setup to be completed.
    context.become(waitForJobSetupMode(pipeline, flinkJobCreatorRef, jobSetupContext))

    // Return value
    Right(pipeline)
  }
}