/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.pipeline.job

import akka.actor.{Actor, ActorLogging, ActorRef, Props, Terminated, Timers}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.flink.actor.FlinkInfraChecker
import com.netapp.df.uba.server.common.flink.actor.FlinkInfraChecker.ContextFlinkInfraCheck
import com.netapp.df.uba.server.common.flink.settings.FlinkSettings
import com.netapp.df.uba.server.common.kafka.KafkaInfraChecker._
import com.netapp.df.uba.server.common.kafka.{KafkaInfraChecker, KafkaSettings}
import com.netapp.df.uba.server.common.misc.model._
import com.netapp.df.uba.server.common.misc.utils.ActorPathUtils

import scala.concurrent.duration.{FiniteDuration, _}

/**
  * PipelineInfraChecker handles the co-ordination of checks against infrastructure components this service depends on.
  * - Kafka checks
  * - Flink checks based on checkFlinkInfra boolean
  *
  * Handles the repeated checks for these components separated by an interval
  */
object PipelineInfraChecker {
  final val ActorName: String = "pipelineInfraChecker"
  def props(requestorRef: ActorRef,
            kafkaSettings: KafkaSettings,
            flinkSettings: FlinkSettings,
            checkFlinkInfra: Boolean): Props = Props(new PipelineInfraChecker(requestorRef, kafkaSettings, flinkSettings, checkFlinkInfra))

  // Overall context used for infrastructure checks for action succeeded and action failed messages
  final val ContextPipelineInfraCheck: String = "pipelineInfraCheck"

  // A utility case class
  case class PipelineInfraCheckContext(attemptsAllowed: Int,
                                       attemptNumber: Int,
                                       attemptInterval: FiniteDuration,
                                       checkerRef: Option[ActorRef])

  // Timer key
  case class PipelineInfraCheckerTimerKey(infraContext: String)

  // Internal message used to handle multiple infrastructure checks
  case class CheckInfraComponent(infraContext: String, attemptNumber: Int)
}

class PipelineInfraChecker(requestorRef: ActorRef,
                           kafkaSettings: KafkaSettings,
                           flinkSettings: FlinkSettings,
                           checkFlinkInfra: Boolean) extends Actor with ActorLogging with Timers {

  import PipelineInfraChecker._

  // Currently, attempts and retry interval between attempts for infra checks is not exposed, so hard-coding the same.
  // This class has the ability to manage the same.
  protected val checkAttempts = 1
  protected val attemptInterval: FiniteDuration = 5.seconds

  // Initial receive method
  override def receive: Receive = initialMode

  def initialMode: Receive = LoggingReceive {

    // Since there are repeated attempts, it just sends a message as if it came from timer also with the initial attempt number as 1
    case CheckInfra =>
      self ! CheckInfraComponent(ContextKafkaInfraCheck, 1)

      //check if at least 1 flink job is enabled, and accordingly decide whether flink infra check is needed
      if (checkFlinkInfra) {
        self ! CheckInfraComponent(ContextFlinkInfraCheck, 1)
        context.become(checkMode(
          Map.empty,
          Map.empty,
          remainingCount = 2
        ))
      } else {
        context.become(checkMode(
          Map.empty,
          Map.empty,
          remainingCount = 1
        ))
      }

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message in initialMode that is not handled: " + somethingElse)
  }

  // Receive method that handles the infrastructure checks
  private def checkMode(infraCheckContextMap: Map[String, PipelineInfraCheckContext],
                        infraCheckResultMap: Map[String, ActionStatus],
                        remainingCount: Int): Receive = LoggingReceive {

    // A message that is for checking a specific infrastructure component
    case CheckInfraComponent(infraContext, attemptNumber) =>
      val infraCheckContext = infraContext match {
        case ContextKafkaInfraCheck =>

          val kafkaInfraChecker: ActorRef = createKafkaInfraCheckerActor(attemptNumber)
          context watch kafkaInfraChecker

          PipelineInfraCheckContext(
            checkAttempts,
            attemptNumber,
            attemptInterval,
            Some(kafkaInfraChecker))

        case ContextFlinkInfraCheck =>

          val flinkInfraChecker: ActorRef = createFlinkInfraCheckerActor(attemptNumber)
          context watch flinkInfraChecker

          PipelineInfraCheckContext(
            checkAttempts,
            attemptNumber,
            attemptInterval,
            Some(flinkInfraChecker))
      }
      val newInfraCheckContextMap = infraCheckContextMap ++ Seq(
        infraContext -> infraCheckContext
      )

      // Ensure the new context is added
      context.become(checkMode(newInfraCheckContextMap,
        infraCheckResultMap,
        remainingCount))

    // Received a result, update the maps accordingly
    case actionStatus: ActionStatus =>
      log.info(s"Infrastructure check result for ${actionStatus.context} - $actionStatus.")

      evaluateResult(actionStatus.context,
        infraCheckContextMap,
        infraCheckResultMap,
        actionStatus,
        remainingCount)

    // Some actor terminated, handle this as action status failed
    case Terminated(infraCheckerChild) =>
      log.error(s"Infra Checker actor (${infraCheckerChild.path.toString}) has terminated.")

      val infraContext = ActorPathUtils.prefixNameFromPath(infraCheckerChild)
      val errorMessage = s"InfraCheck actor (${infraCheckerChild.path.toString}) has terminated."
      log.error(errorMessage)

      // Create the actionStatus for this scenario
      val actionFailed = ActionFailed(infraContext, Error(ErrorCode.InternalError, errorMessage, infraContext))

      // Evaluate the result and decide whether to remain in context or switch
      evaluateResult(
        infraContext,
        infraCheckContextMap,
        infraCheckResultMap,
        actionFailed,
        remainingCount)

    case somethingElse =>
      log.warning(s"${this.getClass.getSimpleName} actor received a message in checkMode that is not handled: " + somethingElse)
  }

  // Utility method to evaluate result and decide whether to continue or not
  private def evaluateResult(infraContext: String,
                             infraCheckContextMap: Map[String, PipelineInfraCheckContext],
                             infraCheckResultMap: Map[String, ActionStatus],
                             actionStatus: ActionStatus,
                             remainingCount: Int): Unit = {

    // Add the current result to the infraCheckResultMap map with key as context name
    val newInfraCheckResultMap = infraCheckResultMap ++ Seq(infraContext -> actionStatus)

    // get the context of the infra check
    val currentCheckContext = infraCheckContextMap(infraContext)

    // Attempt configuration
    val attemptsAllowed = currentCheckContext.attemptsAllowed
    val attemptNumber = currentCheckContext.attemptNumber
    val checkInterval = currentCheckContext.attemptInterval
    val remainingAttempts = attemptsAllowed - attemptNumber

    // Stop watching the infra actor and stop it
    context unwatch currentCheckContext.checkerRef.get
    context stop currentCheckContext.checkerRef.get

    // Check different unique conditions that needs different handling
    (remainingCount - 1, actionStatus, remainingAttempts) match {
      case (0, ActionSucceeded(_), _) =>
        // remaining count is 0 and the current action has succeeded. All done
        createResponseAction(newInfraCheckResultMap)

      case (0, _, 0) =>
        // Scenario where all infra services have been checked (first 0), status does not matter, remainingAttempts is 0 as this was the last one to be checked
        // Return result to requestorRef
        createResponseAction(newInfraCheckResultMap)

      case (_, ActionFailed(_, _), rem) if rem > 0=>
        // Scenario where there is a failure and there are still remaining attempts for that service
        // Start the timer and send a self message to create an actor for CheckInfraComponent for that service, also increase the attempt number
        timers.startSingleTimer(PipelineInfraCheckerTimerKey(infraContext), CheckInfraComponent(infraContext, attemptNumber + 1), checkInterval)
      // No change in context as the values continue, result is discarded in this case as in case of failure only the last one matters
      // No change in remaining count as there is more attempts remaining here

      case (_, _, _) =>
        // Change mode to wait for more action statuses, change the remaining count
        context.become(checkMode(
          infraCheckContextMap,
          newInfraCheckResultMap,
          remainingCount - 1))
    }
  }

  // Utility method to create a response action to the parent
  protected def createResponseAction(infraCheckResultMap: Map[String, ActionStatus]): Unit = {

    // Check if infra checks succeeded for all
    if (infraCheckResultMap.values.forall(_.isInstanceOf[ActionSucceeded])) {

      // Indicate success of creation of all indices to the parent
      requestorRef ! ActionSucceeded(ContextPipelineInfraCheck)
      context.become(initialMode)

    } else {

      // Get the list of all errors in infra checks
      val allErrors = infraCheckResultMap
        .values
        .filterNot(_.isInstanceOf[ActionSucceeded])
        .map(_.asInstanceOf[ActionFailed].error)
        .toList

      // Create an error object
      val errorObject = Error(ErrorCode.InternalError, "Infra check failed.", ContextPipelineInfraCheck, Some(allErrors))

      // Indicate to parent about failure and go back to initialMode
      requestorRef ! ActionFailed(ContextPipelineInfraCheck, errorObject)
      context.become(initialMode)
    }
  }

  // Separate methods to create other actors to allow for overriding in unit tests
  protected def createKafkaInfraCheckerActor(attemptNumber: Int): ActorRef =
    context.actorOf(KafkaInfraChecker.props(kafkaSettings, self), s"$ContextKafkaInfraCheck-$attemptNumber")

  protected def createFlinkInfraCheckerActor(attemptNumber: Int): ActorRef =
    context.actorOf(FlinkInfraChecker.props(flinkSettings, self), s"$ContextFlinkInfraCheck-$attemptNumber")
}
