/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.pipeline.job

import akka.actor.{Actor, ActorLogging, ActorRef, Props}
import akka.event.LoggingReceive
import com.netapp.df.uba.server.common.misc.actor.StateManager
import com.netapp.df.uba.server.common.misc.model.ServiceName.ServiceName
import com.netapp.df.uba.server.common.misc.model._
import com.netapp.df.uba.server.common.misc.utils._
import com.netapp.df.uba.server.common.pipeline.data.{Pipeline, SavePoint}
import com.netapp.df.uba.server.common.pipeline.job.PipelineActor.{CreatePipeline, CreateSavepoint, ResetPipeline}
import PipelineJobManagerRbKeys._

import java.util.ResourceBundle

/**
  * PipelineJobManagerApiHandler is the actor that is used as an API handler
  * in any of the job manager micro-services.
  * It is instantiated for every API call received by the the job manager
  */
object PipelineJobManagerApiHandler {
  def props(serviceNameOfJobManager: ServiceName): Props = Props(new PipelineJobManagerApiHandler(serviceNameOfJobManager))
}

class PipelineJobManagerApiHandler(serviceNameOfJobManager: ServiceName)
  extends Actor
    with ActorLogging
    with PipelineJobManagerResourceBundleUser
    with ExecContextAndTimeOutImplicits
    with AppActorUser {

  import context._

  // Initially receive the request from the route handlers
  override def receive: Receive = routeRequestReceive.orElse(timedOutReceive)

  // Request from route handlers
  private def routeRequestReceive: Receive = LoggingReceive {
    case createPipelineJobMessage: CreatePipeline =>
      // Set an overall receive timeout so that the actor is stopped on its own in case there is no response from the underlying application actors
      setReceiveTimeout(timeout.duration)

      // Check state of the service before sending the payload to the backend actor
      askStateManager(
        sender(),
        getPipelineActorPath,
        createPipelineJobMessage,
        pipelineManagerResultMode(sender()),
        apiContext = "Create Pipeline"
      )
    case resetPipelineJobMessage: ResetPipeline =>
      // Set an overall receive timeout so that the actor is stopped on its own in case there is no response from the underlying application actors
      setReceiveTimeout(timeout.duration)

      // Check state of the service before sending the payload to the backend actor
      askStateManager(
        sender(),
        getPipelineActorPath,
        resetPipelineJobMessage,
        pipelineManagerResultMode(sender()),
        apiContext = "Reset Pipeline"
      )
    case createSavepointMessage: CreateSavepoint =>
      // Set an overall receive timeout so that the actor is stopped on its own in case there is no response from the underlying application actors
      setReceiveTimeout(timeout.duration)

      // Check state of the service before sending the payload to the backend actor
      askStateManager(
        sender(),
        getPipelineActorPath,
        createSavepointMessage,
        pipelineManagerResultMode(sender()),
        apiContext = "Take Savepoint"
      )
  }

  // Check on state of all services first
  private def askStateManager(routeHandler: ActorRef,
                              backendManagerPath: String,
                              backendMessagePayload: Any,
                              backendResponseReceiveMethod: Receive,
                              apiContext: String): Unit = {
    // Send a message to StateManager to get hold of the service state
    askDataFromActor(StateManager.ActorPath, GetStateDetails)

    context.become(
      stateResponseReceive(
        routeHandler,
        backendManagerPath,
        backendMessagePayload,
        backendResponseReceiveMethod,
        apiContext
      ).orElse(timedOutReceive)
    )
  }

  // Message received from state manager
  private def stateResponseReceive(routeHandler: ActorRef,
                                   backendActorPath: String,
                                   backendPayload: Any,
                                   backendResponseReceiveMethod: Receive,
                                   apiContext: String): Receive = LoggingReceive {
    // Response from StateManager
    case stateDetails: StateDetails =>
      // Check if metadata service is running and set the flag
      stateDetails.serviceStatusMap(serviceNameOfJobManager) match {
        case Running =>
          // Send a message to backend actor to create/reset pipeline
          askDataFromActor(backendActorPath, backendPayload)

          // Change the receive method to such that it receives responses from backend actor
          context.become(backendResponseReceiveMethod.orElse(timedOutReceive))

        case actualStatus =>
          // Send error back to caller
          val errorObject: Error = serviceStatusToError(serviceNameOfJobManager, actualStatus, resourceBundle)
          log.error(s"API Context: $apiContext - Returning error $errorObject for API call.")
          sendResponse(routeHandler, errorObject)
      }
  }

  // Handles the response from pipelineActor
  def pipelineManagerResultMode(routeHandler: ActorRef): Receive = LoggingReceive {

    case resultPipeline: Pipeline =>
      log.info(s"Received result from PipelineActor for pipeline of type: ${resultPipeline.getPipelineType}.")
      sendResponse(routeHandler, resultPipeline)

    case savepoint: SavePoint =>
      log.info(s"Received result from PipelineActor for savepoint creation with id: ${savepoint.getId}, location: ${savepoint.getLocation}.")
      sendResponse(routeHandler, savepoint)

    case error: Error =>
      log.error(s"Received error from PipelineActor - $error.")
      sendResponse(routeHandler, error)
  }

  /**
    * Utility method to convert a service status to Error object
    */
  def serviceStatusToError(serviceName: ServiceName, serviceStatus: ServiceStatus, resourceBundle: ResourceBundle): Error = {
    serviceStatus match {
      case Initializing | Stopping | Stopped =>
        Error(ErrorCode.ServiceNotRunning, RbUtils.rbMessage(resourceBundle, RBKEY_API_SERVICE_NOT_RUNNING, serviceName.toString, serviceStatus.statusString), serviceName.toString)

      case Failed(reason, errorCode) =>
        Error(errorCode, RbUtils.rbMessage(resourceBundle, RBKEY_API_SERVICE_FAILED, serviceName.toString, reason), serviceName.toString)

      case NotEnabled =>
        Error(ErrorCode.ServiceDisabled, RbUtils.rbMessage(resourceBundle, RBKEY_API_SERVICE_DISABLED, serviceName.toString), serviceName.toString)

      case Running =>
        // Will actually never reach here. Just to remove the warning
        Error(ErrorCode.NoError, "", "")
    }
  }

  // Code refactored as a method so that unit tests can override it
  protected def getPipelineActorPath: String = PipelineActor.ActorPath
}