/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.pipeline.job

import akka.actor.ActorRef
import akka.http.scaladsl.model.StatusCodes
import akka.http.scaladsl.server.{Directive1, Route}
import akka.http.scaladsl.unmarshalling.FromRequestUnmarshaller
import akka.pattern.ask
import akka.util.Timeout
import com.netapp.df.uba.server.common.misc.actor.StateApiHandler.GetState
import com.netapp.df.uba.server.common.misc.actor.{StateApiHandler, StateManager}
import com.netapp.df.uba.server.common.misc.api.{AuthDirectives, BaseRoute, ErrorUtils}
import com.netapp.df.uba.server.common.misc.apikey.ApiKey
import com.netapp.df.uba.server.common.misc.model.ServiceName.ServiceName
import com.netapp.df.uba.server.common.misc.model.{ComponentStateApi, Error, JwtTokenContext, Role, TenantContext}
import com.netapp.df.uba.server.common.pipeline.data.{Pipeline, ResetPipelineInfo, SavePoint}
import com.netapp.df.uba.server.common.pipeline.job.PipelineActor.{CreatePipeline, CreateSavepoint, ResetPipeline}

import scala.concurrent.duration._

/**
  * Main set of routes of the micro-service
  */
trait PipelineJobManagerRoute
  extends BaseRoute
    with PipelineJobManagerRouteHandlerImplicits
    with PipelineJsonFormats
    with AuthDirectives {

  implicit val timeout1: Timeout = 60.seconds

  val jobManagerRoutes: Route =
    handleRejections(customApiRejectionHandler) {
      pathPrefix(basePath / jobManagerRouteName()) {
        authenticateAndExtractTenant(envName(), allowedRoles = Set(Role.SystemAdmin)) { (_: TenantContext, _: JwtTokenContext, _: Option[ApiKey]) =>
          pathPrefix("state") {
            Route.seal {
              pathEndOrSingleSlash {
                get {
                  // Spawn and ask the StateApiHandler for the current state of the component
                  val statusActorRef = createStateApiHandlerActorRef()
                  onSuccess(statusActorRef ? GetState) {
                    case response: ComponentStateApi => complete(response)
                    // error cases
                    case errorObject: Error =>
                      complete(ErrorUtils.errorCodeToStatusCodes(errorObject.code) -> errorObject)
                    // Debug this
                    case somethingElse =>
                      complete("Something else came" + somethingElse)
                  }
                }
              }
            }
          } ~
            pathPrefix("pipelines") {
              Route.seal {
                pathEndOrSingleSlash {
                  post {
                    entity(as[Pipeline]) {
                      inputPipeline: Pipeline =>

                        // Pass the pipeline to PipelineJobManagerApiHandler actor
                        onSuccess(createPipelineJobManagerApiHandlerActorRef(jobManagerServiceName()) ? CreatePipeline(inputPipeline)) {

                          // error cases
                          case errorObject: Error =>
                            complete(ErrorUtils.errorCodeToStatusCodes(errorObject.code) -> errorObject)

                          case pipelineResponse: Pipeline =>
                            complete(StatusCodes.Accepted -> pipelineResponse)

                          // Debug this
                          case somethingElse =>
                            complete("Something else came" + somethingElse)
                        }
                    }
                  }
                } ~
                  path(Segment / "reset") { pipelineId =>
                    post {
                      optionalEntity(as[ResetPipelineInfo]) {
                        resetPipelineInfoOption: Option[ResetPipelineInfo] =>

                          val resetPipelineInfo = if (resetPipelineInfoOption.isEmpty) new ResetPipelineInfo()
                          else resetPipelineInfoOption.get

                          // Pass the reset pipeline info to PipelineJobManagerApiHandler actor
                          onSuccess(createPipelineJobManagerApiHandlerActorRef(jobManagerServiceName()) ? ResetPipeline(pipelineId, resetPipelineInfo)) {

                            // error cases
                            case errorObject: Error =>
                              complete(ErrorUtils.errorCodeToStatusCodes(errorObject.code) -> errorObject)

                            case pipelineResponse: Pipeline =>
                              complete(StatusCodes.Accepted -> pipelineResponse)

                            // Debug this
                            case somethingElse =>
                              complete("Something else came" + somethingElse)
                          }
                      }
                    }
                  } ~
                  path(Segment / "savepoints") { pipelineId =>
                    post {
                      onSuccess(createPipelineJobManagerApiHandlerActorRef(jobManagerServiceName()) ? CreateSavepoint(pipelineId)) {
                        // error cases
                        case errorObject: Error =>
                          complete(ErrorUtils.errorCodeToStatusCodes(errorObject.code) -> errorObject)

                        case savepoint: SavePoint =>
                          complete(StatusCodes.Accepted -> savepoint)

                        // Debug this
                        case somethingElse =>
                          complete("Something else came" + somethingElse)
                      }
                    }
                  }
              }
            }
        }
      }
    }

  /**
    * A custom directive to handle the optional reset body to be recognized as Option[ResetPipelineInfo]
    * https://github.com/akka/akka-http/issues/284
    */
  def optionalEntity[T](implicit um: FromRequestUnmarshaller[T]): Directive1[Option[T]] =
    entity(as[String]).flatMap { stringBody =>
      if (stringBody == null || stringBody.isEmpty) provide(Option.empty[T])
      else entity(um).flatMap(obj => provide(Some(obj)))
    }

  // Code to create child actors so that unit tests can override it
  protected def createStateApiHandlerActorRef(): ActorRef =
    system.actorOf(StateApiHandler.props(jobManagerServiceName().toString, StateManager.ActorPath))

  protected def createPipelineJobManagerApiHandlerActorRef(serviceNameOfJobManager: ServiceName): ActorRef = {
    system.actorOf(PipelineJobManagerApiHandler.props(serviceNameOfJobManager))
  }
}