/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.pipeline.job

import akka.actor.ActorSystem
import akka.http.scaladsl.model.StatusCodes._
import akka.http.scaladsl.server.Directives.{complete, extractUri}
import akka.http.scaladsl.server.{ExceptionHandler, MethodRejection, RejectionHandler, ValidationRejection}
import akka.pattern.AskTimeoutException
import akka.util.Timeout
import com.netapp.df.uba.server.common.misc.api.JsonFormats
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model.{Error, ErrorCode}
import com.netapp.df.uba.server.common.misc.utils.RbUtils
import PipelineJobManagerRbKeys._
import com.netapp.df.uba.server.common.misc.model.ServiceName.ServiceName

import java.util.ResourceBundle
import scala.concurrent.ExecutionContext
import scala.concurrent.duration._

/**
  * A set of implicits and definitions required by ApiServer and route implementors.
  */
trait PipelineJobManagerRouteHandlerImplicits extends JsonFormats with LogTrait {
  // ActorSystem
  implicit val system: ActorSystem

  // Execution context implicit using actor system dispatcher
  implicit val executionContext: ExecutionContext = system.dispatcher

  // Implicit definition of timeout required in ask pattern calls
  implicit val timeout: Timeout = 60.seconds

  // Resource bundle akka extension
  val rbBundle = PipelineJobManagerResBundle(system)

  // Actual resource bundle
  implicit val resourceBundle: ResourceBundle = rbBundle.resourceBundle

  /**
    * Concrete classes to provide the micro-service name that acts as
    * the job manager for a group of related jobs.
    * This is used as a prefix to all HTTP REST API routes.
    */
  protected def jobManagerRouteName(): String

  /**
    * Concrete classes to provide the micro-service name
    */
  protected def jobManagerServiceName(): ServiceName

  /**
    * Override to provide the environment in which the micro-service is running
    * Generally derived from the settings
    */
  protected def envName(): String

  /**
    * Custom exception handler used for APIs
    * Done only to ensure the message is got from the resource key.
    */
  implicit def componentApiExceptionHandler: ExceptionHandler =
    ExceptionHandler {
      case ex: AskTimeoutException =>
        extractUri { uri =>
          if (logger.isDebugEnabled) ex.printStackTrace()
          logger.error(s"Request to $uri could not be handled in time. Exception: ${ex.getClass.getSimpleName} Message: ${ex.getMessage}")
          complete(ServiceUnavailable -> Error(ErrorCode.InternalError, RbUtils.rbMessage(resourceBundle, RBKEY_API_INTERNAL_ERROR_TIMEOUT), jobManagerServiceName().toString))
        }

      case ex: Exception =>
        extractUri { uri =>
          if (logger.isDebugEnabled) ex.printStackTrace()
          logger.error(s"Internal error in handling the request $uri. Reason: ${ex.getMessage}")
          complete(InternalServerError -> Error(ErrorCode.InternalError, RbUtils.rbMessage(resourceBundle, RBKEY_API_INTERNAL_ERROR_REASON, ex.getMessage), jobManagerServiceName().toString))
        }
    }

  /**
    * Custom rejection handler used for APIs
    * Default exception handler provided by Akka HTTP is sophisticated enough and does not ideally need much modifications.
    * Done here just to handle some common custom scenarios and also to get messages from the resource bundle
    */
  implicit def customApiRejectionHandler: RejectionHandler =
    RejectionHandler.newBuilder()
      .handle {
        case ValidationRejection(msg, _) =>
          complete(BadRequest -> Error(ErrorCode.BadRequest, RbUtils.rbMessage(resourceBundle, RBKEY_API_VALIDATION_ERROR, msg), jobManagerServiceName().toString))
      }
      .handleAll[MethodRejection] { methodRejections =>
        val names = methodRejections.map(_.supported.name)
        // Not using error entity here as this is not application level validation issue
        complete((MethodNotAllowed, RbUtils.rbMessage(resourceBundle, RBKEY_API_METHOD_NOT_SUPPORTED_ERROR, names.mkString(", "))))
      }
      .handleNotFound { complete((NotFound, RbUtils.rbMessage(resourceBundle, RBKEY_API_ROUTE_NOT_FOUND_ERROR))) }
      .result()
}
