/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.pipeline.job

import com.netapp.df.uba.server.common.flink.cluster.{FlinkApiClusterHandler, FlinkApiJobHandler}
import com.netapp.df.uba.server.common.flink.settings.{FlinkCluster, FlinkJobConfig, FlinkProtocol, FlinkSettings}
import com.netapp.df.uba.server.common.flink.utils.CheckOrSavePointWithCreateTime
import com.netapp.df.uba.server.common.kafka._
import com.netapp.df.uba.server.common.misc.log.LogTrait
import com.netapp.df.uba.server.common.misc.model.Error
import com.netapp.df.uba.server.common.misc.model.ErrorCode.ErrorCode
import com.netapp.df.uba.server.common.misc.utils.GenericUtils
import com.netapp.df.uba.server.common.pipeline.client.PipelineClient
import com.netapp.df.uba.server.common.pipeline.data._
import org.apache.http.HttpHost
import org.apache.kafka.clients.admin.AdminClient

import scala.collection.JavaConverters._
import scala.collection.mutable
import scala.concurrent.duration.FiniteDuration
import scala.util.{Failure, Success, Try}

/**
  * A set of methods used during pipeline management
  */
//noinspection SimplifyBooleanMatch
trait PipelineJobUtils extends LogTrait {
  /*
    Utility method to create a FlinkCluster object using the flink server
    string obtained from the Pipeline object
   */
  def createFlinkCluster(pipeline: Pipeline): FlinkCluster = {
    // A host is created here to ease the split of hostname, port and protocol scheme
    val flinkHttpHost = HttpHost.create(pipeline.getFlinkServer)
    FlinkCluster(
      flinkHttpHost.getHostName,
      flinkHttpHost.getPort,
      FlinkProtocol.withName(flinkHttpHost.getSchemeName)
    )
  }

  /**
    * Validate the input pipeline object for mandatory parameters
    */
  def checkBasicPipelineParameters(inputPipeline: Pipeline): Try[Boolean] = Try {

    // Pipeline object itself should not be null - not sure if it can come like this anyway
    if (inputPipeline == null) throw new Exception("Pipeline definition should be non-null for create API.")

    // Check for infra components as specified
    val kafkaServers = inputPipeline.getKafkaServers
    if (kafkaServers == null || kafkaServers.trim.isEmpty) throw new Exception("Pipeline definition requires Kafka server input parameter.")

    val schemaRegistryServers = inputPipeline.getSchemaRegistryServers
    if (schemaRegistryServers == null || schemaRegistryServers.trim.isEmpty) throw new Exception("Pipeline definition requires Schema Registry server input parameter.")

    val zookeeperServers = inputPipeline.getZookeeperServers
    if (zookeeperServers == null || zookeeperServers.trim.isEmpty) throw new Exception(s"Pipeline definition requires Zookeeper server input parameter.")

    val flinkServer = inputPipeline.getFlinkServer
    if (flinkServer == null || flinkServer.trim.isEmpty) throw new Exception(s"Pipeline definition requires Flink server input parameter.")

    // Ensure parallelism, partition and replica to be > 0 if provided
    if (inputPipeline.getParallelism != null && inputPipeline.getParallelism <= 0) throw new Exception(s"Pipeline definition can not have <= 0 parallelism value.")
    if (inputPipeline.getPartitions != null && inputPipeline.getPartitions <= 0) throw new Exception(s"Pipeline definition can not have <= 0 partitions value.")
    if (inputPipeline.getReplicas != null && inputPipeline.getReplicas <= 0) throw new Exception(s"Pipeline definition can not have <= 0 replicas value.")

    // Note: Fields like active, current, assignable, isTopicOwner is not considered from input. It is directly decided by internal logic
    true
  }

  /**
    * Validate the pipeline type and load the default job config for the pipeline type
    */
  def checkAndGetDefaultJobConfig(inputPipeline: Pipeline,
                                  flinkSettings: FlinkSettings): Try[FlinkJobConfig] = {
    // Extract the pipeline type
    val pipelineType = inputPipeline.getPipelineType

    // Get the baseline flink job config
    val jobBaseName = pipelineType.getJobBaseName
    val jobDefaultConfigOption: Option[FlinkJobConfig] = flinkSettings
      .flinkJobs
      .find(_.jobName == jobBaseName)

    jobDefaultConfigOption match {
      // Error in case pipeline type is not found in conf file
      case None => Failure(new Exception(s"Base job configuration for pipeline type $pipelineType not found."))

      case Some(jobDefaultConfig) => Success(jobDefaultConfig)
    }
  }

  /**
    * Validate if topic owner rules are followed in input
    */
  def validateNonTopicOwnerConfig(inputPipeline: Pipeline,
                                  defaultJobConfig: FlinkJobConfig,
                                  kafkaSettings: KafkaSettings): Try[Boolean] = Try {
    if (defaultJobConfig.isJobTopicOwner) {
      //If topicNames are explicitly specified, validate right topic types are provided.
      if (inputPipeline.getTopicNames != null && !inputPipeline.getTopicNames.isEmpty) {
        val topicNameKeysFromDefaultJobConfig = defaultJobConfig.inputTopicNameKeys
        topicNameKeysFromDefaultJobConfig.foreach(topicNameKey => {
          val topicType = PipelineTopicType.getTypeByKey(topicNameKey)
          if (!inputPipeline.getTopicNames.containsKey(topicType)) {
            throw new Exception(s"Missing topic for type $topicType, pipeline type ${inputPipeline.getPipelineType}.")
          }
        })
      }
      true
    } else {
      // Error checks for the case when pipeline is not the topic owner

      // Check 1: Topic name(s) is mandatory
      if ((inputPipeline.getTopicName == null || inputPipeline.getTopicName.trim.isEmpty) &&
        (inputPipeline.getTopicNames == null || inputPipeline.getTopicNames.isEmpty)) {
        throw new Exception(s"Pipeline where isTopicOwner = false should provide the topic name.")
      }
      // Check 2: Input should not have partitions and replicas option
      if (inputPipeline.getPartitions != null || inputPipeline.getReplicas != null) {
        throw new Exception(s"Pipeline where isTopicOwner = false should not provide partitions or replica count for input topic.")
      }
      // Check 3: Ensure that the topic name provided is already present in the kafka cluster
      if (inputPipeline.getTopicNames != null && !inputPipeline.getTopicNames.isEmpty) {
        val topicNameKeysFromDefaultJobConfig = defaultJobConfig.inputTopicNameKeys
        topicNameKeysFromDefaultJobConfig.foreach(topicNameKey => {
          val topicType = PipelineTopicType.getTypeByKey(topicNameKey)
          val topicName = inputPipeline.getTopicNames.get(topicType)
          if (null == topicName) {
            throw new Exception(s"Missing topic for type $topicType, pipeline type ${inputPipeline.getPipelineType}.")
          }
          val checkTopicExistenceTry = manageTopic(
            inputPipeline,
            kafkaSettings.adminClientRequestTimeout,
            ConsumerTopic(topicName)
          )
          checkTopicExistenceTry match {
            case Success(_) =>
              logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Topic existence checked for $topicName.")
            case Failure(ex) =>
              throw new Exception(ex.getMessage)
          }
        })

        true
      } else {
        val checkTopicExistenceTry = manageTopic(
          inputPipeline,
          kafkaSettings.adminClientRequestTimeout,
          ConsumerTopic(inputPipeline.getTopicName)
        )
        checkTopicExistenceTry match {
          case Success(_) =>
            logger.info(s"Pipeline: ${pipelineString(inputPipeline)} - Topic existence checked for ${inputPipeline.getTopicName}.")
            true
          case Failure(ex) =>
            throw new Exception(ex.getMessage)
        }
      }
    }
  }

  /**
    * Obtain the default topic config for creation using the base name
    */
  def checkAndGetDefaultTopicConfig(defaultJobConfig: FlinkJobConfig,
                                    kafkaSettings: KafkaSettings): Try[Map[PipelineTopicType, ProducerTopic]] = Try {
    val topicNameKeysFromDefaultJobConfig = defaultJobConfig.inputTopicNameKeys
    topicNameKeysFromDefaultJobConfig.map(topicNameKeyFromDefaultJobConfig => {
      val baseProducerTopicConfigOption = kafkaSettings.topicMap.get(topicNameKeyFromDefaultJobConfig)

      val producerTopic = baseProducerTopicConfigOption match {
        case None =>
          // Error in case pipeline type is not found in reference.conf
          val errorMessage = s"Unable to find the base topic configuration for topic name $topicNameKeyFromDefaultJobConfig."
          logger.error(errorMessage)
          throw new Exception(errorMessage)

        case Some(producerTopic: ProducerTopic) => producerTopic

        case Some(_) =>
          // Error in case topic configuration is not defined as a producer in reference.conf
          val errorMessage = s"Unable to find the base topic producer configuration for topic name $topicNameKeyFromDefaultJobConfig."
          logger.error(errorMessage)
          throw new Exception(errorMessage)
      }

      val pipelineTopicType = PipelineTopicType.getTypeByKey(topicNameKeyFromDefaultJobConfig)
      if (null == pipelineTopicType) {
        throw new Exception(s"Unable to find the topic type for base topic name $topicNameKeyFromDefaultJobConfig.")
      }
      pipelineTopicType -> producerTopic
    }).toMap
  }

  /**
    * Utility method to create an updated pipeline object to be sent to service-pipelines
    */
  def buildPipelineObjectForCreation(inputPipeline: Pipeline,
                                     defaultJobConfig: FlinkJobConfig,
                                     defaultTopicConfig: Map[PipelineTopicType, ProducerTopic]): Try[Pipeline] = Try {
    val updatedPipeline = new Pipeline(inputPipeline, inputPipeline.getTenantId)

    /*
     Fields same as in input pipeline:
     1. basics: id, tenantId, pipelineType
     2. infra components: kafka, zk, flink, schema registry
     3. assigned tenants
     4. additional job parameters
     */
    // Set initial status explicitly
    updatedPipeline.setPipelineStatus(PipelineStatus.UNKNOWN)

    /*
     Flag values are as follows:
     current = true is set by service-pipelines
     active = true is set by service-pipelines
     assignable = false is set by SFJM/TJM as of now
     Initially making assignable as false. And once the Pipeline is up and running, this will be set based on supplied
     pipeline input.
     */
    updatedPipeline.setIsAssignable(false)

    // Parallelism if not provided by SRE call explicitly, takes the default which is at job level
    val inputParallelismOption = Option(inputPipeline.getParallelism)
    val parallelismConfigName = s"${defaultJobConfig.jobName}.flink.job.parallelism"
    val parallelismInDefaultJobConfigString = defaultJobConfig.jobParameters.getOrElse(parallelismConfigName, "1")
    updatedPipeline.setParallelism(inputParallelismOption.getOrElse(Integer.parseInt(parallelismInDefaultJobConfigString)))

    // Set the topic owner flag as per the job configuration
    updatedPipeline.setTopicOwner(defaultJobConfig.isJobTopicOwner)

    /*
      Partition and replicas should be provided only when isTopicOwner = true
     */
    if (updatedPipeline.isTopicOwner) {
      // Partitions and replicas if not provided takes the default which is configured at topic level in base topic config
      //In case of multiple topics for a pipeline, partitions and replicas are picked randomly from one of the base topic config.
      val inputPartitionsOption = Option(inputPipeline.getPartitions)
      updatedPipeline.setPartitions(inputPartitionsOption.getOrElse(defaultTopicConfig.head._2.partitionCount))

      val inputReplicasOption = Option(inputPipeline.getReplicas)
      updatedPipeline.setReplicas(inputReplicasOption.getOrElse(defaultTopicConfig.head._2.replicationFactor))
    }

    //If either of topic name, CG name, or  job name are provided, use those as is. Else we generate these below.
    updatedPipeline.setTopicName(inputPipeline.getTopicName)
    updatedPipeline.setTopicNames(inputPipeline.getTopicNames)
    updatedPipeline.setConsumerGroupName(inputPipeline.getConsumerGroupName)
    updatedPipeline.setJobName(inputPipeline.getJobName)

    //Only the job manager (SFJM/TJM) has metadata for number of topics used by a given job. Since topic(s), CG and job name has pipelineId embedded,
    //need to generate these here only, instead of service-pipelines. service-pipelines can then use these pipeline attributes as is.
    val pipelineId = java.util.UUID.randomUUID.toString
    updatedPipeline.setId(pipelineId)

    val jobBaseName = updatedPipeline.getPipelineType.getJobBaseName

    //Generate the topicNames field if not already provided by user. There are 2 possibilities here:
    //1. If the scalar topicName is provided by user, but the vector topicNames is NOT: In this case, use the scalar topicName to populate topicNames field
    //as well, so that both the fields are in sync. For multi-topics pipeline, the 1st topic is picked based on user provided topicName field. While
    // rest are auto-generated.
    //2. If Neither the scalar topicName, NOR vector topicNames are provided by user: Auto-generate all the required topic names for this pipeline.
    //Populate the scalar topicName field as well, so that both the fields are in sync.
    //3. If the vector topicNames is provided by user, but scalar topicName is NOT: In this case, use the topicNames field to auto-populate topicName field
    //as well, so that both are in sync.
    val topicNameKeysFromDefaultJobConfig = defaultJobConfig.inputTopicNameKeys
    if (null == updatedPipeline.getTopicNames || updatedPipeline.getTopicNames.isEmpty) {
      //If user has explicitly provided a topic name, use that to derive topicNames field.
      //For multi-topics pipeline, the 1st topic is picked based on user provided topicName field. While rest are auto-generated.
      if (updatedPipeline.getTopicName != null && updatedPipeline.getTopicName.nonEmpty) {
        val pipelineTopicNamesMap = new java.util.HashMap[PipelineTopicType, String]
        val firstTopicNameKey = topicNameKeysFromDefaultJobConfig.head
        val pipelineTopicType = PipelineTopicType.getTypeByKey(firstTopicNameKey)
        if (null == pipelineTopicType) throw new Exception(s"Invalid base topic configuration for topic name $firstTopicNameKey")

        pipelineTopicNamesMap.put(pipelineTopicType, updatedPipeline.getTopicName)

        val restTopicNameKeysFromDefaultJobConfig = topicNameKeysFromDefaultJobConfig.drop(1)
        populateTopicNamesMap(updatedPipeline, restTopicNameKeysFromDefaultJobConfig, pipelineTopicNamesMap)
      } else {
        val pipelineTopicNamesMap = new java.util.HashMap[PipelineTopicType, String]
        populateTopicNamesMap(updatedPipeline, topicNameKeysFromDefaultJobConfig, pipelineTopicNamesMap)
      }
    } else {
      if (null == updatedPipeline.getTopicName) { //Sync the missing topicName field with the user provided topicNames field.
        updatedPipeline.setTopicName(updatedPipeline.getTopicNames.values().asScala.head)
      }
    }

    if (null == updatedPipeline.getConsumerGroupName) {
      updatedPipeline.setConsumerGroupName(s"${Pipeline.PROCESSING_PREFIX}cg_${pipelineId}_$jobBaseName")
    }

    if (null == updatedPipeline.getJobName) {
      updatedPipeline.setJobName(s"${Pipeline.PROCESSING_PREFIX}${pipelineId}_$jobBaseName")
    }

    updatedPipeline
  }

  private def populateTopicNamesMap(pipeline: Pipeline,
                                    topicNameKeysFromDefaultJobConfig: List[String],
                                    pipelineTopicNamesMap: java.util.HashMap[PipelineTopicType, String]): Unit = {
    //Topic names generation logic. Example topicname for multi topics pipeline:
    //pipeline_177d5cc7-3b42-4ef0-98ad-281e15e81221_0_alerts-sink
    //pipeline_177d5cc7-3b42-4ef0-98ad-281e15e81221_1_alerts-sink
    //And for single topic pipeline:
    //pipeline_7ca2bcc5-5cf4-4f8a-b227-8f036816c00b_0_events-sink
    val pipelineId = pipeline.getId
    val jobBaseName = pipeline.getPipelineType.getJobBaseName
    var topicIdx = 0
    for (topicNameKey <- topicNameKeysFromDefaultJobConfig) {
      val pipelineTopicType = PipelineTopicType.getTypeByKey(topicNameKey)

      if (null == pipelineTopicType) throw new Exception(s"Invalid base topic configuration for topic name $topicNameKey")

      val pipelineTopicName = s"${Pipeline.PROCESSING_PREFIX}${pipelineId}_${topicIdx}_$jobBaseName"

      pipelineTopicNamesMap.put(pipelineTopicType, pipelineTopicName)
      if (null == pipeline.getTopicName) { //Sync the topicName field with the generate topic name using pipeline Id.
        pipeline.setTopicName(pipelineTopicName)
      }
      topicIdx += 1
    }

    pipeline.setTopicNames(pipelineTopicNamesMap)
  }

  /**
    * Create a pipeline that in service-pipelines and get the pipeline
    * object back which will have id, topic name etc filled up
    */
  def createPipeline(pipelineClient: PipelineClient,
                     pipelineRequest: Pipeline): Try[Pipeline] = Try {
    pipelineClient.addPipeline(pipelineRequest)
  }

  /**
    * Obtain the default topic config for creation using the base name
    */
  def createTopicForPipeline(pipeline: Pipeline,
                             defaultTopicConfigs: Map[PipelineTopicType, ProducerTopic],
                             kafkaSettings: KafkaSettings): Try[List[String]] = Try {

    /*
     In case the pipeline is not owner of the topic, it means the topic is already created.
     Return the topic name directly.
     For Reset case, some of the existing pipelines may not have topicNames field present. Hence support both old and new fields.
     */
    if (!pipeline.isTopicOwner) {
      if (pipeline.getTopicNames != null && !pipeline.getTopicNames.isEmpty) {
        return Success(pipeline.getTopicNames.values().asScala.toList)
      } else return Success(List(pipeline.getTopicName))
    }

    defaultTopicConfigs.map(defaultTopicConfigEntry => {
      val topicType = defaultTopicConfigEntry._1
      val defaultTopicConfig = defaultTopicConfigEntry._2

      // Case where the pipeline is owner of the topic.
      val pipelineTopicName = if (pipeline.getTopicNames != null && !pipeline.getTopicNames.isEmpty) {
        pipeline.getTopicNames.get(topicType)
      } else pipeline.getTopicName

      if (null == pipelineTopicName) {
        throw new Exception(s"Internal Error, missing topic name for topic type $topicType")
      }

      /*
       Actual topic configuration is copied from default topic config.
       Copy over the over-riding values from pipeline
       which is topic name, partitions and replica count
       */
      val partitionCountForTopic = Option(pipeline.getPartitions).getOrElse(defaultTopicConfig.partitionCount).asInstanceOf[Int]
      val replicaCountForTopic = Option(pipeline.getReplicas.toShort).getOrElse(defaultTopicConfig.replicationFactor)
      val actualTopicConfiguration = defaultTopicConfig
        .copy(
          name = pipelineTopicName,
          partitionCountForTopic,
          replicaCountForTopic
        )

      val topicCreationResponseTry = manageTopic(
        pipeline,
        kafkaSettings.adminClientRequestTimeout,
        actualTopicConfiguration
      )
      topicCreationResponseTry match {
        case Success(_) => pipelineTopicName
        case Failure(ex) => throw new Exception(ex.getMessage)
      }
    }).toList
  }

  /**
    * Get checkpoints dir name from flink cluster directly.
    * This API is called only when pipelines are created or reset and is not very huge.
    * There are around 30-40 config entries only.
    * In a way, this validates the flink cluster's endpoint provided in the pipeline payload
    */
  def getFlinkClusterCheckpointDirPath(flinkCluster: FlinkCluster): Try[String] = {
    val flinkApiClusterHandler = FlinkApiClusterHandler(flinkCluster)
    flinkApiClusterHandler.retrieveJobManagerConfigForKey("state.checkpoints.dir")
  }

  /**
    * Utility to create the actual job config object used in rest of the code
    * and derived from base job config object
    */
  def buildPipelineJobConfig(pipeline: Pipeline,
                             defaultJobConfig: FlinkJobConfig,
                             defaultKafkaSettings: KafkaSettings,
                             servicePipelinesUrl: String,
                             vaultServiceName: String,
                             tenantServiceUrl: String,
                             serviceConfigMgmtUrl: String,
                             serviceDbsUrl: Option[String],
                             checkpointDirPath: String,
                             jobExternalSavePointFileOption: Option[String] = None): FlinkJobConfig = {
    /*
      Job parameters that needs an over-ride for the pipeline jobs include the following:
      These are the ones that come from the pipeline definition
      1. Kafka cluster
      2. ZK cluster
      3. Schema registry server
      4. Parallelism
      5. consumer group name
      6. topic name(s) to process on
      7. Any additional job level parameters that comes from pipeline
     */
    val jobBaseName = defaultJobConfig.jobName

    val dynamicParameters = mutable.Map[String, String](elems =
    /*
      Default or the main kafka and ZK servers as defined in the main 3rd party helm chart.
      Used in jobs which needs to refer to global caches like pipeline and tenant caches
      or needs to use a distributed lock for its operation
     */
      s"$jobBaseName.global.config.bootstrap.servers" -> defaultKafkaSettings.bootstrapServers,
      s"$jobBaseName.global.config.zookeeper.servers" -> defaultKafkaSettings.zookeeperServers,
      s"$jobBaseName.global.config.service.pipelines.url" -> servicePipelinesUrl,
      s"$jobBaseName.global.config.vault.service.name" -> vaultServiceName,
      s"$jobBaseName.global.config.service.tenants.url" -> tenantServiceUrl,
      s"$jobBaseName.global.config.service.config.mgmt.url" -> serviceConfigMgmtUrl,

      /*
       Kafka and ZK used for job's data flow i.e
       - Kafka cluster which hosts the topic the job consumes records from
       - ZK cluster associated with the Kafka cluster
       - Schema registry cluster associated with the Kafka cluster
       */
      s"$jobBaseName.kafka.config.bootstrap.servers" -> pipeline.getKafkaServers,
      s"$jobBaseName.kafka.config.zookeeper.servers" -> pipeline.getZookeeperServers,
      s"$jobBaseName.kafka.config.schema.registry.url" -> pipeline.getSchemaRegistryServers,

      // Flink job's dynamic parameters - Job name, parallelism and config group id
      s"$jobBaseName.flink.job.name" -> pipeline.getJobName,
      s"$jobBaseName.flink.job.parallelism" -> pipeline.getParallelism.toString,
      s"$jobBaseName.flink.job.source.kafka.consumer.config.group.id" -> pipeline.getConsumerGroupName,
      // Irrespective of job type, change the state.backend.dirname to the correct value depending on the flink cluster
      s"$jobBaseName.flink.job.state.backend.dirname" -> checkpointDirPath,
      s"$jobBaseName.flink.job.restart-strategy.fixed-delay.attempts" -> defaultJobConfig.pipelineRestartStrategyFixedDelayAttempts.toString,
      s"$jobBaseName.flink.job.restart-strategy.fixed-delay.delay" -> s"${defaultJobConfig.pipelineRestartStrategyFixedDelayDelay.toSeconds}s"
    )

    //DBs endpoint is optional.
    if (serviceDbsUrl.isDefined) {
      dynamicParameters += s"$jobBaseName.global.config.service.dbs.url" -> serviceDbsUrl.get
    }

    //Construct all the topic name arguments for the pipeline job.
    // Note: All jobs use config.topic.info as there are no topic existence checks or topic creation in job code
    //In case of Reset, existing pipelines may not have new topicNames attribute. Hence support both old and new topicNames
    //field for passing topic information to job.
    if (pipeline.getTopicNames != null && !pipeline.getTopicNames.isEmpty) {
      defaultJobConfig.inputTopicNameKeys.foreach(inputTopicNameKey => {
        val pipelineTopicType = PipelineTopicType.getTypeByKey(inputTopicNameKey)
        val pipelineTopicName = pipeline.getTopicNames.get(pipelineTopicType)
        dynamicParameters += (s"$jobBaseName.kafka.config.topic.info.$inputTopicNameKey.name" -> pipelineTopicName)
      })
    } else {
      dynamicParameters += s"$jobBaseName.kafka.config.topic.info.${defaultJobConfig.inputTopicNameKeys.head}.name" -> pipeline.getTopicName
    }

    /*
      There can be a set of additional job parameters to be sent to the job
      which would come from the pipeline definition
     */
    val additionalJobParametersAsMap: Map[String, String] = if (pipeline.getAdditionalJobParameters != null) {
      pipeline
        .getAdditionalJobParameters
        .split(",")
        .map(_.trim)
        .filter(_.contains(":"))
        .filterNot(entry => entry.indexOf(':') == entry.length - 1)
        .map(entry => {
          val parameterName = entry.substring(0, entry.indexOf(':')).trim
          val parameterValue = entry.substring(entry.indexOf(':') + 1).trim

          parameterName -> parameterValue
        })
        .toMap
    } else Map.empty

    // Final set of parameters sent to the job
    val finalJobParameters = defaultJobConfig.jobParameters ++
      dynamicParameters ++
      additionalJobParametersAsMap

    // A job configuration that over-rides necessary fields in the default job configuration
    defaultJobConfig
      .copy(
        jobName = pipeline.getJobName,
        jobJarName = pipeline.getJobJarName,
        jobParameters = finalJobParameters,
        jobSavePointTargetDir = checkpointDirPath,
        jobExternalSavePointFileOption = jobExternalSavePointFileOption
      )
  }

  /**
    * Utility to generate the right return value for error case
    */
  def buildErrorObject(inputPipeline: Pipeline,
                       throwable: Throwable,
                       context: String,
                       errorCode: ErrorCode,
                       microServiceName: String): Either[Error, Pipeline] = {
    buildErrorObject(
      pipelineString(inputPipeline),
      throwable,
      context,
      errorCode,
      microServiceName
    )
  }

  /**
    * Another version of buildErrorObject
    */
  def buildErrorObject(pipelineString: String,
                       throwable: Throwable,
                       context: String,
                       errorCode: ErrorCode,
                       microServiceName: String): Either[Error, Pipeline] = {
    logger.error(s"Pipeline: $pipelineString - Creation failed - $context. Reason: ${throwable.getMessage}")
    Left(Error(errorCode, throwable.getMessage, microServiceName))
  }

  /**
    * Another version of buildErrorObject
    */
  def buildErrorObjectForSavePoint(pipeline: Pipeline,
                                   throwable: Throwable,
                                   context: String,
                                   errorCode: ErrorCode,
                                   microServiceName: String): Either[Error, SavePoint] = {
    buildErrorObjectForSavePoint(
      pipelineString(pipeline),
      throwable,
      context,
      errorCode,
      microServiceName
    )
  }

  def buildErrorObjectForSavePoint(pipelineString: String,
                                   throwable: Throwable,
                                   context: String,
                                   errorCode: ErrorCode,
                                   microServiceName: String): Either[Error, SavePoint] = {
    logger.error(s"Pipeline: $pipelineString - Savepoint operation failed - $context. Reason: ${throwable.getMessage}")
    Left(Error(errorCode, throwable.getMessage, microServiceName))
  }

  /**
    * Small utility to get a string representing the pipleine for log messages
    */
  def pipelineString(pipeline: Pipeline): String = {
    if (pipeline == null) "null"
    else if (pipeline.getId != null && pipeline.getId.nonEmpty) s"id: ${pipeline.getId}, type: ${pipeline.getPipelineType}"
    else s"type: ${pipeline.getPipelineType}"
  }

  // Method to do an action using the kafka admin client
  protected def manageTopic(pipeline: Pipeline,
                            adminClientRequestTimeout: FiniteDuration,
                            topicConfig: TopicTrait): Try[Boolean] = {
    // Create the admin client for kafka
    val adminClientTry: Try[AdminClient] = KafkaUtils.createAdminClient(
      pipeline.getKafkaServers,
      adminClientRequestTimeout
    )
    if (adminClientTry.isFailure) throw new Exception(adminClientTry.failed.get.getMessage)
    val adminClient = adminClientTry.get

    val manageTopicResponseTry = topicConfig match {
      case producerTopic: ProducerTopic => KafkaUtils.handleProducerTopic(
        producerTopic,
        adminClient,
        zookeeperServers = pipeline.getZookeeperServers
      )
      case consumerTopic: ConsumerTopic => KafkaUtils.handleConsumerTopic(
        consumerTopic,
        adminClient,
        zookeeperServers = pipeline.getZookeeperServers
      )
    }

    // Close the admin client
    KafkaUtils.closeAdminClient(adminClientTry)

    manageTopicResponseTry
  }

  // A protected method to retry immediately for 'x' number of times for deletion of a pipeline
  protected def deletePipeline(pipelineClient: PipelineClient, pipeline: Pipeline, attempts: Int = 3): Try[Boolean] = Try {

    val context = s"Pipeline: ${pipelineString(pipeline)} - Deletion of pipeline"

    GenericUtils.retry(attempts, context) {
      pipelineClient.deletePipeline(pipeline.getId)
    }
    logger.info(s"Pipeline: ${pipelineString(pipeline)} - Delete API invoked for the pipeline.")
    true
  }

  /**
    * A protected method to retry immediately for 'x' number of times for updating pipeline properties
    * - status - generally changes from resetting to running or reset_failed
    * - isAssignable - set to true post pipeline job creation being successful
    * - parallelism - changed during reset if parallelism changes
    */
  protected def updatePipelineProperties(pipelineClient: PipelineClient,
                                         pipeline: Pipeline,
                                         attempts: Int = 3): Try[Pipeline] = Try {

    val context = s"Pipeline: ${pipelineString(pipeline)} - Update status to ${pipeline.getPipelineStatus}, " +
      s"isAssignable to ${pipeline.getIsAssignable} and parallelism to ${pipeline.getParallelism}"

    GenericUtils.retry(attempts, context) {
      pipelineClient.updatePipeline(pipeline)
    }
    logger.info(s"$context - Completed")
    pipeline
  }

  // A protected method to retry immediately for 'x' number of times for assigning tenants to a pipeline
  protected def assignTenants(pipelineClient: PipelineClient,
                              pipeline: Pipeline,
                              attempts: Int = 3): Try[Pipeline] = Try {
    val context = s"Pipeline: ${pipelineString(pipeline)} - Assign pipeline tenants"

    val returnValue = GenericUtils.retry(attempts, context) {
      val tenantIds = pipeline
        .getAssignedTenants
        .stream()
        .iterator()
        .asScala
        .map[String](tenantToPipeline => tenantToPipeline.getTenantId)
        .toArray

      pipelineClient.addTenantPipeline(pipeline, tenantIds: _*)
    }
    logger.info(s"$context - Successful")
    returnValue
  }

  /**
    * 1. Validate the reset parameters in a reset pipeline API call
    * 2. Get valid savepoint information to be used in the reset call
    */
  protected def checkAndGetSavePointForResetRequest(pipeline: Pipeline,
                                                    pipelineClient: PipelineClient,
                                                    resetPipelineInfo: ResetPipelineInfo,
                                                    defaultJobConfig: FlinkJobConfig): Try[SavepointForResetInfo] = Try {
    val pipelineStatus = pipeline.getPipelineStatus

    // Basic checks for reset request
    resetPipelineInfo.isForce match {
      case false =>
        // If the reset mode is not force, then the pipeline and the job should be in RUNNING state
        if (pipelineStatus != PipelineStatus.RUNNING) {
          throw new Exception(s"Pipeline: ${pipelineString(pipeline)} - Pipeline is not in valid state to reset. Current status: " + pipeline.getPipelineStatus)
        }
        // Check if the job exists for the pipeline
        if (pipeline.getJobs.isEmpty) {
          throw new Exception(s"Pipeline: ${pipelineString(pipeline)} - No job is associated with the pipeline.")
        }
        // Check if the job is in RUNNING status or not
        val pipelineJob = pipeline.getJobs.get(0)
        if (pipelineJob.getState != FlinkApiJobHandler.JobStateRunning) {
          throw new Exception(s"Pipeline: ${pipelineString(pipeline)} - Job ${pipelineJob.getId} is in ${pipelineJob.getState} state.")
        }
      case true => // No basic checks
    }

    // Additional checks and savepoint infor extraction needed for stateful pipelines
    defaultJobConfig.jobSavePointEnabled match {
      case true => // Stateful pipeline
        deriveSavepointInfoToUse(pipeline, pipelineClient, resetPipelineInfo, defaultJobConfig)

      case false => // Stateless pipeline
        SavepointForResetInfo(None, None)
    }
  }

  /**
    * Derive the savepoint info to be used for reset of a stateful job
    * 1. external savepoint path from reset pipeline info if present
    * 2. last savepoint information if it can be considered for restart of the job
    */
  protected def deriveSavepointInfoToUse(pipeline: Pipeline,
                                         pipelineClient: PipelineClient,
                                         resetPipelineInfo: ResetPipelineInfo,
                                         defaultJobConfig: FlinkJobConfig): SavepointForResetInfo = {
    // Should never get called for stateless
    if (!defaultJobConfig.jobSavePointEnabled) return SavepointForResetInfo(None, None)

    // Irrespective of force=true/false, if an external savepoint path is provided, use it and consider only that for state restore
    if (resetPipelineInfo.getSavePointPath != null && resetPipelineInfo.getSavePointPath.nonEmpty) {
      return SavepointForResetInfo(Some(resetPipelineInfo.getSavePointPath), None)
    }

    // Get a valid last savepoint if present for the pipeline
    val lastSavepoint = Try {
      pipelineClient.getLastSavePoint(pipeline.getId)
    } match {
      case Success(last) => last
      case Failure(ex) =>
        logger.error(s"Pipeline: ${pipelineString(pipeline)} - Failure to load last savepoint. Reason: ${ex.getMessage}")
        null
    }
    lastSavepoint match {
      case null =>
        logger.info(s"Pipeline: ${pipelineString(pipeline)} - Last savepoint is null in deriveSavepointInfoToUse.")
        SavepointForResetInfo(None, None)

      // If it is not force and the last savepoint is in progress, throw an error
      case savepoint if !resetPipelineInfo.isForce && savepoint.getStatus.equalsIgnoreCase(SavePoint.IN_PROGRESS_STATUS) =>
        logger.error(s"Pipeline: ${pipelineString(pipeline)} - Last savepoint $lastSavepoint is in-progress and reset is being done with force=false.")
        throw new Exception(s"Last savepoint $lastSavepoint is in-progress and reset is being done with force=false.")

      case savepoint if savepoint.getStatus.equalsIgnoreCase(SavePoint.COMPLETED_STATUS) =>
        logger.info(s"Pipeline: ${pipelineString(pipeline)} - Last savepoint in deriveSavepointInfoToUse with completed status: ${lastSavepoint.getLocation}")
        SavepointForResetInfo(None, Some(CheckOrSavePointWithCreateTime(lastSavepoint.getLocation, lastSavepoint.getCreated.toInstant.toEpochMilli)))

      case _ => // Probably failed status or in_progress status and force=true, continue with retained checkpoints
        logger.info(s"Pipeline: ${pipelineString(pipeline)} - Last savepoint $lastSavepoint - Not valid for reset")
        SavepointForResetInfo(None, None)
    }
  }
}
