/*
 * $Id$
 *
 * Copyright (c) 2022 NetApp, Inc.
 * All rights reserved.
 */
package com.netapp.df.uba.server.common.pipeline.job

import com.netapp.df.uba.server.common.flink.utils.SavepointCallbackHandler
import com.netapp.df.uba.server.common.misc.utils.GenericUtils
import com.netapp.df.uba.server.common.pipeline.client.PipelineClient
import com.netapp.df.uba.server.common.pipeline.data.{Pipeline, SavePoint}

/**
  * An implementation of the trait that calls the pipeline service to
  * register the savepoint in the database
  */
class PipelineSavepointCallbackHandler(pipeline: Pipeline, pipelineClient: PipelineClient)
  extends SavepointCallbackHandler
    with PipelineJobUtils  {

  private var savepointOption: Option[SavePoint] = None
  private var triggerFailureExceptionOption: Option[Exception] = None

  override def savepointTriggered(triggerId: String): Unit = {
    val context = s"Pipeline: ${pipelineString(pipeline)} - Savepoint trigger for pipeline"
    try {
      GenericUtils.retry(attempts = 3, context) {
        savepointOption = Some(
          pipelineClient.addSavePoint(
            pipeline.getId,
            null,
            triggerId,
            SavePoint.IN_PROGRESS_STATUS
          )
        )
      }
      logger.info(s"Pipeline: ${pipelineString(pipeline)} - Savepoint add operation for savepoitn with triggerId: $triggerId.")
    } catch {
      case ex: Exception =>
        triggerFailureExceptionOption = Some(ex)
    }
  }

  override def savepointSucceeded(triggerId: String, location: String): Unit = {

    if (savepointOption.isEmpty) {
      logger.error(s"Pipeline: ${pipelineString(pipeline)} - Savepoint object is not defined in savepointSucceeded call.")
      return
    }
    try {
      val context = s"Pipeline: ${pipelineString(pipeline)} - Savepoint succeeded for pipeline"
      GenericUtils.retry(attempts = 3, context) {
        pipelineClient.updateSavePoint(
          pipeline.getId,
          savepointOption.get.getId,
          location,
          SavePoint.COMPLETED_STATUS
        )
      }
      logger.info(s"Pipeline: ${pipelineString(pipeline)} - Savepoint update operation for savepoint with triggerId: $triggerId and status as COMPLETED_STATUS.")
    } catch {
      case _: Exception => // Already logged
    }
  }

  override def savepointFailed(triggerId: String): Unit = {
    if (savepointOption.isEmpty) {
      logger.error(s"Pipeline: ${pipelineString(pipeline)} - Savepoint object is not defined in savepointSucceeded call.")
      return
    }
    try {
      val context = s"Pipeline: ${pipelineString(pipeline)} - Savepoint failed for pipeline"
      GenericUtils.retry(attempts = 3, context) {
        pipelineClient.updateSavePoint(
          pipeline.getId,
          savepointOption.get.getId,
          null,
          SavePoint.FAILED_STATUS
        )
      }
      logger.info(s"Pipeline: ${pipelineString(pipeline)} - Savepoint update operation for savepoint with triggerId: $triggerId and status as FAILED_STATUS.")
    } catch {
      case _: Exception => // Already logged
    }
  }

  def getSavepointOption: Option[SavePoint] = savepointOption

  def getTriggerFailureExceptionOption: Option[Exception] = triggerFailureExceptionOption
}